#ifndef BASICRESOURCEMANAGER_H
#define BASICRESOURCEMANAGER_H

#include "ResourceManager.h"
#include "Logger.h"

template<typename T> class BasicResourceManager : public ResourceManager
{
private:
	typedef std::multimap<std::string, T*> MapType;

public:
	BasicResourceManager( Engine& engine,
						  const std::string& componentName ) : ResourceManager( engine, componentName )
	{
	}

	virtual ~BasicResourceManager()
	{
		runGC();

		// alle Ressourcen lschen
		std::list<Resource*> resources = getResources();
		for( std::list<Resource*>::const_iterator it = resources.begin();
			 it != resources.end();
			 it++ )
		{
			std::ostringstream msg;
			msg << "Resource " << (*it)->getFilename() << " wird automatisch freigegeben";
			getEngine().getLogger().warning( msg.str(), LOGGER_AUTO );

			destroyResource( *it );
		}
	}

	ResourceHandle get( const std::string& filename,
						bool unique = false,
						const ParameterMap& options = ParameterMap() )
	{
		if( !unique )
		{
			// Gibt es bereits eine nicht-einzigartige Ressource mit diesem Dateinamen und diesen Optionen?
			std::pair<MapType::iterator, MapType::iterator> range = resources.equal_range( filename );
			for( MapType::iterator it = range.first;
				 it != range.second;
				 it++ )
			{
				if( !it->second->isUnique() &&
					it->second->getOptions() == options )
				{
					// Gefunden!
					return ResourceHandle( it->second );
				}
			}
		}

		// Die Ressource ist noch nicht geladen.
		T* p_resource = new T( *this, filename, unique, options );
		if( p_resource->getError() )
		{
			// Fehler!
			delete p_resource;
			std::ostringstream msg;
			msg << "Ressource " << filename << " konnte nicht geladen werden";
			getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
			return ResourceHandle( 0 );
		}

		// neue Ressource eintragen und Handle liefern
		std::ostringstream msg;
		if( unique ) msg << "Einzigartige ";
		msg << "Ressource " << filename << " wurde geladen";
		getEngine().getLogger().fine( msg.str(), LOGGER_AUTO );
		addResource( p_resource );

		return ResourceHandle( p_resource );
	}

	void destroyResource( Resource* p_resource )
	{
		// sicherstellen, dass die Ressource vom richtigen Typ ist
		T* p_myResource = dynamic_cast<T*>( p_resource );
		assert( p_myResource );

		// die Ressource suchen und entfernen
		for( MapType::iterator it = resources.begin();
			 it != resources.end();
			 it++ )
		{
			if( it->second == p_resource )
			{
				resources.erase( it );
				break;
			}
		}

		// Ressource lschen
		delete p_myResource;
	}

	std::list<Resource*> getResources() const
	{
		std::list<Resource*> theList;
		for( MapType::const_iterator it = resources.begin();
			 it != resources.end();
			 it++ )
		{
			theList.push_back( it->second );
		}

		return theList;
	}

	const MapType& getResourcesMap() const
	{
		return resources;
	}

protected:
	void addResource( Resource* p_resource )
	{
		resources.insert( std::pair<std::string, T*>( p_resource->getFilename(), static_cast<T*>( p_resource ) ) );
	}

private:
	MapType resources;
};

#endif