#ifndef ENGINE_H
#define ENGINE_H

#include "ParameterMap.h"
#include "Logger.h"
#include "Texture.h"
#include "Font.h"
#include "Sound.h"

class Input;
class Component;
class Logger;
class Video;
class Audio;
class GameState;

enum EngineMode
{
	EM_NORMAL,
	EM_PROCESS_ONLY,

	EM_LAST
};

// Hauptklasse der Engine
class ENGINE_API Engine : public Base
{
public:
	Engine( const ParameterMap& options = ParameterMap() );
	~Engine();

	// registriert eine Komponente - die Engine ist nicht fr die Zerstrung des Objekts verantwortlich
	int registerComponent( const std::string& name, Component* p_component );

	// trgt die Komponente aus der Liste aus
	int unregisterComponent( const std::string& name );

	// liefert die Komponente mit dem angegebenen Namen oder null, falls es diese nicht gibt
	Component* getComponentByName( const std::string& name ) const;

	// setzt den Modus der Engine
	void setMode( EngineMode mode );

	// liefert den Modus der Engine zurck
	EngineMode getMode() const;

	// liefert den Logger der Engine zurck
	Logger& getLogger();

	// liefert eine Referenz auf die Grafikkomponente
	Video& getVideo() const;

	// liefert eine Referenz auf die Audiokomponente
	Audio& getAudio() const;

	// liefert die Input-Komponente zurck
	Input& getInput() const;

	// liefert eine Referenz auf den Texturmanager
	TextureManager& getTextureManager() const;

	// liefert eine Referenz auf den Font-Manager
	FontManager& getFontManager() const;

	// liefert eine Referenz auf den Sound-Manager
	SoundManager& getSoundManager() const;

	// registriert einen Spielzustand - die Engine ist nicht fr die Zerstrung des Objekts verantwortlich
	int registerGameState( const std::string& name, GameState* p_state );

	// entfernt den Spielzustand aus der Liste
	int unregisterGameState( const std::string& name );

	// liefert den Spielzustand mit dem angegebenen Namen oder null, falls es diesen nicht gibt
	GameState* getGameStateByName( const std::string& name ) const;

	// liefert den obersten Spielzustand auf dem Stack
	GameState* getCurrentGameState() const;

	// betritt einen neuen Spielzustand und legt ihn auf den Stack
	int pushGameState( const std::string& name, const ParameterMap& params = ParameterMap() );

	// verlsst den obersten Spielzustand und entfernt ihn vom Stack
	int popGameState( bool deactivateCurrent = true, bool activateBelow = true );

	// ersetzt den gesamten Spielzustands-Stack durch einen einzigen neuen Zustand
	int replaceGameStates( const std::string& name, const ParameterMap& params = ParameterMap() );

	// Main-Loop der Engine ausfhren, Rckgabewert != 0 bei Fehler
	int run();

	// veranlasst das Beenden des Main-Loops
	void quit();

	// liefert die Optionen, mit denen die Engine initialisiert wurde
	const ParameterMap& getOptions() const;

	// liefert die Tick-Dauer zurck, in Millisekunden
	unsigned int getTickDuration() const;

	// liefert die Tick-Dauer zurck, in Sekunden
	float getTickDurationInSeconds() const;

	// liefert die Zeit seit Beginn des Main-Loops, in Millisekunden
	unsigned int getTime() const;

	// liefert die Zeit seit Beginn des Main-Loops, in Sekunden
	float getTimeInSeconds() const;

	// liefert die aktuelle Zeit in hoher Auflsung
	unsigned long long getHighResTime() const;

	// liefert die Auflsung des Hochauflsungs-Timers (in Ticks pro Sekunde)
	unsigned long long getHighResTimerResolution() const;

	DEFINE_SIGNAL0( EngineShutDown );
	DEFINE_SIGNAL0( AppGotFocus );
	DEFINE_SIGNAL0( AppLostFocus );
	DEFINE_SIGNAL1( SDLEvent, const SDL_Event& );
	DEFINE_SIGNAL0( FrameBegin );
	DEFINE_SIGNAL0( PreRender );
	DEFINE_SIGNAL0( PostRender );
	DEFINE_SIGNAL0( PostRenderFinal );
	DEFINE_SIGNAL0( FrameEnd );
	DEFINE_SIGNAL0( Update );

private:
	// Datentyp fr Zustandsoperationen
	struct StateOperation
	{
		// Welcher Zustand ist betroffen (nur wichtig bei Push)?
		GameState* p_state;

		// Push oder Pop?
		bool push;

		// Soll der aktuelle Zustand deaktiviert werden (nur wichtig bei Pop)?
		bool deactivateCurrent;

		// Soll der darunter liegende Zustand wieder aktiviert werden (nur wichtig bei Pop)?
		bool activateBelow;

		// Parameter (nur wichtig bei Push)
		ParameterMap params;
	};

	// fhrt alle anstehenden Zustandsbergnge aus
	void performStateTransitions();

	// interne Methode zum Konfigurieren des Loggers - aufgerufen im Konstruktor
	void initLogger();

	// interne Methode zum Freigeben der Logger-Objekte - aufgerufen im Destruktor
	void cleanUpLogger();

	// Modus
	EngineMode mode;

	// der Logger der Engine
	Logger logger;

	// Konsolen-Log-Handler
	LogHandler* p_consoleLogHandler;

	// Map aller Komponenten
	std::map<std::string, Component*> componentRegistry;

	// Grafikkomponente
	Video* p_video;

	// Audiokomponente
	Audio* p_audio;

	// Input-Komponente
	Input* p_input;

	// Texturmanager
	TextureManager* p_textureManager;

	// Font-Manager
	FontManager* p_fontManager;

	// Sound-Manager
	SoundManager* p_soundManager;

	// Stack der Spielzustnde, die gerade laufen
	std::stack<GameState*> currentStates;

	// Map aller zur Verfgung stehenden Spielzustnde
	std::map<std::string, GameState*> stateRegistry;

	// Liste von anstehenden Zustandsoperationen
	std::queue<StateOperation> pendingStateOperations;

	// Gre des Spielzustand-Stacks nach Durchfhrung aller anstehenden Operationen
	int newStackSize;

	// Optionen der Engine
	ParameterMap options;

	// Tick-Dauer in Millisekunden
	unsigned int tickDuration;

	// Zeit seit Beginn des Main-Loops in Millisekunden
	unsigned int time;
};

#endif