#ifndef FONT_H
#define FONT_H

#include "Texture.h"

struct FontStyle
{
	FontStyle( int characterSpacing = 0,
			   int lineSpacing = 0,
			   int italic = 0 ) : characterSpacing( characterSpacing ),
								  lineSpacing( lineSpacing ),
								  italic( italic )
	{
	}

	// zustzlicher Abstand zwischen zwei Zeichen (kann auch negativ sein)
	int characterSpacing;

	// zustzlicher Zeilenabstand (kann auch negativ sein)
	int lineSpacing;

	// Verschiebung der oberen Buchstabenkante nach rechts (kann auch negativ sein)
	int italic;

	bool operator == ( const FontStyle& rhs ) const
	{
		return characterSpacing == rhs.characterSpacing &&
			   lineSpacing == rhs.lineSpacing &&
			   italic == rhs.italic;
	}
};

enum FormattingFlags
{
	FF_LEFT		= 0x0,	// links ausrichten (Standard)
	FF_TOP		= 0x0,	// oben ausrichten (Standard)
	FF_RIGHT	= 0x1,	// rechts ausrichten
	FF_BOTTOM	= 0x2,	// unten ausrichten
	FF_HCENTER	= 0x4,	// horizontal zentrieren
	FF_VCENTER	= 0x8,	// vertikal zentrieren

	FF_LAST
};

class ENGINE_API Font : public Resource
{
	friend class BasicResourceManager<Font>;

public:
	// gibt einen Text aus
	void print( const std::string& text, float lineHeight = 0.0f, const Vec4f& color = Vec4f( 1.0f, 1.0f, 1.0f, 1.0f ), unsigned int formattingFlags = 0, const FontStyle& style = FontStyle(), bool useCache = true, int maxWidth = 0 );

	// bereitet einen Text vor und liefert die berechnete Position jedes Zeichens
	void prepareText( const std::string& text, const FontStyle& style = FontStyle(), bool useCache = true, int maxWidth = 0, Vec2i* p_sizeOut = 0, std::vector<Vec2i>* p_charPositionsOut = 0, unsigned int* p_listIDOut = 0 );

	// fgt automatische Zeilenumbrche in einen Text ein
	std::string adjustText( const std::string& text, const FontStyle& style, int maxWidth ) const;

	// liefert die notwendige Skalierung fr eine bestimmte Texthhe
	float getScaling( float lineHeight ) const;

	// leert den Text-Cache
	void clearCache();

	// lscht den ltesten Cache-Eintrag
	void purgeCache();

	// liefert die aktuelle Cache-Gre
	unsigned int getCacheSize() const;

	// setzt die maximale Cache-Gre
	void setMaxCacheSize( unsigned int maxCacheSize );

	// liefert die maximale Cache-Gre
	unsigned int getMaxCacheSize() const;

private:
	struct CharacterInfo
	{
		Vec2i position;
		Vec2i size;
		Vec2i offset;
		int advanceX;
	};

	struct CacheEntry
	{
		unsigned int lastUsed;
		std::string text;
		FontStyle style;
		int maxWidth;
		Vec2i size;
		std::vector<Vec2i> charPositions;
		unsigned int listID;
	};

	Font( ResourceManager& manager, const std::string& filename, bool unique, const ParameterMap& options );
	~Font();

	int getKerning( unsigned char char1, unsigned char char2 ) const;
	void deleteCacheEntry( const CacheEntry* p_entry );

	// die Textur, die alle Zeichen enthlt
	TextureHandle texture;

	int lineHeight;
	int base;
	CharacterInfo* characterInfo[256];
	std::map<std::pair<int, int>, int> kerning;
	std::list<CacheEntry*> cache;
	unsigned int maxCacheSize;
};

// Farbcode-Funktionen
ENGINE_API std::string beginColor( const Vec4f& color );
ENGINE_API std::string beginColor( float r, float g, float b, float a );
ENGINE_API std::string endColor();

typedef BasicResourceManager<Font> FontManager;
typedef BasicResourceHandle<Font> FontHandle;

#endif