#ifndef MAT33_H
#define MAT33_H

#include "Vec.h"

// 3x3-Matrixklasse
template<typename T> class Mat33
{
public:
	Mat33()
	{
	}

	Mat33( const Mat33& rhs )
	{
		for( int i = 0; i < 9; i++ )
		{
			value[i] = rhs.value[i];
		}
	}

	Mat33( const T* p_values )
	{
		for( int i = 0; i < 9; i++ )
		{
			value[i] = p_values[i];
		}
	}

	~Mat33()
	{
	}

	T determinant() const
	{
		return m11 * ( m22 * m33 - m23 * m32 ) +
			   m12 * ( m23 * m31 - m21 * m33 ) +
			   m13 * ( m21 * m32 - m22 * m31 );
	}

	Mat33 inverse() const
	{
		Mat33 r;
		r.m11 = m22 * m33 - m23 * m32;
		r.m12 = m13 * m32 - m12 * m33;
		r.m13 = m12 * m23 - m13 * m22;
		r.m21 = m23 * m31 - m21 * m33;
		r.m22 = m11 * m33 - m13 * m31;
		r.m23 = m13 * m21 - m11 * m23;
		r.m31 = m21 * m32 - m22 * m31;
		r.m32 = m12 * m31 - m11 * m32;
		r.m33 = m11 * m22 - m12 * m21;
		return r / determinant();
	}

	Mat33& invert()
	{
		return *this = inverse();
	}

	Mat33 transposed() const
	{
		Mat33 r;
		for( int i = 0; i < 3; i++ )
		{
			for( int j = 0; j < 3; j++ )
			{
				r.m[i][j] = m[j][i];
			}
		}

		return r;
	}

	Mat33& transpose()
	{
		return *this = transposed();
	}

	// transformiert einen 2D-Positionsvektor, ohne Projektion
	Vec<2, T> transformPosition( const Vec<2, T>& rhs ) const
	{
		return Vec<2, T>( m11 * rhs.x + m12 * rhs.y + m13,
						  m21 * rhs.x + m22 * rhs.y + m23 );
	}

	// transformiert einen 2D-Richtungsvektor
	Vec<2, T> transformDirection( const Vec<2, T>& rhs ) const
	{
		return Vec<2, T>( m11 * rhs.x + m12 * rhs.y,
						  m21 * rhs.x + m22 * rhs.y );
	}

	// liefert die n-te Achse
	Vec<2, T> getAxis( unsigned int n ) const
	{
		return Vec<2, T>( m[n][0], m[n][1] );
	}

	// liefert die Identittsmatrix
	static Mat33 identity()
	{
		Mat33 r;
		r.m11 = r.m22 = r.m33 = static_cast<T>( 1 );
		r.m12 = r.m13 = r.m21 = r.m23 = r.m31 = r.m32 = static_cast<T>( 0 );
		return r;
	}

	// erzeugt eine Translationsmatrix
	static Mat33 translation( const Vec<2, T>& v )
	{
		Mat33 r( identity() );
		r.m13 = v.x;
		r.m23 = v.y;
		return r;
	}

	// erzeugt eine Rotationsmatrix
	static Mat33 rotation( float angle )
	{
		Mat33 r( identity() );
		r.m11 = static_cast<T>( cosf( angle ) );
		r.m21 = static_cast<T>( sinf( angle ) );
		r.m12 = -r.m21;
		r.m22 = r.m11;
		return r;
	}

	// erzeugt eine Skalierungsmatrix
	static Mat33 scaling( T scaling )
	{
		Mat33 r;
		r.m11 = r.m22 = r.m33 = scaling;
		r.m12 = r.m13 = r.m21 = r.m23 = r.m31 = r.m32 = static_cast<T>( 0 );
		return r;
	}

	// erzeugt eine kombinierte Translations- Skalierungs- und Rotationsmatrix
	static Mat33 transformation( const Vec<2, T>& position,
								 float angle,
								 T scaling )
	{
		Mat33 r;
		r.m11 = scaling * static_cast<T>( cosf( angle ) );
		r.m21 = scaling * static_cast<T>( sinf( angle ) );
		r.m12 = -r.m21;
		r.m22 = r.m11;
		r.m13 = position.x;
		r.m23 = position.y;
		r.m31 = r.m32 = static_cast<T>( 0 );
		r.m33 = scaling;
		return r;
	}

	const T& operator () ( unsigned int row,
						   unsigned int col ) const
	{
		return m[row][col];
	}

	T& operator () ( unsigned int row,
					 unsigned int col )
	{
		return m[row][col];
	}

	Mat33 operator + ( const Mat33& rhs ) const
	{
		Mat33 r;
		for(int i = 0; i < 9; i++)
		{
			r.value[i] = value[i] + rhs.value[i];
		}

		return r;
	}

	Mat33 operator - ( const Mat33& rhs ) const
	{
		Mat33 r;
		for( int i = 0; i < 9; i++ )
		{
			r.value[i] = value[i] - rhs.value[i];
		}

		return r;
	}

	Mat33 operator - () const
	{
		Mat33 r;
		for( int i = 0; i < 9; i++ )
		{
			r.value[i] = -value[i];
		}

		return r;
	}

	Mat33 operator * ( const Mat33& rhs ) const
	{
		Mat33 r;
		for( int i = 0; i < 3; i++ )
		{
			for( int j = 0; j < 3; j++ )
			{
				r.m[i][j] = m[i][0] * rhs.m[0][j] +
					        m[i][1] * rhs.m[1][j] +
							m[i][2] * rhs.m[2][j];
			}
		}

		return r;
	}

	Mat33 operator * ( T rhs ) const
	{
		Mat33 r;
		for( int i = 0; i < 9; i++ )
		{
			r.value[i] = value[i] * rhs;
		}

		return r;
	}

	// Transformation eines 3D-Vektors
	Vec<3, T> operator * ( const Vec<3, T>& rhs ) const
	{
		return Vec<3, T>( m11 * rhs.x + m12 * rhs.y + m13 * rhs.z,
						  m21 * rhs.x + m22 * rhs.y + m23 * rhs.z,
						  m31 * rhs.x + m32 * rhs.y + m33 * rhs.z );
	}

	Mat33 operator / ( T rhs ) const
	{
		Mat33 r;
		for( int i = 0; i < 9; i++ )
		{
			r.value[i] = value[i] / rhs;
		}

		return r;
	}

	Mat33& operator = ( const Mat33& rhs )
	{
		for( int i = 0; i < 9; i++ )
		{
			value[i] = rhs.value[i];
		}

		return *this;
	}

	Mat33& operator += ( const Mat33& rhs )
	{
		return *this = *this + rhs;
	}

	Mat33& operator -= ( const Mat33& rhs )
	{
		return *this = *this - rhs;
	}

	Mat33& operator *= ( const Mat33& rhs )
	{
		return *this = *this * rhs;
	}

	Mat33& operator *= (T rhs)
	{
		return *this = *this * rhs;
	}

	Mat33& operator /= ( T rhs )
	{
		return *this = *this / rhs;
	}

	bool operator == ( const Mat33& rhs ) const
	{
		for( int i = 0; i < 9; i++ )
		{
			if( value[i] != rhs.value[i] ) return false;
		}

		return true;
	}

	bool operator != ( const Mat33& rhs ) const
	{
		return !( *this == rhs );
	}

	// Konvertierung in anderes Matrixformat
	template<typename U> operator Mat33<U> () const
	{
		Mat33<U> r;
		for( int i = 0; i < 9; i++ )
		{
			r.value[i] = static_cast<U>( value[i] );
		}

		return r;
	}

	operator const T* () const
	{
		return value;
	}

	operator T* ()
	{
		return value;
	}

	// die Elemente
	union
	{
		T m[3][3];
		T value[9];

		struct
		{
			T m11;
			T m12;
			T m13;
			T m21;
			T m22;
			T m23;
			T m31;
			T m32;
			T m33;
		};
	};
};

// Multipliktion mit Skalar von links
template<typename T> inline Mat33<T> operator * ( T lhs,
												  const Mat33<T>& rhs )
{
	return rhs * lhs;
}

// in Stream schreiben
template<typename T> inline std::ostream& operator << ( std::ostream& lhs,
														const Mat33<T>& rhs )
{
	for( int i = 0; i < 9; i++ )
	{
		lhs << rhs.value[i];
		if( i != 8 ) lhs << " ";
	}

	return lhs;
}

// aus Stream lesen
template<typename T> inline std::istream& operator >> ( std::istream& lhs,
														Mat33<T>& rhs )
{
	for( int i = 0; i < 9; i++ )
	{
		lhs >> rhs.value[i];
	}

	return lhs;
}

// Standard-Matrixtypen
typedef Mat33<int> Mat33i;
typedef Mat33<float> Mat33f;
typedef Mat33<double> Mat33d;

#endif