#ifndef PARTICLESYSTEM_H
#define PARTICLESYSTEM_H

#include "Texture.h"

class Video;
class ParticleEmitter;
class SimpleEmitter;

struct Particle
{
	// restliche Lebenszeit
	float timeToLive;

	// Position
	Vec2f position;

	// Bewegungsvektor
	Vec2f velocity;

	// Drehung
	float angle;

	// Drehgeschwindigkeit
	float angularVelocity;

	// Farbe
	Vec4f color;

	// Farbnderung pro Sekunde
	Vec4f deltaColor;

	// Gre
	float size;

	// Grennderung pro Sekunde
	float deltaSize;

	// Reibung (1 = kein Verlust, 0 = sofortiger Stillstand)
	float friction;

	// Gravitationseinfluss
	float gravity;

	// aktuelles Animations-Frame
	float animFrame;
};

enum ParticleMode
{
	PM_NORMAL,
	PM_ROTATE_BY_VELOCITY,
	PM_STRETCH_BY_VELOCITY
};

struct ParticleStyle
{
	// Modus
	ParticleMode mode;

	// Streckfaktor fr PM_STRETCH_BY_VELOCITY
	float stretch;

	// Textur
	TextureHandle texture;

	// Animation
	std::vector<Vec2i> frames;

	// Gre der Frames auf der Textur
	Vec2i frameSize;

	// Animationsgeschwindigkeit (in Frames pro Sekunde)
	float animSpeed;

	// Quell-Blendfaktor
	unsigned int srcBlendFactor;

	// Ziel-Blendfaktor
	unsigned int destBlendFactor;
};

class ENGINE_API ParticleSystem : public Base
{
public:
	ParticleSystem( Video& video, unsigned int tickDuration );
	~ParticleSystem();

	// registriert einen Partikel-Stil
	int registerParticleStyle( unsigned int styleID, const ParticleStyle& style );

	// lscht einen Partikel-Stil
	int unregisterParticleStyle( unsigned int styleID );

	// fgt einen Partikel hinzu
	int addParticle( unsigned int styleID, const Particle& particle );

	// erzeugt einen einfachen Emitter
	SimpleEmitter* createSimpleEmitter();

	// lscht einen Emitter
	int destroyEmitter( ParticleEmitter* p_emitter );

	// prft, ob der Emitter gltig ist
	bool isEmitterValid( ParticleEmitter* p_emitter ) const;

	// leert die Partikel
	void clearParticles();

	// rendert das Partikelsystem
	void render( bool renderMultiplicative = true, bool renderAdditive = true );

	// bewegt alle Partikel und verarbeitet die Emitter (nicht wirklich!)
	void update();

	// setzt den Gravitationsvektor
	void setGravity( const Vec2f& gravity );

	// liefert den Gravitationsvektor
	const Vec2f& getGravity() const;

	// setzt die globale Partikelrate
	void setGlobalRate( float globalRate );

	// liefert die globale Partikelrate
	float getGlobalRate() const;

	// liefert die aktuelle Partikelanzahl
	unsigned int getNumParticles() const;

	// liefert die aktuelle Anzahl der Emitter
	unsigned int getNumEmitters() const;

	// setzt die maximale Partikelanzahl
	void setCapacity( unsigned int capacity );

	// liefert die maximale Partikelanzahl
	unsigned int getCapacity() const;

	// liefert die Tick-Dauer
	unsigned int getTickDuration() const;

	// liefert die Tick-Dauer in Sekunden
	float getTickDurationInSeconds() const;

	// liefert das Video-Objekt
	Video& getVideo() const;

private:
	void renderParticles( const ParticleStyle& style, const std::list<Particle>& particles );

	// das Video-Objekt
	Video& video;

	// Partikelanzahl
	unsigned int numParticles;

	// maximale Partikelanzahl
	unsigned int capacity;

	// Partikel-Stile
	std::map<unsigned int, ParticleStyle> particleStyles;

	// Liste der Partikel jedes Stils
	std::map<unsigned int, std::list<Particle> > particleLists;

	// Partikel-Emitter
	std::set<ParticleEmitter*> emitters;

	// Gravitationsvektor
	Vec2f gravity;

	// globale Partikelrate
	float globalRate;

	// Tick-Dauer (damit kann das Partikelsystem langsamer laufen als die Engine)
	const unsigned int tickDuration;

	// Tick-Dauer in Sekunden
	const float tickDurationF;

	// Kehrwert der Tick-Dauer in Sekunden
	const float oneOverTickDurationF;

	// Zeit, die noch verarbeitet werden muss
	unsigned int timeToProcess;
};

#endif