#ifndef SERIALIZER_H
#define SERIALIZER_H

#include "ObjectSerializer.h"

class ENGINE_API Serializer
{
public:
	static Serializer& inst();

	// registriert einen Objektserialisierer fr einen bestimmten Typ
	void registerObjectSerializer( const std::string& type, ObjectSerializer* p_serializer );

	// entfernt einen Objektserialisierer
	void unregisterObjectSerializer( const std::string& type );

	// liefert den Objektserialisierer fr einen bestimmten Typ
	ObjectSerializer* getObjectSerializerByType( const std::string& type ) const;

	// legt den lesbaren Typennamen fr einen Objektserialisierer fest
	void setReadableName( const std::string& type, const std::string& readableName );

	// serialisiert das bergebene Objekt
	template<typename T> int serialize( const T& object,
										SerializationStream& target )
	{
		// Typeninformationen holen
		const type_info& ti = typeid( T );

		// Objektserialisierer holen
		ObjectSerializer* p_serializer = getObjectSerializerByType( ti.name() );
		if( !p_serializer ) return 1;

		// Objekt serialisieren
		target.beginObject( p_serializer->getReadableName() );
		p_serializer->serialize( &object, target );
		target.endObject();

		return 0;
	}

	// deserialisiert ein Objekt
	Variant* deserialize( const SerializationStream& source ) const
	{
		// Typ lesen
		std::string type = source.openObject();
		if( type.empty() ) return 0;

		// Objektserialisierer holen
		ObjectSerializer* p_serializer = getObjectSerializerByType( type );
		if( !p_serializer ) return 0;

		// Objekt deserialisieren
		Variant* p_object = p_serializer->deserialize( source );
		source.closeObject();

		return p_object;
	}

	// deserialisiert ein Objekt
	template<typename T> int deserialize( T& outObject,
										  const SerializationStream& source ) const
	{
		// Typeninformationen holen
		const type_info& ti = typeid( T );

		// Objektserialisierer holen
		ObjectSerializer* p_serializer = getObjectSerializerByType( ti.name() );
		if( !p_serializer ) return 1;

		// Objekt ffnen und prfen, ob der Typ richtig ist
		std::string type = source.openObject();
		if( type != ti.name() && type != p_serializer->getReadableName() )
		{
			source.closeObject();
			return 2;
		}

		// Objekt deserialisieren
		BasicVariant<T>* p_object = dynamic_cast<BasicVariant<T>*>( p_serializer->deserialize( source ) );
		source.closeObject();

		if( !p_object ) return 3;

		// Wert kopieren und das Objekt lschen
		outObject = p_object->getValue();
		delete p_object;

		return 0;
	}

private:
	Serializer();
	Serializer( const Serializer& copy );
	~Serializer();

	// die registrierten Objektserialisierer
	std::map<std::string, ObjectSerializer*> objectSerializerRegistry;
};

#endif