#ifndef UTILS_H
#define UTILS_H

// diese Funktion erwartet einen Pfad und liefert nur den Dateinamen
ENGINE_API std::string getPathFilename( const std::string& path );

// diese Funktion erwartet einen Pfad und liefert nur das Verzeichnis
ENGINE_API std::string getPathDirectory( const std::string& path );

// diese Funktion entfernt die Dateiendung
ENGINE_API std::string removeExtension( const std::string& filename );

// listet alle Dateien auf, die dem Muster entsprechen
ENGINE_API std::list<std::string> listFiles( const std::string& pattern );

// liefert die nchste Zweierpotenz
ENGINE_API unsigned int nextPow2( unsigned int value );

// int-Zufallszahl
ENGINE_API int randInt( int min, int max );

// unsigned int-Zufallszahl im ganzen Wertebereich
ENGINE_API unsigned int randInt();

// float-Zufallszahl
ENGINE_API float randFloat( float min = 0.0f, float max = 1.0f );

// double-Zufallszahl
ENGINE_API double randDouble( double min = 0.0, double max = 1.0 );

// zuflliger 2D-Richtungsvektor
ENGINE_API Vec2f randVec2fDir();

// liefert die reellen Lsungen einer Gleichung vierten Grades
ENGINE_API int solveQuartic( double a, double b, double c, double d, double e, double* p_outRealRoots );

// passt einen Wert linear an einen Zielwert an
template<typename T> void adjustLinear( T& value,
										const T& target,
										float speed,
										float dt)
{
	if( value < target )
	{
		value += dt * speed;
		if( value > target ) value = target;
	}
	else if( value > target )
	{
		value -= dt * speed;
		if( value < target ) value = target;
	}
}

// passt einen Wert weich an einen Zielwert an
template<typename T> void adjustSmooth( T& value,
										const T& target,
										float speed,
										float dt )
{
	float step = speed * dt;
	if( step >= 1.0f ) value = target;
	else value = value * ( 1.0f - step ) + target * step;
}

// grenzt einen Wert ein
template<typename T> T clamp( const T& value,
							  const T& min,
							  const T& max )
{
	if( value < min ) return min;
	else if( value > max ) return max;
	else return value;
}

// berechnet Ort und Zeitpunkt eines mglichen Treffers sowie den Abschussvektor, ohne Gravitation
template<unsigned int DIM, typename T> bool computeHitNoGravity( const Vec<DIM, T>& shooterPosition,
																 const Vec<DIM, T>& shooterVelocity,
																 const Vec<DIM, T>& targetPosition,
																 const Vec<DIM, T>& targetVelocity,
																 T projectileSpeed,
																 Vec<DIM, T>* p_outHitPoint,
																 T* p_outHitTime,
																 Vec<DIM, T>* p_outShootVector )
{
	// Vorberechnungen
	const Vec<DIM, T> targetRelPosition( targetPosition - shooterPosition );
	const Vec<DIM, T> targetRelVelocity( targetVelocity - shooterVelocity );
	const T projectileSpeedSq = projectileSpeed * projectileSpeed;

	// a, b und c berechnen
	const T a = targetRelVelocity.lengthSq() - projectileSpeedSq;
	const T b = static_cast<T>( 2 ) * ( targetRelVelocity ^ targetRelPosition );
	const T c = targetRelPosition.lengthSq();

	// Diskriminante berechnen
	const T discriminant = b * b - static_cast<T>( 4 ) * a * c;

	// Wenn die Diskriminante negativ ist, kann das Ziel nicht getroffen werden.
	if( discriminant < static_cast<T>( 0 ) ) return false;

	// Zeitpunkt des Treffers berechnen
	const T root = static_cast<T>( sqrt( static_cast<double>( discriminant ) ) );
	const T denom = static_cast<T>( 2 ) * a;
	T t = ( -b - root ) / denom;
	if( t <= static_cast<T>( 0 ) )
	{
		t = ( -b + root ) / denom;
		if( t <= static_cast<T>( 0 ) )
		{
			// Es gibt keine positive Lsung - Treffer nicht mglich!
			return false;
		}
	}

	// Ort des Treffers berechnen
	const Vec<DIM, T> hitPoint( targetPosition + t * targetVelocity );

	if( p_outHitPoint ) *p_outHitPoint = hitPoint;
	if( p_outHitTime ) *p_outHitTime = t;

	if( p_outShootVector )
	{
		// Abschussvektor bestimmen
		*p_outShootVector = ( hitPoint - shooterPosition ) / t - shooterVelocity;
	}

	return true;
}

// berechnet Ort und Zeitpunkt eines mglichen Treffers sowie den Abschussvektor, mit Gravitation beim Geschoss
template<unsigned int DIM, typename T> bool computeHitWithGravity( const Vec<DIM, T>& shooterPosition,
																   const Vec<DIM, T>& shooterVelocity,
																   const Vec<DIM, T>& targetPosition,
																   const Vec<DIM, T>& targetVelocity,
																   T projectileSpeed,
																   const Vec<DIM, T>& gravity,
																   Vec<DIM, T>* p_outHitPoint,
																   T* p_outHitTime,
																   Vec<DIM, T>* p_outShootVector )
{
	if( gravity.lengthSq() == static_cast<T>( 0 ) )
	{
		return computeHitNoGravity( shooterPosition, shooterVelocity,
									targetPosition, targetVelocity,
									projectileSpeed,
									p_outHitPoint, p_outHitTime, p_outShootVector );
	}

	// Vorberechnungen
	const Vec<DIM, T> targetRelPosition( targetPosition - shooterPosition );
	const Vec<DIM, T> targetRelVelocity( targetVelocity - shooterVelocity );
	const T projectileSpeedSq = projectileSpeed * projectileSpeed;

	// a, b, c, d und e berechnen
	const double a = static_cast<double>( static_cast<T>( 0.25 ) * gravity.lengthSq() );
	const double b = static_cast<double>( -( targetRelVelocity ^ gravity) );
	const double c = static_cast<double>( targetRelVelocity.lengthSq() - ( targetRelPosition ^ gravity ) - projectileSpeedSq );
	const double d = static_cast<double>( static_cast<T>( 2 ) * ( targetRelVelocity ^ targetRelPosition ) );
	const double e = static_cast<double>( targetRelPosition.lengthSq() );

	// Gleichung lsen
	double r[4];
	int n = solveQuartic( a, b, c, d, e, r );
	if( !n )
	{
		// Kein Treffer mglich!
		return false;
	}

	// die kleinste positive Lsung suchen
	double smallestPositive = DBL_MAX;
	for( int i = 0; i < n; i++ )
	{
		if( r[i] >= 0.0 &&
			r[i] < smallestPositive )
		{
			smallestPositive = r[i];
		}
	}

	if( smallestPositive == DBL_MAX )
	{
		// Kein Treffer mglich!
		return false;
	}

	// Ort des Treffers berechnen
	const T t = static_cast<T>( smallestPositive );
	const Vec<DIM, T> hitPoint( targetPosition + t * targetVelocity );

	if( p_outHitPoint ) *p_outHitPoint = hitPoint;
	if( p_outHitTime ) *p_outHitTime = t;

	if( p_outShootVector )
	{
		// Abschussvektor bestimmen
		*p_outShootVector = ( hitPoint - shooterPosition ) / t - shooterVelocity - t * static_cast<T>( 0.5 ) * gravity;
	}

	return true;
}

// bestimmt den Schnittpunkt von zwei 2D-Geraden
template<typename T> bool intersect2DLines( const Vec<2, T>& line1A,
											const Vec<2, T>& line1B,
											const Vec<2, T>& line2A,
											const Vec<2, T>& line2B,
											Vec<2, T>* p_outIntersectionPoint )
{
	// Richtungen bestimmen
	const Vec<2, T> dir1( line1B - line1A );
	const Vec<2, T> dir2( line2B - line2A );

	// Nenner bestimmen
	const T denominator = dir1.x * dir2.y - dir1.y * dir2.x;
	if( fabsf( denominator ) < 1.0e-10 ) return false;

	if( p_outIntersectionPoint )
	{
		// Geradenabschnitt des Schnittpunkts auf Gerade 1 bestimmen
		const T s = ( dir2.perp() ^ ( line1A - line2A ) ) / denominator;
		*p_outIntersectionPoint = line1A + s * dir1;
	}

	return true;
}

// bestimmt, ob ein Punkt in einem Polygon liegt
template<typename T> bool pointInPolygon( const Vec<2, T>& point,
										  const std::vector<Vec<2, T> >& polygon )
{
	if( polygon.size() < 3 ) return false;

	std::vector<Vec<2, T> >::const_iterator oldIt = polygon.end() - 1;
	std::vector<Vec<2, T> >::const_iterator p1;
	std::vector<Vec<2, T> >::const_iterator p2;
	bool inside = false;

	for( std::vector<Vec<2, T> >::const_iterator it = polygon.begin();
		 it != polygon.end();
		 it++ )
	{
		if( it->x > oldIt->x )
		{
			p1 = oldIt;
			p2 = it;
		}
		else
		{
			p1 = it;
			p2 = oldIt;
		}

		if( ( it->x < point.x ) == ( point.x < oldIt->x ) &&
			( point.y - p1->y ) * ( p2->x - p1->x ) < ( p2->y - p1->y ) * ( point.x - p1->x ) )
		{
			inside = !inside;
		}

		oldIt = it;
	}

	return inside;
}

// bestimmt die Flche eines Polygons
template<typename T> T computePolygonArea( const std::vector<Vec<2, T> >& polygon )
{
	T area = static_cast<T>( 0 );

	for( size_t i = 0; i < polygon.size(); i++ )
	{
		const Vec<2, T>& v1 = polygon[i];
		const Vec<2, T>& v2 = polygon[( i + 1 ) % polygon.size()];
		area += v1.x * v2.y - v2.x * v1.y;
	}

	area /= static_cast<T>( 2 );

	return area;
}

// approximiert Sinus und Cosinus
template<typename T> void fastSinCos( T x,
									  T& outSin,
									  T& outCos )
{
#if 0
	// 14 Multiplikationen, 7 Additionen
	const T x1 = x;
	const T x2 = x1 * x1;
	const T x3 = x1 * x2;
	const T x4 = x2 * x2;
	const T x5 = x2 * x3;
	const T x6 = x3 * x3;
	const T x7 = x3 * x4;
	const T x8 = x4 * x4;

	outSin = static_cast<T>( -0.0001984126984 ) * x7 + static_cast<T>( 0.008333333333 ) * x5 + static_cast<T>( -0.1666666666 ) * x3 + x1;
	outCos = static_cast<T>( 2.480158730e-5 ) * x8 + static_cast<T>( -0.001388888888 ) * x6 + static_cast<T>( 0.04166666666 ) * x4 + static_cast<T>( -0.5 ) * x2 + 1;
#else
	// 9 Multiplikationen, 7 Additionen
	const T x2 = x * x;
	outSin = x * ( static_cast<T>( 1.0 ) + x2 * ( static_cast<T>( -0.1666666666 ) + x2 * ( static_cast<T>( 0.008333333333 ) + x2 * static_cast<T>( -0.0001984126984 ) ) ) );
	outCos = static_cast<T>( 1.0 ) + x2 * ( static_cast<T>( -0.5 ) + x2 * ( static_cast<T>( 0.04166666666 ) + x2 * ( static_cast<T>( -0.001388888888 ) + x2 * static_cast<T>( 2.480158730e-5 ) ) ) );
#endif
}

#ifdef min
#undef min
#endif

template<typename T> T min( const T& a,
						    const T& b )
{
	return a < b ? a : b;
}

#ifdef max
#undef max
#endif

template<typename T> T max( const T& a,
						    const T& b )
{
	return a > b ? a : b;
}

#endif