#ifndef VEC_H
#define VEC_H

// Helfer fr das Wurzelziehen
template<typename T> struct SqrtHelper
{
	static T sqrt( T x )
	{
		return static_cast<T>( ::sqrt( x ) );
	}
};

template<> struct SqrtHelper<float>
{
	static float sqrt( float x )
	{
		return sqrtf( x );
	}
};

template<> struct SqrtHelper<int>
{
	static int sqrt( int x )
	{
		return static_cast<int>( ::sqrt( static_cast<float>( x ) ) );
	}
};

// Helfer fr das schnelle Berechnen der inversen Wurzel
template<typename T> struct FastInvSqrtHelper
{
	static T fastInvSqrt( T x )
	{
		return static_cast<T>( 1 ) / SqrtHelper<T>::sqrt( x );
	}
};

template<> struct FastInvSqrtHelper<float>
{
	static float fastInvSqrt( float x )
	{
		float xHalf = 0.5f * x;
		int i = *reinterpret_cast<int*>( &x );
		i = 0x5F3759DF - ( i >> 1 );
		x = *reinterpret_cast<float*>( &i );
		x = x * ( 1.5f - xHalf * x * x );
		return x;
	}
};

// Helfer fr das schnelle Berechnen der Wurzel
template<typename T> struct FastSqrtHelper
{
	static T fastSqrt( T x )
	{
		return FastInvSqrtHelper<T>::fastInvSqrt( x ) * x;
	}
};

template<> struct FastInvSqrtHelper<double>
{
	static double fastInvSqrt( double x )
	{
		double xHalf = 0.5 * x;
		long long i = *reinterpret_cast<long long*>( &x );
		i = 0x5FE6EC85E7DE30DA - ( i >> 1 );
		x = *reinterpret_cast<double*>( &i );
		x = x * ( 1.5 - xHalf * x * x );
		return x;
	}
};

template<typename T, unsigned int DIM, typename VEC> struct VecBase
{
	T value[DIM];
};

template<typename T, typename VEC> struct VecBase<T, 2, VEC>
{
	union
	{
		T value[2];
		struct { T x, y; };
		struct { T u, v; };
	};

	// Senkrechte (nur sinnvoll fr DIM = 2)
	VEC perp() const
	{
		return VEC( -y, x );
	}

	// Vertauschung der Koordinaten (nur sinnvoll fr DIM = 2)
	VEC swapped() const
	{
		return VEC( y, x );
	}
};

template<typename T, typename VEC> struct VecBase<T, 3, VEC>
{
	union
	{
		T value[3];
		struct { T x, y, z; };
		struct { T r, g, b; };
	};

	// Kreuzprodukt (nur sinnvoll fr DIM = 3)
	VEC operator % ( const VEC& rhs ) const
	{
		return VEC( y * rhs.z - z * rhs.y,
					z * rhs.x - x * rhs.z,
					x * rhs.y - y * rhs.x );
	}
};

template<typename T, typename VEC> struct VecBase<T, 4, VEC>
{
	union
	{
		T value[4];
		struct { T x, y, z, w; };
		struct { T r, g, b, a; };
	};
};

// allgemeine Vektorklasse fr Vektoren ber beliebigen Typen mit beliebiger Dimension
template<unsigned int DIM, typename T> class Vec : public VecBase<T, DIM, Vec<DIM, T> >
{
public:
	Vec()
	{
	}

	Vec( const Vec& rhs )
	{
		for( int i = 0; i < DIM; i++ )
		{
			value[i] = rhs.value[i];
		}
	}

	Vec( T value )
	{
		for( int i = 0; i < DIM; i++ )
		{
			this->value[i] = value;
		}
	}

	Vec( const T* p_values )
	{
		for( int i = 0; i < DIM; i++ )
		{
			this->value[i] = p_values[i];
		}
	}

	// Initialisierung mit 2 Parametern
	Vec( T x, T y )
	{
		value[0] = x;
		value[1] = y;
	}

	// Initialisierung mit 3 Parametern
	Vec( T x, T y, T z )
	{
		value[0] = x;
		value[1] = y;
		value[2] = z;
	}

	// Initialisierung mit 4 Parametern
	Vec( T x, T y, T z, T w )
	{
		value[0] = x;
		value[1] = y;
		value[2] = z;
		value[3] = w;
	}

	template<unsigned int OTHER_DIM> Vec( const Vec<OTHER_DIM, T>& rhs )
	{
		int minDim = min( DIM, OTHER_DIM );
		for( int i = 0; i < minDim; i++ ) value[i] = rhs.value[i];
		for( int i = minDim; i < DIM; i++ ) value[i] = static_cast<T>( 0 );
	}

	~Vec()
	{
	}

	T length() const
	{
		return SqrtHelper<T>::sqrt( lengthSq() );
	}

	T lengthSq() const
	{
		T r = static_cast<T>( 0 );
		for( int i = 0; i < DIM; i++ )
		{
			r += value[i] * value[i];
		}

		return r;
	}

	T lengthFast() const
	{
		return FastSqrtHelper<T>::fastSqrt( lengthSq() );
	}

	Vec normalized() const
	{
		const T myLength = length();
		return myLength ? *this / myLength : *this;
	}

	Vec& normalize()
	{
		return *this = normalized();
	}

	Vec normalizedFast() const
	{
		return *this * FastInvSqrtHelper<T>::fastInvSqrt( lengthSq() );
	}

	Vec& normalizeFast()
	{
		return *this = normalizedFast();
	}

	bool isZero() const
	{
		T zero = static_cast<T>( 0 );
		for( int i = 0; i < DIM; i++ )
		{
			if( value[i] != zero ) return false;
		}

		return true;
	}

	const T& operator [] ( unsigned int index ) const
	{
		return value[index];
	}

	T& operator [] ( unsigned int index )
	{
		return value[index];
	}

	Vec operator + ( const Vec& rhs ) const
	{
		Vec r;
		for(int i = 0; i < DIM; i++)
		{
			r.value[i] = value[i] + rhs.value[i];
		}

		return r;
	}

	Vec operator - ( const Vec& rhs ) const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = value[i] - rhs.value[i];
		}

		return r;
	}

	Vec operator - () const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = -value[i];
		}

		return r;
	}

	// komponentenweise Multiplikation
	Vec operator * ( const Vec& rhs ) const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = value[i] * rhs.value[i];
		}

		return r;
	}

	Vec operator * ( T rhs ) const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = value[i] * rhs;
		}

		return r;
	}

	// komponentenweise Division
	Vec operator / ( const Vec& rhs ) const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = value[i] / rhs.value[i];
		}

		return r;
	}

	Vec operator / ( T rhs ) const
	{
		Vec r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = value[i] / rhs;
		}

		return r;
	}

	// Skalarprodukt
	T operator ^ ( const Vec& rhs ) const
	{
		T r = static_cast<T>( 0 );
		for( int i = 0; i < DIM; i++ )
		{
			r += value[i] * rhs.value[i];
		}

		return r;
	}

	Vec& operator = ( const Vec& rhs )
	{
		for( int i = 0; i < DIM; i++ )
		{
			value[i] = rhs.value[i];
		}

		return *this;
	}

	Vec& operator += ( const Vec& rhs )
	{
		return *this = *this + rhs;
	}

	Vec& operator -= ( const Vec& rhs )
	{
		return *this = *this - rhs;
	}

	Vec& operator *= ( const Vec& rhs )
	{
		return *this = *this * rhs;
	}

	Vec& operator *= (T rhs)
	{
		return *this = *this * rhs;
	}

	Vec& operator /= ( const Vec& rhs )
	{
		return *this = *this / rhs;
	}

	Vec& operator /= ( T rhs )
	{
		return *this = *this / rhs;
	}

	bool operator == ( const Vec& rhs ) const
	{
		for( int i = 0; i < DIM; i++ )
		{
			if( value[i] != rhs.value[i] ) return false;
		}

		return true;
	}

	bool operator != ( const Vec& rhs ) const
	{
		return !( *this == rhs );
	}

	// Konvertierung in anderes Vektorformat (gleiche Dimension)
	template<typename U> operator Vec<DIM, U> () const
	{
		Vec<DIM, U> r;
		for( int i = 0; i < DIM; i++ )
		{
			r.value[i] = static_cast<U>( value[i] );
		}

		return r;
	}

	operator const T* () const
	{
		return value;
	}

	operator T* ()
	{
		return value;
	}
};

// Multipliktion mit Skalar von links
template<unsigned int DIM, typename T> inline Vec<DIM, T> operator * ( T lhs,
																	   const Vec<DIM, T>& rhs )
{
	return rhs * lhs;
}

// in Stream schreiben
template<unsigned int DIM, typename T> inline std::ostream& operator << ( std::ostream& lhs,
																		  const Vec<DIM, T>& rhs )
{
	for( int i = 0; i < DIM; i++ )
	{
		lhs << rhs.value[i];
		if( i != DIM - 1 ) lhs << " ";
	}

	return lhs;
}

// aus Stream lesen
template<unsigned int DIM, typename T> inline std::istream& operator >> ( std::istream& lhs,
																		  Vec<DIM, T>& rhs )
{
	for( int i = 0; i < DIM; i++ )
	{
		lhs >> rhs.value[i];
	}

	return lhs;
}

// Standard-Vektortypen
typedef Vec<2, int> Vec2i;
typedef Vec<3, int> Vec3i;
typedef Vec<4, int> Vec4i;
typedef Vec<2, float> Vec2f;
typedef Vec<3, float> Vec3f;
typedef Vec<4, float> Vec4f;
typedef Vec<2, double> Vec2d;
typedef Vec<3, double> Vec3d;
typedef Vec<4, double> Vec4d;

#endif