#ifndef VIDEO_H
#define VIDEO_H

#include "Component.h"
#include "Texture.h"

struct SDL_Surface;

// Grafikkomponente
class ENGINE_API Video : public Component
{
	friend class Engine;

public:
	// leert den Frame-Buffer
	void clear( const Vec3f& color = Vec3f( 0.0f, 0.0f, 0.0f ) );

	// setzt die Projektionsmatrix
	void setupProjection( const Vec2f& topLeft, const Vec2f& bottomRight );

	// setzt die Kamera
	void setupCamera( const Vec2f& position, float angle, float scale );

	// setzt die Kamera zurck
	void resetCamera();

	// liefert die Kameraparameter
	void getCamera( Vec2f* p_outPosition, float* p_outAngle, float* p_outScale ) const;

	// liefert die Kameramatrix
	const Mat44f& getCameraMatrix() const;

	// stellt Motion-Blurring ein oder aus
	void setupMotionBlurring( bool enable, float decay = 0.05f );

	// fhrt das Motion-Blurring jetzt aus, und nicht am Ende des Frames
	void doMotionBlurring();

	// liefert die Auflsung
	const Vec2i& getResolution() const;

	// liefert die aktive Textur
	TextureHandle getTexture() const;

	// setzt die aktive Textur
	void setTexture( TextureHandle texture, bool force = false );

	// legt den Alpha-Blending-Modus fest
	void setupAlphaBlending( bool enable, unsigned int source = 0, unsigned int destination = 0 );

	// rendert ein Sprite
	void drawSprite( const Vec2i& positionOnTexture, const Vec2i& sizeOnTexture, const Vec2f& sizeOnScreen = Vec2f( 1.0f, 1.0f ) );

	// zeichnet ein Rechteck, das den ganzen Bildschirm ausfllt
	void drawFullScreenQuad( const Vec4f& color = Vec4f( 1.0f, 1.0f, 1.0f, 1.0f ), const Vec2i& topLeftUV = Vec2i( 0, 0 ), const Vec2i& bottomRightUV = Vec2i( 0, 0 ) );

	// testet, ob ein Kreis sichtbar ist
	bool isCircleVisible( const Vec2f& center, float radius ) const;

	// testet, ob ein Kreis sichtbar ist, mit Transformation
	bool isCircleVisible( const Vec2f& center, float radius, const Mat33f& transformation ) const;

	// testet, ob ein Rechteck sichtbar ist
	bool isRectVisible( const Vec2f& boundMin, const Vec2f& boundMax ) const;

	// testet, ob ein Rechteck sichtbar ist, mit Transformation
	bool isRectVisible( const Vec2f& boundMin, const Vec2f& boundMax, const Mat33f& transformation ) const;

	// aktualisiert das View-Frustum
	void updateViewFrustum();

	// liefert die Dauer des letzten Frames
	float getLastFrameDuration() const;

private:
	Video( Engine& engine );
	~Video();

	void onEngineShutDown();
	void onFrameBegin();
	void onFrameEnd();

	// primre Oberflche
	SDL_Surface* p_screen;

	// Auflsung
	Vec2i resolution;

	// Zeitpunkt des letzten Frame-Endes
	float lastFrameEnd;

	// Dauer des letzten Frames
	float lastFrameDuration;

	// aktive Textur
	TextureHandle currentTexture;

	// Projektionsmatrix
	Mat44f projectionMatrix;

	// Kameraposition
	Vec2f cameraPosition;

	// Kamerarotation
	float cameraAngle;

	// Kameraskalierung
	float cameraScale;

	// Kameramatrix
	Mat44f cameraMatrix;

	// Ebenen des View-Frustums
	Vec3f frustumPlane[4];

	// Motion-Blurring
	bool mbEnabled;
	float mbDecay;
	bool mbFrameDone;
	bool mbIsCurrent;
	TextureHandle mbTexture;
};

#endif