#include <PCH.h>
#include <Engine.h>
#include <Audio.h>
#include <SoundManager.h>
#include <SoundInstance.h>
#include <SoundGroup.h>
#include <Listener.h>
#include <fmod.hpp>
#include <fmod_errors.h>

Audio::Audio( Engine& engine ) : Component( engine, "Audio" ),
								 p_system( 0 ),
								 p_listener( 0 ),
								 p_masterSoundGroup( 0 ),
								 p_musicSoundGroup( 0 ),
								 p_musicInst( 0 )
{
	FMOD_RESULT r = FMOD::System_Create( &p_system );
	if( r != FMOD_OK )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << "FMOD-System konnte nicht erzeugt werden: " << FMOD_ErrorString( r );
		engine.getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	noSound = engine.getOptions().get<bool>( "Audio.NoSound", false );
	if( noSound ) p_system->setOutput( FMOD_OUTPUTTYPE_NOSOUND );

	hardware = engine.getOptions().get<bool>( "Audio.UseHardware", false );

	software3DQuality = engine.getOptions().get<unsigned int>( "Audio.Software3DQuality", 1 );

	FMOD_INITFLAGS flags = 0;

	if( !noSound )
	{
		// Unter Windows kann der Sound auf DirectSound laufen, andernfalls FMOD_INIT_NORMAL fr Linux.
#ifdef WIN32
		if( hardware )
		{
			// Bei Hardware wird automatisch die hchste Qualitt genommen.
		}
		else
		{
			// Qualittsstufe
			switch( software3DQuality )
			{
			case 0:
				flags |= FMOD_INIT_DSOUND_HRTFNONE;
				break;
			case 1:
				flags |= FMOD_INIT_DSOUND_HRTFLIGHT;
				break;
			case 2:
				flags |= FMOD_INIT_DSOUND_HRTFFULL;
				break;
			default:
				// Unzulssiger Wert!
				setError( 2 );
				engine.getLogger().error( "Unzulssiger Wert fr Audio.Software3DQuality", LOGGER_AUTO );
				return;
			}
		}
#else
		flags |= FMOD_INIT_NORMAL;
#endif
	}

	r = p_system->init( 256, flags, 0 );
	if( r != FMOD_OK )
	{
		// Fehler!
		setError( 3 );
		std::ostringstream msg;
		msg << "FMOD-System konnte nicht initialisiert werden: " << FMOD_ErrorString( r );
		engine.getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	p_system->set3DSettings( 1.0f, 1.0f, 1.0f );

	// Hrer erzeugen
	p_listener = new Listener( *this );

	// Master-Channel-Group abfragen und daraus die Master-Soundgruppe erzeugen
	FMOD::ChannelGroup* p_masterChannelGroup;
	p_system->getMasterChannelGroup( &p_masterChannelGroup );
	p_masterSoundGroup = new SoundGroup( *this, p_masterChannelGroup );
	soundGroups[p_masterSoundGroup->getName()] = p_masterSoundGroup;

	engine.connectAppGotFocus( this, &Audio::onAppGotFocus );
	engine.connectAppLostFocus( this, &Audio::onAppLostFocus );
	engine.connectUpdate( this, &Audio::onUpdate );
	engine.connectFrameEnd( this, &Audio::onFrameEnd );

	engine.getLogger().fine( "Audio-Komponente initialisiert", LOGGER_AUTO );
}

Audio::~Audio()
{
	// alle Soundgruppen lschen, zuletzt die Master-Soundgruppe
	for( std::map<std::string, SoundGroup*>::const_iterator it = soundGroups.begin();
		 it != soundGroups.end();
		 it++)
	{
		if( it->second != p_masterSoundGroup ) delete it->second;
	}

	delete p_masterSoundGroup;

	// Hrer lschen
	delete p_listener;

	// FMOD herunterfahren
	if( p_system )
	{
		p_system->close();
		p_system->release();
	}

	getEngine().getLogger().fine( "Audio-Komponente heruntergefahren", LOGGER_AUTO );
}

void Audio::onAppGotFocus()
{
	// Master-Soundgruppe weiter spielen lassen
	p_masterSoundGroup->setPaused( false );
}

void Audio::onAppLostFocus()
{
	// Master-Soundgruppe pausieren
	p_masterSoundGroup->setPaused( true );
}

void Audio::onUpdate()
{
	// Hrer aktualisieren
	p_listener->update();
}

void Audio::onFrameEnd()
{
	// alle Soundgruppen aktualisieren
	for( std::map<std::string, SoundGroup*>::const_iterator it = soundGroups.begin();
		 it != soundGroups.end();
		 it++ )
	{
		it->second->update();
	}

	p_system->update();
}

Listener* Audio::getListener() const
{
	return p_listener;
}

SoundGroup* Audio::getMasterSoundGroup() const
{
	return p_masterSoundGroup;
}

SoundGroup* Audio::createSoundGroup( const std::string& name,
									 SoundGroup* p_parent )
{
	// Doppelte Namen sind verboten!
	if( getSoundGroupByName( name ) ) return 0;

	// Master-Gruppe als Vatergruppe, wenn null angegeben
	if( !p_parent ) p_parent = p_masterSoundGroup;

	SoundGroup* p_soundGroup = new SoundGroup( *this, name, p_parent );
	if( p_soundGroup->getError() )
	{
		// Fehler!
		delete p_soundGroup;
		return 0;
	}

	// alles OK - Gruppe eintragen
	soundGroups[name] = p_soundGroup;

	return p_soundGroup;
}

SoundGroup* Audio::getSoundGroupByName( const std::string& name ) const
{
	std::map<std::string, SoundGroup*>::const_iterator it = soundGroups.find( name );
	return it == soundGroups.end() ? 0 : it->second;
}

int Audio::destroySoundGroup( const std::string& name )
{
	SoundGroup* p_soundGroup = getSoundGroupByName( name );
	if( !p_soundGroup ) return 1;

	// Die Master-Gruppe kann nicht gelscht werden!
	if( p_soundGroup == p_masterSoundGroup ) return 2;

	// lschen und aus der Map entfernen
	delete p_soundGroup;
	soundGroups.erase( name );

	return 0;
}

SoundInstance* Audio::playSound( const std::string& filename,
								 const std::string& soundGroup,
								 const std::string& type,
								 const std::string& mode,
								 float volume,
								 float pitch )
{
	ParameterMap options;
	options.set<std::string>( "Type", type );
	options.set<std::string>( "Mode", mode );
	SoundHandle sound = getEngine().getSoundManager().get( filename, false, options );
	if( !sound ) return 0;

	sound->setDefaultSoundGroupName( soundGroup );

	SoundInstance* p_soundInst = sound->createInstance( true );
	if( !p_soundInst ) return 0;

	p_soundInst->setVolume( volume );
	p_soundInst->setPitch( pitch );
	p_soundInst->play();

	return p_soundInst;
}

SoundInstance* Audio::playMusic( const std::string& filename,
								 bool restartIfAlreadyPlaying,
								 float fadeInSpeed,
								 float fadeOutSpeed )
{
	if( p_musicInst )
	{
		if( !restartIfAlreadyPlaying )
		{
			if( p_musicInst->getSound()->getFilename() == filename )
			{
				// Diese Musik wird bereits abgespielt.
				return p_musicInst;
			}
		}

		// alte Musik ausblenden und dann stoppen
		stopMusic( fadeOutSpeed );
	}

	if( filename.empty() ) return 0;

	ParameterMap options;
	options.set<std::string>( "Type", "Stream" );
	options.set<std::string>( "Mode", "2D" );
	SoundHandle music = getEngine().getSoundManager().get( filename, true, options );
	if( !music ) return 0;

	p_musicInst = music->createInstance();
	if( !p_musicInst ) return 0;

	// Soundgruppe setzen
	p_musicInst->setSoundGroup( p_musicSoundGroup );

	// neue Musik einblenden
	p_musicInst->setVolume( 0.0f );
	p_musicInst->setVolumeTarget( 1.0f, fadeInSpeed );
	p_musicInst->play( true );

	return p_musicInst;
}

int Audio::pauseMusic()
{
	if( !p_musicInst ) return 1;
	
	p_musicInst->setPaused( true );

	return 0;
}

int Audio::resumeMusic()
{
	if( !p_musicInst ) return 1;
	
	p_musicInst->setPaused( false );

	return 0;
}

int Audio::stopMusic( float fadeOutSpeed )
{
	if( !p_musicInst ) return 1;

	p_musicInst->setVolumeTarget( 0.0f, fadeOutSpeed );
	p_musicInst->setStopCountDown( -1.0f );
	p_musicInst = 0;

	return 0;
}

int Audio::setMusicSoundGroup( SoundGroup* p_soundGroup )
{
	p_musicSoundGroup = p_soundGroup;
	if( p_musicInst ) p_musicInst->setSoundGroup( p_soundGroup );

	return 0;
}

SoundGroup* Audio::getMusicSoundGroup() const
{
	return p_musicSoundGroup;
}

bool Audio::getNoSoundMode() const
{
	return noSound;
}

bool Audio::getHardwareMode() const
{
	return hardware;
}

unsigned int Audio::getSoftware3DQuality() const
{
	return software3DQuality;
}