#include <PCH.h>
#include <Engine.h>
#include <Logger.h>
#include <ConsoleLogHandler.h>
#include <Video.h>
#include <Audio.h>
#include <TextureManager.h>
#include <Font.h>
#include <SoundManager.h>
#include <GameState.h>
#include <SDL.h>
#include <Input.h>
#include <Profiler.h>

Engine::Engine( const ParameterMap& options ) : mode( EM_NORMAL ),
												p_video( 0 ),
												p_audio( 0 ),
												p_input( 0 ),
												p_textureManager( 0 ),
												p_fontManager( 0 ),
												p_soundManager( 0 ),
												options( options ),
												time( 0 ),
												newStackSize( 0 )
{
	// Logger initialisieren
	initLogger();

	// Tick-Dauer abfragen
	tickDuration = options.get<unsigned int>( "Engine.TickDuration", 20 );

	// SDL initialisieren (nur Timer - Grafik wird von der Video-Klasse initialisiert)
	int r = SDL_Init( SDL_INIT_TIMER );
	if( r )
	{
		// Fehler!
		setError( 1 );
		logger.error("SDL-Initialisierung fehlgeschlagen", LOGGER_AUTO );
		return;
	}

	// Video-Komponente initialisieren (sie registriert sich selbst)
	p_video = new Video( *this );
	r = p_video->getError();
	if( r )
	{
		// Fehler!
		setError( 2 );
		logger.error( "Initialisierung der Video-Komponente fehlgeschlagen", LOGGER_AUTO );
		return;
	}

	// Audio-Komponente initialisieren (sie registriert sich auch selber)
	p_audio = new Audio( *this );
	r = p_audio->getError();
	if( r )
	{
		// Fehler!
		setError( 3 );
		logger.error( "Initialisierung der Audio-Komponente fehlgeschlagen", LOGGER_AUTO );
		return;
	}

	// Input-Komponente initialisieren
	p_input = new Input( *this );
	r = p_input->getError();
	assert( !r );

	// Texturmanager erzeugen
	p_textureManager = new TextureManager( *this, "TextureManager" );
	r = p_textureManager->getError();
	assert( !r );

	// Font-Manager erzeugen
	p_fontManager = new FontManager( *this, "FontManager" );
	r = p_fontManager->getError();
	assert( !r );

	// Sound-Manager erzeugen
	p_soundManager = new SoundManager( *this, "SoundManager" );
	r = p_soundManager->getError();
	assert( !r );

	// UNICODE und Tastendruckwiederholung aktivieren
	SDL_EnableKeyRepeat( 160, 80 );
	SDL_EnableUNICODE( 1 );

	logger.fine( "Engine initalisiert", LOGGER_AUTO );
}

Engine::~Engine()
{
	// Nachricht an alle Interessenten
	EMIT_SIGNAL0( EngineShutDown );

	// Ressourcen-Manager zerstren
	delete p_fontManager;
	delete p_textureManager;
	delete p_soundManager;

	// eigene Komponenten lschen
	delete p_video;
	delete p_audio;
	delete p_input;

	// SDL herunterfahren
	SDL_Quit();

	logger.fine( "Engine heruntergefahren", LOGGER_AUTO );

	// Logger herunterfahren
	cleanUpLogger();
}

int Engine::registerComponent( const std::string& name,
							   Component* p_component )
{
	if( getComponentByName( name ) )
	{
		std::ostringstream msg;
		msg << "Komponente " << name << " ist bereits registriert";
		logger.error( msg.str(), LOGGER_AUTO );
		return 1;
	}

	componentRegistry[name] = p_component;

	std::ostringstream msg;
	msg << "Komponente " << name << " wurde registriert";
	logger.fine( msg.str(), LOGGER_AUTO );

	return 0;
}

int Engine::unregisterComponent( const std::string& name )
{
	if( !getComponentByName( name ) ) return 1;
	componentRegistry.erase( name );

	std::ostringstream msg;
	msg << "Komponente " << name << " wurde entfernt";
	logger.fine( msg.str(), LOGGER_AUTO );

	return 0;
}

Component* Engine::getComponentByName( const std::string& name ) const
{
	// Komponente mit diesem Namen suchen - falls nicht vorhanden, Nullzeiger liefern
	std::map<std::string, Component*>::const_iterator it = componentRegistry.find( name );
	return it == componentRegistry.end() ? 0 : it->second;
}

void Engine::setMode( EngineMode mode )
{
	this->mode = mode;
}

EngineMode Engine::getMode() const
{
	return mode;
}

Logger& Engine::getLogger()
{
	return logger;
}

Video& Engine::getVideo() const
{
	assert( p_video );
	return *p_video;
}

Audio& Engine::getAudio() const
{
	assert( p_audio );
	return *p_audio;
}

Input& Engine::getInput() const
{
	return *p_input;
}

TextureManager& Engine::getTextureManager() const
{
	assert( p_textureManager );
	return *p_textureManager;
}

FontManager& Engine::getFontManager() const
{
	assert( p_fontManager );
	return *p_fontManager;
}

SoundManager& Engine::getSoundManager() const
{
	assert( p_soundManager );
	return *p_soundManager;
}

int Engine::registerGameState( const std::string& name,
							   GameState* p_state )
{
	if( getGameStateByName( name ) )
	{
		std::ostringstream msg;
		msg << "Spielzustand " << name << " ist bereits registriert";
		logger.error( msg.str(), LOGGER_AUTO );
		return 1;
	}

	stateRegistry[name] = p_state;

	std::ostringstream msg;
	msg << "Spielzustand " << name << " wurde registriert";
	logger.fine( msg.str(), LOGGER_AUTO );

	return 0;
}

int Engine::unregisterGameState( const std::string& name )
{
	if( !getGameStateByName( name ) ) return 1;
	stateRegistry.erase( name );

	std::ostringstream msg;
	msg << "Spielzustand " << name << " wurde entfernt";
	logger.fine( msg.str(), LOGGER_AUTO );

	return 0;
}

GameState* Engine::getGameStateByName( const std::string &name ) const
{
	// Spielzustand mit diesem Namen suchen - falls nicht vorhanden, Nullzeiger liefern
	std::map<std::string, GameState*>::const_iterator it = stateRegistry.find( name );
	return it == stateRegistry.end() ? 0 : it->second;
}

GameState* Engine::getCurrentGameState() const
{
	// obersten Spielzustand vom Stack holen - falls leer, Nullzeiger liefern
	if( currentStates.empty() ) return 0;
	return currentStates.top();
}

int Engine::pushGameState( const std::string& name,
						   const ParameterMap& params )
{
	GameState* p_state = getGameStateByName( name );
	if( !p_state ) return 1;

	// vormerken
	StateOperation op;
	op.p_state = p_state;
	op.push = true;
	op.params = params;
	pendingStateOperations.push( op );

	newStackSize++;

	return 0;
}

int Engine::popGameState( bool deactivateCurrent,
						  bool activateBelow )
{
	// verhindern, dass der Stack unterluft
	if( !newStackSize ) return 1;

	// vormerken
	StateOperation op;
	op.p_state = 0;
	op.push = false;
	op.deactivateCurrent = deactivateCurrent;
	op.activateBelow = activateBelow;
	pendingStateOperations.push( op );

	newStackSize--;

	return 0;
}

int Engine::replaceGameStates( const std::string& name,
							   const ParameterMap& params )
{
	GameState* p_state = getGameStateByName( name );
	if( !p_state ) return 1;

	// zuerst alle Zustnde verlassen, ...
	int n = newStackSize;
	for( int i = 0; i < n; i++ )
	{
		popGameState( i == 0, false );
	}

	// ... dann den Neuen betreten
	return pushGameState( name, params );
}

int Engine::run() 
{
	// Zeitkonto
	Uint32 timeToProcess = 0;
	Uint32 timeProcessed = ~0;

	int r = 0;
	bool done = false;
	bool focus = true;
	Profiler& profiler = Profiler::inst();

	do
	{
		// Startzeitpunkt festhalten
		Uint32 start = SDL_GetTicks();

		GameState* p_currentState = getCurrentGameState();

		if( focus && timeProcessed && mode == EM_NORMAL )
		{
			// Nachricht an alle Interessenten
			EMIT_SIGNAL0( FrameBegin );
			EMIT_SIGNAL0( PreRender );

			// Spielzustand rendern
			if( p_currentState ) p_currentState->render();

			// Nachricht an alle Interessenten
			EMIT_SIGNAL0( PostRender );
			EMIT_SIGNAL0( PostRenderFinal );
		}

		// SDL-Ereignisse verarbeiten
		SDL_Event e;
		while( SDL_PollEvent( &e ) )
		{
			switch( e.type )
			{
			case SDL_QUIT:
				done = true;
				r = 1;
				break;
			case SDL_ACTIVEEVENT:
				if( e.active.state & SDL_APPACTIVE ||
					e.active.state & SDL_APPINPUTFOCUS )
				{
					std::ostringstream msg;
					msg << "ActiveEvent: (Gain=" << static_cast<int>( e.active.gain ) << ", State=" << static_cast<int>( e.active.state ) << ")";
					logger.finest( msg.str(), LOGGER_AUTO );

					if( e.active.gain )
					{
						if( !focus )
						{
							focus = true;
							EMIT_SIGNAL0( AppGotFocus );
						}
					}
					else
					{
						if( focus )
						{
							focus = false;
							EMIT_SIGNAL0( AppLostFocus );
						}
					}
				}
				break;
			}

			// Nachricht an alle Interessenten
			EMIT_SIGNAL1( SDLEvent, e );
		}

		if( !focus && mode == EM_NORMAL )
		{
			logger.finest( "Schlafe fr 100 ms", LOGGER_AUTO );
			SDL_Delay( 100 );

			// Schleifendurchlauf beenden
			continue;
		}

		if( mode == EM_PROCESS_ONLY )
		{
			// in jedem Durchlauf ein Update durchfhren
			timeToProcess = tickDuration;
		}

		// bewegen, solange noch Zeit zu verarbeiten ist
		timeProcessed = 0;
		while( timeToProcess >= tickDuration )
		{
			// Spielzustand aktualisieren
			if( p_currentState ) p_currentState->update();

			// Zustandsbergnge durchfhren
			performStateTransitions();
			p_currentState = getCurrentGameState();
			if( !p_currentState )
			{
				done = true;
				break;
			}

			// Nachricht an alle Interessenten
			EMIT_SIGNAL0( Update );

			timeToProcess -= tickDuration;
			timeProcessed += tickDuration;
			time += tickDuration;
		}

		if( mode == EM_NORMAL )
		{
			if( timeProcessed )
			{
				// Nachricht an alle Interessenten
				EMIT_SIGNAL0( FrameEnd );
			}

			Uint32 end = SDL_GetTicks();
			Uint32 dt = end - start;
			if( timeToProcess < tickDuration )
			{
				// keine Rechenzeit verschwenden
				SDL_Delay( 1 );
				end = SDL_GetTicks();
				dt = end - start;
			}

			timeToProcess += dt;
			timeToProcess = min( timeToProcess, 10 * tickDuration );
		}

		Profiler::inst().getStatistics();

	} while( !done );

	// alle noch aktiven Spielzustnde verlassen
	int n = newStackSize;
	for( int i = 0; i < n; i++ ) popGameState( i == 0, false );
	performStateTransitions();

	return r;
}

void Engine::quit()
{
	SDL_Event e;
	e.type = SDL_QUIT;
	SDL_PushEvent( &e );
}

const ParameterMap& Engine::getOptions() const
{
	return options;
}

unsigned int Engine::getTickDuration() const
{
	return tickDuration;
}

float Engine::getTickDurationInSeconds() const
{
	return static_cast<float>( tickDuration ) / 1000.0f;
}

unsigned int Engine::getTime() const
{
	return time;
}

float Engine::getTimeInSeconds() const
{
	return static_cast<float>( time ) / 1000.0f;
}

unsigned long long Engine::getHighResTime() const
{
#ifdef WIN32
	LARGE_INTEGER result;
	if( QueryPerformanceCounter( &result ) ) return result.QuadPart;
	else return SDL_GetTicks();
#else
	return SDL_GetTicks();
#endif
}

unsigned long long Engine::getHighResTimerResolution() const
{
#ifdef WIN32
	LARGE_INTEGER result;
	if( QueryPerformanceFrequency( &result ) ) return result.QuadPart;
	else return 1000;
#else
	return 1000;
#endif
}

void Engine::performStateTransitions()
{
	while( !pendingStateOperations.empty() )
	{
		// Operation holen
		const StateOperation& op = pendingStateOperations.front();

		// aktuellen Spielzustand holen
		GameState* p_currentState = getCurrentGameState();

		if( op.push )
		{
			// Ein neuer Zustand soll auf den Stack gelegt werden.
			// Falls es einen alten Zustand gibt, wird dieser zunchst deaktiviert.
			if( p_currentState )
			{
				p_currentState->onDeactivate();
			}

			// Nun wird der neue Zustand betreten, aktiviert und auf den Stack gelegt.
			assert( op.p_state );
			op.p_state->onEnter( op.params );
			op.p_state->onActivate();
			currentStates.push( op.p_state );
		}
		else
		{
			// Der oberste Zustand soll vom Stack genommen werden.
			// Dazu wird er zunchst deaktiviert (es sei denn es ist anders gewnscht), dann verlassen und vom Stack geholt.
			assert( p_currentState );
			if( op.deactivateCurrent ) p_currentState->onDeactivate();
			p_currentState->onLeave();
			currentStates.pop();

			if( op.activateBelow )
			{
				// Falls darunter ein anderer Zustand liegt, wird dieser wieder aktiviert.
				p_currentState = getCurrentGameState();
				if( p_currentState ) p_currentState->onActivate();
			}
		}

		pendingStateOperations.pop();
	}

	// sicherstellen, dass der Stack nun die korrekte Gre hat
	assert( static_cast<int>( currentStates.size() ) == newStackSize );
}

void Engine::initLogger()
{
	// Log-Level setzen: Wenn nichts anderes angegeben ist, loggen wir nur die interessanten Sachen: INFO, WARNING, ERROR
	LogLevel logLevel = static_cast<LogLevel>( options.get<int>( "Engine.LogLevel", LL_INFO ) );
	logger.setLogLevel( logLevel );

	// Standard-Log-Handler einfgen
	p_consoleLogHandler = new ConsoleLogHandler( logger );
}

void Engine::cleanUpLogger()
{
	delete p_consoleLogHandler;
}
