#include <PCH.h>
#include <Font.h>
#include <Engine.h>
#include <Video.h>
#include <TextureManager.h>
#include <Utils.h>
#include <tinyxml.h>
#include <SDL_OpenGL.h>

Font::Font( ResourceManager& manager,
		    const std::string& filename,
			bool unique,
			const ParameterMap& options) : Resource( manager, filename, unique, options ),
										   characterInfo(),
										   maxCacheSize( 16 )
{
	Logger& logger = manager.getEngine().getLogger();

	// XML-Dokument laden
	TiXmlDocument doc( filename.c_str() );
	if( !doc.LoadFile() )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << "Datei " << filename << " konnte nicht geladen werden";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	TiXmlElement* p_font = doc.FirstChildElement( "font" );
	if( !p_font )
	{
		// Fehler!
		setError( 2 );
		std::ostringstream msg;
		msg << filename << ": <font>-Element fehlt";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// allgemeine Informationen laden
	TiXmlElement* p_common = p_font->FirstChildElement( "common" );
	if( p_common )
	{
		int pages;
		if( !p_common->Attribute( "lineHeight", &lineHeight ) ||
			!p_common->Attribute( "base", &base ) ||
			!p_common->Attribute( "pages", &pages ) )
		{
			// Fehler!
			setError( 3 );
			std::ostringstream msg;
			msg << filename << ": Attribute im <common>-Tag fehlen";
			logger.error( msg.str(), LOGGER_AUTO );
			return;
		}

		// Es werden nur Schriften mit einer einzigen Textur untersttzt.
		if( pages != 1 )
		{
			std::ostringstream msg;
			msg << filename << " enthlt mehr als eine Seite, nur die erste wird bercksichtigt";
			logger.warning( msg.str(), LOGGER_AUTO );
		}
	}
	else
	{
		// Fehler!
		setError( 4 );
		std::ostringstream msg;
		msg << filename << ": <common>-Tag fehlt";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Seiteninformationen laden
	TiXmlElement* p_pages = p_font->FirstChildElement( "pages" );
	if( p_pages )
	{
		// erste Seite suchen
		TiXmlElement* p_page = p_pages->FirstChildElement( "page" );
		while( p_page )
		{
			int id = -1;
			p_page->QueryIntAttribute( "id", &id );

			if( id == 0 )
			{
				// Erste Seite gefunden!
				const char* p_filename = p_page->Attribute( "file" );
				if( p_filename )
				{
					// Textur fr Schriftart laden
					std::string textureFilename = getPathDirectory( filename ) + "\\" + p_filename;
					ParameterMap texOptions;
					texOptions.set<bool>( "UseForBitmapFont", true );
					texOptions.set<bool>( "UseMIPMapping", options.get<bool>( "UseMIPMapping", false ) );
					texture = manager.getEngine().getTextureManager().get( textureFilename, false, texOptions );
					if( !texture )
					{
						// Fehler!
						setError( 5 );
						std::ostringstream msg;
						msg << filename << ": Textur fr diese Schriftart konnte nicht geladen werden";
						logger.error( msg.str(), LOGGER_AUTO );
						return;
					}
				}
				else
				{
					// Fehler!
					setError( 6 );
					std::ostringstream msg;
					msg << filename << ": Attribut im <page>-Tag fehlt";
					logger.error( msg.str(), LOGGER_AUTO );
					return;
				}

				break;
			}

			p_page = p_page->NextSiblingElement( "page" );
		}

		if( !p_page )
		{
			// Fehler!
			setError( 7 );
			std::ostringstream msg;
			msg << filename << ": Seite mit ID 0 wurde nicht gefunden";
			logger.error( msg.str(), LOGGER_AUTO );
			return;
		}
	}
	else
	{
		// Fehler!
		setError( 8 );
		std::ostringstream msg;
		msg << filename << ": <pages>-Tag fehlt";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Zeicheninformationen laden
	TiXmlElement* p_chars = p_font->FirstChildElement( "chars" );
	if( p_chars )
	{
		int n = 0;
		TiXmlElement* p_char = p_chars->FirstChildElement( "char" );
		while( p_char )
		{
			CharacterInfo* p_info = new CharacterInfo;

			// Informationen ber dieses Zeichen abfragen
			int id;
			if( p_char->Attribute( "id", &id ) &&
				p_char->Attribute( "x", &p_info->position.x ) &&
				p_char->Attribute( "y", &p_info->position.y ) &&
				p_char->Attribute( "width", &p_info->size.x ) &&
				p_char->Attribute( "height", &p_info->size.y ) &&
				p_char->Attribute( "xoffset", &p_info->offset.x ) &&
				p_char->Attribute( "yoffset", &p_info->offset.y ) &&
				p_char->Attribute( "xadvance", &p_info->advanceX ) )
			{
				// eintragen
				assert( id >= 0 && id < 256 );
				characterInfo[id] = p_info;
				n++;
			}
			else
			{
				// Dieses Zeichen ist ungltig!
				std::ostringstream msg;
				msg << filename << ": Attribute in einem <char>-Element fehlen - Zeichen wird ignoriert";
				logger.warning( msg.str(), LOGGER_AUTO );
				delete p_info;
			}

			p_char = p_char->NextSiblingElement( "char" );
		}

		if( !n )
		{
			std::ostringstream msg;
			msg << filename << ": Es wurden keine Zeichen geladen";
			logger.warning( msg.str(), LOGGER_AUTO );
		}
		else
		{
			std::ostringstream msg;
			msg << filename << ": Es wurden " << n << " Zeichen geladen";
			logger.finer( msg.str(), LOGGER_AUTO );
		}
	}
	else
	{
		// Fehler!
		setError( 9 );
		std::ostringstream msg;
		msg << filename << ": <chars>-Tag fehlt";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Kerning-Informationen laden (optional)
	TiXmlElement* p_kernings = p_font->FirstChildElement( "kernings" );
	if( p_kernings )
	{
		int n = 0;
		TiXmlElement* p_kerning = p_kernings->FirstChildElement( "kerning" );
		while( p_kerning )
		{
			int first, second, amount;
			if( p_kerning->Attribute( "first", &first ) &&
				p_kerning->Attribute( "second", &second ) &&
				p_kerning->Attribute( "amount", &amount ) )
			{
				// eintragen
				kerning[std::pair<int, int>( first, second )] = amount;
				n++;
			}
			else
			{
				// Dieses Kerning-Paar ist ungltig.
				std::ostringstream msg;
				msg << filename << ": Attribute in einem <kerning>-Element fehlen - Kerning-Paar wird ingoriert";
				logger.warning( msg.str(), LOGGER_AUTO );
			}

			p_kerning = p_kerning->NextSiblingElement( "kerning" );
		}

		if( n )
		{
			std::ostringstream msg;
			msg << filename << ": Es wurden " << n << " Kerning-Paare geladen";
			logger.finer( msg.str(), LOGGER_AUTO );
		}
	}
}

Font::~Font()
{
	// alle Zeichen lschen
	for( int i = 0; i < 256; i++ ) delete characterInfo[i];

	// Cache lschen
	clearCache();
}

void Font::print( const std::string& text,
				  float lineHeight,
				  const Vec4f& color,
				  unsigned int formattingFlags,
				  const FontStyle& style,
				  bool useCache,
				  int maxWidth )
{
	Video& video = getManager().getEngine().getVideo();

	// Alpha-Blending aktivieren und Textur setzen
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	video.setTexture( texture );

	// Enthlt der Text Farbcodes?
	std::string newText;
	if( text.find_first_of( -1 ) != std::string::npos ) newText = beginColor( color ) + text + endColor();
	else newText = text;

	// Text vorbereiten
	unsigned int listID;
	Vec2i size;
	prepareText( newText, style, useCache, maxWidth, &size, 0, &listID );

	// Farbe setzen und Liste ausfhren
	glColor4fv( color );

	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();

	const float scaling = getScaling( lineHeight );
	glScalef( scaling, scaling, 1.0f );

	if( formattingFlags )
	{
		const Vec2f sizeF( size );
		Vec2f offset( 0.0f, 0.0f );

		if( formattingFlags & FF_RIGHT ) offset.x = -sizeF.x;
		else if( formattingFlags & FF_HCENTER ) offset.x = -sizeF.x / 2.0f;
		if( formattingFlags & FF_BOTTOM ) offset.y = -sizeF.y;
		else if( formattingFlags & FF_VCENTER ) offset.y = -sizeF.y / 2.0f;

		glTranslatef( offset.x, offset.y, 0.0f );
	}

	glCallList( listID );

	glPopMatrix();
	glPopAttrib();
}

void Font::prepareText( const std::string& text,
					    const FontStyle& style,
					    bool useCache,
						int maxWidth,
					    Vec2i* p_sizeOut,
						std::vector<Vec2i>* p_charPositionsOut,
						unsigned int* p_listIDOut )
{
	if( useCache )
	{
		// Ist der Text im Cache?
		for( std::list<CacheEntry*>::const_iterator it = cache.begin();
			 it != cache.end();
			 it++ )
		{
			if( (*it)->text == text &&
				(*it)->style == style &&
				(*it)->maxWidth == maxWidth )
			{
				// Wir haben den Eintrag im Cache gefunden
				(*it)->lastUsed = getManager().getEngine().getTime();
				if( p_sizeOut ) *p_sizeOut = (*it)->size;
				if( p_charPositionsOut ) *p_charPositionsOut = (*it)->charPositions;
				if( p_listIDOut ) *p_listIDOut = (*it)->listID;
				return;
			}
		}
	}

	Vec2i cursor( 0, 0 );
	int startX = cursor.x;
	unsigned char prevChar = 0;
	unsigned char currentChar = 0;

	// Cache-Eintrag erzeugen
	CacheEntry* p_entry = new CacheEntry;
	CacheEntry& entry = *p_entry;
	entry.lastUsed = 0;
	entry.text = text;
	entry.style = style;
	entry.maxWidth = maxWidth;
	entry.size = Vec2i( 0, text.empty() ? 0 : lineHeight );
	entry.listID = glGenLists( 1 );

	glNewList( entry.listID, GL_COMPILE );
	glBegin( GL_QUADS );

	std::stack<Vec4f> colors;

	std::string textToPrint;

	if( maxWidth )
	{
		// automatische Zeilenumbrche einfgen
		textToPrint = adjustText( text, style, maxWidth );
	}
	else
	{
		textToPrint = text;
	}

	for( size_t i = 0; i < textToPrint.length(); i++ )
	{
		entry.charPositions.push_back( cursor );

		// Zeichen holen
		prevChar = currentChar;
		currentChar = textToPrint[i];
		if( currentChar == '\n' )
		{
			// neue Zeile
			cursor.x = startX;
			cursor.y += lineHeight + style.lineSpacing;
			entry.size.y += lineHeight + style.lineSpacing;
		}
		else if( currentChar == '\t' )
		{
			const int tabSize = lineHeight * 4;
			cursor.x /= tabSize;
			cursor.x += 1;
			cursor.x *= tabSize;
		}
		else if( currentChar == 255 )
		{
			// Es folgt ein Farbcode!
			unsigned char r = textToPrint[i + 1];
			unsigned char g = textToPrint[i + 2];
			unsigned char b = textToPrint[i + 3];
			unsigned char a = textToPrint[i + 4];
			Vec4f c( static_cast<float>( r ) / 255.0f, static_cast<float>( g ) / 255.0f, static_cast<float>( b ) / 255.0f, static_cast<float>( a ) / 255.0f );
			colors.push( c );
			glColor4fv( colors.top() );
			for( int j = 0; j < 4; j++ ) entry.charPositions.push_back( cursor );
			i += 4;
			currentChar = prevChar;
		}
		else if( currentChar == 254 )
		{
			if( !colors.empty() )
			{
				colors.pop();
				if( !colors.empty() ) glColor4fv( colors.top() );
			}

			currentChar = prevChar;
		}
		else
		{
			CharacterInfo* p_char = characterInfo[currentChar];
			if( p_char )
			{
				const Vec2i position( cursor + p_char->offset );
				glTexCoord2iv( p_char->position );
				glVertex2iv( position + Vec2i( style.italic, 0 ) );
				glTexCoord2iv( p_char->position + Vec2i( p_char->size.x, 0 ) );
				glVertex2iv( position + Vec2i( p_char->size.x + style.italic, 0 ) );
				glTexCoord2iv( p_char->position + p_char->size );
				glVertex2iv( position + p_char->size );
				glTexCoord2iv( p_char->position + Vec2i( 0, p_char->size.y ) );
				glVertex2iv( position + Vec2i( 0, p_char->size.y ) );

				// Position dieses Zeichens korrigieren
				entry.charPositions.back() = position;

				// rechten Rand des Zeichens in die Grenberechnung einbeziehen
				entry.size.x = max( entry.size.x, position.x + p_char->size.x + max( 0, style.italic ) );

				// Cursor bewegen
				cursor.x += p_char->advanceX + getKerning( prevChar, currentChar ) + style.characterSpacing;
			}
		}
	}

	glEnd();
	glEndList();

	if( p_sizeOut ) *p_sizeOut = entry.size;
	if( p_charPositionsOut ) *p_charPositionsOut = entry.charPositions;
	if( p_listIDOut ) *p_listIDOut = entry.listID;

	if( useCache )
	{
		// Ist der Cache voll?
		if( cache.size() == maxCacheSize )
		{
			// den ltesten Eintrag entfernen
			purgeCache();
		}

		// Cache-Eintrag schreiben
		cache.push_back( p_entry );
	}
	else
	{
		// Eintrag wieder lschen
		delete p_entry;
	}
}

std::string Font::adjustText( const std::string& text,
							  const FontStyle& style,
							  int maxWidth ) const
{
	int cursorX = 0;
	int startX = cursorX;
	unsigned char prevChar = 0;
	unsigned char currentChar = 0;

	std::string result( text );

	for( size_t i = 0; i < result.length(); i++ )
	{
		// Zeichen holen
		prevChar = currentChar;
		currentChar = result[i];
		if( currentChar == '\n' )
		{
			// neue Zeile
			cursorX = startX;
		}
		else if( currentChar == '\t' )
		{
			const int tabSize = lineHeight * 4;
			cursorX /= tabSize;
			cursorX += 1;
			cursorX *= tabSize;
		}
		else if( currentChar == 255 )
		{
			// Es folgt ein Farbcode!
			i += 4;
			currentChar = prevChar;
		}
		else if( currentChar == 254 )
		{
			currentChar = prevChar;
		}
		else
		{
			CharacterInfo* p_char = characterInfo[currentChar];
			if( p_char )
			{
				bool newlineInserted = false;

				if( currentChar != ' ' )
				{
					// Ein druckbares Zeichen!
					int positionX = cursorX + p_char->offset.x;
					if( maxWidth &&
						( positionX + p_char->size.x + style.italic > maxWidth ||
						  positionX + p_char->size.x > maxWidth ) )
					{
						// Die Zeile wird zu lang! Das letzte Leerzeichen wird durch einen Zeilenumbruch ersetzt.
						for( size_t j = i; j != ~0; j-- )
						{
							if( result[j] == ' ' )
							{
								// Hier den Zeilenumbruch einfgen!
								result[j] = '\n';

								// Alles, was danach kommt, muss neu berechnet werden.
								i = j;
								cursorX = startX;
								currentChar = '\n';
								newlineInserted = true;
								break;
							}
						}
					}
				}

				if( !newlineInserted )
				{
					// Cursor bewegen
					cursorX += p_char->advanceX + getKerning( prevChar, currentChar ) + style.characterSpacing;
				}
			}
		}
	}

	return result;
}

float Font::getScaling( float lineHeight ) const
{
	return lineHeight <= 0.0f ? 1.0f : ( lineHeight / static_cast<float>( this->lineHeight ) );
}

void Font::clearCache()
{
	for( std::list<CacheEntry*>::const_iterator it = cache.begin();
		 it != cache.end();
		 it++ )
	{
		deleteCacheEntry( *it );
	}

	cache.clear();
}

void Font::purgeCache()
{
	if( cache.empty() ) return;

	// den ltesten Eintrag suchen
	std::list<CacheEntry*>::iterator oldest = cache.begin();
	for( std::list<CacheEntry*>::iterator it = oldest;
		 it != cache.end();
		 it++ )
	{
		if( (*it)->lastUsed < (*oldest)->lastUsed )
		{
			// Dieser Eintrag ist noch lter!
			oldest = it;
		}
	}

	// ltesten Eintrag lschen
	deleteCacheEntry( *oldest );
	cache.erase( oldest );
}

unsigned int Font::getCacheSize() const
{
	return static_cast<unsigned int>( cache.size() );
}

void Font::setMaxCacheSize( unsigned int maxCacheSize )
{
	this->maxCacheSize = maxCacheSize;
	while( cache.size() > maxCacheSize ) purgeCache();
}

unsigned int Font::getMaxCacheSize() const
{
	return maxCacheSize;
}

int Font::getKerning( unsigned char char1,
					  unsigned char char2 ) const
{
	// Existiert ein passender Eintrag in der Kerning-Tabelle?
	std::map<std::pair<int, int>, int>::const_iterator it = kerning.find( std::pair<int, int>( char1, char2 ) );
	return it == kerning.end() ? 0 : it->second;
}

void Font::deleteCacheEntry( const CacheEntry* p_entry )
{
	assert( p_entry );

	glDeleteLists( p_entry->listID, 1 );
	delete p_entry;
}

std::string beginColor( const Vec4f& color )
{
	return beginColor( color.r, color.g, color.b, color.a );
}

std::string beginColor( float r,
					    float g,
						float b,
						float a )
{
	unsigned char s[6];
	s[0] = 255;
	s[1] = static_cast<unsigned char>( clamp( r, 0.0f, 1.0f ) * 255.0f );
	s[2] = static_cast<unsigned char>( clamp( g, 0.0f, 1.0f ) * 255.0f );
	s[3] = static_cast<unsigned char>( clamp( b, 0.0f, 1.0f ) * 255.0f );
	s[4] = static_cast<unsigned char>( clamp( a, 0.0f, 1.0f ) * 255.0f );
	s[5] = 0;
	return std::string( reinterpret_cast<const char*>( s ) );
}

std::string endColor()
{
	unsigned char s[2];
	s[0] = 254;
	s[1] = 0;
	return std::string( reinterpret_cast<const char*>( s ) );
}