#include <PCH.h>
#include <Input.h>
#include <Engine.h>
#include <SDL.h>

static const char BIT_DOWN = 1;
static const char BIT_PRESSED = 2;
static const char BIT_RELEASED = 4;

Input::Input( Engine& engine ) : Component( engine, "Input" ),
								 mousePosition( 0, 0 ),
								 mouseMotion( 0, 0 ),
								 statesChanged( false )
{
	// Array fr die Tasten und Mausknpfe anlegen
	p_states = new char[KEY_LAST + MBTN_LAST];

	// Array leeren
	for( int i = 0; i < KEY_LAST + MBTN_LAST; i++ ) p_states[i] = 0;

	updateMousePosition();

	engine.connectSDLEvent( this, &Input::onSDLEvent );
	engine.connectUpdate( this, &Input::onUpdate );

	engine.getLogger().fine( "Input-Komponente initialisiert", LOGGER_AUTO );
}

Input::~Input()
{
	// Array freigeben
	delete[] p_states;

	getEngine().getLogger().fine( "Input-Komponente heruntergefahren", LOGGER_AUTO );
}

bool Input::isKeyDown( Key key ) const
{
	assert( key < KEY_LAST );
	return ( p_states[key] & BIT_DOWN ) ? true : false;
}

bool Input::wasKeyPressed( Key key ) const
{
	assert( key < KEY_LAST );
	return ( p_states[key] & BIT_PRESSED ) ? true : false;
}

bool Input::wasKeyReleased( Key key ) const
{
	assert( key < KEY_LAST );
	return ( p_states[key] & BIT_RELEASED ) ? true : false;
}

bool Input::isMouseButtonDown( MouseButton mouseButton ) const
{
	assert( mouseButton < MBTN_LAST );
	return ( p_states[KEY_LAST + mouseButton] & BIT_DOWN ) ? true : false;
}

bool Input::wasMouseButtonPressed( MouseButton mouseButton ) const
{
	assert( mouseButton < MBTN_LAST );
	return ( p_states[KEY_LAST + mouseButton] & BIT_PRESSED ) ? true : false;
}

bool Input::wasMouseButtonReleased( MouseButton mouseButton ) const
{
	assert( mouseButton < MBTN_LAST );
	return ( p_states[KEY_LAST + mouseButton] & BIT_RELEASED ) ? true : false;
}

const Vec2i& Input::getMousePosition() const
{
	return mousePosition;
}

const Vec2i& Input::getMouseMotion() const
{
	return mouseMotion;
}

void Input::onSDLEvent( const SDL_Event& e )
{
	switch( e.type )
	{
	case SDL_KEYDOWN:
	case SDL_KEYUP:
	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
		{
			// Eine Taste oder ein Mausknopf wurde gedrckt oder losgelassen!
			bool wasKey = ( e.type == SDL_KEYDOWN || e.type == SDL_KEYUP );
			int state = wasKey ? e.key.state : e.button.state;
			unsigned int index = wasKey ? e.key.keysym.sym : KEY_LAST + e.button.button;
			assert( index < KEY_LAST + MBTN_LAST );

			switch( state )
			{
			case SDL_PRESSED:
				// "gedrckt"- und "wurde gerade gedrckt"-Bits setzen
				p_states[index] |= BIT_DOWN | BIT_PRESSED;
				break;
			case SDL_RELEASED:
				// "gedrckt"-Bit lschen und "wurde gerade losgelassen"-Bit setzen
				p_states[index] = ( p_states[index] & ~BIT_DOWN ) | BIT_RELEASED;
				break;
			}

			statesChanged = true;
			break;
		}

	case SDL_MOUSEMOTION:
		// relative Mausbewegung und absolute Mausposition speichern
		mouseMotion = Vec2i( e.motion.xrel, e.motion.yrel );
		mousePosition = Vec2i( e.motion.x, e.motion.y );
		break;
	}
}

void Input::onUpdate()
{
	updateMousePosition();

	// Mausbewegung zurcksetzen
	mouseMotion = Vec2i( 0, 0 );

	if( statesChanged )
	{
		// "wurde gedrckt"- und "wurde losgelassen"-Bits lschen
		for( int i = 0; i < KEY_LAST + MBTN_LAST; i++ ) p_states[i] &= ~( BIT_PRESSED | BIT_RELEASED );

		statesChanged = false;
	}
}

void Input::updateMousePosition()
{
	// aktuelle Mausposition holen
	SDL_GetMouseState( &mousePosition.x, &mousePosition.y );
}