#include <PCH.h>
#include <ParticleSystem.h>
#include <SimpleEmitter.h>
#include <Engine.h>
#include <Video.h>
#include <Utils.h>
#include <SDL_OpenGL.h>

ParticleSystem::ParticleSystem( Video& video,
							    unsigned int tickDuration ) : video( video ),
															  gravity( 0.0f, 0.0f ),
															  globalRate( 1.0f ),
															  timeToProcess( 0 ),
															  numParticles( 0 ),
															  capacity( 4096 ),
															  tickDuration( tickDuration ),
															  tickDurationF( static_cast<float>( tickDuration ) / 1000.0f ),
															  oneOverTickDurationF( 1000.0f / static_cast<float>( tickDuration ) )
{
}

ParticleSystem::~ParticleSystem()
{
	// alle Emitter lschen
	for( std::set<ParticleEmitter*>::const_iterator it = emitters.begin();
		 it != emitters.end();
		 it++ )
	{
		delete *it;
	}
}

int ParticleSystem::registerParticleStyle( unsigned int styleID,
										   const ParticleStyle& style )
{
	std::map<unsigned int, ParticleStyle>::const_iterator it = particleStyles.find( styleID );
	if( it != particleStyles.end() ) return 1;

	particleStyles[styleID] = style;
	particleLists[styleID] = std::list<Particle>();

	return 0;
}

int ParticleSystem::unregisterParticleStyle( unsigned int styleID )
{
	std::map<unsigned int, ParticleStyle>::iterator it = particleStyles.find( styleID );
	if( it == particleStyles.end() ) return 1;

	particleStyles.erase( it );
	particleLists.erase( styleID );

	return 0;
}

int ParticleSystem::addParticle( unsigned int styleID,
								 const Particle& particle )
{
	if( globalRate < 1.0f )
	{
		// Partikel zufllig verwerfen
		if( randFloat() > globalRate ) return 0;
	}

	// Liste fr diesen Stil suchen
	std::map<unsigned int, std::list<Particle> >::iterator it = particleLists.find( styleID );
	if( it == particleLists.end() ) return 1;

	// Voll?
	if( numParticles == capacity )
	{
		// ltesten Partikel lschen
		it->second.pop_front();
	}
	else
	{
		numParticles++;
	}

	// Zeit einrechnen
	const float dt = tickDurationF;
	Particle p;
	p.timeToLive = particle.timeToLive;
	p.position = particle.position;
	p.velocity = particle.velocity * dt;
	p.angle = particle.angle;
	p.angularVelocity = particle.angularVelocity * dt;
	p.color = particle.color;
	p.deltaColor = particle.deltaColor * dt;
	p.size = particle.size;
	p.deltaSize = particle.deltaSize * dt;
	p.friction = particle.friction == 1.0f ? 1.0f : powf( particle.friction, dt );
	p.gravity = particle.gravity * dt * dt;
	p.animFrame = particle.animFrame;

	// Partikel einfgen
	it->second.push_back( p );

	return 0;
}

SimpleEmitter* ParticleSystem::createSimpleEmitter()
{
	SimpleEmitter* p_emitter = new SimpleEmitter( *this );
	emitters.insert( p_emitter );

	return p_emitter;
}

int ParticleSystem::destroyEmitter( ParticleEmitter* p_emitter )
{
	assert( p_emitter );

	std::set<ParticleEmitter*>::iterator it = emitters.find( p_emitter );
	if( it == emitters.end() ) return 1;

	emitters.erase( it );
	delete p_emitter;

	return 0;
}

bool ParticleSystem::isEmitterValid( ParticleEmitter* p_emitter ) const
{
	std::set<ParticleEmitter*>::const_iterator it = emitters.find( p_emitter );
	return it != emitters.end();
}

void ParticleSystem::clearParticles()
{
	for( std::map<unsigned int, std::list<Particle> >::iterator it = particleLists.begin();
		 it != particleLists.end();
		 it++ )
	{
		it->second.clear();
	}
}

void ParticleSystem::render( bool renderMultiplicative,
							 bool renderAdditive )
{
	if( !numParticles ) return;

	for( std::map<unsigned int, ParticleStyle>::const_iterator it = particleStyles.begin();
		 it != particleStyles.end();
		 it++ )
	{
		if( ( renderMultiplicative && it->second.destBlendFactor != GL_ONE ) ||
			( renderAdditive && it->second.destBlendFactor == GL_ONE ) )
		{
			renderParticles( it->second, particleLists[it->first] );
		}
	}
}

void ParticleSystem::update()
{
	timeToProcess += video.getEngine().getTickDuration();

	unsigned int timeProcessed = 0;
	while( timeToProcess >= tickDuration )
	{
		// Emitter aktualisieren
		for( std::set<ParticleEmitter*>::iterator it = emitters.begin();
			 it != emitters.end(); )
		{
			(*it)->update();

			if( (*it)->isToBeDeleted() )
			{
				delete *it;
				it = emitters.erase( it );
			}
			else
			{
				it++;
			}
		}

		if( numParticles )
		{
			// Partikel aktualisieren
			const float dt = tickDurationF;
			for( std::map<unsigned int, std::list<Particle> >::iterator itList = particleLists.begin();
				 itList != particleLists.end();
				 itList++ )
			{
				const ParticleStyle& style = particleStyles[itList->first];
				std::list<Particle>& theList = itList->second;

				for( std::list<Particle>::iterator it = theList.begin();
					 it != theList.end(); )
				{
					Particle& p = *it;

					if( p.timeToLive < 0.0f || p.size <= 0.0f )
					{
						it = theList.erase( it );
						numParticles--;
					}
					else
					{
						p.timeToLive -= dt;
						p.position += p.velocity;

						if( style.mode == PM_NORMAL )
						{
							p.angle += p.angularVelocity;

							if( fabsf( p.angle ) > 3.14159f )
							{
								if( p.angle < 0.0f ) p.angle = fmodf( p.angle, 3.14159f ) + 3.14159f;
								else p.angle = fmodf( p.angle, 3.14159f ) - 3.14159f;
							}
						}

						p.color += p.deltaColor;
						p.size += p.deltaSize;
						p.animFrame += dt * style.animSpeed;

						p.velocity += p.gravity * gravity;

						if( p.friction != 1.0f )
						{
							p.velocity *= p.friction;
							if( style.mode == PM_NORMAL ) p.angularVelocity *= p.friction;
						}

						it++;
					}
				}
			}
		}

		timeProcessed += tickDuration;
		timeToProcess -= tickDuration;
	}
}

void ParticleSystem::setGravity( const Vec2f& gravity )
{
	this->gravity = gravity;
}

const Vec2f& ParticleSystem::getGravity() const
{
	return gravity;
}

void ParticleSystem::setGlobalRate( float globalRate )
{
	this->globalRate = globalRate;
}

float ParticleSystem::getGlobalRate() const
{
	return globalRate;
}

unsigned int ParticleSystem::getNumParticles() const
{
	return numParticles;
}

unsigned int ParticleSystem::getNumEmitters() const
{
	return static_cast<unsigned int>( emitters.size() );
}

void ParticleSystem::setCapacity( unsigned int capacity )
{
	this->capacity = capacity;
}

unsigned int ParticleSystem::getCapacity() const
{
	return capacity;
}

unsigned int ParticleSystem::getTickDuration() const
{
	return tickDuration;
}

float ParticleSystem::getTickDurationInSeconds() const
{
	return tickDurationF;
}

Video& ParticleSystem::getVideo() const
{
	return video;
}

void ParticleSystem::renderParticles( const ParticleStyle& style,
									  const std::list<Particle>& particles )
{
	video.setTexture( style.texture );
	video.setupAlphaBlending( true, style.srcBlendFactor, style.destBlendFactor );

	bool hasAnim = style.frames.size() > 1;

	glBegin( GL_QUADS );

	for( std::list<Particle>::const_iterator it = particles.begin();
		 it != particles.end();
		 it++ )
	{
		const Particle& p = *it;
		const Vec2i& frame = hasAnim ? ( style.frames[static_cast<int>( p.animFrame ) % style.frames.size()] ) : style.frames[0];
		const float halfSize = 0.5f * p.size;

		if( style.mode == PM_NORMAL )
		{
			float sinR, cosR;
			float angle = p.angle;

			if( fabsf( angle ) > 3.14159f )
			{
				if( angle < 0.0f ) angle = fmodf( angle, 3.14159f ) + 3.14159f;
				else angle = fmodf( angle, 3.14159f ) - 3.14159f;
			}

			fastSinCos( p.angle, sinR, cosR );

			const Vec2f axisX = Vec2f( cosR, sinR ) * halfSize;
			const Vec2f axisY = Vec2f( -sinR, cosR ) * halfSize;

			glColor4fv( p.color );
			glTexCoord2iv( frame );
			glVertex2fv( p.position - axisX - axisY );
			glTexCoord2i( frame.x + style.frameSize.x, frame.y );
			glVertex2fv( p.position + axisX - axisY );
			glTexCoord2iv( frame + style.frameSize );
			glVertex2fv( p.position + axisX + axisY );
			glTexCoord2i( frame.x, frame.y + style.frameSize.y );
			glVertex2fv( p.position - axisX + axisY );
		}
		else if( style.mode == PM_ROTATE_BY_VELOCITY )
		{
			const Vec2f axisX( p.velocity.normalizedFast() * halfSize );
			const Vec2f axisY( -axisX.y, axisX.x );

			glColor4fv( p.color );
			glTexCoord2iv( frame );
			glVertex2fv( p.position - axisX - axisY );
			glTexCoord2i( frame.x + style.frameSize.x, frame.y );
			glVertex2fv( p.position + axisX - axisY );
			glTexCoord2iv( frame + style.frameSize );
			glVertex2fv( p.position + axisX + axisY );
			glTexCoord2i( frame.x, frame.y + style.frameSize.y );
			glVertex2fv( p.position - axisX + axisY );
		}
		else if( style.mode == PM_STRETCH_BY_VELOCITY )
		{
			const float speed = p.velocity.lengthFast();
			Vec2f axisX( p.velocity / speed * halfSize );
			Vec2f axisY( -axisX.y, axisX.x );
			axisX *= speed * oneOverTickDurationF * style.stretch;

			glColor4fv( p.color );
			glTexCoord2iv( frame );
			glVertex2fv( p.position - axisX - axisY );
			glTexCoord2i( frame.x + style.frameSize.x, frame.y );
			glVertex2fv( p.position + axisX - axisY );
			glTexCoord2iv( frame + style.frameSize );
			glVertex2fv( p.position + axisX + axisY );
			glTexCoord2i( frame.x, frame.y + style.frameSize.y );
			glVertex2fv( p.position - axisX + axisY );
		}
	}

	glEnd();
}