#include <PCH.h>
#include <Placeable.h>
#include <SDL_OpenGL.h>

Placeable::Placeable() : p_parent( 0 ),
						 transform( Mat33f::identity() ),
						 invTransform( Mat33f::identity() ),
						 invTransformDirty( false )
{
}

Placeable::~Placeable()
{
	// alle Kinder entfernen
	while( !children.empty() ) detachChild( *children.begin(), true );

	// mich selbst entfernen
	if( p_parent ) p_parent->detachChild( this, false );
}

void Placeable::attachTo( Placeable* p_parent,
						  bool maintainAbsTransform )
{
	if( p_parent )
	{
		p_parent->attachChild( this, maintainAbsTransform );
	}
}

void Placeable::detachFromParent( bool maintainAbsTransform )
{
	if( p_parent )
	{
		p_parent->detachChild( this, maintainAbsTransform );
	}
}

void Placeable::attachChild( Placeable* p_child,
							 bool maintainAbsTransform )
{
	assert( p_child );

	if( children.find( p_child ) != children.end() ) return;

	Mat33f absTransform;
	if( maintainAbsTransform )
	{
		// absolute Transformation des Kinds holen
		absTransform = p_child->getAbsTransform();
	}

	// Kind vom alten Vaterobjekt entfernen
	if( p_child->p_parent ) p_child->p_parent->detachChild( p_child, false );

	p_child->p_parent = this;

	if( maintainAbsTransform )
	{
		// absolute Transformation des Kinds wiederherstellen
		p_child->setAbsTransform( absTransform );
	}

	// in die eigene Kindermenge aufnehmen
	children.insert( p_child );
}

void Placeable::detachChild( Placeable* p_child,
							 bool maintainAbsTransform )
{
	assert( p_child );

	// Kind aus der Menge entfernen
	if( children.erase( p_child ) )
	{
		if( maintainAbsTransform )
		{
			// absolute Transformation des Kinds wird zur relativen Transformation
			p_child->setTransform( p_child->getAbsTransform() );
		}

		p_child->p_parent = 0;
	}
}

Placeable* Placeable::getParent() const
{
	return p_parent;
}

const std::set<Placeable*>& Placeable::getChildren() const
{
	return children;
}

void Placeable::setPosition( const Vec2f& position )
{
	Mat33f transform( getTransform() );

	transform.m13 = position.x;
	transform.m23 = position.y;

	setTransform( transform );
}

Vec2f Placeable::getPosition() const
{
	Mat33f transform( getTransform() );
	return Vec2f( transform.m13, transform.m23 );
}

void Placeable::setAbsPosition( const Vec2f& position )
{
	if( !p_parent ) setPosition( position );
	else setPosition( p_parent->getInvAbsTransform().transformPosition( position ) );
}

Vec2f Placeable::getAbsPosition() const
{
	if( !p_parent ) return getPosition();
	else return p_parent->getAbsTransform().transformPosition( getPosition() );
}

void Placeable::setAngle( float angle )
{
	Mat33f transform( getTransform() );

	transform.m11 = cosf( angle );
	transform.m21 = sinf( angle );
	transform.m12 = -transform.m21;
	transform.m22 = transform.m11;

	setTransform( transform );
}

float Placeable::getAngle() const
{
	Mat33f transform( getTransform() );
	return -atan2f( transform.m12, transform.m11 );
}

void Placeable::setAbsAngle( float angle )
{
	if( !p_parent ) setAngle( angle );
	else setAngle( angle - p_parent->getAbsAngle() );
}

float Placeable::getAbsAngle() const
{
	if( !p_parent ) return getAngle();
	else return p_parent->getAbsAngle() + getAngle();
}

void Placeable::setDirection( const Vec2f& direction )
{
	Mat33f transform( getTransform() );

	const Vec2f& d( direction.normalized() );
	transform.m12 = d.x;
	transform.m22 = d.y;
	transform.m11 = transform.m22;
	transform.m21 = -transform.m12;

	setTransform( transform );
}

Vec2f Placeable::getDirection() const
{
	Mat33f transform( getTransform() );
	return Vec2f( transform.m12, transform.m22 );
}

void Placeable::setAbsDirection( const Vec2f& direction )
{
	if( !p_parent ) setDirection( direction );
	else setDirection( p_parent->getInvAbsTransform().transformDirection( direction ) );
}

Vec2f Placeable::getAbsDirection() const
{
	if( !p_parent ) return getDirection();
	else return p_parent->getAbsTransform().transformDirection( getDirection() );
}

void Placeable::setTransform( const Mat33f& transform )
{
	this->transform = transform;

	invTransformDirty = true;
}

Mat33f Placeable::getTransform() const
{
	return transform;
}

void Placeable::setAbsTransform( const Mat33f& transform )
{
	if( !p_parent ) setTransform( transform );
	else setTransform( p_parent->getInvAbsTransform() * transform );
}

Mat33f Placeable::getAbsTransform() const
{
	if( !p_parent ) return getTransform();
	else return p_parent->getAbsTransform() * getTransform();
}

const Mat33f& Placeable::getInvTransform() const
{
	if( invTransformDirty ) updateInvTransform();
	return invTransform;
}

Mat33f Placeable::getInvAbsTransform() const
{
	return getAbsTransform().inverse();
}

Vec2f Placeable::relPositionToAbsPosition( const Vec2f& position ) const
{
	return getAbsTransform().transformPosition( position );
}

Vec2f Placeable::absPositionToRelPosition( const Vec2f& position ) const
{
	return getInvAbsTransform().transformPosition( position );
}

float Placeable::relAngleToAbsAngle( float angle ) const
{
	const float r = getAbsAngle() + angle;
	if( r < -3.1416f ) return r + 6.2832f;
	else if( r > 3.1416f ) return r - 6.2832f;
	else return r;
}

float Placeable::absAngleToRelAngle( float angle ) const
{
	const float r = angle - getAbsAngle();
	if( r < -3.1416f ) return r + 6.2832f;
	else if( r > 3.1416f ) return r - 6.2832f;
	else return r;
}

Vec2f Placeable::relDirectionToAbsDirection( const Vec2f& direction ) const
{
	return getAbsTransform().transformDirection( direction );
}

Vec2f Placeable::absDirectionToRelDirection( const Vec2f& direction ) const
{
	return getInvAbsTransform().transformDirection( direction );
}

Mat33f Placeable::relTransformToAbsTransform( const Mat33f& transform ) const
{
	return getAbsTransform() * transform;
}

Mat33f Placeable::absTransformToRelTransform( const Mat33f& transform ) const
{
	return getInvAbsTransform() * transform;
}

void Placeable::setupGLMatrix()
{
	// Transformationsmatrix fr OpenGL aus Orientierung und Position zusammenbasteln
	const Mat33f& t = getTransform();

	float matrix[16] = { t.m11, t.m21, 0.0f, 0.0f,
						 t.m12, t.m22, 0.0f, 0.0f,
						 0.0f,  0.0f,  1.0f, 0.0f,
						 t.m13, t.m23, 0.0f, 1.0f };

	glMultMatrixf( matrix );
}

void Placeable::updateInvTransform() const
{
	invTransform = transform.inverse();

	invTransformDirty = false;
}