#include <PCH.h>
#include <Profiler.h>
#include <SDL.h>

#ifdef WIN32
#include <Windows.h>
#endif

Profiler::Profiler()
{
#ifdef WIN32
	// Ist der Hochleistungs-Timer verfgbar?
	LARGE_INTEGER f;
	BOOL available = QueryPerformanceFrequency( &f );
	if( available )
	{
		// Ja!
		timerResolution = f.QuadPart;
	}
	else
	{
		// Nein! SDL verwenden!
		timerResolution = 1000;
	}
#else
	// SDL verwenden
	timerResolution = 1000;
#endif

	oneOverTimerResolution = static_cast<long double>( 1.0 ) / static_cast<long double>( timerResolution );
	setUpdateCycle( 1000 );

	// Wurzelknoten erzeugen
	p_root = new ProfileInfo;
	p_root->name = "OVERALL";
	p_root->displayedName = "OVERALL";
	profiles["OVERALL"] = p_root;

	reset();
}

Profiler::Profiler( const Profiler& copy )
{
}

Profiler::~Profiler()
{
	// aufrumen
	for( std::map<std::string, ProfileInfo*>::iterator it = profiles.begin();
		 it != profiles.end();
		 it++ )
	{
		delete it->second;
	}
}

Profiler& Profiler::inst()
{
	static Profiler inst;
	return inst;
}

void Profiler::setUpdateCycle( unsigned int updateCycle )
{
	// Wenn der Zyklus gendert wurde, mssen die Statistiken neu generiert werden.
	if( this->updateCycle != updateCycle ) reset();

	this->updateCycle = updateCycle;
}

void Profiler::setupProfile( const std::string& name,
							 const std::string& displayedName,
							 const std::string& parentName )
{
	// Schon vorhanden?
	if( profiles.find( name ) != profiles.end() ) return;

	ProfileInfo* p_parent = parentName.empty() ? p_root : profiles[parentName];

	// Profil eintragen
	ProfileInfo* p_profile = new ProfileInfo;
	p_profile->name = name;
	p_profile->displayedName = displayedName.empty() ? name : displayedName;
	p_profile->p_parent = p_parent;
	profiles[name] = p_profile;

	if( p_parent ) p_parent->children.push_back( p_profile );
}

void Profiler::begin( const std::string& name )
{
	std::map<std::string, ProfileInfo*>::iterator it = profiles.find( name );
	if( it == profiles.end() )
	{
		reset();
		statistics = std::string( "FEHLER: Unbekannter Profilname \"" ) + name + "\"";
		return;
	}

	RunningProfile entry;
	entry.p_profile = it->second;
	entry.startTime = getTimer();
	running.push( entry );
}

void Profiler::end()
{
	if( running.empty() )
	{
		reset();
		statistics = "FEHLER: end() fter aufgerufen als begin()";
		return;
	}

	RunningProfile& entry = running.top();

	unsigned long long endTime = getTimer();
	if( endTime > entry.startTime )
	{
		unsigned long long duration = endTime - entry.startTime;

		// gebrauchte Zeit auf dieses Profil und alle Vorfahren aufaddieren
		ProfileInfo* p_profile = entry.p_profile;
		while( p_profile )
		{
			p_profile->currentTicksConsumed += duration;
			p_profile->totalTicksConsumed += duration;
			p_profile->profiled = true;

			p_profile = p_profile->p_parent;
		}
	}

	running.pop();
}

std::string Profiler::getStatistics()
{
	if( profiles.empty() )
	{
		return "Keine Profile";
	}

	if( !running.empty() )
	{
		return "FEHLER: begin() fter aufgerufen als end()";
	}

	if( !lastUpdate ) lastUpdate = SDL_GetTicks();
	else
	{
		// Ist die Messperiode abgelaufen?
		if( SDL_GetTicks() > lastUpdate + updateCycle )
		{
			// die Statistiken aktualisieren
			updateStatistics();
			lastUpdate = SDL_GetTicks();
		}
	}

	return statistics;
}

void Profiler::reset()
{
	lastUpdate = 0;
	numUpdates = 0;
	statistics = "Bitte warten ...";

	for( std::map<std::string, ProfileInfo*>::iterator it = profiles.begin();
		 it != profiles.end();
		 it++ )
	{
		ProfileInfo& profile = *it->second;
		profile.profiled = false;
		profile.numSamples = 0;
		profile.currentTicksConsumed = 0;
		profile.totalTicksConsumed = 0;
	}
}

unsigned long long Profiler::getTimer() const
{
#ifdef WIN32
	if( timerResolution == 1000 ) return SDL_GetTicks();
	else
	{
		LARGE_INTEGER t;
		QueryPerformanceCounter( &t );
		return t.QuadPart;
	}
#else
	return SDL_GetTicks();
#endif
}

unsigned long long Profiler::getTimerResolution() const
{
	return timerResolution;
}

void Profiler::updateStatistics()
{
	bool anyProfiled = false;

	// Zeitangaben in Sekunden umrechnen und Zhler erhhen
	for( std::map<std::string, ProfileInfo*>::iterator it = profiles.begin();
		 it != profiles.end();
		 it++ )
	{
		ProfileInfo* p_profile = it->second;
		if( p_profile->profiled )
		{
			p_profile->currentTimeConsumed = static_cast<double>( static_cast<long double>( p_profile->currentTicksConsumed ) * oneOverTimerResolution );
			p_profile->totalTimeConsumed = static_cast<double>( static_cast<long double>( p_profile->totalTicksConsumed ) * oneOverTimerResolution );
			p_profile->numSamples++;
			anyProfiled = true;
		}
	}

	if( !anyProfiled )
	{
		statistics = "Keine Daten aufgezeichnet";
		return;
	}

	numUpdates++;

	// Infos ausgeben
	std::ostringstream info;
	info << "T = " << updateCycle << " ms, N = " << numUpdates << std::endl << std::endl;

	// Kopfzeile ausgeben
	std::ostringstream header;
	header << "| ";
	header.width( 30 );
	header.setf( std::ios::left );
	header << "Name";
	header << " | ";
	header.width( 10 );
	header.setf( std::ios::right );
	header << "ms";
	header << " | ";
	header.width( 10 );
	header << "%";
	header << " | ";
	header.width( 10 );
	header << "Total %";
	header << " | ";
	header.width( 10 );
	header << "Avg ms";
	header << " | ";
	header.width( 10 );
	header << "Avg %";
	header << " | ";
	header.width( 11 );
	header << "Avg Total %";
	header << " |";
	size_t length = header.tellp();
	header << std::endl;

	header << std::string( length, '-' );
	header << std::endl;

	std::string result( info.str() + header.str() );

	// Profile rekursiv abarbeiten
	result += buildStatistics( p_root, 0 );

	for( std::map<std::string, ProfileInfo*>::iterator it = profiles.begin();
		 it != profiles.end();
		 it++ )
	{
		it->second->profiled = false;
	}

	statistics = result;
}

std::string Profiler::buildStatistics( ProfileInfo* p_profile,
									   int indent )
{
	ProfileInfo& profile = *p_profile;

	std::ostringstream line;

	line.setf( std::ios::fixed );
	line.precision( 3 );
	line << "| ";
	line.width( 30 );
	line.setf( std::ios::left );
	line << std::string( indent * 2, ' ' ) + profile.displayedName;
	line << " | ";
	line.width( 10 );
	line.setf( std::ios::right );

	if( profile.profiled ) line << profile.currentTimeConsumed * 1000.0;
	else line << "N/A";

	line << " | ";
	line.width( 10 );

	if( profile.profiled )
	{
		if( profile.p_parent )
		{
			if( profile.p_parent->currentTicksConsumed ) line << 100.0 * profile.currentTimeConsumed / profile.p_parent->currentTimeConsumed;
			else line << 0.0;
		}
		else
		{
			line << "-";
		}
	}
	else line << "N/A";

	line << " | ";
	line.width( 10 );

	if( profile.profiled ) line << 100.0 * profile.currentTimeConsumed / ( static_cast<double>( updateCycle ) / 1000.0 );
	else line << "N/A";

	line << " | ";

	line.width( 10 );

	if( profile.profiled ) line << profile.totalTimeConsumed * 1000.0 / static_cast<double>( profile.numSamples );
	else line << "N/A";

	line << " | ";
	line.width( 10 );

	if( profile.profiled )
	{
		if( profile.p_parent )
		{
			if( profile.p_parent->totalTicksConsumed ) line << 100.0 * profile.totalTimeConsumed / profile.p_parent->totalTimeConsumed;
			else line << 0.0;
		}
		else
		{
			line << "-";
		}
	}
	else line << "N/A";

	line << " | ";
	line.width( 11 );

	if( profile.profiled ) line << 100.0 * profile.totalTimeConsumed / ( static_cast<double>( updateCycle ) / 1000.0 ) / static_cast<double>( profile.numSamples );
	else line << "N/A";

	line << " |";

	line << std::endl;

	std::string result( line.str() );

	for( std::vector<ProfileInfo*>::iterator it = profile.children.begin();
		 it != profile.children.end();
		 it++ )
	{
		result += buildStatistics( *it, indent + 1 );
	}

	// Zhler fr die aktuelle Zeit zurcksetzen
	profile.currentTicksConsumed = 0;

	return result;
}