#include <PCH.h>
#include <SimpleEmitter.h>
#include <ParticleSystem.h>
#include <Utils.h>

SimpleEmitter::SimpleEmitter( ParticleSystem& ps ) : ParticleEmitter( ps ),
													 particlesToEmit( 0.0f ),
													 lastPosition( FLT_MAX, FLT_MAX ),
													 lastAngle( 0.0f )
{
}

SimpleEmitter::~SimpleEmitter()
{
}

void SimpleEmitter::setSettings( const SimpleEmitterSettings& settings )
{
	this->settings = settings;
}

const SimpleEmitterSettings& SimpleEmitter::getSettings() const
{
	return settings;
}

void SimpleEmitter::addParticleStyle( unsigned int style )
{
	settings.partStyles.push_back( style );
}

void SimpleEmitter::setParticleTimeToLive( float timeToLive,
										   float variation )
{
	settings.partTimeToLive = timeToLive;
	settings.partTimeToLiveVar = variation;
}

void SimpleEmitter::setParticlePositionRadius( float radius )
{
	settings.partPositionRadius = radius;
}

void SimpleEmitter::setParticleDirection( const Vec2f& direction,
										  float variation )
{
	const Vec2f d( direction.normalized() );
	settings.partDirection =	-atan2f( d.y, d.x );
	settings.partDirectionVar = variation;
}

void SimpleEmitter::setParticleSpeed( float speed,
									  float variation )
{
	settings.partSpeed = speed;
	settings.partSpeedVar = variation;
}

void SimpleEmitter::setParticleAngle( float angle,
									  float variation )
{
	settings.partAngle = angle;
	settings.partAngleVar = variation;
}

void SimpleEmitter::setParticleAngularVelocity( float angularVelocity,
											    float variation )
{
	settings.partAngularVelocity = angularVelocity;
	settings.partAngularVelocityVar = variation;
}

void SimpleEmitter::setParticleStartColor( const Vec4f& startColor,
										   const Vec4f& variation )
{
	settings.partStartColor = startColor;
	settings.partStartColorVar = variation;
}

void SimpleEmitter::setParticleEndColor( const Vec4f& endColor,
										 const Vec4f& variation )
{
	settings.partEndColor = endColor;
	settings.partEndColorVar = variation;
}

void SimpleEmitter::setParticleStartSize( float startSize,
										  float variation )
{
	settings.partStartSize = startSize;
	settings.partStartSizeVar = variation;
}

void SimpleEmitter::setParticleEndSize( float endSize,
										float variation )
{
	settings.partEndSize = endSize;
	settings.partEndSizeVar = variation;
}

void SimpleEmitter::setParticleFriction( float friction,
										 float variation )
{
	settings.partFriction = friction;
	settings.partFrictionVar = variation;
}

void SimpleEmitter::setParticleGravity( float gravity,
										float variation )
{
	settings.partGravity = gravity;
	settings.partGravityVar = variation;
}

void SimpleEmitter::setParticleAnimFrame( float animFrame,
										  float variation )
{
	settings.partAnimFrame = animFrame;
	settings.partAnimFrameVar = variation;
}

void SimpleEmitter::setUseRelativeVelocity( bool useRelativeVelocity )
{
	settings.useRelativeVelocity = useRelativeVelocity;
}

unsigned int SimpleEmitter::doWork( unsigned int maxNumParticles )
{
	ParticleSystem& ps = getParticleSystem();
	unsigned int n = 0;

	const float dt = ps.getTickDurationInSeconds();

	const Vec2f currentPosition( getAbsPosition() );
	const float currentAngle = getAbsAngle();
	if( lastPosition.x == FLT_MAX )
	{
		lastPosition = currentPosition;
		lastAngle = currentAngle;
	}

	const Vec2f currentVelocity( ( currentPosition - lastPosition ) / ps.getTickDurationInSeconds() );

	if( !settings.partStyles.empty() )
	{
		particlesToEmit += getRate() * dt;
	}

	while( particlesToEmit >= 1.0f && n < maxNumParticles )
	{
		// Die Partikel werden zufllig zwischen der aktuellen und der letzten Emitterposition verteilt.
		float sample = randFloat( 0.0f, 1.0f );

		Particle p;

		p.timeToLive = settings.partTimeToLive;
		if( settings.partTimeToLiveVar > 0.0f ) p.timeToLive += randFloat( -settings.partTimeToLiveVar, settings.partTimeToLiveVar );

		p.position = lastPosition + sample * ( currentPosition - lastPosition );
		if( settings.partPositionRadius > 0.0f ) p.position += randVec2fDir() * randFloat( 0.0f, settings.partPositionRadius );

		p.angle = lastAngle + sample * ( currentAngle - lastAngle ) + settings.partAngle;
		if( settings.partAngleVar > 0.0f ) p.angle += randFloat( -settings.partAngleVar, settings.partAngleVar );

		float dir = lastAngle + sample * ( currentAngle - lastAngle ) + settings.partDirection;
		if( settings.partDirectionVar > 0.0f ) dir += randFloat( -settings.partDirectionVar, settings.partDirectionVar );
		float speed = settings.partSpeed;
		if( settings.partSpeedVar > 0.0f ) speed += randFloat( -settings.partSpeedVar, settings.partSpeedVar );
		p.velocity = speed * Vec2f( cosf( dir ), sinf( dir ) );
		if( settings.useRelativeVelocity ) p.velocity += currentVelocity;

		p.angularVelocity = settings.partAngularVelocity;
		if( settings.partAngularVelocityVar > 0.0f ) p.angularVelocity += randFloat( -settings.partAngularVelocityVar, settings.partAngularVelocityVar );

		p.color = settings.partStartColor;
		if( !settings.partStartColorVar.isZero() )
		{
			p.color.r += randFloat( -settings.partStartColorVar.r, settings.partStartColorVar.r );
			p.color.g += randFloat( -settings.partStartColorVar.g, settings.partStartColorVar.g );
			p.color.b += randFloat( -settings.partStartColorVar.b, settings.partStartColorVar.b );
			p.color.a += randFloat( -settings.partStartColorVar.a, settings.partStartColorVar.a );
		}

		Vec4f endColor( settings.partEndColor );
		if( !settings.partEndColorVar.isZero() )
		{
			endColor.r += randFloat( -settings.partEndColorVar.r, settings.partEndColorVar.r );
			endColor.g += randFloat( -settings.partEndColorVar.g, settings.partEndColorVar.g );
			endColor.b += randFloat( -settings.partEndColorVar.b, settings.partEndColorVar.b );
			endColor.a += randFloat( -settings.partEndColorVar.a, settings.partEndColorVar.a );
		}

		p.deltaColor = ( endColor - p.color ) / p.timeToLive;

		p.size = settings.partStartSize;
		if( settings.partStartSizeVar > 0.0f ) p.size += randFloat( -settings.partStartSizeVar, settings.partStartSizeVar );

		float endSize = settings.partEndSize;
		if( settings.partEndSizeVar > 0.0f ) endSize += randFloat( -settings.partEndSizeVar, settings.partEndSizeVar );

		p.deltaSize = ( endSize - p.size ) / p.timeToLive;

		p.friction = settings.partFriction;
		if( settings.partFrictionVar ) p.friction += randFloat( -settings.partFrictionVar, settings.partFrictionVar );

		p.gravity = settings.partGravity;
		if( settings.partGravityVar ) p.gravity += randFloat( -settings.partGravityVar, settings.partGravityVar );

		p.animFrame = settings.partAnimFrame;
		if( settings.partAnimFrameVar ) p.animFrame += randFloat( -settings.partAnimFrameVar, settings.partAnimFrameVar );

		unsigned int styleIndex = 0;
		if( settings.partStyles.size() > 1 ) styleIndex = randInt( 0, static_cast<int>( settings.partStyles.size() - 1 ) );
		ps.addParticle( settings.partStyles[styleIndex], p );

		particlesToEmit -= 1.0f;
		n++;
	}

	lastPosition = currentPosition;
	lastAngle = currentAngle;

	return n;
}