#include <PCH.h>
#include <Sound.h>
#include <SoundInstance.h>
#include <Audio.h>
#include <Engine.h>
#include <fmod.hpp>
#include <fmod_errors.h>

Sound::Sound( ResourceManager& manager,
			  const std::string& filename,
			  bool unique,
			  const ParameterMap& options ) : Resource( manager, filename, unique, options ),
											  audio( manager.getEngine().getAudio() ),
											  p_sound( 0 ),
											  p_defaultSoundGroup( 0 ),
											  instanceCreationInterval( -1.0f ),
											  lastInstanceCreation( -FLT_MAX ),
											  maxPlayingInstances( ~0 ),
											  nativeFrequency( 0.0f )
{
	if( audio.getNoSoundMode() ) return;

	FMOD_MODE soundMode = audio.getHardwareMode() ? FMOD_HARDWARE : FMOD_SOFTWARE;

	// in den Optionen schauen, ob es ein Stream sein soll
	const std::string type = options.get<std::string>( "Type", "Sample" );
	if( type == "Stream" ) soundMode |= FMOD_CREATESTREAM;
	else if( type == "Sample" ) soundMode |= FMOD_CREATESAMPLE;
	else
	{
		setError( 1 );
		audio.getEngine().getLogger().error( filename + ": Unbekannter Typ: " + type, LOGGER_AUTO );
		return;
	}

	// 2D oder 3D?
	const std::string mode = options.get<std::string>( "Mode", "3D" );
	if( mode == "2D" ) soundMode |= FMOD_2D;
	else if( mode == "3D" ) soundMode |= FMOD_3D;
	else
	{
		setError( 2 );
		audio.getEngine().getLogger().error( filename + ": Unbekannter Modus: " + mode, LOGGER_AUTO );
		return;
	}

	FMOD_RESULT r = audio.p_system->createSound( filename.c_str(), soundMode, 0, &p_sound );
	if( r )
	{
		setError( 3 );
		std::ostringstream msg;
		msg << filename << ": FMOD-Sound konnte nicht erzeugt werden: " << FMOD_ErrorString( r );
		audio.getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	// minimale und maximale Hrreichweite des Sounds setzen
	setDefaultDistances( 15.0f, 10000.0f );

	// native Frequenz holen
	p_sound->getDefaults( &nativeFrequency, 0, 0, 0 );

	int channels, bits;
	p_sound->getFormat( 0, 0, &channels, &bits );
	std::ostringstream msg;
	msg << filename << ": " << channels << " Kanle, " << bits << " Bits pro Sample, " << static_cast<int>(nativeFrequency) << " Hz";
	audio.getEngine().getLogger().finer( msg.str(), LOGGER_AUTO );

	setDefaultSoundGroup( audio.getMasterSoundGroup() );
}

Sound::~Sound()
{
	if( p_sound )
	{
		p_sound->release();
	}
}

SoundInstance* Sound::createInstance( bool forceCreation )
{
	if( !p_sound ) return 0;

	const float time = audio.getEngine().getTimeInSeconds();
	const float dt = time - lastInstanceCreation;

	if( forceCreation ||
		( dt >= instanceCreationInterval &&
		  getNumInstances() < maxPlayingInstances ) )
	{
		SoundInstance* p_instance = new SoundInstance( audio, *this );
		if( p_instance->getError() )
		{
			std::ostringstream msg;
			msg << getFilename() << ": " << "Konnte keine Instanz erzeugen";
			audio.getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
			delete p_instance;
			return 0;
		}

		instances.insert( p_instance );
		lastInstanceCreation = time;

		return p_instance;
	}
	else
	{
		return 0;
	}
}

bool Sound::isInstanceValid( SoundInstance* p_soundInstance ) const
{
	std::set<SoundInstance*>::const_iterator it = instances.find( p_soundInstance );
	return it != instances.end();
}

void Sound::update()
{
	// Liste der Instanzen durchgehen und update aufrufen,
	// auerdem prfen, ob sie gelscht werden drfen (und dies dann auch tun)
	for( std::set<SoundInstance*>::iterator it = instances.begin();
		 it != instances.end(); )
	{
		(*it)->update();

		if( (*it)->isToBeDeleted() )
		{
			delete *it;
			it = instances.erase( it );
		}
		else
		{
			it++;
		}
	}
}

void Sound::destroyAll()
{
	// alle Instanzen lschen
	for( std::set<SoundInstance*>::iterator it = instances.begin();
		 it != instances.end();
		 it++ )
	{
		delete *it;
	}

	instances.clear();
}

void Sound::setDefaultSoundGroup( SoundGroup* p_soundGroup )
{
	if( !p_soundGroup ) p_soundGroup = audio.getMasterSoundGroup();
	p_defaultSoundGroup = p_soundGroup;
}

void Sound::setDefaultSoundGroupName( const std::string& soundGroup )
{
	setDefaultSoundGroup( audio.getSoundGroupByName( soundGroup ) );
}

SoundGroup* Sound::getDefaultSoundGroup() const
{
	return p_defaultSoundGroup;
}

int Sound::setDefaultDistances( float minDist,
							    float maxDist )
{
	if( !p_sound ) return 0;

	FMOD_RESULT r = p_sound->set3DMinMaxDistance( minDist, maxDist );
	if( r ) return 1;

	return 0;
}

int Sound::getDefaultDistances( float* p_minDistOut,
							    float* p_maxDistOut ) const
{
	if( !p_sound ) return 0;

	FMOD_RESULT r = p_sound->get3DMinMaxDistance( p_minDistOut, p_maxDistOut );
	if( r ) return 1;

	return 0;
}

void Sound::setInstanceCreationInterval( float instanceCreationInterval )
{
	this->instanceCreationInterval = instanceCreationInterval;
}

float Sound::getInstanceCreationInterval() const
{
	return instanceCreationInterval;
}

void Sound::setMaxPlayingInstances( unsigned int maxPlayingInstances )
{
	this->maxPlayingInstances = maxPlayingInstances;
}

unsigned int Sound::getMaxPlayingInstances() const
{
	return maxPlayingInstances;
}

float Sound::getNativeFrequency() const
{
	return nativeFrequency;
}

unsigned int Sound::getNumInstances() const
{
	return static_cast<unsigned int>( instances.size() );
}