#include <PCH.h>
#include <SoundGroup.h>
#include <SoundInstance.h>
#include <Audio.h>
#include <Engine.h>
#include <fmod.hpp>
#include <fmod_errors.h>

SoundGroup::SoundGroup( Audio& audio,
					    FMOD::ChannelGroup* p_masterGroup ) : SoundControllerTarget( audio ),
															  audio( audio ),
															  p_channelGroup( p_masterGroup ),
															  name( "Master" ),
															  p_parent( 0 ),
															  isMaster( true ),
															  flags( 0 )
{
	std::ostringstream msg;
	msg << "Soundgruppe " << name << " wurde erzeugt";
	audio.getEngine().getLogger().fine( msg.str(), LOGGER_AUTO );
}

SoundGroup::SoundGroup( Audio& audio,
					    const std::string& name,
						SoundGroup* p_parent ) : SoundControllerTarget( audio ),
												 audio( audio ),
												 p_channelGroup( 0 ),
												 name( name ),
												 p_parent( p_parent ),
												 isMaster( false ),
												 flags( 0 )
{
	assert( p_parent );

	// FMOD-Channel-Group erzeugen
	FMOD_RESULT r = audio.p_system->createChannelGroup( name.c_str(), &p_channelGroup );
	if( r )
	{
		setError( 1 );
		std::ostringstream msg;
		msg << "FMOD-Channel-Group konnte nicht erzeugt werden: " << FMOD_ErrorString( r );
		audio.getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	std::ostringstream msg;
	msg << "Soundgruppe " << name << " wurde erzeugt";
	audio.getEngine().getLogger().fine( msg.str(), LOGGER_AUTO );

	// die neue Channel-Group zur gewhlten Parent-Group hinzufgen
	p_parent->p_channelGroup->addGroup( p_channelGroup );

	// in die Kinderliste einfgen
	p_parent->children.insert( this );
}

SoundGroup::~SoundGroup()
{
	std::ostringstream msg;
	msg << "Soundgruppe " << name << " wurde zerstrt";
	audio.getEngine().getLogger().fine( msg.str(), LOGGER_AUTO );

	if( !isMaster )
	{
		p_channelGroup->release();

		SoundGroup* p_masterSoundGroup = audio.getMasterSoundGroup();

		// Die Kinder und die Sound-Instanzen dieser Gruppe gehren jetzt zur Master-Gruppe.
		for( std::set<SoundGroup*>::const_iterator it = children.begin();
			 it != children.end();
			 it++ )
		{
			(*it)->p_parent = p_masterSoundGroup;
			p_masterSoundGroup->children.insert( *it );
		}

		while( !soundInstances.empty() )
		{
			std::set<SoundInstance*>::iterator it = soundInstances.begin();
			( *it )->setSoundGroup( p_masterSoundGroup );
		}

		// alle Sounds neu auswerten
		p_masterSoundGroup->evaluateAll();

		// aus der Kinderliste entfernen
		p_parent->children.erase( this );
	}
}

int SoundGroup::stop()
{
	FMOD_RESULT r = p_channelGroup->stop();
	if( r ) return 1;

	return 0;
}

int SoundGroup::setPaused( bool paused )
{
	if( paused ) flags |= 1;
	else flags &= ~1;

	evaluateAll();

	return 0;
}

int SoundGroup::setMuted( bool muted )
{
	if( muted ) flags |= 2;
	else flags &= ~2;

	evaluateAll();

	return 0;
}

int SoundGroup::setVolume( float volume )
{
	FMOD_RESULT r = p_channelGroup->setVolume( volume );
	if( r ) return 1;

	return 0;
}

int SoundGroup::setPitch( float pitch )
{
	FMOD_RESULT r = p_channelGroup->setPitch( pitch );
	if( r ) return 1;

	return 0;
}

bool SoundGroup::isPaused() const
{
	return ( flags & 1 ) ? true : false;
}

bool SoundGroup::isMuted() const
{
	return ( flags & 2 ) ? true : false;
}

float SoundGroup::getVolume() const
{
	float volume;
	p_channelGroup->getVolume( &volume );

	return volume;
}

float SoundGroup::getPitch() const
{
	float pitch;
	p_channelGroup->getPitch( &pitch );

	return pitch;
}

const Audio& SoundGroup::getAudio() const
{
	return audio;
}

const std::string& SoundGroup::getName() const
{
	return name;
}

SoundGroup* SoundGroup::getParent() const
{
	return p_parent;
}

void SoundGroup::evaluateAll()
{
	for( std::set<SoundInstance*>::const_iterator it = soundInstances.begin();
		 it != soundInstances.end();
		 it++ )
	{
		(*it)->evaluate();
	}

	for( std::set<SoundGroup*>::const_iterator it = children.begin();
		 it != children.end();
		 it++ )
	{
		(*it)->evaluateAll();
	}
}