#include <PCH.H>
#include <SoundInstance.h>
#include <SoundGroup.h>
#include <Engine.h>
#include <Audio.h>
#include <fmod.hpp>
#include <fmod_errors.h>

SoundInstance::SoundInstance( Audio& audio,
							  Sound& sound ) : SoundControllerTarget( audio ),
											   audio( audio ),
											   sound( &sound ),
											   p_channel( 0 ),
											   p_soundGroup( 0 ),
											   flags( 0 ),
											   z( 0.0f ),
											   calculateVelocity( true ),
											   lastPosition( FLT_MAX, FLT_MAX )
{
	// Channel holen, aber noch nicht abspielen
	FMOD_RESULT r = audio.p_system->playSound( FMOD_CHANNEL_FREE, sound.p_sound, true, &p_channel );
	if( r )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << sound.getFilename() << ": Instanz kann nicht abgespielt werden: " << FMOD_ErrorString( r );
		audio.getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Standard-Soundgruppe setzen
	setSoundGroup( sound.p_defaultSoundGroup );

	// Pause
	setPaused( true );
}

SoundInstance::~SoundInstance()
{
	// Channel freigeben
	stop();

	p_soundGroup->soundInstances.erase( this );
}

void SoundInstance::update()
{
	SoundControllerTarget::update();

	FMOD_MODE mode;
	p_channel->getMode( &mode );
	if( mode & FMOD_3D )
	{
		// aktuelle Position und Geschwindigkeit berechnen
		const Vec2f currentPosition( getAbsPosition() );
		if( lastPosition.x == FLT_MAX ) lastPosition = currentPosition;
		const Vec2f currentVelocity = calculateVelocity ? ( ( currentPosition - lastPosition ) / audio.getEngine().getTickDurationInSeconds() ) : Vec2f( 0.0f, 0.0f );

		// an FMOD bermitteln
		FMOD_VECTOR pos = { currentPosition.x, currentPosition.y, z };
		FMOD_VECTOR vel = { currentVelocity.x, currentVelocity.y, 0.0f };
		p_channel->set3DAttributes( &pos, calculateVelocity ? &vel : 0 );

		lastPosition = currentPosition;
	}
}

int SoundInstance::play()
{
	update();

	// Abspielposition an den Beginn setzen
	FMOD_RESULT r = p_channel->setPosition( 0, FMOD_TIMEUNIT_MS );
	if( r ) return 1;

	// Sound weiterlaufen lassen
	setPaused( false );

	return 0;
}

int SoundInstance::play( bool looping )
{
	update();

	// Abspielposition an den Beginn setzen
	FMOD_RESULT r = p_channel->setPosition( 0, FMOD_TIMEUNIT_MS );
	if( r ) return 1;

	// Looping einrichten
	setLooping( looping );

	// Sound weiterlaufen lassen
	setPaused( false );

	return 0;
}

int SoundInstance::stop()
{
	FMOD_RESULT r = p_channel->stop();
	if( r ) return 1;

	return 0;
}

int SoundInstance::setLooping( bool looping )
{
	FMOD_MODE mode = looping ? FMOD_LOOP_NORMAL : FMOD_LOOP_OFF;
	FMOD_RESULT r;

	if( isPlaying() )
	{
		// erst pausieren, dann Modus ndern, anschlieend weiterspielen
		setPaused( true );
		r = p_channel->setMode( mode );
		setPaused( false );
	}
	else
	{
		r = p_channel->setMode( mode );
	}

	if( r ) return 1;

	return 0;
}

int SoundInstance::setPaused( bool paused )
{
	if( paused ) flags |= 1;
	else flags &= ~1;

	evaluate();

	return 0;
}

int SoundInstance::setMuted( bool muted )
{
	if( muted ) flags |= 2;
	else flags &= ~2;

	evaluate();

	return 0;
}

int SoundInstance::setVolume( float volume )
{
	FMOD_RESULT r = p_channel->setVolume( volume );
	if( r ) return 1;

	return 0;
}

int SoundInstance::setPitch( float pitch )
{
	// Die Hhe bezieht sich auf die native Frequenz des Sounds.
	FMOD_RESULT r = p_channel->setFrequency( sound->getNativeFrequency() * pitch );
	if( r ) return 1;

	return 0;
}

int SoundInstance::setPlayCursor( float seconds )
{
	FMOD_RESULT r = p_channel->setPosition( static_cast<unsigned int>( seconds * 1000.0f ), FMOD_TIMEUNIT_MS );
	if( r ) return 1;

	return 0;
}

int SoundInstance::setSoundGroup( SoundGroup* p_newSoundGroup )
{
	if( !p_newSoundGroup ) p_newSoundGroup = audio.getMasterSoundGroup();

	if( p_soundGroup == p_newSoundGroup ) return 0;

	if( p_soundGroup )
	{
		// Instanz aus der alten Gruppe entfernen
		p_soundGroup->soundInstances.erase( this );
	}

	// der neuen Gruppe hinzufgen
	if( p_channel->setChannelGroup( p_newSoundGroup->p_channelGroup ) )
	{
		return 1;
	}

	p_soundGroup = p_newSoundGroup;
	p_soundGroup->soundInstances.insert( this );

	evaluate();

	return 0;
}

int SoundInstance::setDistances( float minDist,
								 float maxDist )
{
	FMOD_RESULT r = p_channel->set3DMinMaxDistance( minDist, maxDist );
	if( r ) return 1;

	return 0;
}

bool SoundInstance::isLooping() const
{
	FMOD_MODE mode;
	p_channel->getMode( &mode );
	return ( mode & FMOD_LOOP_NORMAL ) ? true : false;
}

bool SoundInstance::isPaused() const
{
	return ( flags & 1 ) ? true : false;
}

bool SoundInstance::isMuted() const
{
	return ( flags & 2 ) ? true : false;
}

bool SoundInstance::isPlaying() const
{
	return !isPaused();
}

bool SoundInstance::isToBeDeleted() const
{
	bool playing;
	p_channel->isPlaying( &playing );

	return !playing;
}

float SoundInstance::getVolume() const
{
	float volume;
	p_channel->getVolume( &volume );

	return volume;
}

float SoundInstance::getPitch() const
{
	float frequency;
	p_channel->getFrequency( &frequency );

	// Die Hhe bezieht sich auf die native Frequenz des Sounds.
	return frequency / sound->getNativeFrequency();
}

float SoundInstance::getPlayCursor() const
{
	unsigned int playCursor;
	FMOD_RESULT r = p_channel->getPosition( &playCursor, FMOD_TIMEUNIT_MS );
	if( r ) return -1.0f;
	return static_cast<float>( playCursor ) / 1000.0f;
}

SoundGroup* SoundInstance::getSoundGroup() const
{
	return p_soundGroup;
}

int SoundInstance::getDistances( float* p_minDistOut,
								 float* p_maxDistOut ) const
{
	FMOD_RESULT r = p_channel->get3DMinMaxDistance( p_minDistOut, p_maxDistOut );
	if( r ) return 1;

	return 0;
}

void SoundInstance::setZ( float z )
{
	this->z = z;
}

float SoundInstance::getZ() const
{
	return z;
}

void SoundInstance::setCalculateVelocity( bool calculateVelocity )
{
	this->calculateVelocity = calculateVelocity;
}

bool SoundInstance::getCalculateVelocity() const
{
	return calculateVelocity;
}

SoundHandle SoundInstance::getSound() const
{
	return sound;
}

void SoundInstance::evaluate()
{
	unsigned int effectiveFlags = flags;

	SoundGroup* p_group = p_soundGroup;
	while( p_group )
	{
		effectiveFlags |= p_group->flags;
		p_group = p_group->p_parent;
	}

	p_channel->setPaused( ( effectiveFlags & 1 ) ? true : false );
	p_channel->setMute( ( effectiveFlags & 2 ) ? true : false );
}