#include <PCH.h>
#include <Texture.h>
#include <Engine.h>
#include <SDL.h>
#include <SDL_OpenGL.h>
#include <SILLY/SILLY.h>

Texture::Texture( ResourceManager& manager,
				  const std::string& filename,
				  bool unique,
				  const ParameterMap& options) : Resource( manager, filename, unique, options ),
												 texID( 0 ),
												 p_surface( 0 )
{
	Logger& logger = manager.getEngine().getLogger();

	// Datei ffnen
	FILE* p_file = fopen( filename.c_str(), "rb" );
	if( !p_file )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << filename << ": Konnte nicht geffnet werden";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	fseek( p_file, 0, SEEK_END );
	size_t fileSize = ftell( p_file );
	fseek( p_file, 0, SEEK_SET );
	SILLY::byte* p_data = new SILLY::byte[fileSize];
	if( fread( p_data, 1, fileSize, p_file ) != fileSize )
	{
		// Fehler!
		setError( 2 );
		std::ostringstream msg;
		msg << filename << ": Konnte nicht gelesen werden";
		logger.error( msg.str(), LOGGER_AUTO );
		fclose( p_file );
		delete[] p_data;
		return;
	}

	fclose( p_file );

	// Speicherquelle fr SILLY erzeugen
	SILLY::MemoryDataSource source( p_data, fileSize );

	// Header laden
	SILLY::Image img( source );
	if( !img.loadImageHeader() )
	{
		// Fehler!
		setError( 3 );
		std::ostringstream msg;
		msg << filename << ": Nicht untersttztes Format";
		logger.error( msg.str(), LOGGER_AUTO );
		delete[] p_data;
		return;
	}

	// Bildinhalt laden
	if( !img.loadImageData( SILLY::PF_RGBA, SILLY::PO_TOP_LEFT ) )
	{
		// Fehler!
		setError( 4 );
		std::ostringstream msg;
		msg << filename << ": Bildinhalt konnte nicht geladen werden";
		logger.error( msg.str(), LOGGER_AUTO );
		delete[] p_data;
		return;
	}

	// Speicher freigeben
	delete[] p_data;

	// Gre des Bilds kopieren
	size = Vec2i( static_cast<int>( img.getWidth() ), static_cast<int>( img.getHeight() ) );

	// Hatte das Originalbild einen Alphakanal?
	SILLY::PixelFormat fmt = img.getSourcePixelFormat();
	bool hadAlphaChannel = fmt == SILLY::PF_RGBA || fmt == SILLY::PF_A1B5G5R5;

	std::ostringstream msg;
	msg << filename << ": " << size.x << "x" << size.y << " (" << img.getLoaderIdentifierString() << ")";
	logger.finer( msg.str(), LOGGER_AUTO );

	// Oberflche mit dem richtigen Pixelformat erzeugen
	p_surface = SDL_CreateRGBSurface( SDL_SWSURFACE, size.x, size.y, 32,
									  0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000 );
	if( !p_surface )
	{
		setError( 5 );
		std::ostringstream msg;
		msg << filename << ": Erzeugen der SDL-Oberflche fehlgeschlagen";
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Inhalt hinein kopieren
	SDL_LockSurface( p_surface );
	memcpy( p_surface->pixels, img.getPixelsDataPtr(), img.getPixelsDataSize() );
	SDL_UnlockSurface( p_surface );

	if( options.get<bool>( "UseForBitmapFont", false ) && !hadAlphaChannel )
	{
		// Rot-Kanal in Alpha-Kanal kopieren, Farbkanle auf Wei setzen
		prepareForBitmapFont();
	}

	bool useMIPMapping = manager.getEngine().getOptions().get<bool>( "Video.UseMIPMapping", true );
	useMIPMapping = options.get<bool>( "UseMIPMapping", useMIPMapping );

	// OpenGL-Textur erzeugen
	glGenTextures( 1, &texID );
	if( !texID )
	{
		setError( 6 );
		std::ostringstream msg;
		msg << filename << ": OpenGL-Textur konnte nicht generiert werden: " << glGetError();
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	glPushAttrib( GL_TEXTURE_BIT | GL_TRANSFORM_BIT );

	// OpenGL-Textur einrichten
	glBindTexture( GL_TEXTURE_2D, texID );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, useMIPMapping ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	// Bilddaten hineinkopieren
	SDL_LockSurface( p_surface );
	glPixelStorei( GL_UNPACK_ROW_LENGTH, p_surface->pitch / p_surface->format->BytesPerPixel );
	if( useMIPMapping ) gluBuild2DMipmaps( GL_TEXTURE_2D, 4, p_surface->w, p_surface->h, GL_RGBA, GL_UNSIGNED_BYTE, p_surface->pixels );
	else glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, p_surface->w, p_surface->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_surface->pixels );
	glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
	SDL_UnlockSurface( p_surface );

	// Matrix erzeugen
	glMatrixMode( GL_TEXTURE );
	glPushMatrix();
	glLoadIdentity();
	float w = static_cast<float>( p_surface->w ), h = static_cast<float>( p_surface->h );
	glScalef( 1.0f / w, 1.0f / h, 1.0f );
	glGetFloatv( GL_TEXTURE_MATRIX, matrix );
	glPopMatrix();

	glPopAttrib();
}

Texture::Texture( ResourceManager& manager,
				  const Vec2i& size,
				  int glFormat ) : Resource( manager, "", true, ParameterMap() ),
								   texID( 0 ),
								   p_surface( 0 ),
								   size( size )
{
	Logger& logger = manager.getEngine().getLogger();

	// OpenGL-Textur erzeugen
	glGenTextures( 1, &texID );
	if( !texID )
	{
		setError( 1 );
		std::ostringstream msg;
		msg << getFilename() << ": OpenGL-Textur konnte nicht generiert werden: " << glGetError();
		logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	std::ostringstream msg;
	msg << getFilename() << ": " << size.x << "x" << size.y;
	logger.finer( msg.str(), LOGGER_AUTO );

	glPushAttrib( GL_TEXTURE_BIT | GL_TRANSFORM_BIT );

	// OpenGL-Textur einrichten
	glBindTexture( GL_TEXTURE_2D, texID );
	glTexImage2D( GL_TEXTURE_2D, 0, glFormat, size.x, size.y, 0, glFormat, GL_UNSIGNED_BYTE, 0 );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	// Matrix erzeugen
	glMatrixMode( GL_TEXTURE );
	glPushMatrix();
	glLoadIdentity();
	float w = static_cast<float>( size.x ), h = static_cast<float>( size.y );
	glScalef( 1.0f / w, 1.0f / h, 1.0f );
	glGetFloatv( GL_TEXTURE_MATRIX, matrix );
	glPopMatrix();

	glPopAttrib();
}

Texture::~Texture()
{
	if( p_surface )
	{
		// SDL-Oberflche freigeben
		SDL_FreeSurface( p_surface );
	}

	if( texID )
	{
		// OpenGL-Textur freigeben
		glDeleteTextures( 1, &texID );
	}
}

const Vec2i& Texture::getSize() const
{
	return size;
}

void Texture::activate()
{
	// Textur aktivieren
	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, texID );

	// Pixel-Texturkoordinaten
	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_TEXTURE );
	glLoadMatrixf( matrix );
	glPopAttrib();
}

void Texture::prepareForBitmapFont()
{
	assert( p_surface );

	SDL_LockSurface( p_surface );

	unsigned char* p_pixels = static_cast<unsigned char*>( p_surface->pixels );
	for( int x = 0; x < p_surface->w; x++ )
	{
		for( int y = 0; y < p_surface->h; y++ )
		{
			// Pixel holen
			unsigned char* p_pixel = p_pixels + y * p_surface->pitch + 4 * x;

			// Rot nach Alpha kopieren
			p_pixel[3] = p_pixel[0];

			// Rot, Grn und Blau auf 255 setzen
			p_pixel[0] = p_pixel[1] = p_pixel[2] = 255;
		}
	}

	SDL_UnlockSurface( p_surface );
}