#include <PCH.h>
#include <Video.h>
#include <Engine.h>
#include <TextureManager.h>
#include <ParticleSystem.h>
#include <Utils.h>
#include <SDL.h>
#include <SDL_OpenGL.h>
#include <SILLY/SILLY.h>

Video::Video( Engine& engine ) : Component( engine, "Video" ),
								 p_screen( 0 ),
								 lastFrameEnd( 0.0f ),
								 lastFrameDuration( 0.0f ),
								 mbFrameDone( false ),
								 mbIsCurrent( false )
{
	// Grafiksystem der SDL initialisieren
	int r = SDL_InitSubSystem( SDL_INIT_VIDEO );
	if( r )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << "Initialisieren des Video-Subsystems fehlgeschlagen: " << SDL_GetError();
		engine.getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Optionen abfragen - falls nicht vorhanden, Standardwerte benutzen
	const ParameterMap& options = engine.getOptions();
	resolution.x = options.get<unsigned int>( "Video.Width", 1024 );
	resolution.y = options.get<unsigned int>( "Video.Height", 768 );
	bool windowed = options.get<bool>( "Video.Windowed", false );
	std::string windowTitle = options.get<std::string>( "Video.WindowTitle", "SDL Window" );

	// Fenstertitel setzen und Cursor ausblenden
	SDL_WM_SetCaption( windowTitle.c_str(), 0 );
	SDL_ShowCursor( 0 );

	// OpenGL-Einstellungen vornehmen
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 32 );
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, options.get<bool>( "Video.VSync", true ) ? 1 : 0 );

	// Fenster und OpenGL-Kontext erzeugen
	p_screen = SDL_SetVideoMode( resolution.x, resolution.y, 32, SDL_OPENGL | ( windowed ? 0 : SDL_FULLSCREEN ) );
	if( !p_screen )
	{
		// Fehler!
		setError( 2 );
		std::ostringstream msg;
		msg << "Setzen des Video-Modus fehlgeschlagen: " << SDL_GetError();
		engine.getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	// SILLY initialisieren
	if( !SILLY::SILLYInit() )
	{
		// Fehler!
		setError( 3 );
		std::ostringstream msg;
		msg << "Initialisierung von SILLY fehlgeschlagen";
		engine.getLogger().error( msg.str(), LOGGER_AUTO );
		return;
	}

	// Standardeinstellungen fr Grafik setzen
	glShadeModel( GL_SMOOTH );
	glEnable( GL_POINT_SMOOTH );
	glEnable( GL_LINE_SMOOTH );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glHint( GL_POINT_SMOOTH_HINT, GL_NICEST );
	glHint( GL_LINE_SMOOTH_HINT, GL_NICEST );
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	// 2D-Projektionsmatrix setzen
	setupProjection( -resolution / 2, resolution / 2 );

	glMatrixMode( GL_MODELVIEW );
	resetCamera();
	setupMotionBlurring( false );

	engine.connectEngineShutDown( this, &Video::onEngineShutDown );
	engine.connectFrameBegin( this, &Video::onFrameBegin );
	engine.connectFrameEnd( this, &Video::onFrameEnd );

	engine.getLogger().fine( "Video-Komponente initialisiert", LOGGER_AUTO );
}

Video::~Video()
{
	// SILLY herunterfahren
	SILLY::SILLYCleanup();

	// SDL wird von der Engine komplett heruntergefahren.

	getEngine().getLogger().fine( "Video-Komponente heruntergefahren", LOGGER_AUTO );
}

void Video::onEngineShutDown()
{
	// Textur zurcksetzen, damit sie freigegeben werden kann
	currentTexture.reset();

	mbTexture = 0;
}

void Video::onFrameBegin()
{
	mbFrameDone = false;
}

void Video::onFrameEnd()
{
	float currentTime = getEngine().getTimeInSeconds();
	lastFrameDuration = lastFrameEnd == 0.0f ? 0.0f : currentTime - lastFrameEnd;
	lastFrameEnd = currentTime;

	doMotionBlurring();
	if( !mbFrameDone ) mbIsCurrent = false;

	// gerendertes Frame anzeigen
	SDL_GL_SwapBuffers();

	// OpenGL-Fehler aufgetreten?
	unsigned int error = glGetError();
	if( error != GL_NO_ERROR )
	{
		std::ostringstream msg;
		msg << "OpenGL-Fehler aufgetreten: " << error;
		getEngine().getLogger().error( msg.str(), LOGGER_AUTO );
	}
}

void Video::clear( const Vec3f& color )
{
	glClearColor( color.r, color.g, color.b, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT );
}

void Video::setupProjection( const Vec2f& topLeft,
							 const Vec2f& bottomRight )
{
	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluOrtho2D( topLeft.x, bottomRight.x, bottomRight.y, topLeft.y );
	glPopAttrib();

	// Matrix holen
	glGetFloatv( GL_PROJECTION_MATRIX, projectionMatrix );
}

void Video::setupCamera( const Vec2f& position,
						 float angle,
						 float scale )
{
	this->cameraPosition = position;
	this->cameraAngle = angle;
	this->cameraScale = scale;

	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glScalef( scale, scale, 1.0f );
	gluLookAt( position.x, position.y, 1.0, position.x, position.y, 0.0, -sinf( angle ), cosf( angle ), 0.0 );
	glPopAttrib();

	updateViewFrustum();
}

void Video::resetCamera()
{
	setupCamera( Vec2f( 0.0f, 0.0f ), 0.0f, 1.0f );
}

void Video::getCamera( Vec2f* p_outPosition,
					   float* p_outAngle,
					   float* p_outScale ) const
{
	if( p_outPosition ) *p_outPosition = cameraPosition;
	if( p_outAngle ) *p_outAngle = cameraAngle;
	if( p_outScale ) *p_outScale = cameraScale;
}

const Mat44f& Video::getCameraMatrix() const
{
	return cameraMatrix;
}

void Video::setupMotionBlurring( bool enable,
								 float decay )
{
	mbEnabled = enable;
	mbDecay = decay;

	if( enable && !mbTexture )
	{
		// Textur erzeugen
		const Vec2i size( nextPow2( p_screen->w ), nextPow2( p_screen->h ) );
		mbTexture = getEngine().getTextureManager().createTexture( size, GL_RGB );
	}
}

void Video::doMotionBlurring()
{
	if( !mbEnabled || mbFrameDone || mbDecay <= 0.0f ) return;

	const float f = powf( mbDecay, lastFrameDuration );
	if( f < 1.0f / 255.0f ) return;

	TextureHandle oldTexture = currentTexture;

	setTexture( mbTexture );

	if( mbIsCurrent )
	{
		// altes Bild ber das aktuelle Bild legen
		setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		drawFullScreenQuad( Vec4f( 1.0f, 1.0f, 1.0f, f ), Vec2i( 0, resolution.y ), Vec2i( resolution.x, 0 ) );
	}

	if( f < 1.0f )
	{
		// zusammengefgtes Bild wieder holen
		glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, resolution.x, resolution.y );
	}

	setTexture( oldTexture );

	mbFrameDone = true;
	mbIsCurrent = true;
}

const Vec2i& Video::getResolution() const
{
	return resolution;
}

TextureHandle Video::getTexture() const
{
	return currentTexture;
}

void Video::setTexture( TextureHandle texture,
					    bool force )
{
	if( !force && texture == currentTexture ) return;

	currentTexture = texture;

	if( currentTexture )
	{
		currentTexture->activate();
	}
	else
	{
		glDisable( GL_TEXTURE_2D );
		glPushAttrib( GL_TRANSFORM_BIT );
		glMatrixMode( GL_TEXTURE );
		glLoadIdentity();
		glPopAttrib();
	}
}

void Video::setupAlphaBlending( bool enable,
								unsigned int source,
								unsigned int destination )
{
	if( enable )
	{
		glEnable( GL_BLEND );
		glBlendFunc( source, destination );

		if( source == GL_SRC_ALPHA && ( destination == GL_ONE || destination == GL_ONE_MINUS_SRC_ALPHA ) )
		{
			glEnable( GL_ALPHA_TEST );
			glAlphaFunc( GL_GREATER, 0.0f );
		}
		else
		{
			glDisable( GL_ALPHA_TEST );
		}
	}
	else
	{
		glDisable( GL_BLEND );
		glDisable( GL_ALPHA_TEST );
	}
}

void Video::drawSprite( const Vec2i& positionOnTexture,
					    const Vec2i& sizeOnTexture,
						const Vec2f& sizeOnScreen )
{
	glBegin( GL_QUADS );
	glTexCoord2iv( positionOnTexture );
	glVertex2f( -sizeOnScreen.x * 0.5f, -sizeOnScreen.y * 0.5f );
	glTexCoord2i( positionOnTexture.x + sizeOnTexture.x, positionOnTexture.y );
	glVertex2f( sizeOnScreen.x * 0.5f, -sizeOnScreen.y * 0.5f );
	glTexCoord2iv( positionOnTexture + sizeOnTexture );
	glVertex2f( sizeOnScreen.x * 0.5f, sizeOnScreen.y * 0.5f );
	glTexCoord2i( positionOnTexture.x, positionOnTexture.y + sizeOnTexture.y );
	glVertex2f( -sizeOnScreen.x * 0.5f, sizeOnScreen.y * 0.5f );
	glEnd();
}

void Video::drawFullScreenQuad( const Vec4f& color,
							    const Vec2i& topLeftUV,
								const Vec2i& bottomRightUV )
{
	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();

	glBegin( GL_QUADS );
	glColor4fv( color );
	glTexCoord2i( topLeftUV.x, topLeftUV.y );
	glVertex2i( -1, 1 );
	glTexCoord2i( bottomRightUV.x, topLeftUV.y );
	glVertex2i( 1, 1 );
	glTexCoord2i( bottomRightUV.x, bottomRightUV.y );
	glVertex2i( 1, -1 );
	glTexCoord2i( topLeftUV.x, bottomRightUV.y );
	glVertex2i( -1, -1 );
	glEnd();

	glPopMatrix();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glPopAttrib();
}

bool Video::isCircleVisible( const Vec2f& center,
							 float radius ) const
{
	for( int i = 0; i < 4; i++ )
	{
		// Prfen, ob der fernste Punkt des Kreises auerhalb liegt. Falls ja, liegt das ganze Objekt auerhalb.
		const float d = frustumPlane[i].x * center.x + frustumPlane[i].y * center.y + frustumPlane[i].z;
		if( d < -radius ) return false;
	}

	return true;
}

bool Video::isCircleVisible( const Vec2f& center,
							 float radius,
							 const Mat33f& transformation ) const
{
	return isCircleVisible( transformation.transformPosition( center ), radius );
}

bool Video::isRectVisible( const Vec2f& boundMin,
						   const Vec2f& boundMax ) const
{
	for( int i = 0; i < 4; i++ )
	{
		// Prfen, ob der fernste Punkt des Rechtecks auerhalb liegt. Falls ja, liegt das ganze Objekt auerhalb.
		const Vec2f farPoint( frustumPlane[i].x > 0.0f ? boundMax.x : boundMin.x,
							  frustumPlane[i].y > 0.0f ? boundMax.y : boundMin.y );
		const float d = frustumPlane[i].x * farPoint.x + frustumPlane[i].y * farPoint.y + frustumPlane[i].z;
		if( d < 0.0f ) return false;
	}

	return true;
}

bool Video::isRectVisible( const Vec2f& boundMin,
						   const Vec2f& boundMax,
						   const Mat33f& transformation ) const
{
	const Vec2f p1( transformation.transformPosition( boundMin ) );
	const Vec2f p2( transformation.transformPosition( boundMax ) );
	const Vec2f p3( transformation.transformPosition( Vec2f( boundMin.x, boundMax.y ) ) );
	const Vec2f p4( transformation.transformPosition( Vec2f( boundMax.x, boundMin.y ) ) );
	const Vec2f worldBoundMin( min( min( p1.x, p2.x ), min( p3.x, p4.x ) ), min( min( p1.y, p2.y ), min( p3.y, p4.y ) ) );
	const Vec2f worldBoundMax( max( max( p1.x, p2.x ), max( p3.x, p4.x ) ), max( max( p1.y, p2.y ), max( p3.y, p4.y ) ) );

	for( int i = 0; i < 4; i++ )
	{
		// Prfen, ob der fernste Punkt des Rechtecks auerhalb liegt. Falls ja, liegt das ganze Objekt auerhalb.
		const Vec2f farPoint( frustumPlane[i].x > 0.0f ? worldBoundMax.x : worldBoundMin.x,
							  frustumPlane[i].y > 0.0f ? worldBoundMax.y : worldBoundMin.y );
		const float d = frustumPlane[i].x * farPoint.x + frustumPlane[i].y * farPoint.y + frustumPlane[i].z;
		if( d < 0.0f ) return false;
	}

	return true;
}

void Video::updateViewFrustum()
{
	// Kameramatrix holen
	glGetFloatv( GL_MODELVIEW_MATRIX, cameraMatrix );

	// Kameramatrix mit Projektionsmatrix multiplizieren
	const Mat44f transform( cameraMatrix * projectionMatrix );
	const float* m = transform;

	// die vier Ebenen berechnen
	frustumPlane[0] = Vec3f( m[3] + m[0], m[7] + m[4], m[15] + m[12] );	// links
	frustumPlane[1] = Vec3f( m[3] - m[0], m[7] - m[4], m[15] - m[12] );	// rechts
	frustumPlane[2] = Vec3f( m[3] - m[1], m[7] - m[5], m[15] - m[13] );	// oben
	frustumPlane[3] = Vec3f( m[3] + m[1], m[7] + m[5], m[15] + m[13] );	// unten

	// normalisieren
	for( int i = 0; i < 4; i++ )
	{
		frustumPlane[i] /= Vec2f( frustumPlane[i].x, frustumPlane[i].y ).length();
	}
}

float Video::getLastFrameDuration() const
{
	return lastFrameDuration;
}