#ifndef CONTROLLABLE_H
#define CONTROLLABLE_H

#include "NeuralNet.h"

enum ParameterType
{
	PT_INPUT,				// Eingabeparameter
	PT_OUTPUT,				// Ausgabeparameter (Schub, Lenkung ...)
	PT_OUTPUT_USER_ONLY,	// Ausgabeparameter, der nur bei der Benutzersteuerung bercksichtigt wird (keine KI)

	PT_LAST
};

enum ControlMode
{
	CM_DEAD,		// keine Eingaben
	CM_USER,		// Eingabe kommt vom Benutzer (ber die Tasten)
	CM_AI_NET,		// Eingabe kommt von der KI (ber neuronales Netz)
	CM_AI_SIMPLE,	// Eingabe kommt von der KI
	CM_PLAYBACK,	// Eingabe ist aufgezeichnet

	CM_LAST
};

class Controllable
{
public:
	Controllable( GameEnvironment& env );
	virtual ~Controllable();

protected:
	// registriert einen Ein- oder Ausgabeparameter
	void registerParameter( const std::string& name, ParameterType type, float minValue = 0.0f, float maxValue = 1.0f, float stdValue = 0.0f );

	// lscht einen Parameter
	void unregisterParameter( const std::string& name );

	// lscht alle Parameter
	void unregisterAllParameters();

	// setzt den Wert eines Parameters direkt
	void setParameter( const std::string& name, float value );

	// fragt den Wert eines Parameters ab
	float getParameter( const std::string& name ) const;

	// wird aufgerufen, wenn die Eingabeparameter bestimmt werden sollen
	virtual void generateInput();

	// wird aufgerufen, wenn die KI denken soll
	virtual void think();

	// erledigt die ganze Arbeit
	void updateControllable();

public:
	// schreibt alle Ein- und Ausgabeparameter in einen String
	std::string dumpParameters() const;

	// liefert den Kontrollmodus
	ControlMode getControlMode() const;

	// legt fest, dass dieses Objekt vom Benutzer ber die Tasten gesteuert werden soll
	void setUserMode();

	// bindet eine Taste an einen Parameter
	void bindKey( Key key, const std::string& parameter, float value );

	// legt fest, dass dieses Objekt durch die KI gesteuert werden soll
	void setAIMode();

	// fgt ein Verhaltensmuster (in Form eines neuronalen Netzes) hinzu
	int registerAIBehavior( const std::string& name, const std::string& filename );

	// legt das aktuelle Verhaltensmuster fr die KI fest
	int setAIBehavior( const std::string& name );

	// liefert das aktuelle Verhaltensmuster fr die KI
	std::string getAIBehavior() const;

	// setzt das Neuronale Netz direkt
	void setAINeuralNet( NeuralNetHandle neuralNet );

	// liefert das Neuronale Netz
	NeuralNetHandle getAINeuralNet() const;

	// setzt den simplen KI-Modus
	void setAISimpleMode();

	// legt fest, dass die Ausgabeparameter aus einer Datei gelesen werden sollen
	void setPlaybackMode( const std::string& filename );

	// zeichnet die Ausgabeparameter in eine Datei auf
	void startRecording( const std::string& filename );

	// stoppt die Aufzeichnung
	void stopRecording();

	// liefert true, wenn eine Aufzeichnung aktiv ist
	bool isRecording() const;

	// legt das Prfix fr die Trainingsdateien fest
	void setTrainingFilenamePrefix( const std::string& trainingFilenamePrefix );

	// beginnt eine Trainingsphase fr die KI
	void startTraining();

	// beendet eine Trainingsphase und schreibt sie fest, falls erwnscht
	void stopTraining( bool wasGood );

	// liefert true, wenn die KI am Trainieren ist
	bool isTraining() const;

private:
	struct Parameter
	{
		ParameterType type;
		float value;
		float minValue;
		float maxValue;
		float stdValue;
		int neuralMapping;
	};

	struct KeyBinding
	{
		Key key;
		std::string param;
		float value;
	};

	GameEnvironment& ctrEnv;
	std::map<std::string, Parameter> params;
	unsigned int numInputs;
	unsigned int numOutputs;
	ControlMode mode;
	std::list<KeyBinding> keyBindings;
	std::ifstream playback;
	bool recording;
	std::ofstream record;
	std::string trainingFilenamePrefix;
	bool training;
	std::vector<float> currentTraining;
	std::map<std::string, NeuralNetHandle> behaviors;
	std::string currentBehaviorName;
	NeuralNetHandle currentBehavior;
};

#endif