#ifndef GAMEOBJECT_H
#define GAMEOBJECT_H

#include "Level.h"
#include "CartoonRenderer.h"

enum CollisionBit
{
	CB_NONE			= 0x0,
	CB_STATIC		= 0x1,
	CB_PLAYER		= 0x2,
	CB_ENEMY		= 0x4,
	CB_PROJECTILE	= 0x8,
	CB_LASER		= 0x10,
	CB_OBJECT		= 0x20,
	CB_BLOCK_VIEW	= 0x40,
	CB_ALL			= ~0
};

enum RenderPass
{
	RP_BACKGROUND_1,
	RP_BACKGROUND_2,
	RP_BACKGROUND_3,
	RP_SHADOW,
	RP_OBJECT_1,
	RP_OBJECT_2,
	RP_FOREGROUND,
	RP_GLOW,
	RP_DEBUG_INFO
};

enum DebrisType
{
	DT_RANDOM,
	DT_NONE,
	DT_FIRST,

	DT_ENEMY_ERASER,
	DT_ENEMY_AASTATION,
	DT_PLAYER,
	DT_ENERGYNODE,

	DT_LAST
};

class GameObject : public Placeable, public Base
{
public:
	GameObject( Level& level, const TiXmlElement* p_root, bool createBody, const std::string& overrideID = "" );
	virtual ~GameObject();

	void setTransform( const Mat33f& transform );
	Mat33f getTransform() const;

	// liefert den Typ dieses Objekts als String
	virtual std::string getType() const = 0;

	// aktualisiert das Objekt (Steuerung des Spielers, KI, ...)
	virtual void update();

	void updatePhysics();

	// rendert das Objekt
	virtual void render( RenderPass pass );

	// wird aufgerufen, wenn das Objekt mit einem anderen Objekt kollidiert
	virtual void onCollision( GameObject* p_other, CollisionInfo& info );

	// wird aufgerufen, wenn das Objekt zerstrt werden soll
	virtual void onKill();

	// wird aufgerufen, wenn das Objekt wiedergeboren werden soll
	virtual void onRebirth();

	// liefert true, wenn das Objekt vom Level gelscht werden soll
	bool isToBeDeleted() const;

	// hier fgt das Objekt seine Gefahrzonen in den Level ein
	virtual void addAvoidanceAreas();

	// setzt die Geschwindigkeit des Objekts
	void setVelocity( const Vec2f& velocity );

	// liefert die Geschwindigkeit des Objekts
	Vec2f getVelocity() const;

	// setzt die Winkelgeschwindigkeit des Objekts
	void setAngularVelocity( float angularVelocity );

	// liefert die Winkelgeschwindigkeit des Objekts
	float getAngularVelocity() const;

	// fgt ein Liniensegment zu den Shapes hinzu
	void addSegment( const Vec2f& from, const Vec2f& to, float radius );

	// fgt einen Linienzug zu den Shapes hinzu
	void addPolyLine( const std::vector<Vec2f>& vertices, bool close, float radius, float shift = 0.0f );

	// fgt einen Kreis zu den Shapes hinzu
	void addCircle( const Vec2f& position, float radius );

	// passt die Gesamtmasse an
	void setTotalMass( float totalMass );

	// liefert die Gesamtmasse
	float getTotalMass() const;

	// setzt die Kollisions-Bits
	void setCollisionBits( unsigned long categoryBits, unsigned long collideBits );
	void setCategoryBits( unsigned long categoryBits );
	void setCollideBits( unsigned long collideBits );
	unsigned long getCategoryBits() const;
	unsigned long getCollideBits() const;

	// setzt die Reibung
	void setFriction( float movementFriction, float rotationFriction );
	void setMovementFriction( float movementFriction );
	void setRotationFriction( float rotationFriction );
	float getMovementFriction() const;
	float getRotationFriction() const;

	// legt fest, ob das Objekt von der Gravitation beeinflusst wird
	void setGravityInfluence( bool gravityInfluence );

	// liefert true, wenn das Objekt von der Gravitation beeinflusst wird
	bool getGravityInfluence() const;

	// legt das Material des Objekts fest
	void setMaterial( const std::string& material );

	// liefert das Material des Objekts
	const std::string& getMaterial() const;

	// setzt das Zielobjekt
	void setTargetID( const std::string& targetID );

	// liefert die ID des Zielobjekts
	const std::string& getTargetID() const;

	// setzt das Zielobjekt
	void setTarget( GameObject* p_target );

	// liefert das Zielobjekt
	GameObject* getTarget() const;

	// wird aufgerufen, wenn das Objekt mit einem anderen Objekt kollidiert
	void processCollision( GameObject* p_other, CollisionInfo& info );

	// schliet eine Kollision mit einem anderen Objekt aus oder aktiviert sie wieder
	void setCollisionEnabled( GameObject* p_other, bool enable );

	// liefert false, wenn das Objekt mit dem angegebenen Objekt nicht kollidieren mchte
	bool isCollisionEnabled( GameObject* p_other ) const;

	// liefert true, wenn das Objekt Laserstrahlen reflektieren kann
	virtual bool reflectsLaser() const;

	// liefert true, wenn dieses Objekt das angegebene Objekt sehen kann
	bool canSee( GameObject* p_other, float maxDistance = 25.0f, const Vec2f& relRayStartPosition = Vec2f( 0.0f, 0.0f ) ) const;
	bool canSee_lua( GameObject* p_other ) const { return canSee( p_other ); }

	// lsst das Objekt explodieren
	void explode( int numDebris, DebrisType debrisType, float radius, float scale = 1.0f, bool makeHole = true, int explosionSound = 1, float volume = 1.0f, float pitch = 1.0f );
	void explode_lua( int numDebris, DebrisType debrisType, float size, bool makeHole, float volume ) { explode( numDebris, debrisType, 4.0f * size, size, makeHole, volume > 0.0f ? 1 : 0, volume, randFloat( 0.9f, 1.1f ) ); }

	// vernichtet das Objekt
	void kill();

	// erlaubt, dass das Objekt wiedergeboren wird
	void rebirth();

	// macht das Objekt (un)sichtbar
	void setVisible( bool visible );

	// liefert true, wenn das Objekt sichtbar ist
	bool isVisible() const;

	// friert das Objekt ein / taut es auf
	void setFrozen( bool frozen );

	// liefert true, wenn das Objekt eigefroren ist
	bool isFrozen() const;

	// setzt den Pass-Through-Modus oder hebt ihn auf (Objekte knnen durch dieses Objekt durchfliegen)
	void setPassThrough( bool enable );

	// liefert true, wenn der Pass-Through-Modus aktiv ist
	bool getPassThrough() const;

	// setzt den Faktor fr den Schaden, den der Spieler bei der Kollision mit dem Objekt erleidet
	void setPlayerCollisionDamageScale( float playerCollisionDamageScale );

	// liefert den Spielerkollisionsfaktor
	float getPlayerCollisionDamageScale() const;

	// setzt den Kollisions-Callback
	void setCollisionCallback( luabind::object callback );

	// liefert den Kollisions-Callback
	luabind::object getCollisionCallback() const;

	void setOnKilled( luabind::object onKilled ) { this->onKilled = onKilled; }
	luabind::object getOnKilled() const { return onKilled; }
	void setOnReanimated( luabind::object onReanimated ) { this->onReanimated = onReanimated; }
	luabind::object getOnReanimated() const { return onReanimated; }

	Level& getLevel() const;
	const std::string& getID() const;
	dBodyID getBody() const;
	const dMass& getMass() const;
	const Vec2f& getBoundCenter() const;
	float getBoundRadius() const;
	const Vec2f& getBoundMin() const;
	const Vec2f& getBoundMax() const;
	bool getTurnToZombie() const { return turnToZombie; }
	void setTurnToZombie( bool v ) { turnToZombie = v; }
	bool isZombie() const { return dead && turnToZombie; }

protected:
	// lscht den ODE-Body
	void deleteBody();

	// lscht die ODE-Geoms
	void deleteGeoms();

	// setzt den Bounding-Circle
	void setBoundingCircle( const Vec2f& boundCenter, float boundRadius );

	// setzt das Bounding-Rect
	void setBoundingRect( const Vec2f& boundMin, const Vec2f& boundMax );

	// erweitert das Bounding-Rect so, dass es den angegebenen Punkt enthlt
	void extendBoundingRect( const Vec2f& point );

	// vergrert das Bounding-Rect
	void enlargeBoundingRect( float value );

	// das zugehrige Levelobjekt
	Level& level;

	// Spielumgebung
	GameEnvironment& env;

	// Shortcut zum Video-Objekt
	Video& video;

	// Menge von Objekten, mit denen dieses Objekt nicht kollidieren kann
	std::set<GameObject*> excludeCollisions;

private:
	// vervollstndigt die Kollisionsinformationen
	void completeCollision( GameObject* p_other, CollisionInfo& info );

	// wendet die Kollisionseffekte an
	void processCollisionEffects( GameObject* p_other, CollisionInfo& info );

	// ID des Objekts
	std::string id;

	// Liste aller Shapes
	std::list<dGeomID> geoms;

	// der ODE-Body
	dBodyID body;

	// 2D-Constraint
	dJointID planeConstraint;

	// Masse
	dMass mass;

	// Materialname
	std::string material;

	// Kollisions-Bits
	unsigned long categoryBits;
	unsigned long collideBits;

	// Reibung fr Bewegungen
	float movementFriction;

	// Reibung fr Rotationen
	float rotationFriction;

	// Zielobjekt
	std::string targetID;

	// Bounding-Circle
	Vec2f boundCenter;
	float boundRadius;

	// Bounding-Rect
	Vec2f boundMin;
	Vec2f boundMax;

	// auf true setzen, damit das Objekt gelscht wird
	bool dead;

	// auf true setzen, damit das Objekt nach seinem Tod zum Zombie wird und wiederbelebt werden kann
	bool turnToZombie;

	// Sichtbarkeit
	bool visible;

	// Eingefroren?
	bool frozen;

	// Pass-Through?
	bool passThrough;

	// Spielerkollisionsfaktor
	float playerCollisionDamageScale;

	// Kollisions-Callback
	luabind::object collisionCallback;

	// Kill-Callback
	luabind::object onKilled;

	// Wiedergeburts-Callback
	luabind::object onReanimated;
};

#endif

