#ifndef LEVEL_H
#define LEVEL_H

#include <Scripting.h>

enum RenderPass;
class GameObject;
class ScriptObject;
class ScriptObj_Counter;
class CartoonRenderer;

enum LevelFlags
{
	LF_RENDER_AVOIDANCE_AREAS = 0x1,
	LF_RENDER_OBJECT_DEBUG_INFO = 0x2,

	LF_LAST
};

struct CollisionParameters
{
	struct Effect
	{
		Effect() : minSpeed( 5.0f ),
			       createEmitter( false )
		{
		}

		float minSpeed;
		SoundHandle sound;
		float soundVolumeFactor;
		bool createEmitter;
		SimpleEmitterSettings emitterSettings;
		float particleRateFactor;
	};

	CollisionParameters() : dontCreateContactJoint( false ),
							bouncyness( 0.25f ),
							minBounceSpeed( 0.1f ),
							friction( 10.0f ),
							threshold( 0.5f )
	{
	}

	// physikalische Parameter
	bool dontCreateContactJoint;
	float bouncyness;
	float minBounceSpeed;
	float friction;

	// Effektparameter
	float threshold;
	Effect frontal;
	Effect scratch;
};

struct CollisionInfo
{
	bool passThrough;
	bool frontal;

	Vec2f collisionPoint;
	Vec2f collisionNormal;
	Vec2f relativeVelocity;
	float speed;
	bool minSpeedExceeded;
	CollisionParameters* p_params;

	std::list<dContact> contacts;
	std::list<dJointID> joints;
};

struct AvoidanceArea
{
	// Prioritt
	int priority;

	// Von wem erzeugt?
	GameObject* p_createdBy;

	// Typ
	// 0: Kreis
	// 1: Rechteck
	int type;

	// Transformationsmatrix
	Mat33f transformation;

	// inverse Transformationsmatrix
	Mat33f inverseTransformation;

	union
	{
		// fr Kreis
		struct
		{
			Vec2f center;
			float radius;
			float radiusSq;
		};

		// fr Rechteck
		struct
		{
			Vec2f min;
			Vec2f max;
		};
	};

	bool operator < ( const AvoidanceArea& rhs ) const
	{
		if( priority < rhs.priority ) return true;
		return p_createdBy < rhs.p_createdBy;
	}
};

class Level : public Base
{
public:
	Level( GameEnvironment& env, const std::string& filename, bool testMode = false );
	~Level();

	// liefert schnelle Informationen ber einen Level
	static void getLevelInfo( const std::string& filename, std::string& outName, std::string& outType, std::string& outDescription );

	// formatiert eine Zeitangabe
	static std::string formatTime( unsigned int msecs );

	void render();
	void renderOverlays();
	void update();

	// erzeugt ein Objekt aus einer XML-Struktur
	GameObject* createObject( const TiXmlElement* p_root );

	// bernimmt die Verantwortung fr ein Objekt
	int adoptObject( GameObject* p_object );

	// bernimmt die Verantwortung fr ein Script-Objekt
	void adoptScriptObject( ScriptObject* p_object );

	// liefert das Objekt mit der angegebenen ID
	GameObject* getObjectByID( const std::string& id ) const;

	// liefert den hchsten erreichten Checkpoint
	GameObject* getHighestCheckpoint() const;

	// setzt das Spiel an einem Checkpoint fort
	void continueAtCheckpoint( GameObject* p_checkpoint );

	// spielt die Musik ab
	void playMusic();

	// wird bei einer Kollision aufgerufen
	void nearCallback( dGeomID geom1, dGeomID geom2 );

	// setzt die Gravitation
	void setGravity( const Vec2f& gravity );

	// liefert die Gravitation
	Vec2f getGravity() const;

	// setzt das Zielobjekt der Kamera
	void setCameraTarget( GameObject* p_cameraTarget, bool moveThereImmediately = false );
	void setCameraTarget_lua( GameObject* p_cameraTarget ) { setCameraTarget( p_cameraTarget ); }

	// liefert das Zielobjekt der Kamera
	GameObject* getCameraTarget() const;

	// registriert die Kollisionsparameter fr ein Paar von Materialtypen
	void registerCollisionPair( const std::string& type1, const std::string& type2, const CollisionParameters& params );

	// liefert die Kollisionsparameter fr ein Paar von Materialtypen
	CollisionParameters* getCollisionPair( const std::string& type1, const std::string& type2 );

	// sorgt dafr, dass ein Sound nicht mehr neu geladen wird
	void cacheSound( SoundHandle sound );

	// erzeugt einen Strahl und liefert den nchsten Schnittpunkt mit dem Level zurck
	GameObject* castRay( const Vec2f& origin, const Vec2f& direction, float maxDistance, unsigned long categoryBits, unsigned long collisionBits, Vec2f* p_outPosition = 0, float* p_outDistance = 0, Vec2f* p_outNormal = 0 );

	// erzeugt einen Zylinder und liefert den nchsten Schnittpunkt mit dem Level zurck
	GameObject* castCylinder( const Vec2f& origin, const Vec2f& direction, float radius, float maxDistance, unsigned long categoryBits, unsigned long collisionBits, Vec2f* p_outPosition = 0, float* p_outDistance = 0, Vec2f* p_outNormal = 0 );

	// fgt dem Spielerobjekt Punkte hinzu
	void addPlayerScore( int value, const Vec2f& position = Vec2f( FLT_MAX, FLT_MAX ), bool immediate = false );
	void addPlayerScore_lua( int value ) { addPlayerScore( value, Vec2f( FLT_MAX, FLT_MAX ), true ); }
	void addPlayerScore_lua2( int value, const Vec2f& position ) { addPlayerScore( value, position, true ); }

	// fgt ein Textobjekt hinzu
	void addText( const std::string& text, const Vec2f& position, const Vec4f& color, float size, float timeToLive );

	// fgt eine SMS hinzu
	void addSMS( const std::string& sender, const std::string& text );

	// fgt eine kreisfrmige Gefahrzone hinzu
	void addAvoidanceCircle( int priority, GameObject* p_me, const Mat33f& transformation, const Vec2f& center, float radius );

	// fgt eine rechteckfrmige Gefahrzone hinzu
	void addAvoidanceRect( int priority, GameObject* p_me, const Mat33f& transformation, const Vec2f& min, const Vec2f& max );

	// fgt eine linienfrmige Gefahrzone hinzu
	void addAvoidanceLine( int priority, GameObject* p_me, const Vec2f& start, const Vec2f& end, float thickness, float lengthen = 0.0f );

	// berechnet einen Fluchtvektor, der aus den Gefahrzonen herausfhrt
	bool getEscapeVector( GameObject* p_me, float myRadius, Vec2f& outEscapeVector ) const;

	// fgt einen Timer-Callback hinzu
	void addTimerCallback( luabind::object callback, float interval );

	// entfernt Timer-Callbacks
	int removeTimerCallback( luabind::object callback );

	// lsst ein Script nach einer bestimmten Zeit wieder aufwachen
	void addResumeCountDown( lua_State* p_lua, float time );

	// erzeugt einen Counter
	ScriptObj_Counter* createCounter( int displayMode, int direction, float duration, luabind::object onElapsed );

	GameEnvironment& getEnv() const;
	const std::string& getFilename() const;
	const std::string& getName() const;
	const std::string& getType() const;
	const std::string& getDescription() const;
	const std::string& getMusicFilename() const;
	bool isInTestMode() const;
	const std::map<std::string, GameObject*>& getObjects() const;
	const std::set<GameObject*>& getObjectsSet() const { return objectsSet; }
	dWorldID getWorld() const;
	dSpaceID getSpace() const;
	TextureHandle getSprites() const;
	TextureHandle getParticles() const;
	CartoonRenderer& getCR() const;
	ParticleSystem& getPS() const;
	bool isCameraTrackingEnabled() const;
	void setCameraTrackingEnabled( bool enable );
	const Vec2f& getCamera() const;
	void setCamera( const Vec2f& camera );
	float getCameraZoom() const;
	void setCameraZoom( float cameraZoom );
	float getCameraAngle() const;
	void setCameraAngle( float cameraAngle );
	const Vec4f& getPassColor() const;
	const std::vector<GameObject*>& getVisibleObjects() const;
	unsigned int getNumVisibleObjects() const;
	bool isFading() const;
	bool isFinished() const;
	float getTime() const;
	unsigned int getRaceTime() const;
	float getRaceTime_lua() const { return static_cast<float>( raceTime ) / 1000.0f; }
	void resetRaceTime();
	void setFinalScore( unsigned int finalScore );
	unsigned int getFinalScore() const;
	void setFlags( unsigned int flags );
	unsigned int getFlags() const;
	float getPhysicsTimePerFrame() const;
	std::string getObjectAutoID();

private:
	struct CollisionPair
	{
		GameObject* p_obj1;
		GameObject* p_obj2;

		bool operator < ( const CollisionPair& rhs ) const
		{
			if( p_obj1 == rhs.p_obj1 ) return p_obj2 < rhs.p_obj2;
			else return p_obj1 < rhs.p_obj1 ;
		}
	};

	struct Text
	{
		std::string text;
		Vec2f position;
		Vec4f color;
		float size;
		float angle;
		float timeToLive;
	};

	struct SMS
	{
		std::string sender;
		std::string text;
		std::vector<size_t> pageStarts;
		float duration;
	};

	struct TimerCallback
	{
		luabind::object callback;
		float interval;
		float t;
	};

	struct ResumeCountDown
	{
		lua_State* p_lua;
		float t;
	};

	int preload( TiXmlElement* p_root );
	int load( TiXmlElement* p_root );
	int loadResources();
	void setupParticleTypes();
	void setupCollisionParameters();
	void renderObjects( RenderPass pass, const Vec4f& color = Vec4f( 1.0f ) );
	void deleteObject( GameObject* p_object );
	void processCollisions();
	void calculateVisibility();
	void renderTexts();
	void updateTexts();
	void updateSMS();
	void updateScriptObjects();
	void renderAvoidanceAreas();
	void renderMobile();
	void renderScriptObjects();

	GameEnvironment& env;
	Video& video;

	// Dateiname des Levels
	const std::string filename;

	// Testmodus?
	bool testMode;

	// Rennmodus?
	bool raceMode;

	// alle Spielobjekte in der aktuellen Karte
	std::map<std::string, GameObject*> objects;

	// Menge der Spielobjekte
	std::set<GameObject*> objectsSet;

	// Menge der Script-Objekte
	std::set<ScriptObject*> scriptObjects;

	// Menge der Gefahrzonen fr Gegner
	std::set<AvoidanceArea> avoidanceAreas;

	// Name des Levels
	std::string name;

	// Typ des Levels
	std::string type;

	// Beschreibungstext, der am Anfang angezeigt wird
	std::string description;

	// Script
	std::string script;

	// Dateiname des Musikstcks
	std::string musicFilename;

	// Dateiname des Hintergrundbilds
	std::string backgroundFilename;

	// ODE-Welt
	dWorldID world;

	// ODE-Space fr schnellere Kollisionserkennung
	dSpaceID space;

	// Gruppe der Kontakt-Joints
	dJointGroupID contactGroup;

	TextureHandle background;
	TextureHandle sprites;
	TextureHandle particles;
	TextureHandle mobile;
	CartoonRenderer* p_cr;
	ParticleSystem* p_ps;
	bool cameraTracking;
	Vec2f camera;
	Vec2f cameraTarget;
	float cameraZoom;
	float cameraAngle;
	GameObject* p_cameraTarget;
	Vec4f passColor;
	std::map<std::string, CollisionParameters> collisionParameters;
	std::set<SoundHandle> soundCache;
	std::map<CollisionPair, CollisionInfo> collisions;
	std::vector<GameObject*> visibleObjects;
	std::list<Text> texts;
	std::queue<SMS> smsInbox;
	std::set<GameObject*> zombies;
	float smsTime;
	float fade;
	float fadeOut;
	float time;
	float fadeTime;
	unsigned int raceTime;
	bool finished;
	unsigned int finalScore;
	int negativeScore;
	unsigned int flags;
	unsigned int nextObjectAutoID;

	lua_State* p_lua;
	int threadRef;
	std::list<TimerCallback> timerCallbacks;
	std::list<ResumeCountDown> resumeCountDowns;
};

#endif