#include <PCH.h>
#include <Campaign.h>
#include <Level.h>
#include <HighscoresDB.h>

Campaign::Campaign( const std::string& filename ) : filename( filename )
{
	// XML-Dokument laden
	TiXmlDocument doc( ( std::string( "data\\levels\\" ) + filename ).c_str() );
	if( !doc.LoadFile() )
	{
		// Fehler!
		setError( 1 );
		return;
	}

	TiXmlElement* p_root = doc.FirstChildElement( "Campaign" );
	if( !p_root )
	{
		// Fehler!
		setError( 2 );
		return;
	}

	// Name holen
	const char* p_name = p_root->Attribute( "name" );
	if( p_name ) name = p_name;

	TiXmlElement* p_levels = p_root->FirstChildElement( "Levels" );
	if( p_levels )
	{
		const char* p_includeAllOfType = p_levels->Attribute( "includeAllOfType" );
		if( p_includeAllOfType )
		{
			// alle Levels vom angegebenen Typ suchen und einfgen
			std::list<std::string> files = listFiles( "data\\levels\\*.level" );
			for( std::list<std::string>::const_iterator it = files.begin();
				 it != files.end();
				 it++ )
			{
				std::string levelName, levelType, levelDesc;
				Level::getLevelInfo( *it, levelName, levelType, levelDesc );
				if( levelType == p_includeAllOfType ) levels.push_back( *it );
			}
		}

		const char* p_includeAll = p_levels->Attribute( "includeAll" );
		if( p_includeAll )
		{
			const std::string pattern( p_includeAll );
			const std::string patternDir( getPathDirectory( pattern ) );

			// alle Levels, deren Dateiname passt, hinzufgen
			std::list<std::string> files = listFiles( std::string( "data\\levels\\" ) + pattern );
			for( std::list<std::string>::const_iterator it = files.begin();
				 it != files.end();
				 it++ )
			{
				levels.push_back( patternDir + "\\" + *it );
			}
		}

		TiXmlElement* p_level = p_levels->FirstChildElement( "Level" );
		while( p_level )
		{
			const char* p_filename = p_level->GetText();
			if( p_filename ) levels.push_back( p_filename );

			p_level = p_level->NextSiblingElement( "Level" );
		}
	}

	// Anzahl der entsperrten Levels holen
	int temp = static_cast<int>( levels.size() );
	p_root->QueryIntAttribute( "numUnlockedLevels", &temp );
	numUnlockedLevels = temp;
}

Campaign::~Campaign()
{
}

const std::string& Campaign::getFilename() const
{
	return filename;
}

const std::string& Campaign::getName() const
{
	return name;
}

unsigned int Campaign::getNumUnlockedLevels() const
{
	return numUnlockedLevels;
}

const std::vector<std::string>& Campaign::getLevels() const
{
	return levels;
}

bool Campaign::isLevelUnlocked( const std::string& level ) const
{
	// Welche Nummer hat dieser Level?
	size_t levelNo = ~0;
	for( size_t i = 0; i < levels.size(); i++ )
	{
		if( levels[i] == level )
		{
			levelNo = i;
			break;
		}
	}

	if( levelNo == ~0 ) return false;

	if( numUnlockedLevels > levelNo ) return true;

	return levelNo < numUnlockedLevels + getNumLevelsCompleted();
}

unsigned int Campaign::getNumLevelsCompleted() const
{
	HighscoresDB& db = HighscoresDB::inst();
	unsigned int n = 0;
	for( size_t i = 0; i < levels.size(); i++ )
	{
		if( db.wasLevelCompleted( levels[i] ) ) n++;
	}

	return n;
}