#include <PCH.h>
#include <CartoonRenderer.h>
#include <xhash>

typedef void ( APIENTRY *TESS_CALLBACK_FN )( void );

static void APIENTRY vertexCallback( void* p_data )
{
	Vec4f* p_vec = static_cast<Vec4f*>( p_data );

	glTexCoord2f( p_vec->z, p_vec->w );
	glVertex2f( p_vec->x, p_vec->y );
}

CartoonRenderer::CartoonRenderer( GameEnvironment& env,
								  const std::string& stylesFilename ) : env( env ),
																		p_currentStyle( 0 ),
																		lastSegmentDir( 0.0f, 0.0f )
{
	std::string directory = getPathDirectory( stylesFilename );
	TextureManager& texMgr = env.engine.getTextureManager();

	// XML-Dokument laden
	TiXmlDocument doc( stylesFilename.c_str() );
	if( !doc.LoadFile() )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << "Datei " << stylesFilename << " konnte nicht geladen werden";
		env.logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	TiXmlElement* p_root = doc.FirstChildElement( "Styles" );
	if( !p_root )
	{
		// Fehler!
		setError( 2 );
		std::ostringstream msg;
		msg << stylesFilename << ": <Styles>-Element fehlt";
		env.logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	int n = 0;
	TiXmlElement* p_style = p_root->FirstChildElement( "Style" );
	while( p_style )
	{
		Style s;
		s.numLines = 0;

		// Name holen
		const char* p_name = p_style->Attribute( "name" );
		if( p_name )
		{
			s.name = p_name;

			TiXmlElement* p_fillElem = p_style->FirstChildElement( "Fill" );
			if( p_fillElem )
			{
				// Fll-Tiling holen
				s.fillTiling = 1.0f;
				p_fillElem->QueryFloatAttribute( "tiling", &s.fillTiling );

				const char* p_fill = p_fillElem->GetText();
				if( p_fill ) s.fill = texMgr.get( directory + "\\" + p_fill );
			}

			TiXmlElement* p_linesElem = p_style->FirstChildElement( "Lines" );
			if( p_linesElem )
			{
				// Anzahl der Linien holen
				s.numLines = 1;
				p_linesElem->QueryIntAttribute( "num", &s.numLines );

				// Offset holen
				s.offset = 0;
				p_linesElem->QueryIntAttribute( "offset", &s.offset );

				// Tiling holen
				s.lineTiling = 1.0f;
				p_linesElem->QueryFloatAttribute( "tiling", &s.lineTiling );

				// Linienbreite holen
				s.lineWidth = 1.0f;
				p_linesElem->QueryFloatAttribute( "width", &s.lineWidth );

				const char* p_lines = p_linesElem->GetText();
				if( p_lines ) s.lines = texMgr.get( directory + "\\" + p_lines );
			}

			styles[s.name] = s;
			n++;
		}

		p_style = p_style->NextSiblingElement( "Style" );
	}

	if( !n ) env.logger.warning( "Es wurden keine Stile erzeugt", LOGGER_AUTO );
	else
	{
		std::ostringstream msg;
		msg << "Es wurden " << n << " Stile erzeugt";
		env.logger.finer( msg.str(), LOGGER_AUTO );
	}

	// Tesselator erzeugen
	p_tess = gluNewTess();
	assert( p_tess );

	gluTessCallback( p_tess, GLU_TESS_BEGIN, ( TESS_CALLBACK_FN )( glBegin ) );
	gluTessCallback( p_tess, GLU_TESS_VERTEX, ( TESS_CALLBACK_FN )( vertexCallback ) );
	gluTessCallback( p_tess, GLU_TESS_END, glEnd );
}

CartoonRenderer::~CartoonRenderer()
{
	// Tesselator lschen
	gluDeleteTess( p_tess );
}

int CartoonRenderer::setStyle( const std::string& name )
{
	if( p_currentStyle )
	{
		if( p_currentStyle->name == name ) return 0;
	}

	// Stil suchen
	std::map<std::string, Style>::iterator it = styles.find( name );
	if( it == styles.end() )
	{
		p_currentStyle = 0;
		return 1;
	}

	p_currentStyle = &it->second;

	return 0;
}

void CartoonRenderer::setContext( const std::string& name )
{
	setContext( static_cast<unsigned int>( stdext::hash_value( name ) ) );
}

void CartoonRenderer::setContext( unsigned int seed )
{
	rng = 0x1000 + ( seed % 0xF000 );

	lastSegmentEnd = Vec2f( FLT_MAX, FLT_MAX );
}

int CartoonRenderer::drawPolyLine( const std::vector<Vec2f>& vertices,
								   bool close,
								   float widthScale,
								   const Vec4f& color,
								   const std::vector<Vec4f>* p_colors )
{
	if( !p_currentStyle ) return 1;
	if( !p_currentStyle->numLines ) return 2;
	if( vertices.size() < 2 ) return 3;

	// Linientextur setzen
	env.engine.getVideo().setTexture( p_currentStyle->lines );

	glBegin( GL_QUADS );

	Vec2f from( 0.0f, 0.0f ), to( 0.0f, 0.0f );

	int lineSubStyle = randInt() % p_currentStyle->numLines;
	int nextLineSubStyle = randInt() % p_currentStyle->numLines;
	int firstLineSubStyle = lineSubStyle;

	size_t i = 0;
	for( std::vector<Vec2f>::const_iterator it = vertices.begin();
		 it != vertices.end();
		 it++, i++ )
	{
		from = to;
		to = *it;

		if( it != vertices.begin() )
		{
			Vec4f color1, color2;
			if( p_colors ) color1 = ( *p_colors )[i - 1], color2 = ( *p_colors )[i];
			else color1 = color2 = color;
			drawSegment( from, to,
						 p_currentStyle->offset + lineSubStyle, p_currentStyle->offset + nextLineSubStyle,
						 widthScale, color1, color2 );
		}

		lineSubStyle = nextLineSubStyle;
		nextLineSubStyle = randInt() % p_currentStyle->numLines;
	}

	if( close && vertices.size() > 2 )
	{
		// Linienzug schlieen
		Vec4f color1, color2;
		if( p_colors ) color1 = p_colors->back(), color2 = p_colors->front();
		else color1 = color2 = color;
		drawSegment( vertices.back(), vertices.front(),
					 p_currentStyle->offset + lineSubStyle, p_currentStyle->offset + firstLineSubStyle,
					 widthScale, color1, color2 );
	}

	glEnd();

	return 0;
}

int CartoonRenderer::fillPolygon( const std::vector<Vec2f>& vertices,
								  bool overrideTexture,
								  TextureHandle texture )
{
	if( !p_currentStyle ) return 1;

	// Flltextur setzen
	env.engine.getVideo().setTexture( overrideTexture ? texture : p_currentStyle->fill );

	// zufllige Drehung der Textur
	glPushAttrib( GL_TRANSFORM_BIT );
	glMatrixMode( GL_TEXTURE );
	glPushMatrix();
	glRotated( static_cast<double>( randInt() % 360 ), 0.0, 0.0, 1.0 );

	// Polygon beginnen
	gluTessBeginPolygon( p_tess, 0 );
	gluTessBeginContour( p_tess );

	Vec4f* p_data = new Vec4f[vertices.size()];

	// an den Tesselator geben
	int index = 0;
	for( std::vector<Vec2f>::const_iterator it = vertices.begin();
		 it != vertices.end();
		 it++, index++ )
	{
		double coords[] = { it->x, it->y, 0.0 };
		p_data[index] = Vec4f( it->x, it->y, it->x * p_currentStyle->fillTiling, it->y * p_currentStyle->fillTiling );
		gluTessVertex( p_tess, coords, &p_data[index] );
	}

	gluTessEndContour( p_tess );
	gluTessEndPolygon( p_tess );

	delete[] p_data;

	glPopMatrix();
	glPopAttrib();

	return 0;
}

void CartoonRenderer::drawSegment( const Vec2f& from,
								   const Vec2f& to,
								   int lineSubStyle,
								   int nextLineSubStyle,
								   float widthScale,
								   const Vec4f& color1,
								   const Vec4f& color2 )
{
	const float length = ( to - from ).length();
	const Vec2f axisX( ( to - from ) / length );
	const Vec2f axisY( Vec2f( -axisX.y, axisX.x ) * p_currentStyle->lineWidth * 0.5f * widthScale );

	const int x = p_currentStyle->lineTiling < 0.0f ? p_currentStyle->lines->getSize().x : static_cast<int>( length * p_currentStyle->lineTiling );
	const float dot = axisX ^ lastSegmentDir;

	Vec2f p1, p2;
	if( dot > 0.65f && from == lastSegmentEnd )
	{
		// Linie weiterzeichnen
		p1 = lastCorner[0];
		p2 = lastCorner[1];
	}
	else
	{
		p1 = from + axisY;
		p2 = from - axisY;
	}

	const Vec2f p3( to - axisY );
	const Vec2f p4( to + axisY );

	if( lineSubStyle == nextLineSubStyle )
	{
		const int y = lineSubStyle * 32;
		glColor4fv( color1 );
		glTexCoord2i( 0, y + 32 );
		glVertex2fv( p1 );
		glTexCoord2i( 0, y );
		glVertex2fv( p2 );
		glColor4fv( color2 );
		glTexCoord2i( x, y );
		glVertex2fv( p3 );
		glTexCoord2i( x, y + 32 );
		glVertex2fv( p4 );
	}
	else
	{
		const Vec4f transparent1( color1.r, color1.g, color1.b, 0.0f );
		const Vec4f transparent2( color2.r, color2.g, color2.b, 0.0f );
		const Vec4f opaque1( color1 );
		const Vec4f opaque2( color2 );

		const int y1 = lineSubStyle * 32;
		glColor4fv( opaque1 );
		glTexCoord2i( 0, y1 + 32 );
		glVertex2fv( p1 );
		glTexCoord2i( 0, y1 );
		glVertex2fv( p2 );
		glColor4fv( transparent2 );
		glTexCoord2i( x, y1 );
		glVertex2fv( p3 );
		glTexCoord2i( x, y1 + 32 );
		glVertex2fv( p4 );

		const int y2 = nextLineSubStyle * 32;
		glColor4fv( transparent1 );
		glTexCoord2i( x, y2 + 32 );
		glVertex2fv( p1 );
		glTexCoord2i( x, y2 );
		glVertex2fv( p2 );
		glColor4fv( opaque2 );
		glTexCoord2i( 0, y2 );
		glVertex2fv( p3 );
		glTexCoord2i( 0, y2 + 32 );
		glVertex2fv( p4 );
	}

	lastSegmentEnd = to;
	lastSegmentDir = axisX;
	lastCorner[0] = p4;
	lastCorner[1] = p3;
}

unsigned int CartoonRenderer::randInt() const
{
	rng *= rng;
	rng >>= 8;
	rng &= 0xFFFF;
	return rng;
}