#include <PCH.h>
#include <Console.h>
#include <Scripting.h>
#include <SDL.h>

Console::Console( GameEnvironment& env ) : Component( env.engine, "Console" ),
										   env( env ),
										   visible( false ),
										   cursor( 0 ),
										   historyCursor( ~0 )
{
	Vec4f c1( 1.0f, 1.0f, 0.25f, 1.0f );
	Vec4f c2( 0.25f, 0.75f, 1.0f, 1.0f );
	print( "***********************\n", c1 );
	print( "*", c1 );
	print( " Rocket Rage Console ", c2 );
	print( "*\n", c1 );
	print( "***********************\n\n", c1 );

	env.engine.connectPostRenderFinal( this, &Console::onPostRenderFinal );
	env.engine.connectSDLEvent( this, &Console::onSDLEvent );
}

Console::~Console()
{
}

void Console::print( const std::string& text,
					 const Vec4f& color )
{
	buffer.append( std::string( 1, 255 ) );
	buffer.append( std::string( 1, static_cast<unsigned char>( clamp( color.r, 0.0f, 1.0f ) * 255.0f ) ) );
	buffer.append( std::string( 1, static_cast<unsigned char>( clamp( color.g, 0.0f, 1.0f ) * 255.0f ) ) );
	buffer.append( std::string( 1, static_cast<unsigned char>( clamp( color.b, 0.0f, 1.0f ) * 255.0f ) ) );
	buffer.append( std::string( 1, static_cast<unsigned char>( clamp( color.a, 0.0f, 1.0f ) * 255.0f ) ) );
	buffer += text;
	buffer.append( std::string( 1, 254 ) );

	// Zeilen zhlen
	size_t i = 0;
	size_t n = 0;
	while( true )
	{
		i = buffer.find( '\n', i );
		if( i == std::string::npos ) break;
		i++;
		n++;
	}

	// Wenn es zu viele Zeilen sind, dann werden die ersten abgeschnitten.
	if( n > 30 )
	{
		n -= 30;
		i = 0;
		for( size_t m = 0; m < n; m++ ) i = buffer.find( '\n', i ) + 1;
		buffer = std::string( buffer, i );
	}
}

void Console::clear()
{
	buffer.clear();
}

void Console::setVisible( bool visible )
{
	this->visible = visible;
}

bool Console::isVisible() const
{
	return visible;
}

void Console::onPostRenderFinal()
{
	if( !visible ) return;

	Video& video = env.engine.getVideo();
	video.setupProjection( Vec2f( 0.0f, 0.0f ), Vec2f( 48.0f, 36.0f ) );
	video.setTexture( 0 );
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glBegin( GL_QUADS );
	glColor4f( 0.0f, 0.25f, 0.0f, 0.75f );
	glVertex2i( 0, 0 );
	glVertex2i( 48, 0 );
	glVertex2i( 48, 24 );
	glVertex2i( 0, 24 );
	glEnd();
	glBegin( GL_LINE_LOOP );
	glColor4f( 0.0f, 1.0f, 0.0f, 0.75f );
	glVertex2i( 0, 0 );
	glVertex2i( 48, 0 );
	glVertex2i( 48, 24 );
	glVertex2i( 0, 24 );
	glEnd();
	glBegin( GL_LINES );
	glVertex2i( 0, 23 );
	glVertex2i( 48, 23 );
	glEnd();

	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glTranslatef( 0.375f, 0.375f, 0.0f );
	env.fixedFont->print( buffer, 0.75f );
	glPopMatrix();
	glPushMatrix();
	glTranslatef( 0.375f, 23.125f, 0.0f );
	env.fixedFont->print( std::string( "> " ) + input, 0.75f );
	env.fixedFont->print( std::string( "  " ) + std::string( cursor, ' ' ) + "_", 0.75f );
	glPopMatrix();
}

void Console::onSDLEvent( const SDL_Event& e )
{
	if( e.type == SDL_KEYDOWN )
	{
		char c = static_cast<char>( e.key.keysym.unicode );
		if( c >= 32 || c < 0 )
		{
			if( c == '' ) setVisible( !isVisible() );
			else if( isVisible() )
			{
				input.insert( cursor, std::string( 1, c ) );
				cursor++;
			}
		}
		else if( isVisible() )
		{
			switch( e.key.keysym.sym )
			{
			case SDLK_LEFT:
				if( cursor ) cursor--;
				break;

			case SDLK_RIGHT:
				if( cursor < input.length() ) cursor++;
				break;

			case SDLK_HOME:
				cursor = 0;
				break;

			case SDLK_END:
				cursor = input.length();
				break;

			case SDLK_DELETE:
				input.erase( cursor, 1 );
				break;

			case SDLK_BACKSPACE:
				if( cursor )
				{
					input.erase( cursor - 1, 1 );
					cursor--;
				}
				break;

			case SDLK_RETURN:
				if( input == "clear" ) buffer.clear();
				else if( input == "luainfo" ) Scripting::inst().luaInfo();
				else Scripting::inst().executeScript( 0, input );

				history.push_back( input );
				while( history.size() > 64 ) history.erase( history.begin() );

				input.clear();
				cursor = 0;
				historyCursor = ~0;
				break;

			case SDLK_UP:
				if( historyCursor == ~0 ) historyCursor = history.size() - 1;
				else if( historyCursor ) historyCursor--;
				if( historyCursor != ~0 )
				{
					input = history[historyCursor];
					cursor = input.length();
				}
				break;

			case SDLK_DOWN:
				if( historyCursor == ~0 || historyCursor == history.size() - 1 )
				{
					historyCursor = ~0;
					input = "";
					cursor = 0;
				}
				else if( historyCursor < history.size() - 1 )
				{
					historyCursor++;
					input = history[historyCursor];
					cursor = input.length();
				}
				break;
			}
		}
	}
}