#include <PCH.h>
#include <Controllable.h>
#include <NeuralNet.h>

Controllable::Controllable( GameEnvironment& env ) : ctrEnv( env ),
												     numInputs( 0 ),
												     numOutputs( 0 ),
												     mode( CM_DEAD ),
												     recording( false ),
												     trainingFilenamePrefix( "training_" ),
												     training( false )
{
}

Controllable::~Controllable()
{
	stopRecording();
	stopTraining( false );
}

void Controllable::registerParameter( const std::string& name,
									  ParameterType type,
									  float minValue,
									  float maxValue,
  									  float stdValue )
{
	Parameter param;
	param.type = type;
	param.value = stdValue;
	param.minValue = minValue;
	param.maxValue = maxValue;
	param.stdValue = stdValue;
	param.neuralMapping = -1;

	if( type == PT_INPUT ) param.neuralMapping = numInputs++;
	else if( type == PT_OUTPUT ) param.neuralMapping = numOutputs++;

	params[name] = param;
}

void Controllable::unregisterParameter( const std::string& name )
{
	params.erase( name );
}

void Controllable::unregisterAllParameters()
{
	params.clear();
}

void Controllable::setParameter( const std::string& name,
								 float value )
{
	std::map<std::string, Parameter>::iterator it = params.find( name );
	if( it == params.end() ) return;

	it->second.value = clamp( value, it->second.minValue, it->second.maxValue );
}

float Controllable::getParameter( const std::string& name ) const
{
	std::map<std::string, Parameter>::const_iterator it = params.find( name );
	return it == params.end() ? 0.0f : it->second.value;
}

void Controllable::generateInput()
{
}

void Controllable::think()
{
	if( currentBehavior )
	{
		// Eingabeparameter in das neuronale Netzwerk stecken
		float* p_input = new float[numInputs];
		for( std::map<std::string, Parameter>::const_iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == PT_INPUT )
			{
				const float v = it->second.value;
				const float tv = clamp( ( v - it->second.minValue ) / ( it->second.maxValue - it->second.minValue ), 0.0f, 1.0f );
				p_input[it->second.neuralMapping] = tv;
			}
		}

		// Simulation durchfhren
		float* p_output = currentBehavior->run( p_input );

		// Ausgabeparameter abgreifen
		for( std::map<std::string, Parameter>::iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == PT_OUTPUT )
			{
				const float v = clamp( p_output[it->second.neuralMapping], 0.0f, 1.0f );
				const float tv = it->second.minValue + v * ( it->second.maxValue - it->second.minValue );
				it->second.value = tv;
			}
		}

		delete[] p_input;
	}
}

void Controllable::updateControllable()
{
	// alle Parameter auf ihre Standardwerte setzen
	for( std::map<std::string, Parameter>::iterator it = params.begin();
		 it != params.end();
		 it++ )
	{
		it->second.value = it->second.stdValue;
	}

	if( mode == CM_AI_NET || training )
	{
		// Die Eingabeparameter mssen bestimmt werden.
		generateInput();
	}

	if( mode == CM_USER )
	{
		Input& input = ctrEnv.engine.getInput();

		// nachschauen, ob irgendwelche Tasten gedrckt sind
		for( std::list<KeyBinding>::const_iterator it = keyBindings.begin();
			 it != keyBindings.end();
			 it++ )
		{
			if( input.isKeyDown( it->key ) )
			{
				// den mit dieser Taste verbundenen Parameterwert setzen
				setParameter( it->param, it->value );
			}
		}
	}
	else if( mode == CM_PLAYBACK )
	{
		if( !playback.eof() )
		{
			// Parameter aus der Wiedergabedatei lesen
			for( std::map<std::string, Parameter>::iterator it = params.begin();
				 it != params.end();
				 it++ )
			{
				if( it->second.type == PT_OUTPUT ||
					it->second.type == PT_OUTPUT_USER_ONLY )
				{
					playback.read( reinterpret_cast<char*>( &it->second.value ), sizeof( it->second.value ) );
				}
			}
		}
	}
	else if( mode == CM_AI_NET || mode == CM_AI_SIMPLE )
	{
		think();
	}

	if( recording )
	{
		// alle Ausgabeparameter in die Datei schreiben
		for( std::map<std::string, Parameter>::const_iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == PT_OUTPUT ||
				it->second.type == PT_OUTPUT_USER_ONLY )
			{
				record.write( reinterpret_cast<const char*>( &it->second.value ), sizeof( it->second.value ) );
			}
		}
	}

	if( training )
	{
		// alle Eingabeparameter merken
		float* p_data = new float[numInputs + numOutputs];
		for( std::map<std::string, Parameter>::const_iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == PT_INPUT )
			{
				const float v = it->second.value;
				const float tv = clamp( ( v - it->second.minValue ) / ( it->second.maxValue - it->second.minValue ), 0.0f, 1.0f );
				p_data[it->second.neuralMapping] = tv;
			}
		}

		// alle Ausgabeparameter merken
		for( std::map<std::string, Parameter>::const_iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == PT_OUTPUT )
			{
				const float v = it->second.value;
				const float tv = clamp( ( v - it->second.minValue ) / ( it->second.maxValue - it->second.minValue ), 0.0f, 1.0f );
				p_data[numInputs + it->second.neuralMapping] = tv;
			}
		}

		for( unsigned int i = 0; i < numInputs + numOutputs; i++ ) currentTraining.push_back( p_data[i] );
		delete[] p_data;
	}
}

std::string Controllable::dumpParameters() const
{
	std::ostringstream text;

	for( int pass = PT_INPUT; pass != PT_LAST; pass++ )
	{
		for( std::map<std::string, Parameter>::const_iterator it = params.begin();
			 it != params.end();
			 it++ )
		{
			if( it->second.type == pass )
			{
				text << it->first << " = " << it->second.value << std::endl;
			}
		}
	}

	return text.str();
}

ControlMode Controllable::getControlMode() const
{
	return mode;
}

void Controllable::setUserMode()
{
	mode = CM_USER;
}

void Controllable::bindKey( Key key,
						    const std::string& parameter,
							float value )
{
	KeyBinding binding;
	binding.key = key;
	binding.param = parameter;
	binding.value = value;
	keyBindings.push_back( binding );
}

void Controllable::setAIMode()
{
	mode = CM_AI_NET;
}

int Controllable::registerAIBehavior( const std::string& name,
									  const std::string& filename )
{
	NeuralNetHandle neuralNet = ctrEnv.neuralNetMgr.get( filename );
	if( !neuralNet ) return 1;

	behaviors[name] = neuralNet;

	return 0;
}

int Controllable::setAIBehavior( const std::string& name )
{
	std::map<std::string, NeuralNetHandle>::const_iterator it = behaviors.find( name );
	if( it == behaviors.end() ) return 1;

	currentBehaviorName = it->first;
	currentBehavior = it->second;

	return 0;
}

std::string Controllable::getAIBehavior() const
{
	return currentBehaviorName;
}

void Controllable::setAINeuralNet( NeuralNetHandle neuralNet )
{
	currentBehavior = neuralNet;
}

NeuralNetHandle Controllable::getAINeuralNet() const
{
	return currentBehavior;
}

void Controllable::setAISimpleMode()
{
	mode = CM_AI_SIMPLE;
}

void Controllable::setPlaybackMode( const std::string& filename )
{
	if( playback.is_open() ) playback.close();

	playback.open( filename.c_str(), std::ios::in | std::ios::binary );

	mode = CM_PLAYBACK;
}

void Controllable::startRecording( const std::string& filename )
{
	if( record.is_open() ) record.close();

	record.open( filename.c_str(), std::ios::out | std::ios::binary );
	recording = true;
}

void Controllable::stopRecording()
{
	if( recording )
	{
		record.close();
		recording = false;
	}
}

bool Controllable::isRecording() const
{
	return recording;
}

void Controllable::setTrainingFilenamePrefix( const std::string& trainingFilenamePrefix )
{
	this->trainingFilenamePrefix = trainingFilenamePrefix;
}

void Controllable::startTraining()
{
	// wenn schon trainiert wird, aktuelles Training abbrechen
	if( training ) stopTraining( false );

	training = true;
}

void Controllable::stopTraining( bool wasGood )
{
	if( training )
	{
		if( wasGood )
		{
			// Dateiname generieren
			time_t t = time( 0 );
			tm* p_time = localtime( &t );
			char filename[256] = "";
			sprintf( filename, "%s%02d-%02d-%d@%02d-%02d-%02d.train",
					 trainingFilenamePrefix.c_str(),
					 p_time->tm_mday, p_time->tm_mon + 1, 1900 + p_time->tm_year,
					 p_time->tm_hour, p_time->tm_min, p_time->tm_sec );

			std::ofstream out( filename );

			// aktuelles Training speichern
			size_t numDataSets = currentTraining.size() / ( numInputs + numOutputs );
			out << numDataSets << " " << numInputs << " " << numOutputs << std::endl;
			size_t index = 0;
			for( size_t i = 0; i < numDataSets; i++ )
			{
				for( unsigned int j = 0; j < numInputs; j++ ) out << currentTraining[index++] << ( j == numInputs - 1 ? "" : " " );
				out << std::endl;
				for( unsigned int j = 0; j < numOutputs; j++ ) out << currentTraining[index++] << ( j == numOutputs - 1 ? "" : " " );
				out << std::endl;
			}

			out.close();
		}

		currentTraining.clear();
		training = false;
	}
}

bool Controllable::isTraining() const
{
	return training;
}