#include <PCH.h>
#include <GS_Config.h>

GS_Config::GS_Config( Engine& engine,
					  GUI& gui,
					  ParameterMap& config,
					  const ParameterMap& defaultConfig ) : GameState( engine, "Config" ),
															gui( gui ),
															config( config ),
															defaultConfig( defaultConfig )
{
	// Layouts laden
	gui.loadLayout( "Config.layout" );
	SUBSCRIBE_GUI_EVENT( "Config/NoSound", CEGUI::Checkbox::EventCheckStateChanged, &GS_Config::handleConfig );
	SUBSCRIBE_GUI_EVENT( "Config/HardwareSound", CEGUI::Checkbox::EventCheckStateChanged, &GS_Config::handleConfig );
	SUBSCRIBE_GUI_EVENT( "Config/Software3DSoundQuality", CEGUI::Slider::EventValueChanged, &GS_Config::handleConfig );
	SUBSCRIBE_GUI_EVENT( "Config/OK", CEGUI::PushButton::EventClicked, &GS_Config::handleConfig );
	SUBSCRIBE_GUI_EVENT( "Config/Quit", CEGUI::PushButton::EventClicked, &GS_Config::handleConfig );
	SUBSCRIBE_GUI_EVENT( "Config/Default", CEGUI::PushButton::EventClicked, &GS_Config::handleConfig );

	// Logo laden
	logo = engine.getTextureManager().get( "data\\logo.png" );
}

GS_Config::~GS_Config()
{
}

void GS_Config::onEnter( const ParameterMap& params )
{
	time = 0;
}

void GS_Config::onLeave()
{
}

void GS_Config::onActivate()
{
	// Fenster anzeigen
	gui.showAndFocusWindow( "Config" );

	updateScreen();
}

void GS_Config::onDeactivate()
{
}

void GS_Config::render()
{
	video.clear( Vec3f( 0.75f, 0.75f, 0.75f ) );

	// Logo rendern
	video.resetCamera();
	video.setupProjection( Vec2f( -1.0f, -1.0f ), Vec2f( 1.0f, 1.0f ) );
	video.setTexture( 0 );
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE );
	video.setTexture( logo );
	glPushMatrix();
	glTranslatef( 0.75f, 0.75f, 0.0f );
	glColor4f( 1.0f, 1.0f, 1.0f, 0.75f );
	video.drawSprite( Vec2i( 0, 0 ), Vec2i( 1024, 512 ), Vec2f( 0.6f, 0.3f ) );
	glPopMatrix();
	glColor4f( 1.0f, 1.0f, 1.0f, 0.1f );
	video.drawSprite( Vec2i( 0, 0 ), Vec2i( 1024, 512 ), Vec2f( 4.0f, 2.0f ) );
}

void GS_Config::update()
{
	time += engine.getTickDuration();
}

bool GS_Config::handleConfig( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "Config/NoSound" )
	{
		updateActivation();
	}
	else if( windowName == "Config/HardwareSound" )
	{
		updateActivation();
	}
	else if( windowName == "Config/Software3DSoundQuality" )
	{
		// Qualittsstufe als Text neben den Schieberegler schreiben
		unsigned int quality = static_cast<unsigned int>( static_cast<CEGUI::Slider*>( we.window )->getCurrentValue() );
		std::ostringstream str;
		str << quality;
		gui.getWindow( "Config/Software3DSoundQualityValue" )->setText( str.str() );
	}
	else if( windowName == "Config/OK" )
	{
		// Konfiguration speichern
		updateConfig();
		XMLSerializationStream str;
		Serializer::inst().serialize( config, str );
		str.save( "engine.cfg" );

		// Spiel starten
		config.set<int>( "Config.Result", 0 );
		engine.popGameState();
	}
	else if( windowName == "Config/Quit" )
	{
		// Spiel beenden
		config.set<int>( "Config.Result", 1 );
		engine.popGameState();
	}
	else if( windowName == "Config/Default" )
	{
		// Standardeinstellungen wiederherstellen
		config = defaultConfig;
		updateScreen();
	}
	else
	{
		return false;
	}

	return true;
}

void GS_Config::updateScreen()
{
	// Auflsung
	std::ostringstream name;
	name << "Config/R" << config.get<unsigned int>( "Video.Width" ) << "x" << config.get<unsigned int>( "Video.Height" );
	static_cast<CEGUI::RadioButton*>( gui.getWindow( name.str() ) )->setSelected( true );

	// Fenstermodus
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/Windowed" ) )->setSelected( config.get<bool>( "Video.Windowed" ) );

	// V-Sync
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/VSync" ) )->setSelected( config.get<bool>( "Video.VSync" ) );

	// MIP-Mapping
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/UseMIPMapping" ) )->setSelected( config.get<bool>( "Video.UseMIPMapping" ) );

	// kein Sound
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/NoSound" ) )->setSelected( config.get<bool>( "Audio.NoSound" ) );

	// Hardware-Sound
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/HardwareSound" ) )->setSelected( config.get<bool>( "Audio.UseHardware" ) );

	// Software-Soundqualitt
	unsigned int quality = config.get<unsigned int>( "Audio.Software3DQuality" );
	static_cast<CEGUI::Slider*>( gui.getWindow( "Config/Software3DSoundQuality" ) )->setCurrentValue( static_cast<float>( quality ) );
}

void GS_Config::updateActivation()
{
	bool hardwareSound = static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/HardwareSound" ) )->isSelected();
	gui.getWindow( "Config/Software3DSoundQuality" )->setEnabled( !hardwareSound );
	gui.getWindow( "Config/Software3DSoundQualityValue" )->setEnabled( !hardwareSound );

	bool noSound = static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/NoSound" ) )->isSelected();
	gui.getWindow( "Config/HardwareSound" )->setEnabled( !noSound );
	if( noSound )
	{
		gui.getWindow( "Config/Software3DSoundQuality" )->disable();
		gui.getWindow( "Config/Software3DSoundQualityValue" )->disable();
	}
}

void GS_Config::updateConfig()
{
	// Auflsung
	Vec2i resolution( 0, 0 );
	if( static_cast<CEGUI::RadioButton*>( gui.getWindow( "Config/R800x600" ) )->isSelected() ) resolution = Vec2i( 800, 600 );
	if( static_cast<CEGUI::RadioButton*>( gui.getWindow( "Config/R1024x768" ) )->isSelected() ) resolution = Vec2i( 1024, 768 );
	if( static_cast<CEGUI::RadioButton*>( gui.getWindow( "Config/R1280x1024" ) )->isSelected() ) resolution = Vec2i( 1280, 1024 );
	config.set<unsigned int>( "Video.Width", static_cast<unsigned int>( resolution.x ) );
	config.set<unsigned int>( "Video.Height", static_cast<unsigned int>( resolution.y ) );

	// Fenstermodus
	config.set<bool>( "Video.Windowed", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/Windowed" ) )->isSelected() );

	// V-Sync
	config.set<bool>( "Video.VSync", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/VSync" ) )->isSelected() );

	// MIP-Mapping
	config.set<bool>( "Video.UseMIPMapping", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/UseMIPMapping" ) )->isSelected() );

	// kein Sound
	config.set<bool>( "Audio.NoSound", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/NoSound" ) )->isSelected() );

	// Hardware-Sound
	config.set<bool>( "Audio.UseHardware", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Config/HardwareSound" ) )->isSelected() );

	// Software-Soundqualitt
	unsigned int quality = static_cast<unsigned int>( static_cast<CEGUI::Slider*>( gui.getWindow( "Config/Software3DSoundQuality" ) )->getCurrentValue() );
	config.set<unsigned int>( "Audio.Software3DQuality", quality );
}