#include <PCH.h>
#include <GS_Game.h>
#include <Level.h>
#include <Campaign.h>
#include <HighscoresDB.h>
#include <Console.h>
#include <Obj_Player.h>
#include <Obj_PlayerGhost.h>
#include <Obj_LevelTarget.h>

GS_Game::GS_Game( GameEnvironment& env ) : GameState( env.engine, "Game" ),
										   env( env ),
										   p_level( 0 ),
										   lastName( "Gib deinen Namen ein!" )
{
	// Layouts laden
	env.gui.loadLayout( "InGameMenu.layout", "GameRoot" );
	SUBSCRIBE_GUI_EVENT( "InGameMenu", CEGUI::Window::EventShown, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window", CEGUI::FrameWindow::EventCloseClicked, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window/Continue", CEGUI::PushButton::EventClicked, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window/Restart", CEGUI::PushButton::EventClicked, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window/RestartFromCheckpoint", CEGUI::PushButton::EventClicked, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window/Options", CEGUI::PushButton::EventClicked, &GS_Game::handleInGameMenu );
	SUBSCRIBE_GUI_EVENT( "InGameMenu/Window/Quit", CEGUI::PushButton::EventClicked, &GS_Game::handleInGameMenu );

	env.gui.loadLayout( "LevelCompleted.layout", "GameRoot" );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted", CEGUI::Window::EventShown, &GS_Game::handleLevelCompleted );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted/Window/EnterName", CEGUI::Editbox::EventTextAccepted, &GS_Game::handleLevelCompleted );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted/Window/EnterNameSubmit", CEGUI::PushButton::EventClicked, &GS_Game::handleLevelCompleted );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted/Window/PlayAgain", CEGUI::PushButton::EventClicked, &GS_Game::handleLevelCompleted );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted/Window/PlayNextLevel", CEGUI::PushButton::EventClicked, &GS_Game::handleLevelCompleted );
	SUBSCRIBE_GUI_EVENT( "LevelCompleted/Window/Quit", CEGUI::PushButton::EventClicked, &GS_Game::handleLevelCompleted );

	SETUP_PROFILE( "OBJECTS",						"Objekte",				"" );
	SETUP_PROFILE( "OBJECTS_UPDATE",				"aktualisieren",		"OBJECTS" );
	SETUP_PROFILE( "OBJECTS_VISIBILITY",			"Sichtbarkeitstest",	"OBJECTS" );
	SETUP_PROFILE( "OBJECTS_RENDER",				"rendern",				"OBJECTS" );
	SETUP_PROFILE( "PARTICLES",						"Partikelsystem",		"" );
	SETUP_PROFILE( "PARTICLES_UPDATE",				"aktualisieren",		"PARTICLES" );
	SETUP_PROFILE( "PARTICLES_RENDER",				"rendern",				"PARTICLES" );
	SETUP_PROFILE( "PHYSICS",						"Physik",				"" );
	SETUP_PROFILE( "PHYSICS_COLLISION_DETECTION",	"Kollisionserkennung",	"PHYSICS" );
	SETUP_PROFILE( "PHYSICS_STEP",					"Step",					"PHYSICS" );
}

GS_Game::~GS_Game()
{
}

void GS_Game::onEnter( const ParameterMap& params )
{
	env.logger.fine( "GS_Game::onEnter", LOGGER_AUTO );

	// HUD und Icons laden
	hud = engine.getTextureManager().get( "data\\hud.png" );
	icons = engine.getTextureManager().get( "data\\icons.png" );

	// Sounds laden
	ParameterMap options;
	options.set<std::string>( "Mode", "2D" );
	countDownSound = engine.getSoundManager().get( "data\\countdown.ogg", false, options );

	// Level laden
	currentLevel = params.get<std::string>( "Level" );
	currentCampaign = params.get<std::string>( "Campaign" );
	loadLevel( currentLevel );
}

void GS_Game::onLeave()
{
	env.logger.fine( "GS_Game::onLeave", LOGGER_AUTO );

	// Level lschen
	loadLevel( "" );

	audio.stopMusic();
}

void GS_Game::onActivate()
{
	env.logger.fine( "GS_Game::onActivate", LOGGER_AUTO );

	env.gui.getWindow( "GameRoot" )->show();

	optionShowPhysics = false;
	optionShowStatistics = false;
	optionSlowMotion = false;
	optionMoveCamera = false;
	paused = false;
}

void GS_Game::onDeactivate()
{
	env.logger.fine( "GS_Game::onDeactivate", LOGGER_AUTO );

	env.gui.getWindow( "GameRoot" )->hide();

	if( p_countDownSoundInst && countDownSound->isInstanceValid( p_countDownSoundInst ) )
	{
		p_countDownSoundInst->stop();
		p_countDownSoundInst = 0;
	}
}

void GS_Game::render()
{
	video.setupProjection( Vec2f( -24.0f, -18.0f ), Vec2f( 24.0f, 18.0f ) );
	video.setupMotionBlurring( env.options.getOptions().get<bool>( "MotionBlurring" ), clamp( motionBlurring / 100.0f, 0.0f, 0.002f ) );

	if( optionMoveCamera )
	{
		p_level->setCamera( camera );
		p_level->setCameraZoom( cameraZoom );
		p_level->setCameraAngle( 0.0f );
	}

	// Level rendern
	p_level->render();
	video.doMotionBlurring();
	video.setupMotionBlurring( false );

	if( optionShowPhysics ) renderPhysicsView();

	video.setupProjection( Vec2f( 0.0f, 0.0f ), Vec2f( 48.0f, 36.0f ) );

	if( optionShowStatistics ) renderStatistics();

	renderHUD();

	p_level->renderOverlays();
}

void GS_Game::update()
{
	if( p_countDownSoundInst )
	{
		if( !countDownSound->isInstanceValid( p_countDownSoundInst ) )
		{
			p_countDownSoundInst = 0;
		}
	}

	if( !env.gui.areWindowsVisible() && !Scripting::inst().getConsole()->isVisible() )
	{
		if( input.wasKeyPressed( KEY_PAUSE ) )
		{
			paused = !paused;
		}

		if( input.wasKeyPressed( KEY_RETURN ) && p_level->getType() == "TimeAttack" )
		{
			// Level neu starten
			loadLevel( currentLevel );
		}

		if( input.wasKeyPressed( KEY_r ) )
		{
			// Statistiken zurcksetzen
			Profiler::inst().reset();
		}

		if( input.wasKeyPressed( KEY_F1 ) ) optionShowPhysics = !optionShowPhysics;
		if( input.wasKeyPressed( KEY_F2 ) ) optionShowStatistics = !optionShowStatistics;
		if( input.wasKeyPressed( KEY_F3 ) ) optionSlowMotion = !optionSlowMotion;
		if( input.wasKeyPressed( KEY_F4 ) )
		{
			optionMoveCamera = !optionMoveCamera;
			if( optionMoveCamera )
			{
				camera = p_level->getCamera();
				cameraZoom = p_level->getCameraZoom();
			}
		}

		if( input.wasKeyPressed( KEY_F5 ) )
		{
			p_level->setFlags( p_level->getFlags() ^ LF_RENDER_AVOIDANCE_AREAS );
			p_level->setFlags( p_level->getFlags() ^ LF_RENDER_OBJECT_DEBUG_INFO );
		}

		if( optionMoveCamera )
		{
			camera += 0.1f * static_cast<Vec2f>( input.getMouseMotion() );
			if( input.isMouseButtonDown( MBTN_LEFT ) ) cameraZoom *= powf( 0.25f, env.dt );
			if( input.isMouseButtonDown( MBTN_RIGHT ) ) cameraZoom /= powf( 0.25f, env.dt );
			cameraZoom = clamp( cameraZoom, 0.25f, 4.0f );
		}

		if( input.wasKeyPressed( KEY_ESCAPE ) )
		{
			// Men anzeigen
			env.gui.getWindow( "InGameMenu" )->show();
		}

		if( paused ) return;

		if( input.wasKeyPressed( KEY_g ) )
		{
			GameObject* p_player = p_level->getObjectByID( "player" );
			if( p_player )
			{
				if( input.isKeyDown( KEY_LSHIFT ) )
				{
					// God Mode deaktivieren
					p_player->setCollisionBits( CB_PLAYER, CB_ALL );
				}
				else
				{
					// God Mode
					p_player->setCollisionBits( CB_NONE, CB_NONE );
				}
			}
		}

		if( input.wasKeyPressed( KEY_F8 ) )
		{
			// Ziel zerstren
			Obj_LevelTarget* p_target = static_cast<Obj_LevelTarget*>( p_level->getObjectByID( "target" ) );
			if( p_target ) p_target->destroy();
		}

		if( input.wasKeyPressed( KEY_BACKSPACE ) )
		{
			// Spieler zerstren
			Obj_Player* p_player = static_cast<Obj_Player*>( p_level->getObjectByID( "player" ) );
			if( p_player )
			{
				p_player->setShield( 0.0f );
				p_player->setHealth( 0.0f );
			}
		}

		if( optionSlowMotion )
		{
			static int frame = 0;
			if( frame++ % 10 ) return;
		}
	}
	else
	{
		// Wenn Fenster sichtbar sind, wird das Spiel solange pausiert.
		if( p_countDownSoundInst ) p_countDownSoundInst->setPaused( true );
		return;
	}

	// Level aktualisieren
	p_level->update();

	Obj_Player* p_player = static_cast<Obj_Player*>( p_level->getObjectByID( "player" ) );

	if( raceMode && raceCountDown > 0.0f )
	{
		// Im Rennmodus lsst sich die Rakete erst nach Ablauf des Count-Downs steuern.
		if( raceCountDown > 1.0f )
		{
			p_level->resetRaceTime();
			if( p_player )
			{
				p_player->setAIMode();

				if( raceCountDown > 1.5f )
				{
					lastPlayerPos = p_player->getPosition();
					lastPlayerVelocity = p_player->getVelocity();
					lastPlayerAngle = p_player->getAngle();
					lastPlayerAngularVelocity = p_player->getAngularVelocity();
					lastPlayerHealth = p_player->getHealth();
				}
				else if( raceCountDown == 1.5f && levelTry )
				{
					p_player->setPosition( lastPlayerPos );
					p_player->setVelocity( lastPlayerVelocity );
					p_player->setAngle( lastPlayerAngle );
					p_player->setAngularVelocity( lastPlayerAngularVelocity );
					p_player->setHealth( lastPlayerHealth );
				}
			}
		}
		else
		{
			if( p_replayGhost->isFrozen() )
			{
				if( p_player ) p_player->setUserMode();
				p_replayGhost->setFrozen( false );
				p_recordGhost->setFrozen( false );
				levelTry++;
			}
		}

		if( raceCountDown <= 4.0f )
		{
			if( !p_countDownSoundInst )
			{
				p_countDownSoundInst = countDownSound->createInstance( true );
				if( p_countDownSoundInst )
				{
					p_countDownSoundInst->play();
					if( levelTry ) p_countDownSoundInst->setPlayCursor( 2.5f );
				}
			}
			else
			{
				p_countDownSoundInst->setPaused( false );
			}
		}

		raceCountDown -= env.dt;
	}

	// Gesundheits- und Tankfllstandsanzeige anpassen
	float currentHealth = 0.0f;
	if( p_player )
	{
		adjustSmooth( healthDisplay, p_player->getHealth(), 10.0f, env.dt );
		adjustSmooth( fuelDisplay, p_player->getFuel(), 10.0f, env.dt );
		currentHealth = p_player->getHealth();
	}
	else
	{
		healthDisplay = 0.0f;
	}

	adjustSmooth( smileyOffset, Vec2f( 0.0f, 0.0f ), 8.0f, env.dt );
	adjustSmooth( motionBlurring, 0.0f, 10.0f, env.dt );
	if( motionBlurring < 1.0e-16f ) motionBlurring = 0.0f;

	if( currentHealth < oldHealth )
	{
		const float loss = oldHealth - currentHealth;

		// dem Smiley einen Sto versetzen
		smileyOffset += randVec2fDir() * clamp( loss * 0.1f, 0.0f, 1.0f );

		if( loss > 5.0f )
		{
			// Motion-Blurring
			motionBlurring += clamp( loss / 15.0f, 0.0f, 1.0f );
		}
	}

	if( !dialogShown && p_level->isFinished() )
	{
		if( p_level->getObjectByID( "target" ) )
		{
			// Level nicht geschafft!
			// Wurde ein Checkpoint erreicht?
			GameObject* p_checkpoint = p_level->getHighestCheckpoint();
			if( p_checkpoint )
			{
				if( rebirthCountDown == 0.0f )
				{
					// in 2 Sekunden beim Checkpoint wieder einsteigen
					rebirthCountDown = 2.0f;
					p_rebirthCheckpoint = p_checkpoint;
					p_level->setCameraTarget( p_checkpoint );
				}
			}
			else
			{
				// Men anzeigen!
				env.gui.getWindow( "InGameMenu" )->show();
				dialogShown = true;
			}
		}
		else
		{
			// Level geschafft - Dialog anzeigen!
			env.gui.getWindow( "LevelCompleted" )->show();
			dialogShown = true;
		}
	}

	oldHealth = currentHealth;

	if( rebirthCountDown > 0.0f )
	{
		rebirthCountDown -= env.dt;
		if( rebirthCountDown <= 0.0f )
		{
			// beim Checkpoint wieder einsteigen
			p_level->continueAtCheckpoint( p_rebirthCheckpoint );
			rebirthCountDown = 0.0f;
			p_rebirthCheckpoint = 0;
			p_level->setCameraTarget( p_level->getObjectByID( "player" ) );
		}
	}

	// Audio-Hrer an die Position der Kamera setzen
	audio.getListener()->setPosition( p_level->getCamera() );
	audio.getListener()->setZ( -10.0f );

	time += engine.getTickDuration();
}

void GS_Game::loadLevel( std::string level )
{
	if( p_level && p_level->getFilename() == level )
	{
		if( !raceMode ) levelTry++;
	}
	else
	{
		levelTry = 0;
	}

	delete p_level;
	p_level = 0;

	if( p_countDownSoundInst && countDownSound->isInstanceValid( p_countDownSoundInst ) )
	{
		p_countDownSoundInst->stop();
		p_countDownSoundInst = 0;
	}

	if( !level.empty() )
	{
		// Level laden
		p_level = new Level( env, level );

		// Musik abspielen
		p_level->playMusic();

		// wichtige Sounds vorbereiten, damit sie nicht whrend des Spiels geladen werden mssen
		SoundManager& soundMgr = engine.getSoundManager();
		SoundHandle sound = soundMgr.get( "data\\explosion1.ogg" );
		sound->setDefaultSoundGroupName( "SFX" );
		sound->setDefaultDistances( 25.0f, 10000.0f );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\explosion2.ogg" );
		sound->setDefaultSoundGroupName( "SFX" );
		sound->setDefaultDistances( 25.0f, 10000.0f );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\explosion3.ogg" );
		sound->setDefaultSoundGroupName( "SFX" );
		sound->setDefaultDistances( 25.0f, 10000.0f );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\vaporize.ogg" );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );

		ParameterMap options;
		options.set<std::string>( "Type", "Sample" );
		options.set<std::string>( "Mode", "2D" );
		sound = soundMgr.get( "data\\sms.ogg", false, options );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\beep.ogg", false, options );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\notify.ogg", false, options );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\succeeded.ogg", false, options );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );
		sound = soundMgr.get( "data\\failed.ogg", false, options );
		sound->setDefaultSoundGroupName( "SFX" );
		p_level->cacheSound( sound );

		// Spielersteuerung festlegen
		Controllable* p_player = dynamic_cast<Controllable*>( p_level->getObjectByID( "player" ) );
		p_player->setUserMode();

		// Die Kamera soll dieses Objekt verfolgen.
		p_level->setCameraTarget( dynamic_cast<GameObject*>( p_player ), true );
	}

	p_replayGhost = 0;
	p_recordGhost = 0;
	healthDisplay = 100.0f;
	fuelDisplay = 0.0f;
	oldHealth = 100.0f;
	smileyOffset = Vec2f( 0.0f, 0.0f );
	motionBlurring = 0.0f;
	dialogShown = false;
	raceMode = false;
	time = 0;
	raceCountDown = levelTry ? 1.5f : 5.0f;
	p_countDownSoundInst = 0;
	rebirthCountDown = 0.0f;
	p_rebirthCheckpoint = 0;

	if( p_level && p_level->getType() == "TimeAttack" )
	{
		// Ghost erzeugen
		const std::string ghostFilename( std::string( "data\\levels\\" ) + removeExtension( p_level->getFilename() ) + ".ghost" );
		p_replayGhost = new Obj_PlayerGhost( *p_level, ghostFilename, false );
		p_recordGhost = new Obj_PlayerGhost( *p_level, ghostFilename, true );
		p_replayGhost->setFrozen( true );
		p_recordGhost->setFrozen( true );

		raceMode = true;
	}
}

void GS_Game::renderHUD()
{
	Obj_Player* p_player = static_cast<Obj_Player*>( p_level->getObjectByID( "player" ) );

	// Hintergrund rendern
	video.setTexture( hud );
	glBegin( GL_QUADS );
	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
	glTexCoord2i( 0, 1 );
	glVertex2i( 0, 27 );
	glTexCoord2i( 1024, 1 );
	glVertex2i( 48, 27 );
	glTexCoord2i( 1024, 256 );
	glVertex2i( 48, 36 );
	glTexCoord2i( 0, 256 );
	glVertex2i( 0, 36 );
	glEnd();

	video.setTexture( icons );

	// Gesundheitsanzeige rendern
	int sprite;
	if( !p_level->getObjectByID( "target" ) ) sprite = 7;
	else if( healthDisplay <= 0.0f ) sprite = 6;
	else sprite = 5 - static_cast<int>( ( healthDisplay / 100.0f ) * 5.0f + 0.5f );
	glPushMatrix();
	glTranslatef( 16.5f + smileyOffset.x, 34.0f + smileyOffset.y, 0.0f );
	float scaling = 3.0f;
	float t = static_cast<float>( time ) / 1000.0f;
	if( sprite == 4 ) scaling += 0.1f * sinf( 15.0f * t );
	else if( sprite == 5 ) scaling += 0.15f * sinf( 20.0f * t );
	glScalef( scaling, scaling, 1.0f );
	video.drawSprite( Vec2i( sprite * 64, 192 ), Vec2i( 64, 64 ) );
	glPopMatrix();

	std::ostringstream lcdText;

	if( p_player )
	{
		// Tankfllstandsanzeige rendern
		glPushMatrix();
		glTranslatef( 21.75f, 33.5f, 0.0f );
		glScalef( 0.7f, 0.7f, 1.0f );
		glPushMatrix();
		glScalef( 9.0f, 4.5f, 1.0f );
		video.drawSprite( Vec2i( 0, 64 ), Vec2i( 192, 96 ) );
		glPopMatrix();
		glPushMatrix();
		glTranslatef( 0.0f, 1.25f, 0.0f );
		float angle = 120.0f * ( ( fuelDisplay / 100.0f ) - 0.5f );
		glTranslatef( 0.0f, 1.6f, 0.0f );
		glRotatef( angle, 0.0f, 0.0f, 1.0f );
		glTranslatef( 0.0f, -1.6f, 0.0f );
		glScalef( 3.0f, 6.0f, 1.0f );
		video.drawSprite( Vec2i( 192, 64 ), Vec2i( 64, 128 ) );
		glPopMatrix();
		glPopMatrix();

		// Waffeninformationen vom Spieler holen
		WeaponType currentWeapon = p_player->getCurrentWeapon();
		const WeaponInfo* p_weapons = p_player->getWeapons();

		glPushMatrix();
		glTranslatef( 27.0f, 33.85f, 0.0f );
		glScalef( 3.0f, 3.0f, 1.0f );

		for( int i = 0; i < WT_LAST_PLAYER; i++ )
		{
			if( p_weapons[i].initialized &&
				p_weapons[i].ammo )
			{
				// Transparenzgrad bestimmen
				float alpha = ( i == currentWeapon ) ? 1.0f : 0.5f;

				// Icon dieser Waffe rendern
				glColor4f( 1.0f, 1.0f, 1.0f, alpha );
				video.setTexture( icons );
				video.drawSprite( Vec2i( i * 64, 0 ), Vec2i( 64, 64 ) );

				if( p_weapons[i].ammo != ~0 )
				{
					// Munitionsanzeige
					std::ostringstream ammo;
					ammo << p_weapons[i].ammo;
					glPushMatrix();
					glTranslatef( -0.25f, 0.5f, 0.0f );
					env.normalFont->print( ammo.str(), 0.25f, Vec4f( 0.05f, 0.05f, 0.1f, alpha ) );
					glPopMatrix();
				}
			}
			else
			{
				// leere Karte rendern
				glColor4f( 1.0f, 1.0f, 1.0f, 0.25f );
				video.setTexture( icons );
				video.drawSprite( Vec2i( 448, 0 ), Vec2i( 64, 64 ) );
			}

			glTranslatef( 1.05f, -0.05f, 0.0f );
			glRotatef( -1.75f, 0.0f, 0.0f, 1.0f );
		}

		glPopMatrix();
	}

	if( raceMode )
	{
		if( raceCountDown > 1.0f && !p_replayGhost->isEmpty() )
		{
			// Ghost-Informationen anzeigen
			lcdText << "GHOST: ";
			lcdText.width( 11 );
			lcdText.flags( std::ios::right );
			lcdText << Level::formatTime( p_replayGhost->getTimeNeeded() );
		}
		else
		{
			lcdText << "TIME: ";
			lcdText.width( 12 );
			lcdText.flags( std::ios::right );
			lcdText << Level::formatTime( p_level->getRaceTime() );
		}
	}
	else
	{
		if( p_player )
		{
			bool healthLow = p_player->getHealth() <= 10.0f;
			bool fuelLow = p_player->getFuel() <= 10.0f;
			unsigned int secs = time / 1000;
			unsigned int quarterSecs = time / 250;

			if( healthLow && !( secs % 3 ) )
			{
				if( quarterSecs % 2 ) lcdText << "ARMOR LOW";
			}
			else if( fuelLow && !( ( secs + 1 ) % 3 ) )
			{
				if( quarterSecs % 2 ) lcdText << "FUEL LOW";
			}
			else
			{
				lcdText << "SCORE: ";
				lcdText.width( 10 );
				lcdText.flags( std::ios::right );
				lcdText << p_player->getScore();
			}
		}
		else
		{
			lcdText << "SCORE: ";
			lcdText.width( 10 );
			lcdText.flags( std::ios::right );
			lcdText << p_level->getFinalScore();
		}
	}

	if( !p_player )
	{
		unsigned int secs = time / 1000;
		if( secs % 2)
		{
			lcdText.str( "" );
			lcdText.width( 16 );
			lcdText.flags( std::ios::right );

			if( !p_level->getObjectByID( "target" ) )
			{
				// Level geschafft!
				lcdText << "LEVEL FINISHED";
			}
			else
			{
				// Spieler tot!
				lcdText << "MATH ERROR";
			}
		}
	}

	// Text fr das LCD-Display rendern
	glPushMatrix();
	glTranslatef( 39.4f, 32.55f, 0.0f );
	glRotatef( -5.0f, 0.0f, 0.0f, 1.0f );
	env.lcdFont->print( lcdText.str(), 0.65f, Vec4f( 0.05f, 0.05f, 0.1f, 0.9f ) );
	glPopMatrix();

	if( raceMode )
	{
		if( raceCountDown <= 4.0f && raceCountDown > 0.0f )
		{
			const int secondsLeft = static_cast<int>( raceCountDown );
			const float fraction = raceCountDown - floorf( raceCountDown );
			const float t = ( 4.0f - raceCountDown ) / 4.0f;

			std::ostringstream str;
			if( secondsLeft )
			{
				if( !levelTry ) str << secondsLeft;
			}
			else
			{
				str << "GO!";
			}

			float size = 20.0f - static_cast<float>( secondsLeft ) - 4.0f * fraction;
			if( !secondsLeft ) size += 2.0f * sinf( fraction * 20.0f );
			const Vec4f color( 1.0f - 0.75f * t, 0.5f + 0.5f * t, 0.1f, fraction * fraction );

			glPushMatrix();
			glTranslatef( 24.0f, 18.0f, 0.0f );
			env.countDownFont->print( str.str(), size * 1.1f, Vec4f( 0.0f, 0.0f, 0.0f, color.a * 0.5f ), FF_HCENTER | FF_VCENTER );
			env.countDownFont->print( str.str(), size, color, FF_HCENTER | FF_VCENTER );
			glPopMatrix();
		}
	}

	if( paused )
	{
		glPushMatrix();
		glTranslatef( 24.0f, 18.0f, 0.0f );
		env.scoreFont->print( "Pause", 4.5f, Vec4f( 0.0f, 0.0f, 0.0f, 0.5f ), FF_HCENTER | FF_VCENTER );
		env.scoreFont->print( "Pause", 4.0f, Vec4f( 0.5f, 1.0f, 0.25f, 1.0f ), FF_HCENTER | FF_VCENTER );
		glPopMatrix();
	}
}

void GS_Game::renderPhysicsView()
{
	double projectionMatrix[16];
	int viewport[4];
	glGetDoublev( GL_PROJECTION_MATRIX, projectionMatrix );
	glGetIntegerv( GL_VIEWPORT, viewport );

	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();

	video.setTexture( 0 );
	glBegin( GL_QUADS );
	glColor4f( 0.0f, 0.0f, 0.0f, 0.75f );
	glVertex2i( -24, -18 );
	glVertex2i( 24, -18 );
	glVertex2i( 24, 18 );
	glVertex2i( -24, 18 );
	glEnd();

	glLoadMatrixf( video.getCameraMatrix() );
	double cameraMatrix[16];
	glGetDoublev( GL_MODELVIEW_MATRIX, cameraMatrix );

	glBegin( GL_POINTS );
	glColor4f( 1.0f, 0.0f, 0.0f, 1.0f );

	const Vec2i resolution( video.getResolution() );
	for( int x = 0; x < resolution.x; x += 4 )
	{
		for( int y = ( ( x / 4 ) % 2 ) * 2; y < resolution.y; y += 4 )
		{
			double wx, wy, wz;
			gluUnProject( x, y, 1.0, cameraMatrix, projectionMatrix, viewport, &wx, &wy, &wz );

			if( p_level->castRay( Vec2d( wx, wy ), Vec2f( 0.0f, 0.0f ), 0.0f, CB_ALL, CB_ALL ) )
			{
				glVertex2d( wx, wy );
			}
		}
	}

	glEnd();
	glPopMatrix();
}

void GS_Game::renderStatistics()
{
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();

	video.setTexture( 0 );
	glBegin( GL_QUADS );
	glColor4f( 0.0f, 0.0f, 0.0f, 0.5f );
	glVertex2i( 0, 0 );
	glVertex2i( 48, 0 );
	glVertex2i( 48, 24 );
	glVertex2i( 0, 24 );
	glEnd();

	glTranslatef( 1.0f, 1.0f, 0.0f );
	std::ostringstream msg;
	msg.setf( std::ios::fixed );
	msg.precision( 3 );
	msg << "Objekte:                " << p_level->getObjects().size() << " (" << p_level->getNumVisibleObjects() << ")" << std::endl;
	msg << "Partikel:               " << p_level->getPS().getNumParticles() << std::endl;
	msg << "Partikel-Emitter:       " << p_level->getPS().getNumEmitters() << std::endl;
	msg << "Frame-Dauer:            " << ( video.getLastFrameDuration() * 1000.0f ) << " ms" << std::endl;
	msg << "Kameraposition:         " << p_level->getCamera().x << ", " << p_level->getCamera().y << std::endl;
	GameObject* p_player = p_level->getObjectByID( "player" );
	if( p_player )
	{
		msg << "Spielerposition:        " << p_player->getPosition().x << ", " << p_player->getPosition().y << std::endl;
		msg << "Spielergeschwindigkeit: " << p_player->getVelocity().length() << " k/s" << std::endl;
	}

	msg.precision( 10 );
	msg << "Motion-Blurring:        " << motionBlurring << std::endl;

	msg << std::endl << std::endl << std::endl << std::endl << std::endl;
	msg << Profiler::inst().getStatistics();

	env.fixedFont->print( msg.str(), 0.8f, Vec4f( 0.25f, 0.9f, 0.25f, 0.9f ) );

	glTranslatef( 18.0f, 0.0f, 0.0f );

	msg.str( "" );
	msg.flags( std::ios::left );
	msg << "Sichtbare Objekte:" << std::endl;
	int i = 0;
	const std::vector<GameObject*>& visibleObjects = p_level->getVisibleObjects();
	for( std::vector<GameObject*>::const_iterator it = visibleObjects.begin();
		 it != visibleObjects.end();
		 it++ )
	{
		msg.width( 30 );
		msg.setf( std::ios::left );
		msg << ( ( *it )->getID() + " (" + ( *it )->getType() + ")" );
		if( ++i == 4 )
		{
			msg << std::endl;
			i = 0;
		}
	}

	env.fixedFont->print( msg.str(), 0.5f, Vec4f( 0.25f, 0.9f, 0.25f, 0.9f ) );

	glPopMatrix();
}

bool GS_Game::handleInGameMenu( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "InGameMenu" )
	{
		env.gui.getWindow( "InGameMenu/Window" )->activate();
		env.gui.getWindow( "InGameMenu/Window/Continue" )->setEnabled( p_level->getObjectByID( "player" ) != 0 || p_level->getHighestCheckpoint() != 0 );
		env.gui.getWindow( "InGameMenu/Window/RestartFromCheckpoint" )->setEnabled( p_level->getObjectByID( "player" ) != 0 && p_level->getHighestCheckpoint() != 0 );
		return true;
	}
	else if( windowName == "InGameMenu/Window" ) ;
	else if( windowName == "InGameMenu/Window/Continue" ) ;
	else if( windowName == "InGameMenu/Window/Restart" )
	{
		// Level neu starten
		loadLevel( currentLevel );
	}
	else if( windowName == "InGameMenu/Window/RestartFromCheckpoint" )
	{
		// Spieler zerstren
		Obj_Player* p_player = static_cast<Obj_Player*>( p_level->getObjectByID( "player" ) );
		if( p_player )
		{
			p_player->setShield( 0.0f );
			p_player->setHealth( 0.0f );
		}
	}
	else if( windowName == "InGameMenu/Window/Options" )
	{
		// Optionen anzeigen
		env.options.show();
		return true;
	}
	else if( windowName == "InGameMenu/Window/Quit" )
	{
		// zurck zum Hauptmen
		engine.popGameState();
	}
	else
	{
		return false;
	}

	// Fenster wieder ausblenden
	env.gui.getWindow( "InGameMenu" )->hide();

	return true;
}

bool GS_Game::handleLevelCompleted( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "LevelCompleted" )
	{
		env.gui.getWindow( "LevelCompleted/Window" )->activate();
		env.gui.getWindow( "LevelCompleted/Window/PlayNextLevel" )->setEnabled( !getNextLevel( currentLevel, currentCampaign ).empty() );

		// bentigte Zeit bestimmen
		const unsigned int timeNeeded = ~0 - p_level->getFinalScore();

		// erreichte Punktzahl anzeigen
		std::ostringstream score;
		if( raceMode ) score << Level::formatTime( timeNeeded );
		else score << p_level->getFinalScore();
		env.gui.getWindow( "LevelCompleted/Window/Score" )->setText( score.str() );

		// Highscores anzeigen
		showHighscores();

		// vermerken, dass der Level geschafft wurde
		HighscoresDB::inst().setLevelCompleted( currentLevel );

		bool saveGhost = false;

		if( p_replayGhost && p_recordGhost )
		{
			if( timeNeeded < p_replayGhost->getTimeNeeded() )
			{
				// Es existiert noch kein Ghost, oder der Spieler war schneller als er.
				// Darum wird er in jedem Fall gespeichert.
				saveGhost = true;
			}
		}

		// War das ein Rekord?
		CEGUI::Editbox* p_enterName = static_cast<CEGUI::Editbox*>( env.gui.getWindow( "LevelCompleted/Window/EnterName" ) );
		const unsigned int better = HighscoresDB::inst().getPlaceInHighscore( currentLevel, p_level->getFinalScore() );
		const bool topScore = better == 0;
		const bool newHighscoreEntry = better < 5;
		env.gui.getWindow( "LevelCompleted/Window/EnterName" )->setEnabled( newHighscoreEntry );
		env.gui.getWindow( "LevelCompleted/Window/EnterNameSubmit" )->setEnabled( newHighscoreEntry );

		if( topScore ) saveGhost = true;

		if( newHighscoreEntry )
		{
			audio.playSound( better ? "data\\highscore.ogg" : "data\\highscore_best.ogg", "SFX" );

			p_enterName->setText( lastName );
			p_enterName->setSelection( 0, p_enterName->getText().length() );
			p_enterName->setCaratIndex( ~0 );
			p_enterName->activate();
		}
		else
		{
			p_enterName->setText( "Das ist nicht genug!" );
		}

		if( saveGhost )
		{
			// Ghost speichern
			if( p_recordGhost ) p_recordGhost->save();
		}
	}
	else if( windowName == "LevelCompleted/Window/EnterName" ||
			 windowName == "LevelCompleted/Window/EnterNameSubmit" )
	{
		CEGUI::Window* p_enterName = env.gui.getWindow( "LevelCompleted/Window/EnterName" );
		const std::string name( p_enterName->getText().c_str() );
		if( !name.empty() )
		{
			// Name merken
			lastName = p_enterName->getText();

			// eintragen
			HighscoresEntry entry;
			entry.name = name;
			entry.score = p_level->getFinalScore();
			HighscoresDB::inst().addHighscoreEntry( currentLevel, entry );

			// Bestenliste aktualisieren
			showHighscores();

			p_enterName->setText( "" );
			env.gui.getWindow( "LevelCompleted/Window/EnterName" )->deactivate();
			env.gui.getWindow( "LevelCompleted/Window/EnterName" )->disable();
			env.gui.getWindow( "LevelCompleted/Window/EnterNameSubmit" )->disable();
		}
	}
	else if( windowName == "LevelCompleted/Window/PlayAgain" )
	{
		// Level neu starten
		loadLevel( currentLevel );
		env.gui.getWindow( "LevelCompleted" )->hide();
	}
	else if( windowName == "LevelCompleted/Window/PlayNextLevel" )
	{
		// den nchsten Level holen
		currentLevel = getNextLevel( currentLevel, currentCampaign );
		loadLevel( currentLevel );
		env.gui.getWindow( "LevelCompleted" )->hide();
	}
	else if( windowName == "LevelCompleted/Window/Quit" )
	{
		// zurck zum Hauptmen
		env.gui.getWindow( "LevelCompleted" )->hide();
		engine.popGameState();
	}
	else
	{
		return false;
	}

	return true;
}

void GS_Game::showHighscores()
{
	// Bestenliste anzeigen
	std::string players, scores;
	HighscoresDB::inst().buildHighscores( currentLevel, players, scores );
	env.gui.getWindow( "LevelCompleted/Window/HighscoresPlayers" )->setText( reinterpret_cast<const CEGUI::utf8*>( players.c_str() ) );
	env.gui.getWindow( "LevelCompleted/Window/HighscoresScores" )->setText( scores );
}

std::string GS_Game::getNextLevel( const std::string& currentLevel,
								   const std::string& campaign ) const
{
	Campaign c( campaign );
	if( c.getError() ) return "";

	const std::vector<std::string>& levels = c.getLevels();
	for( size_t i = 0; i < levels.size(); i++ )
	{
		if( levels[i] == currentLevel )
		{
			if( i < levels.size() - 1 )
			{
				// Es gibt einen nchsten Level!
				return levels[i + 1];
			}
			else
			{
				// Das war der letzte Level!
				return "";
			}
		}
	}

	return "";
}