#include <PCH.h>
#include <GS_Menu.h>
#include <Level.h>
#include <Campaign.h>
#include <HighscoresDB.h>

struct MyItem : public CEGUI::ListboxTextItem
{
	MyItem( const std::string& text,
			const std::string& data,
			bool completed = false,
			bool disabled = false ) : CEGUI::ListboxTextItem( text ),
									  data( data )
	{
		setSelectionColours( CEGUI::colour( 1.0f, 1.0f, 1.0f, 1.0f ) );
		setSelectionBrushImage( "RRLook", "ListboxSelection" );
		setDisabled( disabled );
		if( disabled ) setTextColours( CEGUI::colour( 0.75f, 0.5f, 0.5f ) );
		else if( completed ) setTextColours( CEGUI::colour( 0.25f, 0.65f, 0.25f ) );
		else setTextColours( CEGUI::colour( 0.0f, 0.0f, 0.0f ) );
	}

	const std::string data;
};

GS_Menu::GS_Menu( GameEnvironment& env ) : GameState( env.engine, "Menu" ),
										   env( env )
{
	// Layouts laden
	env.gui.loadLayout( "MainMenu.layout", "MainMenuRoot" );
	SUBSCRIBE_GUI_EVENT( "MainMenu/StartGame", CEGUI::PushButton::EventClicked, &GS_Menu::handleMainMenu );
	SUBSCRIBE_GUI_EVENT( "MainMenu/Options", CEGUI::PushButton::EventClicked, &GS_Menu::handleMainMenu );
	SUBSCRIBE_GUI_EVENT( "MainMenu/Quit", CEGUI::PushButton::EventClicked, &GS_Menu::handleMainMenu );

	env.gui.loadLayout( "LevelSelection.layout", "MainMenuRoot" );
	SUBSCRIBE_GUI_EVENT( "LevelSelection", CEGUI::Window::EventShown, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window", CEGUI::FrameWindow::EventCloseClicked, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window/CampaignList", CEGUI::Combobox::EventListSelectionChanged, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window/LevelList", CEGUI::Listbox::EventSelectionChanged, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window/ResetLevelHighscores", CEGUI::PushButton::EventClicked, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window/Play", CEGUI::PushButton::EventClicked, &GS_Menu::handleLevelSelection );
	SUBSCRIBE_GUI_EVENT( "LevelSelection/Window/Quit", CEGUI::PushButton::EventClicked, &GS_Menu::handleLevelSelection );
}

GS_Menu::~GS_Menu()
{
}

void GS_Menu::onEnter( const ParameterMap& params )
{
	env.logger.fine( "GS_Menu::onEnter", LOGGER_AUTO );

	// Hintergrundbilder laden
	for( int i = 1; i <= 3; i++ )
	{
		std::ostringstream filename;
		filename << "data\\mainmenu" << i << ".jpg";
		ParameterMap options;
		options.set<bool>( "UseMIPMapping", false );
		images.push_back( engine.getTextureManager().get( filename.str(), false, options ) );
	}

	// Logo laden
	logo = engine.getTextureManager().get( "data\\logo.png" );

	// Papierbild laden
	paper = engine.getTextureManager().get( "data\\paper1.png" );
}

void GS_Menu::onLeave()
{
	env.logger.fine( "GS_Menu::onLeave", LOGGER_AUTO );
}

void GS_Menu::onActivate()
{
	env.logger.fine( "GS_Menu::onActivate", LOGGER_AUTO );

	env.gui.getWindow( "MainMenuRoot" )->show();
	env.gui.getWindow( "MainMenu" )->show();
	if( env.gui.getWindow( "LevelSelection" )->isVisible() ) env.gui.getWindow( "LevelSelection/Window" )->activate();
	listLevels();
	logoX = -4.0f;

	audio.playMusic( "data\\mainmenu.ogg", false );
}

void GS_Menu::onDeactivate()
{
	env.logger.fine( "GS_Menu::onDeactivate", LOGGER_AUTO );

	env.gui.getWindow( "MainMenuRoot" )->hide();
}

void GS_Menu::render()
{
	video.clear();
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE );

	// Hintergrundbilder rendern
	const float t = 0.25f * engine.getTimeInSeconds();
	for( size_t i = 0; i < images.size(); i++ )
	{
		const float img = static_cast<float>( i );
		Vec4f color( 1.0f + 0.1f * sinf( t ), 1.0f + 0.1f * sinf( 1.3f * t ), 1.0f + 0.1f * sinf( 1.7f * t ), 0.5f + 0.5f * cosf( t + img ) );

		video.setTexture( images[i] );
		video.drawFullScreenQuad( color, Vec2i( 0, 0 ), Vec2i( 1024, 768 ) );
	}

	// Papier rendern
	video.resetCamera();
	video.setupProjection( Vec2f( -1.0f, -1.0f ), Vec2f( 1.0f, 1.0f ) );
	video.setTexture( paper );
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glMatrixMode( GL_TEXTURE );
	glRotatef( 15.0f, 0.0f, 0.0f, 1.0f );
	glBegin( GL_QUADS );
	glColor4f( 1.0f, 1.0f, 1.0f, 0.5f );
	glTexCoord2i( 128 * 3, 0 );
	glVertex2f( 1.0f, -1.0f );
	glTexCoord2i( 128 * 3, 128 * 5 );
	glVertex2f( 1.0f, 1.0f );
	glColor4f( 1.0f, 1.0f, 1.0f, 0.0f );
	glTexCoord2i( 0, 128 * 5 );
	glVertex2f( 0.0f, 1.0f );
	glTexCoord2i( 0, 0 );
	glVertex2f( 0.0f, -1.0f );
	glEnd();
	glMatrixMode( GL_MODELVIEW );

	// Logo rendern
	video.setTexture( 0 );
	glPushMatrix();
	glTranslatef( -logoX, 0.0f, 0.0f );
	glBegin( GL_QUADS );
	glColor4f( 0.0f, 0.0f, 0.0f, 0.25f );
	glVertex2f( -1.0f, -0.5f );
	glVertex2f( 1.0f, -0.5f );
	glVertex2f( 1.0f, 0.5f );
	glVertex2f( -1.0f, 0.5f );
	glEnd();
	glPopMatrix();
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE );
	video.setTexture( logo );
	glColor4f( 1.0f, 1.0f, 1.0f, 0.75f );
	glPushMatrix();
	glTranslatef( logoX, 0.0f, 0.0f );
	video.drawSprite( Vec2i( 0, 0 ), Vec2i( 1024, 512 ), Vec2f( 2.0f, 1.0f ) );
	glPopMatrix();
}

void GS_Menu::update()
{
	adjustSmooth( logoX, 0.0f, 8.0f, env.dt );
}

bool GS_Menu::handleMainMenu( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "MainMenu/StartGame" )
	{
		// Levelauswahlbildschirm anzeigen
		env.gui.getWindow( "LevelSelection" )->show();
	}
	else if( windowName == "MainMenu/Options" )
	{
		// Optionen anzeigen
		env.options.show();
	}
	else if( windowName == "MainMenu/Quit" )
	{
		// Spiel beenden
		engine.popGameState();
	}
	else
	{
		return false;
	}

	return true;
}

bool GS_Menu::handleLevelSelection( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "LevelSelection" )
	{
		// Fenster fokussieren und Kampagnen auflisten
		env.gui.getWindow( "LevelSelection/Window" )->activate();
		listCampaigns( false );
	}
	else if( windowName == "LevelSelection/Window/CampaignList" )
	{
		// Levels auflisten
		listLevels( false );
	}
	else if( windowName == "LevelSelection/Window/LevelList" )
	{
		// ausgewhlten Level darstellen
		displayLevel();
	}
	else if( windowName == "LevelSelection/Window/ResetLevelHighscores" )
	{
		// Highscores zurcksetzen
		resetHighscores();
		displayLevel();
	}
	else if( windowName == "LevelSelection/Window" ||
			 windowName == "LevelSelection/Window/Quit" )
	{
		// Fenster wieder ausblenden
		env.gui.getWindow( "LevelSelection" )->hide();
	}
	else if( windowName == "LevelSelection/Window/Play" )
	{
		CEGUI::Listbox* p_levelList = static_cast<CEGUI::Listbox*>( env.gui.getWindow( "LevelSelection/Window/LevelList" ) );
		CEGUI::Combobox* p_campaignList = static_cast<CEGUI::Combobox*>( env.gui.getWindow( "LevelSelection/Window/CampaignList" ) );
		MyItem* p_levelItem = static_cast<MyItem*>( p_levelList->getFirstSelectedItem() );
		MyItem* p_campaignItem = static_cast<MyItem*>( p_campaignList->getSelectedItem() );
		if( p_levelItem && p_campaignItem )
		{
			// Parameter an das Spiel bergeben
			ParameterMap params;
			params.set<std::string>( "Level", p_levelItem->data );
			params.set<std::string>( "Campaign", p_campaignItem->data );
			params.set<std::string>( "Redirect.GameState", "Game" );
			engine.pushGameState( "Loading", params );
		}
	}
	else
	{
		return false;
	}

	return true;
}

void GS_Menu::listCampaigns( bool keepSelection )
{
	CEGUI::Combobox* p_list = static_cast<CEGUI::Combobox*>( env.gui.getWindow( "LevelSelection/Window/CampaignList" ) );
	size_t oldSelection = p_list->getSelectedItem() ? p_list->getItemIndex( p_list->getSelectedItem() ) : ~0;

	// Liste leeren
	p_list->resetList();

	// alle Kampagnen auflisten
	std::list<std::string> campaignFilenames = listFiles( "data\\levels\\*.campaign" );
	for( std::list<std::string>::const_iterator it = campaignFilenames.begin();
		 it != campaignFilenames.end();
		 it++ )
	{
		Campaign campaign( *it );
		if( !campaign.getError() )
		{
			// zur Liste hinzufgen
			MyItem* p_item = new MyItem( campaign.getName(), campaign.getFilename() );
			p_list->addItem( p_item );
		}
	}

	if( p_list->getItemCount() )
	{
		if( keepSelection && oldSelection != ~0 )
		{
			// den alten Eintrag auswhlen
			p_list->setItemSelectState( oldSelection, true );
		}
		else
		{
			// den ersten Eintrag auswhlen
			p_list->setItemSelectState( size_t( 0 ), true );
		}

		p_list->setText( p_list->getSelectedItem()->getText() );
	}

	listLevels( keepSelection );
}

void GS_Menu::listLevels( bool keepSelection )
{
	CEGUI::Listbox* p_list = static_cast<CEGUI::Listbox*>( env.gui.getWindow( "LevelSelection/Window/LevelList" ) );
	size_t oldSelection = p_list->getFirstSelectedItem() ? p_list->getItemIndex( p_list->getFirstSelectedItem() ) : ~0;

	// Liste leeren
	p_list->resetList();

	// aktuelle Campagne holen
	CEGUI::Combobox* p_campaignList = static_cast<CEGUI::Combobox*>( env.gui.getWindow( "LevelSelection/Window/CampaignList" ) );
	MyItem* p_selection = static_cast<MyItem*>( p_campaignList->getSelectedItem() );
	if( p_selection )
	{
		Campaign campaign( p_selection->data );
		if( !campaign.getError() )
		{
			// Levels auflisten
			const std::vector<std::string>& levels = campaign.getLevels();
			for( size_t i = 0; i < levels.size(); i++ )
			{
				std::string levelName, levelType, levelDesc;
				Level::getLevelInfo( std::string( levels[i] ), levelName, levelType, levelDesc );
				if( !levelName.empty() )
				{
					// zur Liste hinzufgen
					if( levelType == "TimeAttack" ) levelName += " (TA)";
					MyItem* p_item = new MyItem( levelName,
												 levels[i],
												 HighscoresDB::inst().wasLevelCompleted( levels[i] ),
												 !campaign.isLevelUnlocked( levels[i] ) );
					p_list->addItem( p_item );
				}
			}
		}
	}
	else
	{
		p_campaignList->setText( "" );
	}

	if( p_list->getItemCount() )
	{
		if( keepSelection && oldSelection != ~0 )
		{
			// den alten Eintrag auswhlen
			p_list->setItemSelectState( oldSelection, true );
		}
		else
		{
			// den ersten Eintrag auswhlen
			p_list->setItemSelectState( size_t( 0 ), true );
		}
	}

	displayLevel();
}

void GS_Menu::displayLevel()
{
	CEGUI::Window* p_levelName = env.gui.getWindow( "LevelSelection/Window/LevelName" );
	CEGUI::Window* p_levelDesc = env.gui.getWindow( "LevelSelection/Window/LevelDescription" );
	CEGUI::Window* p_levelHighscoresPlayers = env.gui.getWindow( "LevelSelection/Window/LevelHighscoresPlayers" );
	CEGUI::Window* p_levelHighscoresScores = env.gui.getWindow( "LevelSelection/Window/LevelHighscoresScores" );

	// zuerst alles zurcksetzen
	p_levelName->setText( "" );
	p_levelDesc->setText( "" );
	p_levelHighscoresPlayers->setText( "" );
	p_levelHighscoresScores->setText( "" );

	CEGUI::Listbox* p_list = static_cast<CEGUI::Listbox*>( env.gui.getWindow( "LevelSelection/Window/LevelList" ) );
	MyItem* p_item = static_cast<MyItem*>( p_list->getFirstSelectedItem() );
	if( p_item )
	{
		// Bestenliste anzeigen
		std::string players, scores;
		HighscoresDB::inst().buildHighscores( p_item->data, players, scores );
		p_levelHighscoresPlayers->setText( reinterpret_cast<const CEGUI::utf8*>( players.c_str() ) );
		p_levelHighscoresScores->setText( scores );

		if( p_item->isDisabled() )
		{
			// Dieser Level ist noch nicht freigeschaltet!
			p_levelName->setText( "???" );
			p_levelDesc->setText( "Dieser Level ist noch nicht freigeschaltet. Du musst zunchst die anderen schaffen!" );
			p_item = 0;
		}
		else
		{
			// Name und Beschreibung holen
			std::string levelName, levelType, levelDesc;
			Level::getLevelInfo( p_item->data, levelName, levelType, levelDesc );
			p_levelName->setText( levelName );
			p_levelDesc->setText( levelDesc );
		}
	}

	env.gui.getWindow( "LevelSelection/Window/Play" )->setEnabled( p_item != 0 );
}

void GS_Menu::resetHighscores()
{
	CEGUI::Listbox* p_list = static_cast<CEGUI::Listbox*>( env.gui.getWindow( "LevelSelection/Window/LevelList" ) );
	MyItem* p_item = static_cast<MyItem*>( p_list->getFirstSelectedItem() );
	if( p_item )
	{
		HighscoresDB::inst().resetHighscores( p_item->data );

		// Ghost lschen
		const std::string ghostFilename( std::string( "data\\levels\\" ) + removeExtension( p_item->data ) + ".ghost" );
		_unlink( ghostFilename.c_str() );
	}
}