#include <PCH.h>
#include <GS_TestArena.h>
#include <Level.h>
#include <Obj_Player.h>
#include <Obj_TestTarget.h>

GS_TestArena::GS_TestArena( GameEnvironment& env ) : GameState( env.engine, "TestArena" ),
													 env( env ),
													 p_level( 0 ),
													 p_candidate( 0 ),
													 p_candidateObj( 0 ),
													 p_testTarget( 0 ),
													 startPopulation( 0 ),
													 maxPopulation( 0 ),
													 currentNet( 0 ),
													 currentPass( 0 ),
													 numPasses( 0 ),
													 time( 0 )
{
}

GS_TestArena::~GS_TestArena()
{
}

void GS_TestArena::onEnter( const ParameterMap& params )
{
	mode = params.get<std::string>( "Mode", "Play" );

	if( mode == "Evolution" )
	{
		engine.setMode( EM_PROCESS_ONLY );
		env.logger.setLogLevel( LL_INFO );
		env.engine.getLogger().setLogLevel( LL_INFO );

		baseNetFilename = params.get<std::string>( "BaseNetFilename" );
		startPopulation = params.get<unsigned int>( "StartPopulation" );
		maxPopulation = params.get<unsigned int>( "MaxPopulation", 100 );
		numPasses = params.get<unsigned int>( "NumPasses", 1 );
		currentNet = 0;
		currentPass = 0;
		currentScore = 0.0f;

		// Population erzeugen
		for( unsigned int i = 0; i < startPopulation; i++ )
		{
			// Netz laden
			char temp[256] = "";
			sprintf( temp, "%s.%03d", baseNetFilename.c_str(), i );
			std::string netFilename( temp );
			std::cout << "Lade " << netFilename << " ..." << std::endl;
			NeuralNetHandle net = env.neuralNetMgr.get( netFilename );
			population.push_back( net );
		}
	}
	else if( mode == "Play" )
	{
		candidateBehavior = params.get<std::string>( "CandidateBehavior", "" );
	}

	setup();
}

void GS_TestArena::onLeave()
{
	cleanUp();
	population.clear();

	if( mode == "Evolution" )
	{
		engine.setMode( EM_NORMAL );
		env.logger.setLogLevel( LL_FINEST );
		env.engine.getLogger().setLogLevel( LL_FINEST );
	}
}

void GS_TestArena::onActivate()
{
}

void GS_TestArena::onDeactivate()
{
}

void GS_TestArena::render()
{
	if( mode == "Play" )
	{
		video.setupProjection( Vec2f( 0.0f, 0.0f ), Vec2f( 48.0f, 36.0f ) );

		// Level rendern
		p_level->render();

		// ein paar Informationen ausgeben
		glPushMatrix();
		glLoadIdentity();
		glTranslatef( 2.0f, 33.0f, 0.0f );
		env.fixedFont->print( "(A)I aktivieren - (U)ser Control - (T)rainieren", 1.25f, Vec4f( 0.05f, 0.05f, 0.2f, 0.9f ) );
		glPopMatrix();

		if( p_candidate )
		{
			if( p_candidate->isTraining() || p_candidate->getControlMode() == CM_AI_NET )
			{
				if( p_candidate->isTraining() )
				{
					glPushMatrix();
					glLoadIdentity();
					glTranslatef( 2.0f, 34.5f, 0.0f );
					env.fixedFont->print( "Trainiere - Enter zum Speichern, Backspace zum Abbrechen", 1.25f, Vec4f( 0.05f, 0.05f, 0.2f, 0.9f ) );
					glPopMatrix();
				}

				glPushMatrix();
				glTranslatef( 2.0f, 2.0f, 0.0f );
				env.fixedFont->print( p_candidate->dumpParameters(), 1.0f, Vec4f( 0.0f, 0.25f, 0.0f, 0.9f ) );
				glPopMatrix();
			}
		}
	}
}

void GS_TestArena::update()
{
	if( input.wasKeyPressed( KEY_ESCAPE ) ) engine.popGameState();

	// Level aktualisieren
	p_level->update();

	if( mode == "Evolution" )
	{
		// eine Minute lang simulieren
		time += env.engine.getTickDuration();
		if( time >= 1 * 60 * 1000 )
		{
			evaluate();
			cleanUp();
			setup();
		}
	}
	else if( mode == "Play" )
	{
		if( input.wasKeyPressed( KEY_u ) ) p_candidate->setUserMode();
		if( input.wasKeyPressed( KEY_a ) ) p_candidate->setAIMode();
		if( input.wasKeyPressed( KEY_t ) ) p_candidate->startTraining();
		if( input.wasKeyPressed( KEY_BACKSPACE ) ) p_candidate->stopTraining( false );
		if( input.wasKeyPressed( KEY_RETURN ) ) p_candidate->stopTraining( true );

		// Audio-Hrer an die Position der Kamera setzen
		audio.getListener()->setPosition( p_level->getCamera() );
		audio.getListener()->setZ( -10.0f );
	}
}

void GS_TestArena::setup()
{
	// Level laden
	bool testMode = false;
	if( mode == "Evolution" ) testMode = true;
	p_level = new Level( env, "data\\test_arena.xml", testMode );

	// Testziel erzeugen
	p_testTarget = new Obj_TestTarget( *p_level, "player" );

	// Spielersteuerung festlegen
	p_candidateObj = p_level->getObjectByID( "candidate" );
	p_candidateObj->setTargetID( "player" );
	p_candidateObj->setPosition( Vec2f( randFloat( 5.0f, 35.0f ), randFloat( 5.0f, 25.0f ) ) );
	p_candidate = dynamic_cast<Controllable*>( p_candidateObj );
	p_candidate->setTrainingFilenamePrefix( "training\\session_" );

	// Die Kamera soll dieses Objekt verfolgen.
	p_level->setCameraTarget( p_candidateObj );

	if( mode == "Evolution" )
	{
		p_candidate->setAIMode();
		p_candidate->setAINeuralNet( population[currentNet] );
	}
	else if( mode == "Play" )
	{
		if( !candidateBehavior.empty() ) p_candidate->setAIBehavior( candidateBehavior );
		p_candidate->setUserMode();
	}

	time = 0;
}

void GS_TestArena::evaluate()
{
	float score = static_cast<float>( p_testTarget->getHitCount() );
	currentScore += score;

	if( numPasses > 1 ) std::cout << ".";

	currentPass++;
	if( currentPass == numPasses )
	{
		float averageScore = currentScore / static_cast<float>( numPasses );
		std::cout << " *** Durchschnitt: " << averageScore << std::endl;

		currentPass = 0;
		currentScore = 0.0f;

		currentNet++;
		if( currentNet == static_cast<unsigned int>( population.size() ) )
		{
			currentNet = 0;
			engine.popGameState();
		}
	}
}

void GS_TestArena::cleanUp()
{
	// Level lschen
	delete p_level;
}