#include <PCH.h>
#include <GameObject.h>
#include <Obj_Debris.h>
#include <Obj_BulletHole.h>
#include <Scripting.h>

GameObject::GameObject( Level& level, 
						const TiXmlElement* p_root,
						bool createBody,
						const std::string& overrideID ) : level( level ),
														  env( level.getEnv() ),
														  video( level.getEnv().engine.getVideo() ),
														  body( 0 ),
														  material( "Unknown" ),
														  categoryBits( ~0 ),
														  collideBits( ~0 ),
														  movementFriction( 1.0f ),
														  boundCenter( 0.0f, 0.0f ),
														  boundRadius( -1.0f ),
														  boundMin( FLT_MAX ),
														  boundMax( -FLT_MAX ),
														  dead( false ),
														  turnToZombie( false ),
														  visible( true ),
														  frozen( false ),
														  passThrough( false ),
														  playerCollisionDamageScale( 1.0f )
{
	// Masse auf null setzen
	dMassSetZero( &mass );

	const char* p_id = 0;
	if( p_root )
	{
		// ID-Attribut vorhanden?
		p_id = p_root->Attribute( "id" );
	}

	if( !overrideID.empty() ) p_id = overrideID.c_str();

	if( p_id ) id = p_id;
	else
	{
		// fortlaufend nummerierte Standardbezeichnung vergeben
		id = level.getObjectAutoID();
	}

	if( createBody )
	{
		// ODE-Body erzeugen
		body = dBodyCreate( level.getWorld() );
		dBodySetData( body, this );

		// 2D-Constraint erzeugen
		planeConstraint = dJointCreatePlane2D( level.getWorld(), 0 );
		dJointAttach( planeConstraint, body, 0 );
	}

	if( p_root )
	{
		// Position lesen
		Vec2f position( 0.0f, 0.0f );
		p_root->QueryFloatAttribute( "x", &position.x );
		p_root->QueryFloatAttribute( "y", &position.y );
		setPosition( position );

		// Ausrichtung (Winkel) lesen
		float angle = 0.0f;
		p_root->QueryFloatAttribute( "angle", &angle );
		angle /= 57.295779513082320876798154814105f;
		setAngle( angle );

		// Gravitationsbeeinflussung lesen
		int gravity = 1;
		p_root->QueryIntAttribute( "gravity", &gravity );
		setGravityInfluence( gravity ? true : false );

		// Reibung lesen
		float movementFriction_ = 0.75f;
		float rotationFriction_ = 0.1f;
		p_root->QueryFloatAttribute( "movementFriction", &movementFriction_ );
		p_root->QueryFloatAttribute( "rotationFriction", &rotationFriction_ );
		setFriction( movementFriction_, rotationFriction_ );

		// Geschwindigkeit lesen
		Vec2f velocity( 0.0f, 0.0f );
		p_root->QueryFloatAttribute( "velocityX", &velocity.x );
		p_root->QueryFloatAttribute( "velocityY", &velocity.y );
		setVelocity( velocity );

		// Winkelgeschwindigkeit lesen
		float angularVelocity = 0.0f;
		p_root->QueryFloatAttribute( "angularVelocity", &angularVelocity );
		angularVelocity /= 57.295779513082320876798154814105f;
		setAngularVelocity( angularVelocity );

		// Sichtbarkeit lesen
		int visible_ = 1;
		p_root->QueryIntAttribute( "visible", &visible_ );
		setVisible( visible_ ? true : false );

		// Eingefroren?
		int frozen_ = 0;
		p_root->QueryIntAttribute( "frozen", &frozen_ );
		setFrozen( frozen_ ? true : false );

		// Pass-Through?
		int passThrough_ = 0;
		p_root->QueryIntAttribute( "passThrough", &passThrough_ );
		setPassThrough( passThrough_ ? true : false );

		// Spielerkollisionsfaktor
		p_root->QueryFloatAttribute( "playerCollisionDamageScale", &playerCollisionDamageScale );
	}

	if( !p_root )
	{
		if( level.adoptObject( this ) )
    {
      delete this;
    }
	}
}

GameObject::~GameObject()
{
	deleteGeoms();
	deleteBody();

	env.logger.finest( "Objekt mit der ID " + id + " zerstrt", LOGGER_AUTO );
}

void GameObject::setTransform( const Mat33f& transform )
{
	if( !body ) Placeable::setTransform( transform );
	else
	{
		// Position setzen
		dBodySetPosition( body, transform.m13, transform.m23, 0.0 );

		// Rotationsmatrix zusammenbasteln
		dMatrix3 r;
		r[0] = transform.m11;	r[1] = transform.m12;	r[2] = 0.0;		r[3] = 0.0;
		r[4] = transform.m21;	r[5] = transform.m22;	r[6] = 0.0;		r[7] = 0.0;
		r[8] = 0.0;				r[9] = 0.0;				r[10] = 1.0;	r[11] = 0.0;

		// Orientierung setzen
		dBodySetRotation( body, r );
	}
}

Mat33f GameObject::getTransform() const
{
	if( !body ) return Placeable::getTransform();
	else
	{
		Mat33f m;

		// Position holen
		const dReal* p_pos = dBodyGetPosition( body );
		m.m13 = static_cast<float>( p_pos[0] );
		m.m23 = static_cast<float>( p_pos[1] );

		// Orientierung holen
		const dReal* p_rot = dBodyGetRotation( body );
		m.m11 = static_cast<float>( p_rot[0] );
		m.m12 = static_cast<float>( p_rot[1] );
		m.m21 = static_cast<float>( p_rot[4] );
		m.m22 = static_cast<float>( p_rot[5] );
		m.m31 = m.m32 = 0.0f;
		m.m33 = 1.0f;

		return m;
	}
}

void GameObject::update()
{
}

void GameObject::updatePhysics()
{
	if( body )
	{
		// Bewegungsreibung
		if( movementFriction != 1.0f )
		{
			const float f = powf( movementFriction, env.dt );
			const dReal* p_vel = dBodyGetLinearVel( body );
			dBodySetLinearVel( body, p_vel[0] * f, p_vel[1] * f, 0.0 );
		}

		// Drehungsreibung
		if( rotationFriction != 1.0f )
		{
			const float f = powf( rotationFriction, env.dt );
			const dReal* p_vel = dBodyGetAngularVel( body );
			dBodySetAngularVel( body, 0.0, 0.0, p_vel[2] * f );
		}
	}
}

void GameObject::render( RenderPass pass )
{
}

void GameObject::onCollision( GameObject* p_other,
							  CollisionInfo& info )
{
	Scripting::inst().call( collisionCallback, this, p_other );
}

void GameObject::onKill()
{
}

void GameObject::onRebirth()
{
}

bool GameObject::isToBeDeleted() const
{
	return dead;
}

void GameObject::addAvoidanceAreas()
{
}

void GameObject::setVelocity( const Vec2f& velocity )
{
	if( !body ) return;
	else dBodySetLinearVel( body, velocity.x, velocity.y, 0.0 );
}

Vec2f GameObject::getVelocity() const
{
	if( !body ) return Vec2f( 0.0f, 0.0f );
	else
	{
		const dReal* p_vel = dBodyGetLinearVel( body );
		return Vec2d( p_vel[0], p_vel[1] );
	}
}

void GameObject::setAngularVelocity( float angularVelocity )
{
	if( !body ) return;
	else dBodySetAngularVel( body, 0.0, 0.0, angularVelocity );
}

float GameObject::getAngularVelocity() const
{
	if( !body ) return 0.0f;
	else
	{
		const dReal* p_vel = dBodyGetAngularVel( body );
		return static_cast<float>( p_vel[2] );
	}
}

void GameObject::addSegment( const Vec2f& from,
							 const Vec2f& to,
							 float radius )
{
	// Punkte, wenn das Objekt statisch ist, transformieren
	const Vec2f from_( body ? from : relPositionToAbsPosition( from ) );
	const Vec2f to_( body ? to : relPositionToAbsPosition( to ) );

	Vec2f dir = to_ - from_;

	// Transformations-Shape erzeugen
	dGeomID transform = dCreateGeomTransform( level.getSpace() );
	dGeomSetData( transform, this );

	// Zylinder mit Halbkugeln an den Seiten erzeugen
	float length = dir.length() - 0.01f;
	dGeomID capsule = dCreateCapsule( 0, radius, length );
	dGeomSetData( capsule, this );

	// Zylinder in die richtige Lage bringen
	dMatrix3 r;
	dRFrom2Axes( r, 0.0, 0.0, -1.0, -dir.y, dir.x, 0.0 );
	dGeomSetRotation( capsule, r );
	const Vec2f position( from_ + 0.5f * dir );
	dGeomSetPosition( capsule, position.x, position.y, 0.0 );

	// Masse anpassen
	dMass m;
	dMassSetCapsule( &m, 1.0, 3, radius, length );
	dMassRotate( &m, r );
	dMassTranslate( &m, position.x, position.y, 0.0 );
	dMassAdd( &mass, &m );

	// Zylinder dem Transformations-Shape zuweisen
	dGeomTransformSetGeom( transform, capsule );

	// Transformations-Shape dem Krper zuweisen, falls es einen gibt
	if( body ) dGeomSetBody( transform, body );

	// Kollisions-Bits setzen
	dGeomSetCategoryBits( transform, categoryBits );
	dGeomSetCollideBits( transform, collideBits );
	dGeomSetCategoryBits( capsule, categoryBits );
	dGeomSetCollideBits( capsule, collideBits );

	geoms.push_back( transform );
	geoms.push_back( capsule );
}

void GameObject::addPolyLine( const std::vector<Vec2f>& vertices,
							  bool close,
							  float radius,
							  float shift )
{
	struct Edge
	{
		size_t v1;
		size_t v2;
		Vec2f normal;
	};

	struct Vertex
	{
		Vec2f position;
		std::vector<unsigned int> edges;
	};

	// Vertizes kopieren
	std::vector<Vertex> newVertices;
	for( size_t i = 0; i < vertices.size(); i++ )
	{
		Vertex v;
		v.position = vertices[i];
		newVertices.push_back( v );
	}

	// Vertex- und Kantenliste erstellen
	std::vector<Edge> edges;
	for( size_t v1 = 0; v1 < vertices.size() - ( close ? 0 : 1 ); v1++ )
	{
		size_t v2 = ( v1 + 1 ) % vertices.size();
		Edge e;
		e.v1 = v1;
		e.v2 = v2;
		e.normal = ( vertices[v1] - vertices[v2] ).perp().normalized();
		edges.push_back( e );

		size_t edgeNo = v1;
		newVertices[v1].edges.push_back( edgeNo );
		newVertices[v2].edges.push_back( edgeNo );
	}

	// Vertizes verschieben
	for( size_t i = 0; i < newVertices.size(); i++ )
	{
		Vertex& v = newVertices[i];
		switch( v.edges.size() )
		{
		case 0:
			// Das drfte nicht passieren!
			assert( false );
			break;
		case 1:
			// Dieser Vertex gehrt nur zu einer Kante!
			v.position += shift * radius * edges[v.edges[0]].normal;
			break;
		case 2:
			// Dieser Vertex ist Teil von zwei Kanten!
			// Beide Kanten werden verschoben.
			// Anschlieend wird der Schnittpunkt bestimmt.
			const Edge& e1 = edges[v.edges[0]];
			const Edge& e2 = edges[v.edges[1]];
			if( ( e1.normal ^ e2.normal ) > 0.99f )
			{
				// Die Kanten sind parallel!
				v.position += shift * radius * e1.normal;
			}
			else
			{
				const Vec2f e1A( vertices[e1.v1] + shift * radius * e1.normal );
				const Vec2f e1B( vertices[e1.v2] + shift * radius * e1.normal );
				const Vec2f e2A( vertices[e2.v1] + shift * radius * e2.normal );
				const Vec2f e2B( vertices[e2.v2] + shift * radius * e2.normal );
				bool r = intersect2DLines( e1A, e1B, e2A, e2B, &v.position );
				assert( r );
			}
			break;
		}
	}

	// Zylinder hinzufgen
	for( size_t i = 0; i < edges.size(); i++ )
	{
		const Edge& e = edges[i];
		const Vec2f& v1 = newVertices[e.v1].position;
		const Vec2f& v2 = newVertices[e.v2].position;

		addSegment( v1, v2, radius );
	}
}

void GameObject::addCircle( const Vec2f& position,
						    float radius )
{
	// Punkt, wenn das Objekt statisch ist, transformieren
	const Vec2f position_( body ? position : relPositionToAbsPosition( position ) );

	// Transformations-Shape erzeugen
	dGeomID transform = dCreateGeomTransform( level.getSpace() );
	dGeomSetData( transform, this );

	// Kugel erzeugen
	dGeomID sphere = dCreateSphere( 0, radius );
	dGeomSetData( sphere, this );

	// Kugel in die richtige Lage bringen
	dGeomSetPosition( sphere, position_.x, position_.y, 0.0 );

	// Masse anpassen
	dMass m;
	dMassSetSphere( &m, 1.0, radius );
	dMassTranslate( &m, position_.x, position_.y, 0.0 );
	dMassAdd( &mass, &m );

	// Kugel dem Transformations-Shape zuweisen
	dGeomTransformSetGeom( transform, sphere );

	// Transformations-Shape dem Krper zuweisen, falls es einen gibt
	if( body ) dGeomSetBody( transform, body );

	// Kollisions-Bits setzen
	dGeomSetCategoryBits( transform, categoryBits );
	dGeomSetCollideBits( transform, collideBits );
	dGeomSetCategoryBits( sphere, categoryBits );
	dGeomSetCollideBits( sphere, collideBits );

	geoms.push_back( transform );
	geoms.push_back( sphere );
}

void GameObject::setTotalMass( float totalMass )
{
	if( body )
	{
		dMassAdjust( &mass, totalMass );
		dBodySetMass( body, &mass );
	}
}

float GameObject::getTotalMass() const
{
	if( body ) return static_cast<float>( mass.mass );
	else return FLT_MAX;
}

void GameObject::setCollisionBits( unsigned long categoryBits,
								   unsigned long collideBits )
{
	setCategoryBits( categoryBits );
	setCollideBits( collideBits );
}

void GameObject::setCategoryBits( unsigned long categoryBits )
{
	if( this->categoryBits == categoryBits ) return;

	this->categoryBits = categoryBits;

	// die neuen Bits fr alle bereits bestehenden Shapes setzen
	for( std::list<dGeomID>::iterator it = geoms.begin();
		 it != geoms.end();
		 it++ )
	{
		dGeomSetCategoryBits( *it, categoryBits );
	}
}

unsigned long GameObject::getCategoryBits() const
{
	return categoryBits;
}

void GameObject::setCollideBits( unsigned long collideBits )
{
	if( this->collideBits == collideBits ) return;

	this->collideBits = collideBits;

	// die neuen Bits fr alle bereits bestehenden Shapes setzen
	for( std::list<dGeomID>::iterator it = geoms.begin();
		 it != geoms.end();
		 it++ )
	{
		dGeomSetCollideBits( *it, collideBits );
	}
}

unsigned long GameObject::getCollideBits() const
{
	return collideBits;
}

void GameObject::setFriction( float movementFriction,
							  float rotationFriction )
{
	setMovementFriction( movementFriction );
	setRotationFriction( rotationFriction );
}

void GameObject::setMovementFriction( float movementFriction )
{
	this->movementFriction = movementFriction;
}

void GameObject::setRotationFriction( float rotationFriction )
{
	this->rotationFriction = rotationFriction;
}

float GameObject::getMovementFriction() const
{
	return movementFriction;
}

float GameObject::getRotationFriction() const
{
	return rotationFriction;
}

void GameObject::setGravityInfluence( bool gravityInfluence )
{
	if( !body ) return;
	dBodySetGravityMode( body, gravityInfluence ? 1 : 0 );
}

bool GameObject::getGravityInfluence() const
{
	if( !body ) return false;
	return dBodyGetGravityMode( body ) != 0;
}

void GameObject::setMaterial( const std::string& material )
{
	this->material = material;
}

const std::string& GameObject::getMaterial() const
{
	return material;
}

void GameObject::setTargetID( const std::string& targetID )
{
	this->targetID = targetID;
}

const std::string& GameObject::getTargetID() const
{
	return targetID;
}

void GameObject::setTarget( GameObject* p_target )
{
	targetID = p_target ? p_target->getID() : "";
}

GameObject* GameObject::getTarget() const
{
	return level.getObjectByID( targetID );
}

void GameObject::processCollision( GameObject* p_other,
								   CollisionInfo& info )
{
	// Daten vervollstndigen
	completeCollision( p_other, info );

	// Kollisionseffekte verarbeiten
	processCollisionEffects( p_other, info );

	// virtuelle Methode aufrufen
	onCollision( p_other, info );
}

void GameObject::setCollisionEnabled( GameObject* p_other,
									  bool enable )
{
	if( enable ) excludeCollisions.erase( p_other );
	else excludeCollisions.insert( p_other );
}

bool GameObject::isCollisionEnabled( GameObject* p_other ) const
{
	return excludeCollisions.find( p_other ) == excludeCollisions.end();
}

bool GameObject::reflectsLaser() const
{
	return false;
}

bool GameObject::canSee( GameObject* p_other,
						 float maxDistance,
						 const Vec2f& relRayStartPosition ) const
{
	if( !p_other ) return false;

	// Ist das Objekt in Reichweite und sichtbar?
	const Vec2f myPosition( relPositionToAbsPosition( relRayStartPosition ) );
	const Vec2f targetPosition( p_other->getPosition() );
	const float distToTarget = ( targetPosition - myPosition ).length();
	if( distToTarget <= maxDistance )
	{
		GameObject* p_hit = level.castRay( myPosition, targetPosition - myPosition, distToTarget, CB_NONE, CB_BLOCK_VIEW, 0, 0 );
		if( !p_hit || p_hit == p_other )
		{
			// Sichtbar!
			return true;
		}
	}

	return false;
}

void GameObject::explode( int numDebris,
						  DebrisType debrisType,
						  float radius,
						  float scale,
						  bool makeHole,
						  int explosionSound,
						  float volume,
						  float pitch )
{
	const Vec2f myPosition( getPosition() );

	for( int i = 0; i < numDebris; i++ )
	{
		int type = debrisType;
		if( type == DT_RANDOM ) type = randInt( DT_FIRST + 1, DT_LAST - 1 );
		Obj_Debris* p_debris = new Obj_Debris( level, static_cast<DebrisType>( type ) );

		const float dist = randFloat( 0.0f, radius );
		const Vec2f dir( randVec2fDir() );
		p_debris->setPosition( myPosition + dist * dir );

		p_debris->setVelocity( dir * randFloat( 10.0f, 30.0f ) * dist );

		p_debris->setAngularVelocity( randFloat( -20.0f, 20.0f ) );
	}

	if( makeHole )
	{
		// Brandfleck erzeugen
		Obj_BulletHole* p_hole = new Obj_BulletHole( level, 3 );
		p_hole->setPosition( myPosition );
	}

	// Emitter erzeugen
	SimpleEmitter* p_explosionEmitter = level.getPS().createSimpleEmitter();
	p_explosionEmitter->setPosition( myPosition );
	p_explosionEmitter->setRate( 75.0f );
	p_explosionEmitter->setTimeToLive( 0.4f );
	p_explosionEmitter->setVisibilityBehavior( false, 10.0f );
	p_explosionEmitter->addParticleStyle( 7 );
	p_explosionEmitter->setParticleTimeToLive( 0.4f, 0.1f );
	p_explosionEmitter->setParticlePositionRadius( radius * 2.0f );
	p_explosionEmitter->setParticleSpeed( 5.0f, 2.0f );
	p_explosionEmitter->setParticleStartColor( Vec4f( 1.0f, 0.5f, 0.0f, 5.0f ), Vec4f( 0.2f, 0.2f, 0.1f, 1.0f ) );
	p_explosionEmitter->setParticleEndColor( Vec4f( 3.0f, 3.0f, 3.0f, 0.0f ), Vec4f( 0.1f, 0.1f, 0.1f, 0.0f ) );
	p_explosionEmitter->setParticleStartSize( 4.0f * scale, 1.0f * scale );
	p_explosionEmitter->setParticleEndSize( 8.0f * scale, 4.0f * scale );
	p_explosionEmitter->setParticleGravity( 0.0f, 0.1f );
	p_explosionEmitter->setParticleAnimFrame( 0.0f, 10.0f );

	p_explosionEmitter = level.getPS().createSimpleEmitter();
	p_explosionEmitter->setPosition( myPosition );
	p_explosionEmitter->setRate( 50.0f );
	p_explosionEmitter->setTimeToLive( 0.4f );
	p_explosionEmitter->setVisibilityBehavior( false, 10.0f );
	p_explosionEmitter->addParticleStyle( 8 );
	p_explosionEmitter->setParticleTimeToLive( 0.4f, 0.1f );
	p_explosionEmitter->setParticlePositionRadius( radius * 2.0f );
	p_explosionEmitter->setParticleSpeed( 4.0f, 1.0f );
	p_explosionEmitter->setParticleStartColor( Vec4f( 1.0f, 1.0f, 1.0f, 1.0f ), Vec4f( 0.35f, 0.35f, 0.35f, 0.1f ) );
	p_explosionEmitter->setParticleEndColor( Vec4f( 0.0f, 0.0f, 0.0f, 0.0f ), Vec4f( 0.0f, 0.0f, 0.0f, 0.0f ) );
	p_explosionEmitter->setParticleStartSize( 3.0f * scale, 1.0f * scale );
	p_explosionEmitter->setParticleEndSize( 7.0f * scale, 2.0f * scale );
	p_explosionEmitter->setParticleGravity( 0.0f, 0.1f );
	p_explosionEmitter->setParticleAnimFrame( 0.0f, 10.0f );

	if( explosionSound )
	{
		std::stringstream filename;
		filename << "data\\explosion" << explosionSound << ".ogg";
		SoundHandle sound = env.engine.getSoundManager().get( filename.str() );
		if( sound )
		{
			SoundInstance* p_soundInst = sound->createInstance( true );
			if( p_soundInst )
			{
				p_soundInst->setPosition( myPosition );
				p_soundInst->setVolume( volume + randFloat( -0.1f, 0.1f ) );
				p_soundInst->setPitch( pitch + randFloat( -0.25f, 0.25f ) );
				p_soundInst->play();
			}
		}
	}
}

void GameObject::kill()
{
	if( dead ) return;

	dead = true;

	if( level.getCameraTarget() == this ) level.setCameraTarget( 0 );

	// Shapes deaktivieren
	for( std::list<dGeomID>::const_iterator it = geoms.begin();
		 it != geoms.end();
		 it++ )
	{
		if( dGeomGetClass( *it ) == dGeomTransformClass ) dSpaceRemove( level.getSpace(), *it );
		dGeomDisable( *it );
	}

	// Krper deaktivieren
	if( body ) dBodyDisable( body );

	onKill();

	Scripting::inst().call( onKilled, this );
}

void GameObject::rebirth()
{
	if( !dead ) return;

	dead = false;

	// Shapes aktivieren
	for( std::list<dGeomID>::const_iterator it = geoms.begin();
		 it != geoms.end();
		 it++ )
	{
		if( dGeomGetClass( *it ) == dGeomTransformClass ) dSpaceAdd( level.getSpace(), *it );
		dGeomEnable( *it );
	}

	// Krper aktivieren
	if( body ) dBodyEnable( body );

	level.adoptObject( this );

	onRebirth();

	Scripting::inst().call( onReanimated, this );
}

void GameObject::setVisible( bool visible )
{
	this->visible = visible;
}

bool GameObject::isVisible() const
{
	return visible;
}

void GameObject::setFrozen( bool frozen )
{
	this->frozen = frozen;
}

bool GameObject::isFrozen() const
{
	return frozen;
}

void GameObject::setPassThrough( bool enable )
{
	this->passThrough = enable;
}

bool GameObject::getPassThrough() const
{
	return passThrough;
}

void GameObject::setPlayerCollisionDamageScale( float playerCollisionDamageScale )
{
	this->playerCollisionDamageScale = playerCollisionDamageScale;
}

float GameObject::getPlayerCollisionDamageScale() const
{
	return playerCollisionDamageScale;
}

void GameObject::setCollisionCallback( luabind::object callback )
{
	collisionCallback = callback;
}

luabind::object GameObject::getCollisionCallback() const
{
	return collisionCallback;
}

Level& GameObject::getLevel() const
{
	return level;
}

const std::string& GameObject::getID() const
{
	return id;
}

dBodyID GameObject::getBody() const
{
	return body;
}

const dMass& GameObject::getMass() const
{
	return mass;
}

const Vec2f& GameObject::getBoundCenter() const
{
	return boundCenter;
}

float GameObject::getBoundRadius() const
{
	return boundRadius;
}

const Vec2f& GameObject::getBoundMin() const
{
	return boundMin;
}

const Vec2f& GameObject::getBoundMax() const
{
	return boundMax;
}

void GameObject::deleteBody()
{
	if( body )
	{
		Mat33f transform = getTransform();

		// 2D-Constraint lschen
		dJointDestroy( planeConstraint );

		// ODE-Body lschen
		dBodyDestroy( body );

		body = 0;

		setTransform( transform );
	}
}

void GameObject::deleteGeoms()
{
	// ODE-Shapes lschen
	for( std::list<dGeomID>::const_iterator it = geoms.begin();
		 it != geoms.end();
		 it++ )
	{
		dGeomDestroy( *it );
	}

	geoms.clear();
}

void GameObject::setBoundingCircle( const Vec2f& boundCenter,
								    float boundRadius )
{
	this->boundCenter = boundCenter;
	this->boundRadius = boundRadius;
}

void GameObject::setBoundingRect( const Vec2f& boundMin,
								  const Vec2f& boundMax )
{
	this->boundMin = boundMin;
	this->boundMax = boundMax;
}

void GameObject::extendBoundingRect( const Vec2f& point )
{
	boundMin = Vec2f( min( boundMin.x, point.x ), min( boundMin.y, point.y ) );
	boundMax = Vec2f( max( boundMax.x, point.x ), max( boundMax.y, point.y ) );
}

void GameObject::enlargeBoundingRect( float value )
{
	boundMin -= Vec2f( value );
	boundMax += Vec2f( value );
}

void GameObject::completeCollision( GameObject* p_other,
									CollisionInfo& info )
{
	// mittlere Position und mittleren Normalenvektor der Kontakte ermitteln
	info.collisionPoint = Vec2f( 0.0f, 0.0f );
	info.collisionNormal = Vec2f( 0.0f, 0.0f );
	double sum = 0.0;
	for( std::list<dContact>::const_iterator it = info.contacts.begin();
		 it != info.contacts.end();
		 it++ )
	{
		info.collisionPoint += it->geom.depth * Vec2d( it->geom.pos[0], it->geom.pos[1] );
		info.collisionNormal += it->geom.depth * Vec2d( it->geom.normal[0], it->geom.normal[1] );
		sum += it->geom.depth;
	}

	info.collisionPoint /= static_cast<float>( sum );
	info.collisionNormal.normalize();

	// Geschwindigkeiten beider Objekte am Punkt der Kollision ermitteln
	dVector3 myVelocity_ = { 0.0, 0.0, 0.0 };
	if( getBody() ) dBodyGetPointVel( getBody(), info.collisionPoint.x, info.collisionPoint.y, 0.0, myVelocity_ );
	dVector3 otherVelocity_ = { 0.0, 0.0, 0.0 };
	if( p_other && p_other->getBody() )	dBodyGetPointVel( p_other->getBody(), info.collisionPoint.x, info.collisionPoint.y, 0.0, otherVelocity_ );

	// relative Geschwindigkeit an diesem Punkt ermitteln
	Vec2d myVelocity( myVelocity_[0], myVelocity_[1] );
	Vec2d otherVelocity( otherVelocity_[0], otherVelocity_[1] );
	info.relativeVelocity = otherVelocity - myVelocity;

	// Gibt es dazu einen Eintrag?
	CollisionParameters stdParams;
	const std::string otherType = p_other ? p_other->getMaterial() : "Unknown";
	info.p_params = level.getCollisionPair( material, otherType );
	if( !info.p_params ) info.p_params = level.getCollisionPair( "All", otherType );
	if( !info.p_params ) info.p_params = level.getCollisionPair( material, "All" );
	if( !info.p_params ) info.p_params = &stdParams;

	// bestimmen, ob dies eine frontale Kollision ist oder mehr ein "Entlangschliddern"
	float dot = info.collisionNormal ^ info.relativeVelocity.normalized();
	info.frontal = fabsf( dot ) > info.p_params->threshold;

	if( info.frontal )
	{
		// Einschlagsgeschwindigkeit bestimmen
		info.speed = fabsf( info.collisionNormal ^ info.relativeVelocity );
	}
	else
	{
		// Schliddergeschwindigkeit bestimmen
		info.speed = fabsf( info.collisionNormal.perp() ^ info.relativeVelocity );
	}

	if( info.p_params == &stdParams ) info.p_params = 0;

	info.minSpeedExceeded = false;
}

void GameObject::processCollisionEffects( GameObject* p_other,
										  CollisionInfo& info )
{
	if( !info.p_params || info.passThrough ) return;

	// Effekt holen
	CollisionParameters::Effect& effect = info.frontal ? info.p_params->frontal : info.p_params->scratch;

	// Reicht die Geschwindigkeit?
	info.minSpeedExceeded = info.speed > effect.minSpeed;
	if( info.minSpeedExceeded )
	{
		if( effect.createEmitter )
		{
			// einen Emitter erstellen, der die gewnschten Partikel erzeugt
			SimpleEmitter* p_emitter = level.getPS().createSimpleEmitter();
			p_emitter->setTimeToLive( 0.1f );
			p_emitter->setRate( info.speed * effect.particleRateFactor );
			p_emitter->setPosition( info.collisionPoint );
			p_emitter->setSettings( effect.emitterSettings );
		}

		if( effect.sound )
		{
			// Sound spielen
			SoundInstance* p_soundInst = effect.sound->createInstance();
			if( p_soundInst )
			{
				p_soundInst->setPosition( info.collisionPoint );
				p_soundInst->setVolume( info.speed * effect.soundVolumeFactor );
				p_soundInst->setPitch( randFloat( 0.75f, 1.25f ) );
				p_soundInst->play();
			}
		}
	}
}