#include <PCH.h>
#include <HighscoresDB.h>
#include <Level.h>

HighscoresDB::HighscoresDB()
{
	load();
}

HighscoresDB::HighscoresDB( const HighscoresDB& copy )
{
}

HighscoresDB::~HighscoresDB()
{
	save();
}

HighscoresDB& HighscoresDB::inst()
{
	static HighscoresDB inst;
	return inst;
}

unsigned int HighscoresDB::getPlaceInHighscore( const std::string& level,
												unsigned int score ) const
{
	HighscoresList list = getHighscores( level );
	unsigned int better = 0;

	for( HighscoresList::const_iterator it = list.begin();
		 it != list.end();
		 it++ )
	{
		if( it->score > score ) better++;
	}

	return better;
}

HighscoresList HighscoresDB::getHighscores( const std::string& level ) const
{
	HighscoresList result;

	const unsigned long long hash = getHashCode( std::string( "data\\levels\\" ) + level );
	std::map<unsigned long long, HighscoresList>::const_iterator it = highscoresLists.find( hash );
	if( it != highscoresLists.end() ) result = it->second;

	return result;
}

void HighscoresDB::buildHighscores( const std::string& level,
								    std::string& outPlayers,
									std::string& outScores ) const
{
	std::ostringstream players, scores;

	std::string levelName, levelType, levelDesc;
	Level::getLevelInfo( level, levelName, levelType, levelDesc );

	HighscoresList highscores = getHighscores( level );
	if( highscores.empty() )
	{
		players << "Bisher noch keine Rekorde!" << std::endl;
	}
	else
	{
		for( size_t i = 0; i < highscores.size(); i++ )
		{
			players << ( i + 1 ) << ". " << highscores[i].name << std::endl;

			if( levelType == "TimeAttack" ) scores << Level::formatTime( ~0 - highscores[i].score ) << std::endl;
			else scores << highscores[i].score << std::endl;
		}
	}

	outPlayers = players.str();
	outScores = scores.str();
}

void HighscoresDB::addHighscoreEntry( const std::string& level,
									  const HighscoresEntry& entry )
{
	const unsigned long long hash = getHashCode( std::string( "data\\levels\\" ) + level );

	HighscoresList& list = highscoresLists[hash];

	if( list.empty() )
	{
		list.push_back( entry );
	}
	else
	{
		// Position suchen
		HighscoresList::iterator it;
		for( it = list.begin();
			 it != list.end();
			 it++ )
		{
			if( entry.score >= it->score )
			{
				// Vor diesem Eintrag einfgen!
				list.insert( it, entry );
				break;
			}
		}

		if( it == list.end() )
		{
			// hinten anhngen
			list.push_back( entry );
		}

		// Liste auf fnf Eintrge verkrzen
		if( list.size() > 5 ) list.resize( 5 );
	}

	save();
}

void HighscoresDB::resetHighscores( const std::string& level )
{
	const unsigned long long hash = getHashCode( std::string( "data\\levels\\" ) + level );
	highscoresLists.erase( hash );
	save();
}

bool HighscoresDB::wasLevelCompleted( const std::string& level ) const
{
	const unsigned long long hash = getHashCode( std::string( "data\\levels\\" ) + level );
	return levelsCompleted.find( hash ) != levelsCompleted.end();
}

void HighscoresDB::setLevelCompleted( const std::string& level,
									  bool completed )
{
	const unsigned long long hash = getHashCode( std::string( "data\\levels\\" ) + level );
	if( completed ) levelsCompleted.insert( hash );
	else levelsCompleted.erase( hash );

	save();
}

unsigned long long HighscoresDB::getHashCode( const std::string& filename,
											  size_t ignoreLastNBytes ) const
{
	FILE* p_file = fopen( filename.c_str(), "rb" );
	if( !p_file ) return 0;

	// Dateigre bestimmen
	fseek( p_file, 0, SEEK_END );
	size_t fileSize = ftell( p_file );
	fseek( p_file, 0, SEEK_SET );

	MTRand rnd( 42 );

	size_t numBytesLeft = fileSize - ignoreLastNBytes;
	unsigned long long hash = 0;
	while( numBytesLeft )
	{
		unsigned long long k = 0;
		for( int i = 0; i < sizeof( unsigned long long ); i++ )
		{
			k |= static_cast<unsigned long long>( rnd.randInt( 255 ) );
			k <<= 8;
		}

		unsigned long long v = 0;
		numBytesLeft -= fread( &v, 1, min( numBytesLeft, sizeof( v ) ), p_file );
		hash ^= v ^ k;
	}

	fclose( p_file );

	return hash;
}

void HighscoresDB::load()
{
	highscoresLists.clear();
	levelsCompleted.clear();

	FILE* p_file = fopen( "highscores.dat", "rb" );
	if( !p_file ) return;

	// Anzahl der Eintrge in der Highscores-Liste lesen
	size_t n = 0;
	readEncrypted( p_file, &n, sizeof( n ) );
	if( !n ) return;

	// Eintrge lesen
	for( size_t i = 0; i < n; i++ )
	{
		HighscoresList list;

		// Hash-Code lesen
		unsigned long long hash;
		readEncrypted( p_file, &hash, sizeof( hash ) );

		// Anzahl der Eintrge in der Liste lesen
		size_t m;
		readEncrypted( p_file, &m, sizeof( m ) );

		// Liste lesen
		for( size_t j = 0; j < m; j++ )
		{
			HighscoresEntry entry;

			// Lnge des Namens lesen
			size_t o;
			readEncrypted( p_file, &o, sizeof( o ) );

			// Name lesen
			char* p_name = new char[o + 1];
			readEncrypted( p_file, p_name, o );
			p_name[o] = 0;
			entry.name = p_name;
			delete[] p_name;

			// Punktzahl lesen
			readEncrypted( p_file, &entry.score, sizeof( entry.score ) );

			// einfgen
			list.push_back( entry );
		}

		// einfgen
		highscoresLists[hash] = list;
	}

	// Anzahl der Eintrge in der Geschafft-Menge lesen
	n = 0;
	readEncrypted( p_file, &n, sizeof( n ) );

	// Eintrge lesen
	for( size_t i = 0; i < n; i++ )
	{
		// Hash-Code lesen
		unsigned long long hash;
		readEncrypted( p_file, &hash, sizeof( hash ) );

		// einfgen
		levelsCompleted.insert( hash );
	}

	// Hash-Code der Datei lesen
	unsigned long long hash;
	fread( &hash, sizeof( hash ), 1, p_file );
	fclose( p_file );

	// berprfen, ob der Hash-Code stimmt
	unsigned long long fileHash = getHashCode( "highscores.dat", sizeof( hash ) );
	if( hash != fileHash )
	{
		// Nein! Alles zurcksetzen.
		highscoresLists.clear();
		levelsCompleted.clear();
	}
}

void HighscoresDB::save() const
{
	FILE* p_file = fopen( "highscores.dat", "wb" );
	if( !p_file ) return;

	// Anzahl der Eintrge in der Highscores-Liste schreiben
	size_t n = highscoresLists.size();
	writeEncrypted( p_file, &n, sizeof( n ) );

	// Eintrge schreiben
	for( std::map<unsigned long long, HighscoresList>::const_iterator it = highscoresLists.begin();
		 it != highscoresLists.end();
		 it++ )
	{
		// Hash-Code schreiben
		writeEncrypted( p_file, &it->first, sizeof( it->first ) );

		// Anzahl der Eintrge in der Liste schreiben
		n = it->second.size();
		writeEncrypted( p_file, &n, sizeof( n ) );

		// Liste schreiben
		for( HighscoresList::const_iterator j = it->second.begin();
			 j != it->second.end();
			 j++ )
		{
			// Lnge des Namens schreiben
			n = j->name.length();
			writeEncrypted( p_file, &n, sizeof( n ) );

			// Name schreiben
			writeEncrypted( p_file, j->name.c_str(), n );

			// Punktzahl schreiben
			writeEncrypted( p_file, &j->score, sizeof( j->score ) );
		}
	}

	// Anzahl der Eintrge in der Geschafft-Menge schreiben
	n = levelsCompleted.size();
	writeEncrypted( p_file, &n, sizeof( n ) );

	// Eintrge schreiben
	for( std::set<unsigned long long>::const_iterator it = levelsCompleted.begin();
		 it != levelsCompleted.end();
		 it++ )
	{
		// Hash-Code schreiben
		writeEncrypted( p_file, &( *it ), sizeof( *it ) );
	}

	fclose( p_file );

	// Hash-Wert anhngen
	unsigned long long hash = getHashCode( "highscores.dat" );
	p_file = fopen( "highscores.dat", "ab" );
	if( !p_file ) return;
	fwrite( &hash, sizeof( hash ), 1, p_file );
	fclose( p_file );
}

size_t HighscoresDB::writeEncrypted( FILE* p_file,
									 const void* p_source,
									 size_t length ) const
{
	long pos = ftell( p_file );

	MTRand rnd;

	unsigned char* p_data = new unsigned char[length];
	for( size_t i = 0; i < length; i++ )
	{
		rnd.seed( static_cast<unsigned int>( pos + i ) );
		unsigned char key = static_cast<unsigned char>( rnd.randInt( 255 ) );
		unsigned char d = ( reinterpret_cast<const unsigned char*>( p_source ) )[i];
		unsigned char e = d ^ key;
		p_data[i] = e;
	}

	size_t r = fwrite( p_data, 1, length, p_file );

	delete[] p_data;

	return r;
}

size_t HighscoresDB::readEncrypted( FILE* p_file,
								    void* p_dest,
									size_t length ) const
{
	long pos = ftell( p_file );

	MTRand rnd;

	unsigned char* p_data = new unsigned char[length];
	size_t r = fread( p_data, 1, length, p_file );

	for( size_t i = 0; i < r; i++ )
	{
		rnd.seed( static_cast<unsigned int>( pos + i ) );
		unsigned char key = static_cast<unsigned char>( rnd.randInt( 255 ) );
		unsigned char e = p_data[i];
		unsigned char d = e ^ key;
		( reinterpret_cast<unsigned char*>( p_dest ) )[i] = d;
	}

	delete[] p_data;

	return r;
}