#include <PCH.h>
#include <Level.h>
#include <CartoonRenderer.h>
#include <Obj_Wall.h>
#include <Obj_Player.h>
#include <Obj_Enemy_Eraser.h>
#include <Obj_Enemy_AAStation.h>
#include <Obj_Dummy.h>
#include <Obj_Garage.h>
#include <Obj_Laser.h>
#include <Obj_EnergyNode.h>
#include <Obj_LevelTarget.h>
#include <Obj_PowerUp.h>
#include <Obj_StaticImage.h>
#include <Obj_Accelerator.h>
#include <Obj_Checkpoint.h>
#include <Obj_Projectile.h>
#include <Obj_Mine.h>
#include <ScriptObj_Counter.h>
using namespace luabind;

Level::Level( GameEnvironment& env,
			  const std::string& filename,
			  bool testMode ) : env( env ),
								video( env.engine.getVideo() ),
								filename( filename ),
								testMode( testMode ),
								p_cr( 0 ),
								p_ps( 0 ),
								cameraTracking( true ),
								camera( 0.0f, 0.0f ),
								cameraZoom( 1.0f ),
								cameraAngle( 0.0f ),
								p_cameraTarget( 0 ),
								p_lua( 0 )
{
	// XML-Dokument laden
	TiXmlDocument doc( ( std::string( "data\\levels\\" ) + filename ).c_str() );
	if( !doc.LoadFile() )
	{
		// Fehler!
		setError( 1 );
		std::ostringstream msg;
		msg << "Datei " << filename << " konnte nicht geladen werden";
		env.logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	TiXmlElement* p_root = doc.FirstChildElement( "Level" );
	if( !p_root )
	{
		// Fehler!
		setError( 2 );
		std::ostringstream msg;
		msg << filename << ": <Level>-Element fehlt";
		env.logger.error( msg.str(), LOGGER_AUTO );
		return;
	}

	// vorladen
	int r = preload( p_root );
	if( r )
	{
		setError( 3 );
		return;
	}

	if( !testMode )
	{
		// Ressourcen laden
		r = loadResources();
		if( r )
		{
			setError( 4 );
			return;
		}
	}
	else
	{
		p_ps = new ParticleSystem( video, ~0 );
	}

	// den Level versuchen zu laden
	r = load( p_root );
	if( r )
	{
		setError( 5 );
		return;
	}

	// Standardpartikeltypen einrichten
	setupParticleTypes();

	// Standardkollisionsparameter einrichten
	setupCollisionParameters();

	// Startskript ausfhren
	Scripting::inst().executeScript( p_lua, script );
}

Level::~Level()
{
	// alle Objekte lschen
	for( std::map<std::string, GameObject*>::const_iterator it = objects.begin();
		 it != objects.end();
		 it++ )
	{
		delete it->second;
	}

	// auch die Zombies lschen
	for( std::set<GameObject*>::const_iterator it = zombies.begin();
		 it != zombies.end();
		 it++ )
	{
		delete *it;
	}

	// Script-Objekte lschen
	for( std::set<ScriptObject*>::const_iterator it = scriptObjects.begin();
		 it != scriptObjects.end();
		 it++ )
	{
		delete *it;
	}

	// Partikelsystem lschen
	delete p_ps;

	// Cartoon-Renderer lschen
	delete p_cr;

	// Gruppe der Kontakt-Joints lschen
	dJointGroupDestroy( contactGroup );

	// ODE-Space zerstren
	dSpaceDestroy( space );

	// ODE-Welt zerstren
	dWorldDestroy( world );

	// Scripting-Objekte lschen
	timerCallbacks.clear();
	resumeCountDowns.clear();

	// lua herunterfahren
	Scripting::inst().popCurrentLuaState();
	lua_close( p_lua );
}

void Level::getLevelInfo( const std::string& filename,
						  std::string& outName,
						  std::string& outType,
						  std::string& outDescription )
{
	outName = "";
	outType = "";
	outDescription = "";

	TiXmlDocument doc( ( std::string( "data\\levels\\" ) + filename ).c_str() );
	if( !doc.LoadFile() ) return;
	TiXmlElement* p_root = doc.FirstChildElement( "Level" );
	if( !p_root ) return;
	const char* p_name = p_root->Attribute( "name" );
	if( p_name ) outName = p_name;
	else outName = "Unbenannter Level";
	const char* p_type = p_root->Attribute( "type" );
	if( p_type ) outType = p_type;
	else outType = "Standard";
	TiXmlElement* p_description = p_root->FirstChildElement( "Description" );
	if( p_description )
	{
		const char* p_descriptionText = p_description->GetText();
		if( p_descriptionText ) outDescription = p_descriptionText;
	}
}

std::string Level::formatTime( unsigned int msecs )
{
	const unsigned int secs = msecs / 1000;
	const unsigned int mins = secs / 60;

	char result[16];
	sprintf( result, "%02d:%02d.%02d", mins, secs % 60, ( msecs / 10 ) % 100 );

	return result;
}

static void nearCallback( void* p_data,
						  dGeomID geom1,
						  dGeomID geom2 )
{
	static_cast<Level*>( p_data )->nearCallback( geom1, geom2 );
}

void Level::update()
{
	if( raceMode )
	{
		// Punktzahl aus der Zeit berechnen
		finalScore = ~0 - raceTime;
	}
	else
	{
		// Punktzahl des Spielers abfragen
		Obj_Player* p_player = static_cast<Obj_Player*>( getObjectByID( "player" ) );
		if( p_player ) finalScore = p_player->getScore();
	}

PROFILE_BEGIN( "OBJECTS_UPDATE" );

	// Gefahrzonen vorbereiten
	avoidanceAreas.clear();
	for( std::map<std::string, GameObject*>::iterator it = objects.begin();
		 it != objects.end();
		 it++ )
	{
		if( !it->second->isFrozen() )
		{
			it->second->addAvoidanceAreas();
		}
	}

	// Objekte aktualisieren
	for( std::map<std::string, GameObject*>::iterator it = objects.begin();
		 it != objects.end(); )
	{
		it->second->updatePhysics();
		it->second->update();

		if( it->second->isToBeDeleted() )
		{
			// Das Objekt mchte gerne gelscht werden.
			deleteObject( it->second );
			objectsSet.erase( it->second );
			it = objects.erase( it );
		}
		else it++;
	}

	if( !testMode )
	{
		// Texte aktualisieren
		updateTexts();

		// SMS aktualisieren
		updateSMS();
	}

	updateScriptObjects();

PROFILE_END();

	if( !testMode )
	{
		if( cameraTracking )
		{
			// Kamerabewegung
			const Vec2f gravity( getGravity() );
			float cameraZoomTarget = 1.0f / 1.5f;
			float cameraAngleTarget = atan2f( -gravity.x, gravity.y );

			if( p_cameraTarget )
			{
				const Vec2f targetVelocity( p_cameraTarget->getVelocity() );
				const Vec2f ct( p_cameraTarget->getAbsPosition() + 1.0f * targetVelocity + Vec2f( 0.0f, 2.0f / cameraZoomTarget ) );
				if( ( camera - ct ).length() * cameraZoom > 5.0f ) adjustSmooth( cameraTarget, ct, 5.0f, env.dt );
				cameraZoomTarget = max( 0.5f, 1.0f / ( 1.5f + 0.01f * targetVelocity.length() ) );
			}

			if( fade > 0.0f ) cameraZoomTarget = 1.0f / ( 1.5f + 2.0f * fade );

			if( fabsf( cameraAngle - cameraAngleTarget + 6.282f ) < fabsf( cameraAngle - cameraAngleTarget ) ) cameraAngle += 6.282f;
			if( fabsf( cameraAngle - cameraAngleTarget - 6.282f ) < fabsf( cameraAngle - cameraAngleTarget ) ) cameraAngle -= 6.282f;

			adjustSmooth( camera, cameraTarget, 1.5f, env.dt );
			adjustSmooth( cameraZoom, cameraZoomTarget, 1.5f, env.dt );
			adjustSmooth( cameraAngle, cameraAngleTarget, 1.5f, env.dt );
		}

		if( fadeOut == -FLT_MAX )
		{
			// Gibt es den Spieler noch?
			if( !getObjectByID( "player" ) )
			{
				// Nein! Ausblenden!
				fadeOut = 0.0f;
			}
		}
		else
		{
			fadeOut += env.dt;
			if( fadeOut > 3.0f ) finished = true;
		}

		if( fade == -FLT_MAX )
		{
			// Gibt es das Ziel noch?
			if( !getObjectByID( "target" ) )
			{
				// Nein! Der Level ist geschafft.
				fade = 0.0f;
				fadeOut = -1.0f;
			}
		}
		else
		{
			fade += env.dt;
		}

PROFILE_BEGIN( "PARTICLES_UPDATE" );

		// Partikelsystem aktualisieren
		p_ps->setGlobalRate( env.options.getOptions().get<float>( "ParticleRate" ) );
		p_ps->update();

PROFILE_END();

	}
	else
	{
		p_ps->setGlobalRate( 0.0f );
		texts.clear();
	}

PROFILE_BEGIN( "PHYSICS_COLLISION_DETECTION" );

	// alle Kontakt-Joints lschen
	dJointGroupEmpty( contactGroup );

	// potenzielle Kollisionen erkennen, sie werden in "nearCallback" verarbeitet
	collisions.clear();
	dSpaceCollide( space, this, ::nearCallback );
	processCollisions();

PROFILE_END();
PROFILE_BEGIN( "PHYSICS_STEP" );

	// Physiksimulation
	dWorldQuickStep( world, env.dt );

PROFILE_END();

	time += env.dt;
	fadeTime += env.dt;

	if( raceMode &&
		getObjectByID( "player" ) &&
		getObjectByID( "target" ) )
	{
		raceTime += env.engine.getTickDuration();
	}
}

void Level::render()
{
	// Kamera setzen
	video.setupCamera( camera, cameraAngle, cameraZoom );

	// Hintergrund rendern
	video.setTexture( background );
	video.setupAlphaBlending( false );
	glBegin( GL_QUADS );
	glColor3f( 1.0f, 1.0f, 1.0f );
	glTexCoord2i( 0, 0 );
	glVertex2i( -400, -400 );
	glTexCoord2i( 128 * 200, 0 );
	glVertex2i( 400, -400 );
	glTexCoord2i( 128 * 200, 128 * 200 );
	glVertex2i( 400, 400 );
	glTexCoord2i( 0, 128 * 200 );
	glVertex2i( -400, 400 );
	glEnd();
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

PROFILE_BEGIN( "OBJECTS_VISIBILITY" );

	// sichtbare Objekte bestimmen
	calculateVisibility();

PROFILE_END();
PROFILE_BEGIN( "OBJECTS_RENDER" );

	// Hintergrundobjekte rendern
	renderObjects( RP_BACKGROUND_1 );
	renderObjects( RP_BACKGROUND_2 );
	renderObjects( RP_BACKGROUND_3 );

	// Schatten-Samples vorbereiten
	std::vector<Vec2f> samples;
	float alpha = 0.0f;
	switch( env.options.getOptions().get<unsigned int>( "ShadowQuality" ) )
	{
	case 1:
		samples.push_back( Vec2f( 0.15f, 0.15f ) );
		alpha = 0.4f;
		break;
	case 2:
		samples.push_back( Vec2f( 0.1f, 0.2f ) );
		samples.push_back( Vec2f( 0.2f, 0.1f ) );
		alpha = 0.25f;
		break;
	case 3:
		samples.push_back( Vec2f( 0.15f, 0.15f ) );
		samples.push_back( Vec2f( 0.1f, 0.2f ) );
		samples.push_back( Vec2f( 0.2f, 0.1f ) );
		alpha = 0.15f;
		break;
	}

	// Schatten der Objekte rendern
	for( size_t i = 0; i < samples.size(); i++ )
	{
		glPushMatrix();
		glTranslatef( samples[i].x, samples[i].y, 0.0f );
		renderObjects( RP_SHADOW, Vec4f( 0.0f, 0.0f, 0.0f, alpha ) );
		glPopMatrix();
	}

	// Objekte rendern
	renderObjects( RP_OBJECT_1 );
	renderObjects( RP_OBJECT_2 );

	// Vordergrundobjekte rendern
	renderObjects( RP_FOREGROUND );

	if( flags & LF_RENDER_AVOIDANCE_AREAS )
	{
		// Gefahrzonen rendern
		renderAvoidanceAreas();
	}

	if( flags & LF_RENDER_OBJECT_DEBUG_INFO )
	{
		// Debug-Informationen der Objekte rendern
		renderObjects( RP_DEBUG_INFO );
	}

	// Texte rendern
	renderTexts();

PROFILE_END();
PROFILE_BEGIN( "PARTICLES_RENDER" );

	// verdunkelnde Partikel rendern
	p_ps->render( true, false );

PROFILE_END();
PROFILE_BEGIN( "OBJECTS_RENDER" );

	// leuchtende Objekte rendern
	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	renderObjects( RP_GLOW );

PROFILE_END();
PROFILE_BEGIN( "PARTICLES_RENDER" );

	// leuchtende Partikel rendern
	p_ps->render( false, true );

PROFILE_END();

	if( fade > 0.0f )
	{
		// Weiblende
		video.setTexture( 0 );
		video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE );
		video.drawFullScreenQuad( Vec4f( fade * 4.0f, fade * 2.0f, fade, fade ) );
	}

	if( fadeTime < 3.0f || fadeOut > 0.0f )
	{
		// Licht anschalten/ausschalten
		float t;
		if( fadeTime < 3.0f ) t = fadeTime;
		else t = max( 0.0f, 3.0f - fadeOut );
		const float x = floorf( 8.0f * t );
		const float light = 0.1f + 1.0f - ( 1.0f / ( x + 1.0f ) ) + randFloat( 0.0f, 0.01f );

		if( light < 1.0f )
		{
			// Licht rendern
			video.setTexture( 0 );
			video.setupAlphaBlending( true, GL_ZERO, GL_SRC_COLOR );
			video.drawFullScreenQuad( Vec4f( light, light, light + 0.05f, 1.0f ), Vec2i( 0, 0 ), Vec2i( 256, 256 ) );
		}
	}

	video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

void Level::renderOverlays()
{
	video.setupProjection( Vec2f( -24.0f, -18.0f ), Vec2f( 24.0f, 18.0f ) );

	renderMobile();
	renderScriptObjects();
}

GameObject* Level::createObject( const TiXmlElement* p_root )
{
	// Objekttyp lesen
	const char* p_type = p_root->Attribute( "type" );
	if( !p_type )
	{
		env.logger.error( "type-Attribut fehlt", LOGGER_AUTO );
		return 0;
	}

	GameObject* p_object = 0;
	std::string type( p_type );

	if( type == "Player" )					p_object = new Obj_Player( *this, p_root );
	else if( type == "Wall" )				p_object = new Obj_Wall( *this, p_root );
	else if( type == "Enemy_Eraser" )		p_object = new Obj_Enemy_Eraser( *this, p_root );
	else if( type == "Enemy_AAStation" )	p_object = new Obj_Enemy_AAStation( *this, p_root );
	else if( type == "Dummy" )				p_object = new Obj_Dummy( *this, p_root );
	else if( type == "Garage" )				p_object = new Obj_Garage( *this, p_root );
	else if( type == "Laser" )				p_object = new Obj_Laser( *this, p_root );
	else if( type == "EnergyNode" )			p_object = new Obj_EnergyNode( *this, p_root );
	else if( type == "LevelTarget" )		p_object = new Obj_LevelTarget( *this, p_root );
	else if( type == "PowerUp" )			p_object = new Obj_PowerUp( *this, p_root );
	else if( type == "StaticImage" )		p_object = new Obj_StaticImage( *this, p_root );
	else if( type == "Accelerator" )		p_object = new Obj_Accelerator( *this, p_root );
	else if( type == "Checkpoint" )			p_object = new Obj_Checkpoint( *this, p_root );
	else if( type == "Mine" )				p_object = new Obj_Mine( *this, p_root );

	if( !p_object )
	{
		env.logger.error( std::string( "Unbekannter Objekttyp: " ) + p_type, LOGGER_AUTO );
		return 0;
	}

	// Objekt eintragen
	if( adoptObject( p_object ) )
	{
		delete p_object;
		return 0;
	}

	return p_object;
}

#define TO_LUA( GLOBALS, ID, OBJ, TYPE ) if( dynamic_cast<TYPE*>( OBJ ) ) { GLOBALS[ID] = static_cast<TYPE*>( OBJ ); }

int Level::adoptObject( GameObject* p_object )
{
	assert( p_object );

	const std::string& id = p_object->getID();

	// Gibt es ein Objekt mit diesem Namen schon?
	if( objects.find( id ) != objects.end() )
	{
		env.logger.error( std::string( "Objekt mit der ID " ) + id + " existiert bereits", LOGGER_AUTO );
		return 1;
	}

	// Objekt eintragen
	objects[id] = p_object;
	objectsSet.insert( p_object );

	// Falls das Objekt ein Zombie war, dann ist es das jetzt nicht mehr.
	zombies.erase( p_object );

	// das Objekt im lua-Script verfgbar machen
	adl::object g = globals( p_lua );
	TO_LUA( g, id, p_object, Obj_Projectile )
	else TO_LUA( g, id, p_object, Obj_Player )
	else TO_LUA( g, id, p_object, Obj_Enemy )
	else TO_LUA( g, id, p_object, Obj_Laser )
	else TO_LUA( g, id, p_object, Obj_EnergyNode )
	else TO_LUA( g, id, p_object, Obj_LevelTarget )
	else TO_LUA( g, id, p_object, Obj_Checkpoint )
	else TO_LUA( g, id, p_object, Obj_Garage )
	else TO_LUA( g, id, p_object, Obj_Mine )
	else TO_LUA( g, id, p_object, Obj_Accelerator )
	else g[id] = p_object;

	env.logger.finest( std::string( "Objekt mit der ID " ) + id + " hinzugefgt", LOGGER_AUTO );

	return 0;
}

void Level::adoptScriptObject( ScriptObject* p_object )
{
	scriptObjects.insert( p_object );
}

GameObject* Level::getObjectByID( const std::string& id ) const
{
	std::map<std::string, GameObject*>::const_iterator it = objects.find( id );
	return it == objects.end() ? 0 : it->second;
}

GameObject* Level::getHighestCheckpoint() const
{
	Obj_Checkpoint* p_highestCheckpoint = 0;
	for( std::map<std::string, GameObject*>::const_iterator it = objects.begin();
		 it != objects.end();
		 it++ )
	{
		if( it->second->getType() == "Checkpoint" )
		{
			Obj_Checkpoint* p_cp = static_cast<Obj_Checkpoint*>( it->second );
			if( p_cp->isActivated() )
			{
				if( !p_highestCheckpoint )
				{
					p_highestCheckpoint = p_cp;
				}
				else if( p_cp->getNumber() > p_highestCheckpoint->getNumber() ||
					     ( p_cp->getNumber() == p_highestCheckpoint->getNumber() && 
						   p_cp->getActivationTime() > p_highestCheckpoint->getActivationTime() ) )
				{
					p_highestCheckpoint = p_cp;
				}
			}
		}
	}

	return p_highestCheckpoint;
}

void Level::continueAtCheckpoint( GameObject* p_checkpoint )
{
	bool found = false;

	// den Zombie-Spieler wieder ausgraben
	for( std::set<GameObject*>::iterator it = zombies.begin();
		 it != zombies.end(); it++ )
	{
		if( ( *it )->getType() == "Player" )
		{
			Obj_Player* p_player = static_cast<Obj_Player*>( *it );
			Obj_Checkpoint* p_cp = static_cast<Obj_Checkpoint*>( p_checkpoint );

			// Wiedergeburt
			p_player->rebirth();

			// Position und Ausrichtung setzen
			p_player->setPosition( p_cp->getPosition() );
			p_player->setVelocity( Vec2f( 0.0f, 0.0f ) );
			p_player->setAngle( p_cp->getAngle() );
			p_player->setAngularVelocity( 0.0f );

			// reparieren und auftanken
			p_player->setHealth( p_cp->getHealthAmount() );
			const float fuel = p_cp->getFuelAmount();
			if( fuel > 0.0f ) p_player->setFuel( fuel );
			else p_player->addFuel( -fuel );

			// Schutzschild
			p_player->setShield( p_cp->getShieldDuration() );

			// Punktzahl halbieren
			addPlayerScore( -static_cast<int>( ( p_player->getScore() + 1 ) / 2 ), p_cp->getPosition(), true );

			// Funken
			SimpleEmitter* p_emitter = p_ps->createSimpleEmitter();
			p_emitter->attachTo( p_player, false );
			p_emitter->setRate( 2500.0f );
			p_emitter->setTimeToLive( 0.1f );
			p_emitter->addParticleStyle( 4 );
			p_emitter->addParticleStyle( 8 );
			p_emitter->setParticleTimeToLive( 0.4f, 0.2f );
			p_emitter->setParticlePositionRadius( 2.0f );
			p_emitter->setParticleSpeed( 8.0f, 2.0f );
			p_emitter->setParticleStartColor( Vec4f( 0.25f, 1.0f, 0.5f, 0.75f ), Vec4f( 0.4f, 0.4f, 0.4f, 0.25f ) );
			p_emitter->setParticleEndColor( Vec4f( 1.0f, 0.0f, 0.0f, 0.0f ), Vec4f( 0.4f, 0.0f, 0.0f, 0.0f ) );
			p_emitter->setParticleStartSize( 2.0f, 0.5f );
			p_emitter->setParticleEndSize( 5.0f, 1.0f );
			p_emitter->setParticleGravity( 4.0f, 1.0f );
			p_emitter->setParticleFriction( 0.25f, 0.1f );

			// Sound
			SoundHandle sound = env.engine.getSoundManager().get( "data\\rebirth.ogg" );
			if( sound )
			{
				sound->setDefaultSoundGroupName( "SFX" );
				SoundInstance* p_soundInst = sound->createInstance();
				if( p_soundInst )
				{
					p_soundInst->attachTo( p_player, false );
					p_soundInst->play();
				}
			}

			fadeTime = 0.0f;
			found = true;
			break;
		}
	}

	if( found )
	{
		fadeOut = -FLT_MAX;
		finished = false;
	}
}

void Level::playMusic()
{
	if( musicFilename.empty() ) env.engine.getAudio().stopMusic();
	else env.engine.getAudio().playMusic( musicFilename, false );
}

void Level::nearCallback( dGeomID geom1,
						  dGeomID geom2 )
{
	// die Krper bestimmen
	dBodyID body1, body2;
	body1 = dGeomGetBody( geom1 );
	body2 = dGeomGetBody( geom2 );

	// Ein Krper soll nicht einem seiner eigenen Teile kollidieren.
	if( body1 == body2 ) return;

	// Krper, die durch Joints verbunden sind, sollen ebenfalls nicht kollidieren.
	if( body1 && body2 && dAreConnectedExcluding( body1, body2, dJointTypeContact ) ) return;

	// maximal 16 Kontakte generieren
	dContact contact[16];
	int numContacts = dCollide( geom1, geom2, 16, &contact[0].geom, sizeof( dContact ) );

	if( numContacts > 0 )
	{
		// die zugehrigen Spielobjekte bestimmen
		GameObject* p_object1 = static_cast<GameObject*>( dGeomGetData( geom1 ) );
		GameObject* p_object2 = static_cast<GameObject*>( dGeomGetData( geom2 ) );
		if( p_object1->isToBeDeleted() ) p_object1 = 0;
		if( p_object2->isToBeDeleted() ) p_object2 = 0;

		bool ignore = false;
		if( ( p_object1 && !p_object1->isCollisionEnabled( p_object2 ) ) ||
			( p_object2 && !p_object2->isCollisionEnabled( p_object1 ) ) )
		{
			ignore = true;
		}

		if( !ignore )
		{
			// Kollisionsparameter holen, ansonsten Standard
			CollisionParameters params;
			const std::string type1 = p_object1 ? p_object1->getMaterial() : "Unknown";
			const std::string type2 = p_object2 ? p_object2->getMaterial() : "Unknown";
			CollisionParameters* p_params = getCollisionPair( type1, type2 );
			if( !p_params ) p_params = getCollisionPair( "All", type2 );
			if( !p_params ) p_params = getCollisionPair( type1, "All" );
			if( p_params ) params = *p_params;

			// Eintrag in der Map holen
			CollisionPair key;
			key.p_obj1 = min( p_object1, p_object2 );
			key.p_obj2 = max( p_object1, p_object2 );
			CollisionInfo& info = collisions[key];
			info.passThrough = true;

			for( int i = 0; i < numContacts; i++ )
			{
				// Parameter fr Kontakt-Joint festlegen
				contact[i].surface.mode = dContactBounce;				// Objekte sollen voneinander abprallen
				contact[i].surface.mu = params.friction;				// Reibung (0 bis unendlich)
				contact[i].surface.bounce = params.bouncyness;			// Abprallfaktor
				contact[i].surface.bounce_vel = params.minBounceSpeed;	// Mindestgeschwindigkeit fr Abprallen

				bool createContact = true;
				if( params.dontCreateContactJoint ) createContact = false;
				else if( p_object1 && p_object1->getPassThrough() ) createContact = false;
				else if( p_object2 && p_object2->getPassThrough() ) createContact = false;

				if( createContact )
				{
					// Kontakt-Joint erstellen und den beiden Krpern zuweisen
					dJointID contactJoint = dJointCreateContact( world, contactGroup, &contact[i] );
					dJointAttach( contactJoint, body1, body2 );
					info.joints.push_back( contactJoint );
					info.passThrough = false;
				}

				// Kontaktdaten speichern
				info.contacts.push_back( contact[i] );
			}

			if( info.contacts.empty() )
			{
				// Kollision wird ignoriert
				collisions.erase( key );
			}
		}
	}
}

void Level::setGravity( const Vec2f& gravity )
{
	dWorldSetGravity( world, gravity.x, gravity.y, 0.0 );
	if( p_ps ) p_ps->setGravity( gravity );
}

Vec2f Level::getGravity() const
{
	dVector3 gravity;
	dWorldGetGravity( world, gravity );
	return Vec2d( gravity[0], gravity[1] );
}

void Level::registerCollisionPair( const std::string& type1,
								   const std::string& type2,
								   const CollisionParameters& params )
{
	const std::string key1( type1 + "&" + type2 );
	const std::string key2( type2 + "&" + type1 );
	collisionParameters[key1] = params;
	collisionParameters[key2] = params;
}

CollisionParameters* Level::getCollisionPair( const std::string& type1,
											  const std::string& type2 )
{
	const std::string key( type1 + "&" + type2 );
	std::map<std::string, CollisionParameters>::iterator it = collisionParameters.find( key );
	return it == collisionParameters.end() ? 0 : &( it->second );
}

void Level::cacheSound( SoundHandle sound )
{
	soundCache.insert( sound );
}

GameObject* Level::castRay( const Vec2f& origin,
							const Vec2f& direction,
							float maxDistance,
							unsigned long categoryBits,
							unsigned long collisionBits,
							Vec2f* p_outPosition,
							float* p_outDistance,
							Vec2f* p_outNormal )
{
	bool zRay = direction == Vec2f( 0.0f, 0.0f );
	const Vec2f normalizedDir( zRay ? Vec2f( 0.0f, 0.0f ) : direction.normalized() );

	// Strahl erzeugen
	dGeomID ray = dCreateRay( 0, zRay ? 2.0 : maxDistance );
	if( zRay ) dGeomRaySet( ray, origin.x, origin.y, -1.0, 0.0, 0.0, 1.0 );
	else dGeomRaySet( ray, origin.x, origin.y, 0.0, normalizedDir.x, normalizedDir.y, 0.0 );
	dGeomSetCategoryBits( ray, categoryBits );
	dGeomSetCollideBits( ray, collisionBits );

	// Kollisionserkennung
	GameObject* p_object = 0;
	dContactGeom contact[256];
	int numContacts = dCollide( ray, reinterpret_cast<dGeomID>( space ), 256, contact, sizeof( dContactGeom ) );
	if( numContacts )
	{
		// den nchsten Kontakt suchen
		int closestIndex = 0;
		double closestDist = contact[0].depth;
		for( int i = 1; i < numContacts; i++ )
		{
			if( contact[i].depth < closestDist )
			{
				closestDist = contact[i].depth;
				closestIndex = i;
			}
		}

		dContactGeom& closest = contact[closestIndex];

		// Welches Objekt gehrt dazu?
		p_object = static_cast<GameObject*>( dGeomGetData( closest.g2 ) );
		if( p_outPosition ) *p_outPosition = Vec2d( closest.pos[0], closest.pos[1] );
		if( p_outDistance ) *p_outDistance = static_cast<float>( closest.depth );
		if( p_outNormal ) *p_outNormal = Vec2d( closest.normal[0], closest.normal[1] );
	}

	// Strahl zerstren
	dGeomDestroy( ray );

	return p_object;
}

GameObject* Level::castCylinder( const Vec2f& origin,
								 const Vec2f& direction,
								 float radius,
								 float maxDistance,
								 unsigned long categoryBits,
								 unsigned long collisionBits,
								 Vec2f* p_outPosition,
								 float* p_outDistance,
								 Vec2f* p_outNormal )
{
	// Transformations-Shape erzeugen
	dGeomID transform = dCreateGeomTransform( 0 );

	// Zylinder mit Halbkugeln an den Seiten erzeugen
	float length = max( 0.0f, maxDistance - 2.0f * radius );
	dGeomID capsule = dCreateCapsule( 0, radius, length );

	// Zylinder in die richtige Lage bringen
	dMatrix3 r;
	dRFrom2Axes( r, 0.0, 0.0, -1.0, -direction.y, direction.x, 0.0 );
	dGeomSetRotation( capsule, r );
	const Vec2f position( origin + 0.5f * maxDistance * direction );
	dGeomSetPosition( capsule, position.x, position.y, 0.0 );

	// Zylinder dem Transformations-Shape zuweisen
	dGeomTransformSetGeom( transform, capsule );

	// Kollisions-Bits setzen
	dGeomSetCategoryBits( transform, categoryBits );
	dGeomSetCollideBits( transform, collisionBits );
	dGeomSetCategoryBits( capsule, categoryBits );
	dGeomSetCollideBits( capsule, collisionBits );

	// Kollisionserkennung
	GameObject* p_object = 0;
	dContactGeom contact[256];
	int numContacts = dCollide( transform, reinterpret_cast<dGeomID>( space ), 256, contact, sizeof( dContactGeom ) );
	if( numContacts )
	{
		// den nchsten Kontakt suchen
		int closestIndex = 0;
		double closestDist = FLT_MAX;
		for( int i = 0; i < numContacts; i++ )
		{
			const Vec2d hit( contact[i].pos );
			double dist = ( ( hit - origin ) ^ direction ) - contact[i].depth;
			if( dist < closestDist )
			{
				closestDist = dist;
				closestIndex = i;
			}
		}

		dContactGeom& closest = contact[closestIndex];

		// Welches Objekt gehrt dazu?
		p_object = static_cast<GameObject*>( dGeomGetData( closest.g2 ) );
		if( p_outPosition ) *p_outPosition = origin + static_cast<float>( closestDist ) * direction;
		if( p_outDistance ) *p_outDistance = static_cast<float>( closestDist );
		if( p_outNormal ) *p_outNormal = Vec2d( closest.normal[0], closest.normal[1] );
	}

	// Zylinder wieder zerstren
	dGeomDestroy( transform );
	dGeomDestroy( capsule );

	return p_object;
}

void Level::addPlayerScore( int value,
						    const Vec2f& position,
							bool immediate )
{
	Obj_Player* p_player = static_cast<Obj_Player*>( getObjectByID( "player" ) );
	if( p_player )
	{
		p_player->addScore( value );
		if( raceMode ) return;

		if( position.x != FLT_MAX )
		{
			Vec2f newPosition( position + randFloat( 0.0f, 3.0f ) * randVec2fDir() );

			if( value > 0 )
			{
				std::ostringstream text;
				text << "+" << value;
				addText( text.str(), newPosition, Vec4f( randFloat( 0.0f, 0.25f ), randFloat( 0.75f, 1.0f ), randFloat( 0.0f, 0.25f ), 1.0f ), randFloat( 1.0f, 1.25f ) + 1.5f * static_cast<float>( min( value, 5000 ) ) / 2000.0f, 1.5f );
				negativeScore = 0;
			}
			else if( value < 0 )
			{
				if( immediate )
				{
					std::ostringstream text;
					text << value;
					addText( text.str(), newPosition, Vec4f( randFloat( 0.75f, 1.0f ), randFloat( 0.0f, 0.25f ), randFloat( 0.0f, 0.25f ), 1.0f ), randFloat( 1.0f, 1.25f ) + 1.5f * static_cast<float>( min( -value, 5000 ) ) / 2000.0f, 1.5f );
				}
				else
				{
					negativeScore -= value;
					if( negativeScore >= 250 )
					{
						addText( "-250", newPosition, Vec4f( randFloat( 0.75f, 1.0f ), randFloat( 0.0f, 0.25f ), randFloat( 0.0f, 0.25f ), 1.0f ), randFloat( 1.0f, 1.25f ) + 1.5f * static_cast<float>( min( negativeScore, 5000 ) ) / 2000.0f, 1.5f );
						negativeScore -= 250;
					}
				}
			}
		}
	}
}

void Level::addText( const std::string& text,
					 const Vec2f& position,
					 const Vec4f& color,
					 float size,
					 float timeToLive )
{
	if( testMode ) return;

	Text t;
	t.text = text;
	t.position = position;
	t.color = color;
	t.size = size;
	t.angle = atan2f( -getGravity().x, getGravity().y );
	t.timeToLive = timeToLive;
	texts.push_back( t );
}

void Level::addSMS( const std::string& sender,
					const std::string& text )
{
	SMS sms;
	sms.sender = sender;
	sms.text = env.mobileFont->adjustText( text, FontStyle(), 230 );

	// Zeilenanfnge suchen
	size_t lines = 0;
	sms.pageStarts.push_back( 0 );
	for( size_t i = 0; i < sms.text.size(); i++ )
	{
		if( sms.text[i] == '\n' && i != sms.text.size() - 1 )
		{
			lines++;
			if( lines == 6 )
			{
				size_t p = i;
				sms.pageStarts.push_back( ++p );
				lines = 0;
			}
		}
	}

	sms.duration = 3.0f + sms.pageStarts.size() * 6.0f;

	smsInbox.push( sms );
}

void Level::addAvoidanceCircle( int priority,
								GameObject* p_me,
								const Mat33f& transformation,
							    const Vec2f& center,
								float radius )
{
	AvoidanceArea area;
	area.priority = priority;
	area.p_createdBy = p_me;
	area.type = 0;
	area.transformation = transformation;
	area.inverseTransformation = transformation.inverse();
	area.center = center;
	area.radius = radius;
	area.radiusSq = radius * radius;
	avoidanceAreas.insert( area );
}

void Level::addAvoidanceRect( int priority,
							  GameObject* p_me,
							  const Mat33f& transformation,
							  const Vec2f& min,
							  const Vec2f& max )
{
	AvoidanceArea area;
	area.priority = priority;
	area.p_createdBy = p_me;
	area.type = 1;
	area.transformation = transformation;
	area.inverseTransformation = transformation.inverse();
	area.min = min;
	area.max = max;
	avoidanceAreas.insert( area );
}

void Level::addAvoidanceLine( int priority,
							  GameObject* p_me,
							  const Vec2f& start,
							  const Vec2f& end,
							  float thickness,
							  float lengthen )
{
	Vec2f yAxis( end - start );
	const float dist = yAxis.length();
	yAxis /= dist;
	const Vec2f xAxis( yAxis.perp() );
	const float halfThickness = 0.5f * thickness;
	Mat33f transformation( Mat33f::identity() );
	transformation.m11 = xAxis.x;
	transformation.m21 = xAxis.y;
	transformation.m12 = yAxis.x;
	transformation.m22 = yAxis.y;
	transformation.m13 = start.x;
	transformation.m23 = start.y;
	addAvoidanceRect( priority, p_me, transformation, Vec2f( -halfThickness, -lengthen ), Vec2f( halfThickness, dist + lengthen ) );
}

bool Level::getEscapeVector( GameObject* p_me,
							 float myRadius,
							 Vec2f& outEscapeVector ) const
{
	Vec2f myPosition( p_me->getPosition() );
	outEscapeVector = Vec2f( 0.0f, 0.0f );
	bool result = false;

	const float myRadiusSq = myRadius * myRadius;

	for( std::set<AvoidanceArea>::const_iterator it = avoidanceAreas.begin();
		 it != avoidanceAreas.end();
		 it++ )
	{
		const AvoidanceArea& area = *it;
		if( area.p_createdBy == p_me ) continue;

		// Position in den Zonenraum umrechnen
		const Vec2f position( area.inverseTransformation.transformPosition( myPosition + outEscapeVector ) );

		switch( area.type )
		{
		case 0:
			{
				// Kreis - Kreis
				// *************

				const Vec2f dir( area.center - position );
				const float distSq = dir.lengthSq();
				const float radii = myRadius + area.radius;

				if( distSq > radii * radii ) break;

				const float dist = sqrtf( distSq );
				const float penetrationDepth = myRadius + area.radius - dist;
				outEscapeVector -= area.transformation.transformDirection( dir.normalized() * penetrationDepth );
				result = true;

				break;
			}

		case 1:
			{
				// Kreis - Rechteck
				// ****************

				const float distXMin = area.min.x - position.x;
				if( distXMin > myRadius ) break;
				const float distXMax = position.x - area.max.x;
				if( distXMax > myRadius ) break;
				const float distYMin = area.min.y - position.y;
				if( distYMin > myRadius ) break;
				const float distYMax = position.y - area.max.y;
				if( distYMax > myRadius ) break;

				if( distXMin <= 0.0f &&
					distXMax <= 0.0f &&
					distYMin <= 0.0f &&
					distYMax <= 0.0f )
				{
					// Der Mittelpunkt liegt im Rechteck. Raus geht es am schnellsten da, wo die Distanz am wenigsten negativ.
					const float nearest = max( max( distXMin, distXMax ), max( distYMin, distYMax ) );
					const float penetrationDepth = -nearest - myRadius;

					if( nearest == distXMin ) outEscapeVector -= penetrationDepth * area.transformation.getAxis( 0 );
					else if( nearest == distXMax ) outEscapeVector += penetrationDepth * area.transformation.getAxis( 0 );
					else if( nearest == distYMin ) outEscapeVector -= penetrationDepth * area.transformation.getAxis( 1 );
					else if( nearest == distYMax ) outEscapeVector += penetrationDepth * area.transformation.getAxis( 1 );

					result = true;
					break;
				}

				// Position des nchsten Punkts auf dem Rechteck zum Kreis bestimmen
				const Vec2f nearest( clamp( position.x, area.min.x, area.max.x ), clamp( position.y, area.min.y, area.max.y ) );

				// Richtung und Distanz zum nchsten Punkt bestimmen
				const Vec2f dir( nearest - position );
				const float distSq = dir.lengthSq();
				if( distSq > myRadiusSq ) break;

				const float dist = sqrtf( distSq );
				const float penetrationDepth = myRadius - dist;
				outEscapeVector -= area.transformation.transformDirection( dir.normalized() * penetrationDepth );
				result = true;

				break;
			}
		}
	}

	return result;
}

void Level::addTimerCallback( luabind::object callback,
							  float interval )
{
	TimerCallback t;
	t.callback = callback;
	t.interval = interval;
	t.t = 0.0f;
	timerCallbacks.push_back( t );
}

int Level::removeTimerCallback( luabind::object callback )
{
	int n = 0;

	for( std::list<TimerCallback>::iterator it = timerCallbacks.begin();
		 it != timerCallbacks.end();
		 it++ )
	{
		TimerCallback& cb = *it;
		if( cb.callback == callback )
		{
			cb.interval = -1.0f;
			n++;
		}
	}

	return n;
}

void Level::addResumeCountDown( lua_State* p_lua,
							    float time )
{
	ResumeCountDown c;
	c.p_lua = p_lua;
	c.t = time;
	resumeCountDowns.push_back( c );
}

ScriptObj_Counter* Level::createCounter( int displayMode,
										 int direction,
										 float duration,
										 luabind::object onElapsed )
{
	return new ScriptObj_Counter( *this, displayMode, direction, duration, onElapsed );
}

GameEnvironment& Level::getEnv() const
{
	return env;
}

const std::string& Level::getFilename() const
{
	return filename;
}

const std::string& Level::getName() const
{
	return name;
}

const std::string& Level::getType() const
{
	return type;
}

const std::string& Level::getDescription() const
{
	return description;
}

const std::string& Level::getMusicFilename() const
{
	return musicFilename;
}

bool Level::isInTestMode() const
{
	return testMode;
}

const std::map<std::string, GameObject*>& Level::getObjects() const
{
	return objects;
}

dWorldID Level::getWorld() const
{
	return world;
}

dSpaceID Level::getSpace() const
{
	return space;
}

TextureHandle Level::getSprites() const
{
	return sprites;
}

TextureHandle Level::getParticles() const
{
	return particles;
}

CartoonRenderer& Level::getCR() const
{
	assert( p_cr );
	return *p_cr;
}

ParticleSystem& Level::getPS() const
{
	assert( p_ps );
	return *p_ps;
}

bool Level::isCameraTrackingEnabled() const
{
	return cameraTracking;
}

void Level::setCameraTrackingEnabled( bool enable )
{
	cameraTracking = enable;
}

const Vec2f& Level::getCamera() const
{
	return camera;
}

void Level::setCamera( const Vec2f& camera )
{
	this->camera = camera;
}

void Level::setCameraTarget( GameObject* p_cameraTarget,
							 bool moveThereImmediately )
{
	this->p_cameraTarget = p_cameraTarget;

	if( moveThereImmediately )
	{
		if( p_cameraTarget ) camera = cameraTarget = p_cameraTarget->getPosition();
	}
}

GameObject* Level::getCameraTarget() const
{
	return p_cameraTarget;
}

float Level::getCameraZoom() const
{
	return cameraZoom;
}

void Level::setCameraZoom( float cameraZoom )
{
	this->cameraZoom = cameraZoom;
}

float Level::getCameraAngle() const
{
	return cameraAngle;
}

void Level::setCameraAngle( float cameraAngle )
{
	this->cameraAngle = cameraAngle;
}

const Vec4f& Level::getPassColor() const
{
	return passColor;
}

const std::vector<GameObject*>& Level::getVisibleObjects() const
{
	return visibleObjects;
}

unsigned int Level::getNumVisibleObjects() const
{
	return static_cast<unsigned int>( visibleObjects.size() );
}

bool Level::isFading() const
{
	return fade >= 0.0f;
}

bool Level::isFinished() const
{
	return finished;
}

float Level::getTime() const
{
	return time;
}

unsigned int Level::getRaceTime() const
{
	return raceTime;
}

void Level::resetRaceTime()
{
	raceTime = 0;
}

void Level::setFinalScore( unsigned int finalScore )
{
	this->finalScore = finalScore;
}

unsigned int Level::getFinalScore() const
{
	return finalScore;
}

void Level::setFlags( unsigned int flags )
{
	this->flags = flags;
}

unsigned int Level::getFlags() const
{
	return flags;
}

std::string Level::getObjectAutoID()
{
	std::ostringstream id;
	id << "Obj_";
	id.setf( std::ios::fixed );
	id.width( 8 );
	id.fill( '0' );
	id << nextObjectAutoID++;
	return id.str();
}

int Level::preload( TiXmlElement* p_root )
{
	// Name lesen
	const char* p_name = p_root->Attribute( "name" );
	if( p_name ) name = p_name;
	else name = "Unbenannter Level";

	// Typ lesen
	const char* p_type = p_root->Attribute( "type" );
	if( p_type ) type = p_type;
	else type = "Standard";

	// Beschreibung lesen
	TiXmlElement* p_descriptionElem = p_root->FirstChildElement( "Description" );
	if( p_descriptionElem )
	{
		const char* p_description = p_descriptionElem->GetText();
		if( p_description ) description = p_description;
	}

	// Script lesen
	TiXmlElement* p_scriptElem = p_root->FirstChildElement( "Script" );
	if( p_scriptElem )
	{
		const char* p_script = p_scriptElem->GetText();
		if( p_script ) script = p_script;
	}

	// Musikdateiname lesen
	TiXmlElement* p_musicElem = p_root->FirstChildElement( "Music" );
	if( p_musicElem )
	{
		const char* p_music = p_musicElem->GetText();
		if( p_music )
		{
			std::string dir( "data\\" );

			int lookInLevelDir = 0;
			p_musicElem->QueryIntAttribute( "lookInLevelDir", &lookInLevelDir );

			if( lookInLevelDir )
			{
				dir += std::string( "levels\\" ) + getPathDirectory( filename ) + "\\";
			}

			// Dateipfad anpassen
			musicFilename = dir + p_music;
		}
	}

	// Dateiname des Hintergrundbilds lesen
	TiXmlElement* p_backgroundElem = p_root->FirstChildElement( "Background" );
	if( p_backgroundElem )
	{
		const char* p_background = p_backgroundElem->GetText();
		if( p_background )
		{
			std::string dir( "data\\" );

			int lookInLevelDir = 0;
			p_backgroundElem->QueryIntAttribute( "lookInLevelDir", &lookInLevelDir );

			if( lookInLevelDir )
			{
				dir += std::string( "levels\\" ) + getPathDirectory( filename ) + "\\";
			}

			// Dateipfad anpassen
			backgroundFilename = dir + + p_background;
		}
	}

	return 0;
}

int Level::load( TiXmlElement* p_root )
{
	using namespace luabind;

	// lua-Umgebung erzeugen
	p_lua = Scripting::inst().createInterpreter();
	Scripting::inst().pushCurrentLuaState( p_lua );
	globals( p_lua )["level"] = this;

	// ODE-Welt erzeugen
	world = dWorldCreate();
	dWorldSetQuickStepNumIterations( world, 32 );

	// ODE-Space erzeugen
	space = dHashSpaceCreate( 0 );
	dHashSpaceSetLevels( space, -4, 5 );

	// Gruppe der Kontakt-Joints erzeugen
	contactGroup = dJointGroupCreate( 256 );

	// Gravitationsvektor lesen
	Vec2f gravity( 0.0f, 0.0f );
	TiXmlElement* p_gravityElem = p_root->FirstChildElement( "Gravity" );
	if( p_gravityElem )
	{
		p_gravityElem->QueryFloatAttribute( "x", &gravity.x );
		p_gravityElem->QueryFloatAttribute( "y", &gravity.y );
	}

	// Gravitation setzen
	setGravity( gravity );

	// Objekte erzeugen
	TiXmlElement* p_objectsElem = p_root->FirstChildElement( "Objects" );
	if( !p_objectsElem )
	{
		env.logger.error( "<Objects>-Tag fehlt", LOGGER_AUTO );
		setError( 1 );
	}

	nextObjectAutoID = 1;
	int n = 0;
	TiXmlElement* p_objectElem = p_objectsElem->FirstChildElement( "Object" );
	while( p_objectElem )
	{
		GameObject* p_obj = createObject( p_objectElem );
		if( p_obj ) n++;

		p_objectElem = p_objectElem->NextSiblingElement( "Object" );
	}

	if( !n ) env.logger.warning( filename + ": Es wurden keine Objekte erzeugt", LOGGER_AUTO );
	else
	{
		std::ostringstream msg;
		msg << filename << ": Es wurden " << n << " Objekte erzeugt";
		env.logger.finer( msg.str(), LOGGER_AUTO );
	}

	visibleObjects.clear();
	texts.clear();
	while( !smsInbox.empty() ) smsInbox.pop();
	zombies.clear();
	smsTime = -1.0f;
	fade = -FLT_MAX;
	fadeOut = -FLT_MAX;
	time = 0.0f;
	fadeTime = 0.0f;
	raceTime = 0;
	finished = false;
	finalScore = 0;
	negativeScore = 0;
	raceMode = type == "TimeAttack";
	flags = 0;

	return 0;
}

int Level::loadResources()
{
	TextureManager& texMgr = env.engine.getTextureManager();

	// Hintergrund laden
	background = texMgr.get( backgroundFilename );
	if( !background ) return 1;

	// Sprites laden
	sprites = texMgr.get( "data\\sprites.png" );
	if( !sprites ) return 2;

	// Partikel laden
	particles = texMgr.get( "data\\particles.png" );
	if( !particles ) return 3;

	// Handy laden
	ParameterMap options;
	options.set<bool>( "UseMIPMapping", false );
	mobile = texMgr.get( "data\\mobile.png", false, options );
	if( !mobile ) return 4;

	// Cartoon-Renderer erzeugen
	p_cr = new CartoonRenderer( env, "data\\styles.xml" );
	if( p_cr->getError() ) return 5;

	// Partikelsystem erzeugen
	p_ps = new ParticleSystem( video, env.engine.getTickDuration() );
	if( p_ps->getError() ) return 6;

	return 0;
}

void Level::setupParticleTypes()
{
	ParticleStyle style;

	// Rauch
	style = ParticleStyle();
	style.mode = PM_NORMAL;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 1, 1 ) );
	style.frames.push_back( Vec2i( 65, 1 ) );
	style.frames.push_back( Vec2i( 129, 1 ) );
	style.frames.push_back( Vec2i( 193, 1 ) );
	style.frameSize = Vec2i( 62, 62 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 10.0f;
	p_ps->registerParticleStyle( 1, style );

	// Stichflamme
	style = ParticleStyle();
	style.mode = PM_STRETCH_BY_VELOCITY;
	style.stretch = 0.1f;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 0, 64 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE;
	style.animSpeed = 0.0f;
	p_ps->registerParticleStyle( 2, style );

	// Trmmer bei Kollisionen
	style = ParticleStyle();
	style.mode = PM_NORMAL;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 0, 0 ) );
	style.frames.push_back( Vec2i( 64, 0 ) );
	style.frames.push_back( Vec2i( 128, 0 ) );
	style.frames.push_back( Vec2i( 192, 0 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 10.0f;
	p_ps->registerParticleStyle( 3, style );

	// Funken
	style = ParticleStyle();
	style.mode = PM_STRETCH_BY_VELOCITY;
	style.stretch = 0.1f;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 0, 64 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE;
	style.animSpeed = 0.0f;
	p_ps->registerParticleStyle( 4, style );

	// Reparatur/Auftanken
	style = ParticleStyle();
	style.mode = PM_STRETCH_BY_VELOCITY;
	style.stretch = 0.5f;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 64, 64 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 0.0f;
	p_ps->registerParticleStyle( 5, style );

	// Tintenspritzer
	style = ParticleStyle();
	style.mode = PM_STRETCH_BY_VELOCITY;
	style.stretch = 0.1f;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 0, 64 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 0.0f;
	p_ps->registerParticleStyle( 6, style );

	// Explosion
	style = ParticleStyle();
	style.mode = PM_NORMAL;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 1, 129 ) );
	style.frames.push_back( Vec2i( 65, 129 ) );
	style.frames.push_back( Vec2i( 129, 129 ) );
	style.frames.push_back( Vec2i( 193, 129 ) );
	style.frameSize = Vec2i( 62, 62 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 2.0f;
	p_ps->registerParticleStyle( 7, style );

	// Feuer
	style = ParticleStyle();
	style.mode = PM_NORMAL;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 1, 193 ) );
	style.frames.push_back( Vec2i( 65, 193 ) );
	style.frames.push_back( Vec2i( 129, 193 ) );
	style.frames.push_back( Vec2i( 193, 193 ) );
	style.frameSize = Vec2i( 62, 62 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE;
	style.animSpeed = 5.0f;
	p_ps->registerParticleStyle( 8, style );

	// "Erbsenspritzer"
	style = ParticleStyle();
	style.mode = PM_STRETCH_BY_VELOCITY;
	style.stretch = 0.1f;
	style.texture = particles;
	style.frames.clear();
	style.frames.push_back( Vec2i( 64, 64 ) );
	style.frameSize = Vec2i( 64, 64 );
	style.srcBlendFactor = GL_SRC_ALPHA;
	style.destBlendFactor = GL_ONE_MINUS_SRC_ALPHA;
	style.animSpeed = 0.0f;
	p_ps->registerParticleStyle( 9, style );
}

void Level::setupCollisionParameters()
{
	SoundManager& soundMgr = env.engine.getSoundManager();
	CollisionParameters params;
	SimpleEmitter* p_emitter;

	// Spieler auf Metall
	params = CollisionParameters();
	params.frontal.minSpeed = 5.0f;
	params.frontal.sound = soundMgr.get( "data\\collision1.ogg" );
	params.frontal.sound->setDefaultSoundGroupName( "SFX" );
	params.frontal.sound->setInstanceCreationInterval( 0.25f );
	params.frontal.soundVolumeFactor = 1.0f / 50.0f;
	params.frontal.createEmitter = true;
	p_emitter = p_ps->createSimpleEmitter();
	p_emitter->addParticleStyle( 3 );
	p_emitter->setParticleTimeToLive( 0.5f, 0.05f );
	p_emitter->setParticlePositionRadius( 0.25f );
	p_emitter->setParticleSpeed( 2.0f, 0.1f );
	p_emitter->setParticleStartColor( Vec4f( 1.0f, 0.85f, 0.85f, 0.75f ), Vec4f( 0.1f, 0.1f, 0.1f, 0.5f ) );
	p_emitter->setParticleEndColor( Vec4f( 0.85f, 1.0f, 0.85f, 0.0f ), Vec4f( 0.1f, 0.1f, 0.1f, 0.0f ) );
	p_emitter->setParticleStartSize( 0.75f, 0.25f );
	p_emitter->setParticleEndSize( 1.5f, 0.5f );
	p_emitter->setParticleGravity( 0.0f, 0.1f );
	p_emitter->setParticleAnimFrame( 0.0f, 10.0f );
	params.frontal.emitterSettings = p_emitter->getSettings();
	params.frontal.particleRateFactor = 5.0f;
	p_ps->destroyEmitter( p_emitter );
	params.scratch.minSpeed = 5.0f;
	params.scratch.sound = soundMgr.get( "data\\scratch1.ogg" );
	params.scratch.sound->setDefaultSoundGroupName( "SFX" );
	params.scratch.sound->setInstanceCreationInterval( 0.25f );
	params.scratch.soundVolumeFactor = 1.0f / 35.0f;
	params.scratch.createEmitter = true;
	p_emitter = p_ps->createSimpleEmitter();
	p_emitter->addParticleStyle( 4 );
	p_emitter->setParticleTimeToLive( 0.5f, 0.2f );
	p_emitter->setParticlePositionRadius( 0.1f );
	p_emitter->setParticleSpeed( 20.0f, 5.0f );
	p_emitter->setParticleStartColor( Vec4f( 1.0f, 1.0f, 1.0f, 0.5f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.25f ) );
	p_emitter->setParticleEndColor( Vec4f( 1.0f, 1.0f, 1.0f, 0.5f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.25f ) );
	p_emitter->setParticleStartSize( 1.0f, 0.05f );
	p_emitter->setParticleEndSize( 0.0f, 0.0f );
	p_emitter->setParticleGravity( 8.0f, 1.0f );
	p_emitter->setParticleFriction( 0.25f, 0.1f );
	params.scratch.emitterSettings = p_emitter->getSettings();
	params.scratch.particleRateFactor = 2.0f;
	p_ps->destroyEmitter( p_emitter );
	registerCollisionPair( "Player", "Metal", params );
	registerCollisionPair( "Metal", "Metal", params );

	// Tankstelle
	params.friction = 50.0f;
	registerCollisionPair( "Player", "Garage", params );
	registerCollisionPair( "Metal", "Garage", params );

	// Spieler auf Power-Up
	params = CollisionParameters();
	params.dontCreateContactJoint = true;
	registerCollisionPair( "Player", "PowerUp", params );

	// Gummi auf Metall/Gummi
	params = CollisionParameters();
	params.bouncyness = 0.75f;
	params.friction = 5.0f;
	params.minBounceSpeed = 0.0f;
	params.frontal.minSpeed = 5.0f;
	params.frontal.sound = soundMgr.get( "data\\collision2.ogg" );
	params.frontal.sound->setDefaultSoundGroupName( "SFX" );
	params.frontal.sound->setInstanceCreationInterval( 0.25f );
	params.frontal.soundVolumeFactor = 1.0f / 50.0f;
	registerCollisionPair( "Rubber", "Metal", params );
	registerCollisionPair( "Rubber", "Rubber", params );

	// Geschoss auf alles
	params = CollisionParameters();
	params.bouncyness = 0.8f;
	params.friction = 0.0f;
	params.minBounceSpeed = 0.0f;
	params.threshold = 0.15f;
	params.frontal.minSpeed = 0.0f;
	params.scratch.minSpeed = 0.0f;
	registerCollisionPair( "Projectile", "All", params );

	// Trigger-Bereiche (wie z.B. Beschleuniger)
	params = CollisionParameters();
	params.dontCreateContactJoint = true;
	registerCollisionPair( "Trigger", "All", params );
}

void Level::renderObjects( RenderPass pass,
						   const Vec4f& color )
{
	// Objekte rendern
	for( std::vector<GameObject*>::iterator it = visibleObjects.begin();
		 it != visibleObjects.end();
		 it++ )
	{
		video.setTexture( sprites );
		glColor4fv( color );
		passColor = color;

		( *it )->render( pass );
	}
}

void Level::deleteObject( GameObject* p_object )
{
	// War dieses Objekt das Zielobjekt der Kamera?
	if( p_cameraTarget == p_object ) p_cameraTarget = 0;

	if( p_object->getTurnToZombie() )
	{
		// Das Objekt wird zum Zombie.
		zombies.insert( p_object );
	}
	else
	{
		// Objekt aus der lua-Umgebung entfernen
		globals( p_lua )[p_object->getID()] = nil;

		// Objekt lschen
		delete p_object;
	}
}

void Level::processCollisions()
{
	// alle Paare von Objekten ber ihre Kollisionen informieren
	for( std::map<CollisionPair, CollisionInfo>::iterator it = collisions.begin();
		 it != collisions.end();
		 it++ )
	{
		// die Spielobjekte ber die Kollision informieren
		if( it->first.p_obj1 ) it->first.p_obj1->processCollision( it->first.p_obj2, it->second );
		if( it->first.p_obj2 ) it->first.p_obj2->processCollision( it->first.p_obj1, it->second );
	}
}

void Level::calculateVisibility()
{
	visibleObjects.clear();

	for( std::map<std::string, GameObject*>::const_iterator it = objects.begin();
		 it != objects.end();
		 it++ )
	{
		GameObject* p_obj = it->second;

		if( !p_obj->isVisible() ) continue;

		bool visible = false;
		float boundRadius = p_obj->getBoundRadius();

		if( boundRadius == -FLT_MAX ) continue;

		if( boundRadius >= 0.0f )
		{
			// Das Objekt benutzt einen Bounding-Circle.
			visible = video.isCircleVisible( p_obj->getBoundCenter(), boundRadius, p_obj->getTransform() );
		}
		else
		{
			// Das Objekt benutzt ein Bounding-Rect.
			visible = video.isRectVisible( p_obj->getBoundMin(), p_obj->getBoundMax(), p_obj->getTransform() );
		}

		if( visible )
		{
			// Das Objekt hat alle vier Tests bestanden und ist somit sichtbar.
			visibleObjects.push_back( p_obj );
		}
	}
}

void Level::renderTexts()
{
	for( std::list<Text>::const_iterator it = texts.begin();
		 it != texts.end();
		 it++ )
	{
		const Text& t = *it;
		Vec4f color( t.color );
		color.a *= min( 1.0f, t.timeToLive );

		glPushMatrix();
		glTranslatef( t.position.x, t.position.y, 0.0f );
		glRotatef( t.angle * 57.2958f, 0.0f, 0.0f, 1.0f );

		env.scoreFont->print( t.text, t.size * 1.1f, Vec4f( 0.0f, 0.0f, 0.0f, color.a * 0.5f ), FF_HCENTER );
		env.scoreFont->print( t.text, t.size, color, FF_HCENTER );

		glPopMatrix();
	}
}

void Level::updateTexts()
{
	for( std::list<Text>::iterator it = texts.begin();
		 it != texts.end(); )
	{
		Text& t = *it;
		const Vec2f yAxis( sinf( -t.angle ), cosf( -t.angle ) );
		const Vec2f xAxis( yAxis.perp() );
		t.position += env.dt * ( 2.5f * sinf( t.timeToLive * 7.5f ) * xAxis - 7.5f * yAxis );
		t.size += 0.5f * env.dt;
		t.timeToLive -= env.dt;
		if( t.timeToLive < 0.0f ) it = texts.erase( it );
		else it++;
	}
}

void Level::updateSMS()
{
	if( smsTime == -1.0f )
	{
		// Neue SMS da?
		if( !smsInbox.empty() )
		{
			smsTime = 0.0f;

			// Vibrationssound abspielen
			env.engine.getAudio().playSound( "data\\sms.ogg", "SFX", "Sample", "2D", 0.7f, randFloat( 0.95f, 1.05f ) );
		}
	}
	else
	{
		smsTime += env.dt;
		if( smsTime >= smsInbox.front().duration )
		{
			smsTime = -1.0f;
			smsInbox.pop();
		}
	}
}

void Level::updateScriptObjects()
{
	for( std::list<TimerCallback>::iterator it = timerCallbacks.begin();
		 it != timerCallbacks.end(); )
	{
		TimerCallback& cb = *it;

		if( cb.interval < 0.0f ) it = timerCallbacks.erase( it );
		else
		{
			cb.t += env.dt;
			if( cb.t >= cb.interval )
			{
				bool result = false;

				try
				{
					result = luabind::call_function<bool>( cb.callback, cb.interval );
				}
				catch( luabind::error& err )
				{
					Scripting::inst().printException( err );
				}

				if( result ) cb.t = 0.0f;
				else cb.interval = -1.0f;
			}

			it++;
		}
	}

	size_t n = resumeCountDowns.size();
	for( std::list<ResumeCountDown>::iterator it = resumeCountDowns.begin();
		 it != resumeCountDowns.end() && n;
		 n-- )
	{
		ResumeCountDown& cd = *it;
		cd.t -= env.dt;
		if( cd.t <= 0.0f )
		{
			// Script fortsetzen
			Scripting::inst().resumeThread( cd.p_lua );
			it = resumeCountDowns.erase( it );
		}
		else it++;
	}

	// Script-Objekte aktualisieren
	for( std::set<ScriptObject*>::iterator it = scriptObjects.begin();
		 it != scriptObjects.end(); )
	{
		( *it )->update();

		if( ( *it )->isToBeDeleted() )
		{
			// Das Objekt mchte gerne gelscht werden.
			delete *it;
			it = scriptObjects.erase( it );
		}
		else it++;
	}
}

void Level::renderAvoidanceAreas()
{
	video.setTexture( 0 );

	for( std::set<AvoidanceArea>::const_iterator it = avoidanceAreas.begin();
		 it != avoidanceAreas.end();
		 it++ )
	{
		const AvoidanceArea& area = *it;

		Placeable p;
		p.setTransform( area.transformation );

		glPushMatrix();
		p.setupGLMatrix();

		switch( area.type )
		{
		case 0:
			// Kreis
			glBegin( GL_LINE_LOOP );
			glColor4f( 1.0f, 0.0f, 0.0f, 0.75f );
			for( float x = 0.0f; x < 6.283f; x += 0.25f )
			{
				glVertex2f( sinf( x ) * area.radius, cosf( x ) * area.radius );
			}
			glEnd();
			break;

		case 1:
			// Rechteck
			glBegin( GL_LINE_LOOP );
			glColor4f( 1.0f, 0.0f, 0.0f, 0.75f );
			glVertex2f( area.min.x, area.min.y );
			glVertex2f( area.max.x, area.min.y );
			glVertex2f( area.max.x, area.max.y );
			glVertex2f( area.min.x, area.max.y );
			glEnd();
			break;
		}

		glPopMatrix();
	}
}

void Level::renderMobile()
{
	if( smsTime == -1.0f || smsInbox.empty() ) return;

	const SMS& sms = smsInbox.front();

	glPushMatrix();
	glLoadIdentity();

	size_t currentPage = min( sms.pageStarts.size() - 1, static_cast<size_t>( ( smsTime - 2.0f ) / 6.0f ) );

	// Handy rein und raus bewegen
	Vec2f pos( 0.0f, 2.5f );
	float angle = 0.0f;
	if( smsTime < 1.0f )
	{
		const float t = 1.0f - smsTime;
		pos = ( 1.0f - t ) * pos + t * Vec2f( -6.0f, 18.0f );
		angle = t * 25.0f;
	}
	else if( smsTime > sms.duration - 1.0f )
	{
		const float t = smsTime - ( sms.duration - 1.0f );
		pos = ( 1.0f - t ) * pos + t * Vec2f( -6.0f, 18.0f );
		angle = t * 25.0f;
	}

	// Vibration
	if( smsTime < 1.0f || ( smsTime > 1.25f && smsTime < 2.25f ) ) pos += randFloat( 0.0f, 0.025f ) * randVec2fDir();

	glTranslatef( -24.0f + pos.x, pos.y, 0.0f );
	glScalef( 9.6f, 9.6f, 1.0f );
	glRotatef( angle, 0.0f, 0.0f, 1.0f );

	video.setTexture( mobile );
	glBegin( GL_QUADS );
	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
	glTexCoord2i( 0, 0 );
	glVertex2f( 0.0f, 0.0f );
	glTexCoord2i( 256, 0 );
	glVertex2f( 1.0f, 0.0f );
	glTexCoord2i( 256, 512 );
	glVertex2f( 1.0f, 2.0f );
	glTexCoord2i( 0, 512 );
	glVertex2f( 0.0f, 2.0f );
	glEnd();

	glPushMatrix();
	glTranslatef( 100.0f / 256.0f, 180.0f / 256.0f, 0.0f );
	env.mobileFont->print( sms.sender, 16.0f / 256.0f, Vec4f( 0.17f, 0.16f, 0.2f, 0.95f ) );
	glPopMatrix();

	if( ( smsTime > 0.9f && smsTime < 3.0f ) ||
		( smsTime >= 3.0f && fmodf( smsTime - 2.0f, 6.0f ) > 0.25f ) ||
		( smsTime > sms.duration - 1.0f ) )
	{
		glPushMatrix();
		glTranslatef( 60.0f / 256.0f, 205.0f / 256.0f, 0.0f );

		size_t length;
		if( currentPage < sms.pageStarts.size() - 1 ) length = sms.pageStarts[currentPage + 1] - sms.pageStarts[currentPage];
		else length = std::string::npos;
		std::string text( sms.text.substr( sms.pageStarts[currentPage], length ) );
		env.mobileFont->print( text,
							   16.0f / 256.0f,
							   Vec4f( 0.17f, 0.16f, 0.2f, 0.95f ),
							   FF_LEFT,
							   FontStyle( 0, -4 ) );
		glPopMatrix();
	}

	glPopMatrix();
}

void Level::renderScriptObjects()
{
	for( std::set<ScriptObject*>::const_iterator it = scriptObjects.begin();
		 it != scriptObjects.end();
		 it++ )
	{
		( *it )->render();
	}
}