#include <PCH.h>
#include <GS_Intro.h>
#include <GS_Menu.h>
#include <GS_Game.h>
#include <GS_TestArena.h>
#include <GS_Loading.h>
#include <GS_Config.h>
#include <CartoonRenderer.h>
#include <NeuralNet.h>
#include <GUI.h>
#include <Console.h>
#include <Scripting.h>
#include <StackWalker.h>

int runConfig( ParameterMap& config,
			   const ParameterMap& defaultConfig )
{
	// konstante Optionen setzen
	ParameterMap newConfig( config );
	newConfig.set<bool>( "Audio.NoSound", true );
	newConfig.set<unsigned int>( "Engine.TickDuration", 10 );
	newConfig.set<std::string>( "Video.WindowTitle", "RocketRage" );

#ifdef _DEBUG
	// Fenstermodus erzwingen
	newConfig.set<bool>( "Video.Windowed", true );
#endif

	// Engine initialisieren
	Engine engine( newConfig );
	if( engine.getError() )
	{
		std::cout << "Fehler beim Initialisieren der Engine" << std::endl;
		return 1;
	}

	// GUI-Komponente erzeugen
	GUI* p_gui = new GUI( engine );
	if( p_gui->getError() )
	{
		std::cout << "Fehler bei der Initialisierung der GUI-Komponente" << std::endl;
		return 2;
	}

	// Konfigurationsspielzustand erzeugen
	GS_Config cfg( engine, *p_gui, config, defaultConfig );
	engine.pushGameState( "Config" );
	int r = engine.run();
	if( r )
	{
		// Das Fenster wurde geschlossen.
		return 1;
	}

	r = config.get<int>( "Config.Result" );

	// aufrumen
	delete p_gui;

	return r;
}

int runTheGame( int argc,
				char** pp_args )
{
	LogLevel logLevel = LL_INFO;

#ifdef _DEBUG
	logLevel = LL_FINEST;
#endif

	bool showConfig = true;

	// Standardkonfiguration
	ParameterMap defaultConfig;
	defaultConfig.set<unsigned int>( "Video.Width", 1024 );
	defaultConfig.set<unsigned int>( "Video.Height", 768 );
	defaultConfig.set<bool>( "Video.Windowed", false );
	defaultConfig.set<bool>( "Video.VSync", false );
	defaultConfig.set<bool>( "Video.UseMIPMapping", true );
	defaultConfig.set<bool>( "Audio.NoSound", false );
	defaultConfig.set<bool>( "Audio.UseHardware", false );
	defaultConfig.set<unsigned int>( "Audio.Software3DQuality", 1 );

	ParameterMap config = defaultConfig;

	// Optionen fr die Engine laden
	XMLSerializationStream str( "engine.cfg" );
	if( !str.getError() )
	{
		// Konfiguration versuchen zu laden
		ParameterMap deserialized;
		if( !Serializer::inst().deserialize( deserialized, str ) )
		{
			// Die Konfiguration ist gltig.
			config += deserialized;
			showConfig = false;
		}
	}

	// Wenn als Parameter "-config" angegeben wurde, wird der Konfigurationsbildschirm
	// in jedem Fall angezeigt.
	for( int i = 0; i < argc; i++ )
	{
		std::string arg( pp_args[i] );
		if( arg == "-config" ) showConfig = true;
	}

	if( showConfig )
	{
		int r = runConfig( config, defaultConfig );
		if( r )
		{
			// Spiel beenden
			return 0;
		}
	}

	// konstante Optionen setzen
	config.set<int>( "Engine.LogLevel", logLevel );
	config.set<unsigned int>( "Engine.TickDuration", 10 );
	config.set<std::string>( "Video.WindowTitle", "RocketRage" );

#ifdef _DEBUG
	// Fenstermodus erzwingen
	config.set<bool>( "Video.Windowed", true );
#endif

	// Engine initialisieren
	Engine engine( config );
	if( engine.getError() )
	{
		std::cout << "Fehler beim Initialisieren der Engine" << std::endl;
		return 1;
	}

	GameConsoleLogHandler h1( engine.getLogger(), 0, Vec4f( 0.75f, 0.75f, 1.0f, 1.0f ) );

	// Motion-Blurring einschalten
	engine.getVideo().setupMotionBlurring( true, 0.0f );

	// Standardsoundgruppen erzeugen
	Audio& audio = engine.getAudio();
	SoundGroup* p_sfxGroup = audio.createSoundGroup( "SFX", 0 );
	SoundGroup* p_musicGroup = audio.createSoundGroup( "Music", 0 );
	audio.setMusicSoundGroup( p_musicGroup );
	audio.getMasterSoundGroup()->setVolume( 0.75f );

	// Spiel-Logger erzeugen
	Logger gameLogger;
	gameLogger.setLogLevel( logLevel );
	LogHandler* p_handler = new ConsoleLogHandler( gameLogger );
	GameConsoleLogHandler h2( gameLogger, 0, Vec4f( 0.75f, 1.0f, 0.75f, 1.0f ) );

	// Manager fr neuronale Netze erzeugen
	NeuralNetManager* p_neuralNetMgr = new NeuralNetManager( engine, "NeuralNetManager" );

	// GUI-Komponente erzeugen
	GUI* p_gui = new GUI( engine );
	if( p_gui->getError() )
	{
		gameLogger.error( "Fehler bei der Initialisierung der GUI-Komponente", LOGGER_AUTO );
		return 2;
	}

	// Optionen erzeugen
	Options* p_options = new Options( engine, *p_gui );

	// Spielumgebung erzeugen
	GameEnvironment env( engine, gameLogger, *p_neuralNetMgr, *p_gui, *p_options );

	// Konsole erzeugen
	Console* p_console = new Console( env );
	h1.setConsole( p_console );
	h2.setConsole( p_console );

	// Scripting-Umgebung initialisieren
	Scripting::inst().init( p_console );

	// Spielzustnde erzeugen (sie registrieren sich selbst)
	GS_Intro intro( env );
	GS_Menu menu( env );
	GS_Game game( env );
	GS_TestArena testArena( env );
	GS_Loading loading( env );

	// Startzustand festlegen
	ParameterMap params;
/*	params.set<std::string>( "Mode", "Play" );
//	params.set<std::string>( "Mode", "Evolution" );
	params.set<std::string>( "CandidateBehavior", "MoveTo" );
	params.set<std::string>( "BaseNetFilename", "training\\player-moveto.net" );
	params.set<unsigned int>( "StartPopulation", 4 );
	params.set<unsigned int>( "MaxPopulation", 100 );
	params.set<unsigned int>( "NumPasses", 50 );
//	engine.pushGameState( "TestArena", params );*/

	params.set<std::string>( "Redirect.GameState", "Menu" );
	engine.pushGameState( "Loading", params );

	// Starten der Engine
	int r = engine.run();

	// Scripting-Umgebung herunterfahren
	Scripting::inst().exit();

	// Objekte lschen
	h1.disconnect();
	h2.disconnect();
	delete p_options;
	delete p_console;
	delete p_gui;
	delete p_neuralNetMgr;
	delete p_handler;

	// ODE herunterfahren
	dCloseODE();

	return r;
}

std::ofstream crashLog;

class MyStackWalker : public StackWalker
{
public:
	MyStackWalker() : StackWalker() { }
	MyStackWalker( DWORD processID, HANDLE process ) : StackWalker( processID, process ) { }
	void OnOutput( LPCSTR p_text ) { crashLog << p_text << std::endl; StackWalker::OnOutput( p_text ); }
};

LONG WINAPI expFilter( EXCEPTION_POINTERS* p_exception,
					   DWORD exceptionCode )
{
	std::cout << "Writing crash information to \"crash.txt\" ..." << std::endl;

	crashLog.open( "crash.txt" );
	crashLog << "**************************************************" << std::endl;
	crashLog << "An exception was thrown!" << std::endl;
	crashLog << "Exception code: " << exceptionCode << std::endl;
	crashLog << "Printing stack trace ..." << std::endl;
	crashLog << "**************************************************" << std::endl << std::endl;

	MyStackWalker sw;
	sw.ShowCallstack( GetCurrentThread(), p_exception->ContextRecord );

	crashLog.close();

	return EXCEPTION_EXECUTE_HANDLER;
}

int main( int argc,
		  char** pp_argv )
{
#ifdef _DEBUG
	return runTheGame( argc, pp_argv );
#else
	__try
	{
		return runTheGame( argc, pp_argv );
	}
	__except( expFilter( GetExceptionInformation(), GetExceptionCode() ) )
	{
		return 1;
	}
#endif
}