#include <PCH.h>
#include <Obj_Enemy.h>

Obj_Enemy::Obj_Enemy( Level& level,
					  const TiXmlElement* p_root,
					  bool createBody ) : GameObject( level, p_root, createBody ),
										  Controllable( level.getEnv() ),
										  Armed( level.getEnv() ),
										  state( ES_READY ),
										  texCoords( 0, 0 ),
										  sizeOnTexture( 0, 0 ),
										  scaling( 4.0f ),
										  range( 25.0f ),
										  currentWaypoint( 0 ),
										  pathDir( 1 ),
										  pathMode( EPM_ONCE ),
										  waiting( 0.0f ),
										  avoidanceRadius( 5.0f ),
										  oldPosition( FLT_MAX ),
										  moved( 0.0f ),
										  movementCheckTimer( 0.0f )
{
	const TiXmlElement* p_path = p_root->FirstChildElement( "Path" );
	if( p_path )
	{
		int pathMode_ = ES_READY;
		p_path->QueryIntAttribute( "mode", &pathMode_ );
		pathMode = static_cast<EnemyPathMode>( pathMode_ );

		// Pfad lesen
		const TiXmlElement* p_waypoint = p_path->FirstChildElement( "Waypoint" );
		while( p_waypoint )
		{
			const char* p_id = p_waypoint->GetText();
			if( p_id ) path.push_back( p_id );
			p_waypoint = p_waypoint->NextSiblingElement( "Waypoint" );
		}
	}

	// Ziel lesen, Standardziel ist der Spieler
	const char* p_target = p_root->Attribute( "target" );
	if( p_target ) setTargetID( p_target );
	else setTargetID( "player" );

	setAIMode();

	// Der Gegner kann mit allen Objekten kollidieren.
	setCollisionBits( CB_ENEMY, CB_ALL );

	// Standardeingabeparameter vorbereiten
	registerParameter( "Speed", PT_INPUT, 0.0f, 50.0f );
	registerParameter( "AngularVelocity", PT_INPUT, -5.0f, 5.0f );
	registerParameter( "Dir-Gravity", PT_INPUT, -10.0f, 10.0f );
	registerParameter( "Dir-FlightDir", PT_INPUT, -1.0f, 1.0f );
	registerParameter( "Dir-DirToTarget", PT_INPUT, -1.0f, 1.0f );
	registerParameter( "Velocity-Gravity", PT_INPUT, -500.0f, 500.0f );
	registerParameter( "Velocity-TargetVelocity", PT_INPUT, -500.0f, 500.0f );
	registerParameter( "Velocity-DirToTarget", PT_INPUT, -50.0f, 50.0f );
	registerParameter( "TargetSpeed", PT_INPUT, 0.0f, 50.0f );
	registerParameter( "TargetDist", PT_INPUT, -25.0f, 25.0f );
	registerParameter( "DistFlightDir", PT_INPUT, 0.0f, 25.0f );
	registerParameter( "RelSpeed", PT_INPUT, -100.0f, 100.0f );
}

Obj_Enemy::~Obj_Enemy()
{
}

void Obj_Enemy::update()
{
	if( isFrozen() ) return;

	GameObject::update();
	updateControllable();
	updateArmed();
}

void Obj_Enemy::render( RenderPass pass )
{
	if( pass == RP_SHADOW || pass == RP_OBJECT_2 )
	{
		// Sprite rendern
		glPushMatrix();
		setupGLMatrix();
		glScalef( scaling, scaling, 1.0f );
		video.drawSprite( texCoords, sizeOnTexture );
		glPopMatrix();
	}
	else if( pass == RP_DEBUG_INFO )
	{
		video.setTexture( 0 );
		const Vec2f myPosition( getPosition() );

		if( state != ES_READY )
		{
			// aktuellen Zielpunkt rendern
			glBegin( GL_LINE_STRIP );
			glColor4f( 0.0f, 0.0f, 1.0f, 0.75f );
			glVertex2fv( myPosition );
			glVertex2fv( movementTargetPosition );
			glEnd();
		}

		glPushMatrix();
		glTranslatef( myPosition.x, myPosition.y, 0.0f );

		// Ausweichradius darstellen
		glBegin( GL_LINE_LOOP );
		glColor4f( 0.0f, 0.0f, 1.0f, 0.25f );
		for( float x = 0.0f; x < 6.283f; x += 0.25f )
		{
			glVertex2f( sinf( x ) * avoidanceRadius, cosf( x ) * avoidanceRadius );
		}
		glEnd();

		// Modus ausgeben
		std::string str = "";
		switch( state )
		{
		case ES_READY: str = "Ready"; break;
		case ES_AVOID: str = "Avoid"; break;
		case ES_CHASE_TARGET: str = "ChaseTarget"; break;
		case ES_FOLLOW_PATH: str = "FollowPath"; break;
		case ES_FREE: str = "Free"; break;
		}

		env.fixedFont->print( str, 1.0f, Vec4f( 0.0f, 0.0f, 1.0f, 0.75f ) );

		glPopMatrix();
	}
}

void Obj_Enemy::generateInput()
{
	if( getControlMode() == CM_AI_SIMPLE ) return;

	const Vec2f position( getPosition() );
	const Vec2f velocity( getVelocity() );
	const Vec2f flightDir( velocity.normalized() );
	const Vec2f direction( -getDirection() );
	const Vec2f gravity( level.getGravity() );
	const float maxDist = 25.0f;
	float distFlightDir = maxDist;
	const Vec2f rayDir( velocity.length() > 0.1f ? flightDir : direction );
	level.castRay( position, rayDir, maxDist, CB_NONE, CB_STATIC | CB_BLOCK_VIEW, 0, &distFlightDir );

	setParameter( "Speed", velocity.length() );
	setParameter( "AngularVelocity", getAngularVelocity() );
	setParameter( "Dir-Gravity", direction.perp() ^ gravity );
	setParameter( "Dir-FlightDir", direction.perp() ^ flightDir );
	setParameter( "Velocity-Gravity", velocity.perp() ^ gravity );
	setParameter( "DistFlightDir", distFlightDir );

	const Vec2f targetPosition( movementTargetPosition );
	const Vec2f vecToTarget( targetPosition - position );
	const Vec2f dirToTarget( vecToTarget.normalized() );
	const float targetDist = vecToTarget.length() * ( ( direction ^ vecToTarget ) >= 0.0f ? 1.0f : -1.0f );
	const Vec2f targetVelocity( movementTargetVelocity );
	const float relSpeed = ( ( position + 0.001f * velocity - targetPosition + 0.001f * targetVelocity ).length() - vecToTarget.length() ) / 0.001f;

	setParameter( "Dir-DirToTarget", direction.perp() ^ dirToTarget );
	setParameter( "Velocity-TargetVelocity", velocity.perp() ^ targetVelocity );
	setParameter( "Velocity-DirToTarget", velocity.perp() ^ dirToTarget );
	setParameter( "TargetSpeed", targetVelocity.length() );
	setParameter( "TargetDist", targetDist );
	setParameter( "RelSpeed", relSpeed );
}

void Obj_Enemy::think()
{
	Controllable::think();

	const Vec2f myPosition( getPosition() );
	const int numWaypoints = static_cast<int>( path.size() );

	if( oldPosition.x != FLT_MAX ) moved += ( oldPosition - myPosition ).length();

	movementCheckTimer += env.dt;
	if( movementCheckTimer > 2.0f )
	{
		if( moved < 4.0f )
		{
			// Der Gegner hngt wahrscheinlich fest, weil er sich in den letzten Sekunden
			// kaum bewegt hat.
			state = ES_FREE;
			movementTargetPosition = myPosition + randFloat( 5.0f, 10.0f ) * randVec2fDir();
			movementTargetVelocity = randFloat( 0.0f, 10.0f ) * randVec2fDir();
		}

		moved = 0.0f;
		movementCheckTimer = 0.0f;
	}

	oldPosition = myPosition;

	// Befindet sich der Gegner in einer Gefahrzone?
	Vec2f escapeVector;
	bool isInAvoidanceArea = level.getEscapeVector( this, avoidanceRadius, escapeVector );

	// Ist das Zielobjekt sichtbar?
	bool canSeeTarget = false;
	GameObject* p_target = getTarget();
	if( p_target && canSee( p_target, range ) ) canSeeTarget = true;

	// Gibt es noch Wegpunkte abzuarbeiten?
	bool hasNextWaypoint = false;
	if( !path.empty() )
	{
		if( currentWaypoint >= 0 && currentWaypoint < numWaypoints ) hasNextWaypoint = true;
	}

	switch( state )
	{
	case ES_READY:
		{
			// zufllige Bewegungen ausfhren
			movementTargetPosition = myPosition + randFloat( 0.0f, 10.0f ) * randVec2fDir();
			movementTargetVelocity = randFloat( 0.0f, 10.0f ) * randVec2fDir();

			// Zustandsbergnge
			if( isInAvoidanceArea ) state = ES_AVOID;
			else if( canSeeTarget ) state = ES_CHASE_TARGET;
			else if( hasNextWaypoint ) state = ES_FOLLOW_PATH;
			break;
		}

	case ES_AVOID:
		{
			if( isInAvoidanceArea )
			{
				// aus der Gefahrzone fliehen
				movementTargetPosition = myPosition + escapeVector;
				movementTargetVelocity = escapeVector + randVec2fDir();
			}

			// Zustandsbergnge
			if( !isInAvoidanceArea ) state = ES_READY;
			break;
		}

	case ES_CHASE_TARGET:
		{
			if( canSeeTarget )
			{
				// zum Ziel fliegen
				movementTargetPosition = p_target->getPosition();
				movementTargetVelocity = p_target->getVelocity() + randVec2fDir();
				waiting = 0.0f;
			}
			else
			{
				waiting += env.dt;
			}

			// Zustandsbergnge
			if( isInAvoidanceArea ) state = ES_AVOID;
			else if( !canSeeTarget && waiting > 5.0f ) state = ES_READY;
			break;
		}

	case ES_FOLLOW_PATH:
		{
			// zum aktuellen Wegpunkt fliegen
			GameObject* p_currentWaypoint = level.getObjectByID( path[currentWaypoint] );
			if( p_currentWaypoint )
			{
				movementTargetPosition = p_currentWaypoint->getPosition();
				movementTargetVelocity = randVec2fDir();

				// Ist der Wegpunkt erreicht?
				if( ( myPosition - movementTargetPosition ).lengthSq() < 25.0f )
				{
					// Ja. Wir gehen zum Nchsten.
					currentWaypoint += pathDir;

					if( currentWaypoint < 0 )
					{
						// vorderes Ende erreicht
						if( pathMode == EPM_LOOP ) currentWaypoint = numWaypoints - 1;
						else if( pathMode == EPM_PING_PONG )
						{
							currentWaypoint = 0;
							pathDir = -pathDir;
						}
					}
					else if( currentWaypoint >= numWaypoints )
					{
						// hinteres Ende erreicht
						if( pathMode == EPM_LOOP ) currentWaypoint = 0;
						else if( pathMode == EPM_PING_PONG )
						{
							currentWaypoint = numWaypoints - 1;
							pathDir = -pathDir;
						}
					}
				}
			}

			// Zustandsbergnge
			if( isInAvoidanceArea ) state = ES_AVOID;
			else if( canSeeTarget ) state = ES_CHASE_TARGET;
			break;
		}

	case ES_FREE:
		// Zustandsbergnge
		if( isInAvoidanceArea ) state = ES_AVOID;
		else if( moved > 6.0f ) state = ES_READY;
		break;
	}

	// Ist das Ziel in Reichweite und sichtbar?
	if( canSeeTarget )
	{
		// Feuern!
		setParameter( "Fire", 1.0f );
	}
}