#include <PCH.h>
#include <Obj_Enemy_AAStation.h>
#include <Obj_Projectile.h>

Obj_Enemy_AAStation::Obj_Enemy_AAStation( Level& level,
										  const TiXmlElement* p_root ) : Obj_Enemy( level, p_root, false ),
																		 recoil( 0.0f )
{
	setMaxHealth( 10.0f );
	setHealth( 10.0f );
	setMaterial( "Metal" );

	// Die Flugabwehrstation kann mit Objekten kollidieren, aber nicht mit Wnden.
	setCollisionBits( CB_STATIC | CB_ENEMY, ~CB_STATIC );

	std::vector<Vec2f> vertices;
	vertices.push_back( Vec2f( -1.5f, 1.0f ) );
	vertices.push_back( Vec2f( 0.0f, -1.25f ) );
	vertices.push_back( Vec2f( 1.5f, 1.0f ) );
	addPolyLine( vertices, true, 0.5f );

	setBoundingCircle( Vec2f( 0.0f, 0.0f ), 3.0f );

	// Kanone einrichten
	gunBase.setPosition( Vec2f( 0.0f, 0.4f ) );
	gun.setPosition( Vec2f( 0.0f, -1.75f ) );
	gunBase.attachChild( &gun, false );
	attachChild( &gunBase, false );

	// Parameter konfigurieren
	registerParameter( "Turning", PT_OUTPUT, -1.0f, 1.0f );
	registerParameter( "Fire", PT_OUTPUT_USER_ONLY );
	registerParameter( "BurstTime", PT_OUTPUT_USER_ONLY, 0.5f, 5.0f );

	// Welche Waffe?
	int weapon_ = 1;
	p_root->QueryIntAttribute( "weapon", &weapon_ );
	if( weapon_ == 1 ) weapon = WT_AA;
	else if( weapon_ == 2 ) weapon = WT_MISSILE;

	// Waffen konfigurieren
	setupWeapon( weapon, Vec2f( 0.0f, -1.5f ), &gun );
	addAmmo( weapon, ~0 );
	selectWeapon( weapon );
}

Obj_Enemy_AAStation::~Obj_Enemy_AAStation()
{
}

void Obj_Enemy_AAStation::update()
{
	Obj_Enemy::update();

	if( getHealth() <= 0.0f )
	{
		level.addPlayerScore( 2500, getPosition() );
		explode( randInt( 15, 20 ), DT_ENEMY_AASTATION, 2.0f );
		kill();
		return;
	}

	if( isFrozen() ) return;

	// Rcksto der Kanone verarbeiten
	gun.setPosition( Vec2f( 0.0f, -1.75f + min( 0.25f, recoil ) ) );
	adjustLinear( recoil, 0.0f, 5.0f, env.dt );

	float turning = getParameter( "Turning" );
	if( turning != 0.0f )
	{
		// Kanone drehen
		float angle = gunBase.getAngle();
		angle += turning * 1.25f * env.dt;
		angle = clamp( angle, -1.0f, 1.0f );
		gunBase.setAngle( angle );
	}

	if( getParameter( "Fire" ) > 0.0f )
	{
		Obj_Projectile* p_projectile = fire( Vec2f( 0.0f, -1.0f ) );
		if( p_projectile )
		{
			// Detonationsszeit des Geschosses programmieren
			if( weapon == WT_AA ) p_projectile->setTimeToLive( getParameter( "BurstTime" ) );
			p_projectile->setTargetID( getTargetID() );

			// Rcksto
			recoil = 1.0f;
		}
	}
}

void Obj_Enemy_AAStation::render( RenderPass pass )
{
	if( pass == RP_SHADOW || pass == RP_OBJECT_2 )
	{
		glPushMatrix();
		setupGLMatrix();

		// Kanone rendern
		glPushMatrix();
		gunBase.setupGLMatrix();
		gun.setupGLMatrix();
		video.drawSprite( Vec2i( 256, 320 ), Vec2i( 64, 96 ), Vec2f( 2.0f, 3.0f ) );
		glPopMatrix();

		// Station rendern
		video.drawSprite( Vec2i( 256, 416 ), Vec2i( 128, 96 ), Vec2f( 4.0f, 3.0f ) );
		glPopMatrix();
	}
}

void Obj_Enemy_AAStation::think()
{
	GameObject* p_target = getTarget();
	if( p_target )
	{
		if( canSee( p_target, 30.0f, Vec2f( 0.0f, -2.5f ) ) )
		{
			// zielen
			float hitTime;
			Vec2f shootVector;
			if( computeHitWithGravity( getPosition(), Vec2f( 0.0f, 0.0f ),
									   p_target->getPosition(), p_target->getVelocity(),
									   weapon == WT_AA ? getWeapons()[getCurrentWeapon()].speed : 40.0f,
									   weapon == WT_AA ? level.getGravity() : 2.5f * level.getGravity(),
									   static_cast<Vec2f*>( 0 ), &hitTime, &shootVector ) )
			{
				// relative Zielrichtung ermitteln
				const Vec2f relDir( gun.absDirectionToRelDirection( shootVector ) );
				const float angle = gun.getAngle();

				// Kanone nachfhren, wenn ntig
				if( fabsf( relDir.x ) > 0.01f ) setParameter( "Turning", relDir.x );

				// Winkel zwischen Kanone und Abschussrichtung
				const float dot = shootVector.normalized() ^ gun.getAbsDirection();
				const float minDot = weapon == WT_AA ? -0.95f : -0.75f;
				if( dot < minDot )
				{
					// Feuer!
					setParameter( "Fire", 1.0f );

					if( weapon == WT_AA )
					{
						// Detonationszeitpunkt des Projektils festlegen
						setParameter( "BurstTime", hitTime );
					}
				}
			}
		}
	}
}

void Obj_Enemy_AAStation::addAvoidanceAreas()
{
	level.addAvoidanceCircle( 1, this, getTransform(), Vec2f( 0.0f, 0.0f ), 6.0f );
}