#include <PCH.h>
#include <Obj_Garage.h>
#include <Obj_Player.h>

Obj_Garage::Obj_Garage( Level& level,
					    const TiXmlElement* p_root ) : GameObject( level, p_root, false ),
													   p_emitter( 0 ),
													   p_soundInst( 0 ),
													   p_neonSoundInst( 0 ),
													   reload( 0.0f ),
													   flash( 0.0f ),
													   release( false )
{
	setMaterial( "Garage" );

	// Die Werkstatt kann mit Objekten kollidieren, aber nicht mit Wnden.
	setCollisionBits( CB_STATIC, ~CB_STATIC );

	std::vector<Vec2f> vertices;
	vertices.push_back( Vec2f( -2.0f, 3.0f ) );
	vertices.push_back( Vec2f( -1.0f, 2.0f ) );
	vertices.push_back( Vec2f( 3.0f, 2.0f ) );
	vertices.push_back( Vec2f( 4.0f, 3.0f ) );
	addPolyLine( vertices, true, 0.5f );

	setBoundingCircle( Vec2f( 0.0f, 0.0f ), 6.0f );

	// Attribute lesen
	repairLeft = 100.0f;
	p_root->QueryFloatAttribute( "repair", &repairLeft );
	fuelLeft = 100.0f;
	p_root->QueryFloatAttribute( "fuel", &fuelLeft );
	repairPrice = 50.0f;
	p_root->QueryFloatAttribute( "repairPrice", &repairPrice );
	fuelPrice = 25.0f;
	p_root->QueryFloatAttribute( "fuelPrice", &fuelPrice );

	// Emitter erzeugen
	p_emitter = level.getPS().createSimpleEmitter();
	p_emitter->setPosition( Vec2f( 2.5f, -0.5f ) );
	p_emitter->setDirection( Vec2f( -2.0f, 1.0f ) );
	p_emitter->addParticleStyle( 5 );
	p_emitter->setParticleTimeToLive( 0.15f, 0.05f );
	p_emitter->setParticlePositionRadius( 0.1f );
	p_emitter->setParticleDirection( Vec2f( 0.0f, -1.0f ), 0.3f );
	p_emitter->setParticleSpeed( 5.0f, 1.0f );
	p_emitter->setParticleStartColor( Vec4f( 0.5f, 0.5f, 1.0f, 1.0f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.5f ) );
	p_emitter->setParticleEndColor( Vec4f( 0.5f, 0.5f, 1.0f, 0.0f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.0f ) );
	p_emitter->setParticleStartSize( 1.0f, 0.1f );
	p_emitter->setParticleEndSize( 4.0f, 0.2f );
	p_emitter->setParticleGravity( 0.0f, 0.1f );
	p_emitter->setParticleAnimFrame( 0.0f, 10.0f );
	attachChild( p_emitter, false );

	// Sounds laden
	sound = env.engine.getSoundManager().get( "data\\garage.ogg" );
	sound->setDefaultSoundGroupName( "SFX" );
	neonSound = env.engine.getSoundManager().get( "data\\neon1.ogg" );
	neonSound->setDefaultSoundGroupName( "SFX" );
	neonSound->setDefaultDistances( 5.0f, 10000.0f );
}

Obj_Garage::~Obj_Garage()
{
	if( p_emitter ) level.getPS().destroyEmitter( p_emitter );
	if( sound->isInstanceValid( p_soundInst ) ) p_soundInst->stop();
}

void Obj_Garage::update()
{
	GameObject::update();

	p_emitter->setRate( 0.0f );

	bool playerOnPlatform = false;
	bool playerServed = false;

	// Spielerobjekt holen
	Obj_Player* p_player = static_cast<Obj_Player*>( level.getObjectByID( "player" ) );
	if( p_player )
	{
		// relative Position, relativen Winkel und Geschwindigkeit des Spielers ermitteln
		const Vec2f relPos( absPositionToRelPosition( p_player->getPosition() ) );
		const float relAngle = absAngleToRelAngle( p_player->getAngle() );
		const Vec2f velocity( p_player->getVelocity() );

		// Nur wenn alle Werte stimmen, wird der Spieler repariert und aufgetankt.
		if( fabsf( relAngle ) < 0.4f &&
			relPos.x > -1.0f &&
			relPos.x < 3.0f &&
			relPos.y > -5.0f &&
			relPos.y < 0.0f )
		{
			if( fabsf( relAngle ) < 0.2f &&
				relPos.y > -0.75f &&
				velocity.lengthSq() < 0.25f )
			{
				reload += env.dt;
				playerOnPlatform = true;

				if( reload >= 0.5f && reload < FLT_MAX )
				{
					float oldHealth = p_player->getHealth();
					float oldFuel = p_player->getFuel();

					if( p_player->getScore() )
					{
						// reparieren und auftanken
						if( repairLeft > 0.0f ) p_player->addHealth( 30.0f * env.dt );
						if( fuelLeft > 0.0f ) p_player->addFuel( 10.0f * env.dt );
					}

					float healthAdded = p_player->getHealth() - oldHealth;
					float fuelAdded = p_player->getFuel() - oldFuel;

					if( healthAdded > 0.0f || fuelAdded > 0.0f )
					{
						repairLeft -= healthAdded;
						fuelLeft -= fuelAdded;

						int price = static_cast<int>( healthAdded * repairPrice + fuelAdded * fuelPrice );
						level.addPlayerScore( -price, getPosition() );

						playerServed = true;
					}
					else
					{
						reload = FLT_MAX;
						release = true;
					}
				}
			}
			else reload = 0.0f;

			if( !release )
			{
				// Spieler heranziehen
				const float gravityStrength = level.getGravity().length();
				const float factor = 1.0f + min( gravityStrength, ( -getDirection() ^ level.getGravity() ) + gravityStrength );
				const Vec2f targetPos( relPositionToAbsPosition( Vec2f( 1.0f, 2.0f ) ) );
				const Vec2f playerPos( p_player->getPosition() );
				const Vec2f force( ( targetPos - playerPos ).normalized() * 3.5f * factor );
				dBodyAddForceAtRelPos( p_player->getBody(), force.x, force.y, 0.0, 0.0, 2.0, 0.0 );
			}
		}
		else
		{
			reload = 0.0f;
			release = false;
		}
	}
	else
	{
		reload = 0.0f;
		release = false;
	}

	if( playerServed )
	{
		p_emitter->setRate( 50.0f );

		// Sound, wenn ntig, neu abspielen
		if( !sound->isInstanceValid( p_soundInst ) )
		{
			p_soundInst = sound->createInstance();
			if( p_soundInst )
			{
				p_soundInst->setPosition( p_emitter->getPosition() );
				p_soundInst->setVolume( 0.0f );
				p_soundInst->setPitch( 0.1f );
				attachChild( p_soundInst, false );
			}
		}

		if( p_soundInst )
		{
			p_soundInst->setVolumeTarget( 0.25f, 10.0f );
			p_soundInst->setPitchTarget( 1.0f, 10.0f );
			p_soundInst->play( true );
		}
	}
	else
	{
		if( sound->isInstanceValid( p_soundInst ) )
		{
			// Sound abschalten und ihn "vergessen"
			p_soundInst->setVolumeTarget( 0.0f, 10.0f );
			p_soundInst->setPitchTarget( 0.1f, 10.0f );
			p_soundInst->setStopCountDown( -1.0f );
			p_soundInst = 0;
		}
	}

	if( flash <= 0.0f )
	{
		flash -= env.dt;
		if( flash <= -0.25f && ( repairLeft > 0.0f || fuelLeft > 0.0f ) )
		{
			flash = 5.0f;
			p_neonSoundInst = neonSound->createInstance();
			if( p_neonSoundInst )
			{
				attachChild( p_neonSoundInst, false );
				p_neonSoundInst->setVolume( randFloat( 0.25f, 0.3f ) );
				p_neonSoundInst->setPitch( randFloat( 1.0f, 1.1f ) );
				p_neonSoundInst->play();
			}
		}
	}
	else
	{
		adjustSmooth( flash, -0.1f, 5.0f, env.dt );
	}
}

void Obj_Garage::render( RenderPass pass )
{
	if( pass == RP_SHADOW || pass == RP_OBJECT_1 )
	{
		// Sprite rendern
		glPushMatrix();
		setupGLMatrix();

		if( pass == RP_OBJECT_1 )
		{
			// Wenn die Werkstatt leer ist, wird sie dunkler.
			float brightness = 1.0f;
			if( repairLeft <= 0.0f ) brightness -= 0.25f;
			if( fuelLeft <= 0.0f ) brightness -= 0.25f;
			glColor3fv( Vec3f( brightness ) );
		}

		video.drawSprite( Vec2i( 0, 320 ), Vec2i( 256, 192 ), Vec2f( 8.0f, 6.0f ) );

		if( pass == RP_OBJECT_1 && flash > 0.0f )
		{
			// Leuchten rendern
			video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE );
			glColor3f( flash + randFloat( -0.1f, 0.1f ), flash + randFloat( -0.1f, 0.1f ), flash + randFloat( -0.1f, 0.1f ) );
			glTranslatef( 0.0f, 0.25f, 0.0f );
			video.drawSprite( Vec2i( 2, 514 ), Vec2i( 284, 220 ), Vec2f( 9.0f, 7.0f ) );
			video.setupAlphaBlending( true, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		}

		glPopMatrix();
	}
}