#include <PCH.h>
#include <Obj_Laser.h>
#include <Obj_Projectile.h>
#include <Damagable.h>

Obj_Laser::Obj_Laser( Level& level,
					  const TiXmlElement* p_root ) : GameObject( level, p_root, true ),
													 active( false ),
													 p_soundInst( 0 )
{
	setMaterial( "Metal" );

	// Sound laden und Instanz erzeugen
	SoundHandle sound = env.engine.getSoundManager().get( "data\\laser.ogg" );
	sound->setDefaultSoundGroupName( "SFX" );
	sound->setDefaultDistances( 10.0f, 10000.0f );
	p_soundInst = sound->createInstance();
	if( p_soundInst )
	{
		p_soundInst->setVolume( 0.0f );
		p_soundInst->setPitch( 0.1f );
		p_soundInst->setCalculateVelocity( false );
	}

	// Dynamisch oder statisch?
	int dynamic_ = 0;
	p_root->QueryIntAttribute( "dynamic", &dynamic_ );
	dynamic = dynamic_ ? true : false;

	if( !dynamic ) deleteBody();

	addSegment( Vec2f( 0.0f, -1.0f ), Vec2f( 0.0f, 1.0f ), 0.5f );

	if( dynamic )
	{
		setCollisionBits( CB_OBJECT, CB_ALL );
		setTotalMass( 1.0f );
	}
	else
	{
		// Der Laser kann mit Objekten kollidieren, aber nicht mit Wnden.
		setCollisionBits( CB_STATIC | CB_OBJECT, ~CB_STATIC );
	}

	// Liste der Energieknoten lesen
	const TiXmlElement* p_energyNodes = p_root->FirstChildElement( "EnergyNodes" );
	if( p_energyNodes )
	{
		const TiXmlElement* p_energyNode = p_energyNodes->FirstChildElement( "EnergyNode" );
		while( p_energyNode )
		{
			const char* p_id = p_energyNode->GetText();
			if( p_id ) energyNodeIDs.insert( p_id );
			p_energyNode = p_energyNode->NextSiblingElement( "EnergyNode" );
		}
	}

	numEnergyNodesAlive = energyNodeIDs.size();

	// Eingeschaltet?
	int active_ = 1;
	p_root->QueryIntAttribute( "active", &active_ );
	setActive( active_ ? true : false );
}

Obj_Laser::~Obj_Laser()
{
	if( p_soundInst )
	{
		p_soundInst->stop();
	}
}

void Obj_Laser::update()
{
	GameObject::update();

	beam.clear();

	if( numEnergyNodesAlive && active )
	{
		lastHitObjects = currentHitObjects;
		currentHitObjects.clear();

		Vec2f beamPos( relPositionToAbsPosition( Vec2f( 0.0f, -1.5f ) ) );
		Vec2f beamDir( -getDirection() );
		beam.push_back( beamPos );

		int numReflections = 0;
		bool quit = false;

		while( !quit && numReflections < 32 )
		{
			// Entfernung bis zum nchsten Hindernis bestimmen
			Vec2f hitPos;
			Vec2f hitNormal;
			GameObject* p_hit = level.castRay( beamPos, beamDir, 250.0f, CB_LASER, CB_NONE, &hitPos, 0, &hitNormal );
			if( p_hit )
			{
				beamPos = hitPos;

				// Kann dieses Objekt Laserstrahlen reflektieren?
				if( p_hit->reflectsLaser() )
				{
					// Jawohl! Dann tut es das jetzt.
					beamDir -= 2.0f * ( beamDir ^ hitNormal ) * hitNormal;
					numReflections++;
				}
				else
				{
					Obj_Projectile* p_projectile = dynamic_cast<Obj_Projectile*>( p_hit );
					if( p_projectile )
					{
						CollisionInfo info;
						info.passThrough = false;
						info.frontal = true;
						p_projectile->onCollision( 0, info );
					}

					// Normale Objekte mssen zweimal hintereinander getroffen werden, damit sie Schaden abbekommen.
					Damagable* p_damagable = dynamic_cast<Damagable*>( p_hit );
					if( p_damagable ) currentHitObjects.insert( p_damagable );

					// Funken sprhen lassen
					SimpleEmitter* p_emitter = level.getPS().createSimpleEmitter();
					p_emitter->setPosition( beamPos );
					p_emitter->setRate( 20.0f );
					p_emitter->setTimeToLive( 0.1f );
					p_emitter->addParticleStyle( 4 );
					p_emitter->setParticleTimeToLive( 0.3f, 0.1f );
					p_emitter->setParticlePositionRadius( 0.1f );
					p_emitter->setParticleSpeed( 20.0f, 5.0f );
					p_emitter->setParticleStartColor( Vec4f( 1.0f, 1.0f, 1.0f, 0.5f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.25f ) );
					p_emitter->setParticleEndColor( Vec4f( 1.0f, 1.0f, 1.0f, 0.5f ), Vec4f( 0.25f, 0.25f, 0.25f, 0.25f ) );
					p_emitter->setParticleStartSize( 1.0f, 0.05f );
					p_emitter->setParticleEndSize( 0.0f, 0.0f );
					p_emitter->setParticleGravity( 8.0f, 1.0f );
					p_emitter->setParticleFriction( 0.25f, 0.1f );

					quit = true;
				}
			}
			else
			{
				// Der Strahl geht von hier aus ins "Unendliche".
				beamPos += 1000.0f * beamDir;
				quit = true;
			}

			beam.push_back( beamPos );

			// Strahl ein bisschen nach vorne verschieben
			beamPos += 0.01f * beamDir.normalized();
		}

		// Wurden Objekte in diesem und im letzten Update getroffen?
		std::vector<Damagable*> objects;
		std::set_intersection( currentHitObjects.begin(), currentHitObjects.end(),
							   lastHitObjects.begin(), lastHitObjects.end(),
							   std::back_inserter( objects ) );
		for( std::vector<Damagable*>::iterator it = objects.begin();
			 it != objects.end();
			 it++ )
		{
			( *it )->addHealth( -100000.0f );
		}
	}

	if( numEnergyNodesAlive )
	{
		// intakte Energieknoten zhlen
		numEnergyNodesAlive = 0;
		for( std::set<std::string>::const_iterator it = energyNodeIDs.begin();
			 it != energyNodeIDs.end();
			 it++ )
		{
			if( level.getObjectByID( *it ) ) numEnergyNodesAlive++;
		}
	}

	updateSound();

	// Bounding-Circle aktualisieren
	if( beam.empty() ) setBoundingCircle( Vec2f( 0.0f, 0.0f ), 2.5f );
	else
	{
		Vec2f brMin = getPosition() - Vec2f( 2.0f, 2.0f );
		Vec2f brMax = getPosition() + Vec2f( 2.0f, 2.0f );

		for( size_t i = 0; i < beam.size(); i++ )
		{
			brMin = Vec2f( min( brMin.x, beam[i].x ), min( brMin.y, beam[i].y ) );
			brMax = Vec2f( max( brMax.x, beam[i].x ), max( brMax.y, beam[i].y ) );
		}

		setBoundingCircle( absPositionToRelPosition( 0.5f * ( brMin + brMax ) ),
			               0.5f * ( brMax - brMin ).length() );
	}
}

void Obj_Laser::render( RenderPass pass )
{
	if( pass == RP_SHADOW || pass == RP_OBJECT_1 )
	{
		// Sprite rendern
		glPushMatrix();
		setupGLMatrix();

		if( pass == RP_OBJECT_1 )
		{
			float c = numEnergyNodesAlive && active ? 1.0f : 0.75f;
			Vec4f color( c, c, c, 1.0f );
			glColor4fv( color );
		}

		video.drawSprite( Vec2i( 384, 0 ), Vec2i( 64, 128 ), Vec2f( 2.0f, 4.0f ) );
		glPopMatrix();
	}
	else if( pass == RP_GLOW )
	{
		if( numEnergyNodesAlive )
		{
			// Laserstrahl rendern
			CartoonRenderer& cr = level.getCR();
			cr.setStyle( "laser_red" );
			cr.setContext( randInt() );
			const float t = level.getTime() + static_cast<float>( ( this - static_cast<Obj_Laser*>( 0 ) ) % 1000 );
			cr.drawPolyLine( beam, false, 0.75f + 0.1f * sinf( 40.0f * t ) + randFloat( -0.05f, 0.05f ),
							 Vec4f( 1.0f, 1.0f, 1.0f, 0.9f + 0.2f * cosf( 15.0f * level.getTime() ) + randFloat( -0.05f, 0.05f ) ) );
		}
	}
}

void Obj_Laser::addAvoidanceAreas()
{
	if( beam.empty() ) return;

	for( size_t i = 0; i < beam.size() - 1; i++ )
	{
		level.addAvoidanceLine( 1000, this, beam[i], beam[i + 1], 6.0f, 3.0f );
	}
}

void Obj_Laser::onKill()
{
	if( p_soundInst ) p_soundInst->setMuted( true );
}

void Obj_Laser::onRebirth()
{
	if( p_soundInst ) p_soundInst->setMuted( false );
}

void Obj_Laser::setActive( bool active )
{
	if( this->active != active )
	{
		this->active = active;
		updateSound();
	}
	else
	{
		this->active = active;
	}
}

bool Obj_Laser::isActive() const
{
	return active;
}

bool Obj_Laser::isPowered() const
{
	return numEnergyNodesAlive > 0;
}

void Obj_Laser::updateSound()
{
	bool soundShouldBeOn = isActive() && isPowered();
	bool soundIsOn = p_soundInst ? p_soundInst->isPlaying() : false;

	if( soundShouldBeOn == soundIsOn )
	{
		if( soundIsOn && !beam.empty() )
		{
			// Sound an die aktuelle "Einschlagsstelle" des Lasers setzen und Lautstrke/Tonhhe fluktuieren lassen
			p_soundInst->setPosition( beam.back() );
			p_soundInst->setVolumeTarget( randFloat( 0.75f, 1.25f ), 15.0f );
			p_soundInst->setPitchTarget( randFloat( 0.75f, 1.25f ), 15.0f );
		}
	}
	else
	{
		if( soundShouldBeOn )
		{
			if( p_soundInst )
			{
				// Sound abspielen
				p_soundInst->play( true );
				p_soundInst->setPauseCountDown( 0.0f );
			}
		}
		else
		{
			if( p_soundInst )
			{
				// Sound pausieren oder ganz stoppen - je nach dem, ob der Laser berhaupt noch Energie hat
				p_soundInst->setVolumeTarget( 0.0f, 1.0f );
				p_soundInst->setPitchTarget( 0.1f, 1.0f );
				if( isPowered() ) p_soundInst->setPauseCountDown( -1.0f );
				else
				{
					p_soundInst->setStopCountDown( -1.0f );
					p_soundInst = 0;
				}
			}
		}
	}
}