#include <PCH.h>
#include <Obj_Wall.h>

Obj_Wall::Obj_Wall( Level& level,
					const TiXmlElement* p_root ) : GameObject( level, p_root, true ),
												   listID( 0 )
{
	setMaterial( "Metal" );

	// Dynamisch oder statisch?
	int dynamic_ = 0;
	p_root->QueryIntAttribute( "dynamic", &dynamic_ );
	dynamic = dynamic_ ? true : false;

	// Soll die Wand Laserstrahlen durchlassen?
	int blockLaser_ = 1;
	p_root->QueryIntAttribute( "blockLaser", &blockLaser_ );
	blockLaser = blockLaser_ ? true : false;

	// Soll die Wand Laserstrahlen reflektieren?
	int reflectLaser_ = 0;
	p_root->QueryIntAttribute( "reflectLaser", &reflectLaser_ );
	reflectLaser = reflectLaser_ ? true : false;

	// Zerstrbare Wand?
	health = 0;
	p_root->QueryIntAttribute( "health", &health );
	if( health )
	{
		// Gesundheit festlegen
		setMaxHealth( static_cast<float>( health ) );
		setHealth( static_cast<float>( health ) );

		// Textur mit den Rissen laden
		cracksTexture = env.engine.getTextureManager().get( "data\\cracks.png" );
	}

	fade = 2.0f;

	float mass = 10.0f;

	if( !dynamic )
	{
		deleteBody();
	}
	else
	{
		// Masse lesen
		p_root->QueryFloatAttribute( "mass", &mass );
	}

	// Stil lesen
	const char* p_style = p_root->Attribute( "style" );
	if( p_style ) style = p_style;

	// Soll die Wand ausgemalt werden?
	int fill_ = 1;
	p_root->QueryIntAttribute( "fill", &fill_ );
	fill = fill_ ? true : false;

	// Soll der Linienzug geschlossen werden?
	int close_ = 1;
	p_root->QueryIntAttribute( "close", &close_ );
	close = close_ ? true : false;

	// Radius lesen
	radius = 0.5f;
	p_root->QueryFloatAttribute( "radius", &radius );

	// Versetzung nach auen lesen
	shift = close ? -0.7f : 0.0f;
	p_root->QueryFloatAttribute( "shift", &shift );

	// die Eckpunkte einlesen
	const TiXmlElement* p_vertex = p_root->FirstChildElement( "Vertex" );
	while( p_vertex )
	{
		Vec2f vertex( 0.0f, 0.0f );
		p_vertex->QueryFloatAttribute( "x", &vertex.x );
		p_vertex->QueryFloatAttribute( "y", &vertex.y );
		vertices.push_back( vertex );
		extendBoundingRect( vertex );
		p_vertex = p_vertex->NextSiblingElement( "Vertex" );
	}

	if( vertices.size() >= 3 )
	{
		int winding = 0;
		for( size_t i = 0; i < vertices.size(); i++)
		{
			// Richtung der Vertizes bestimmen
			const size_t j = ( i + 1 ) % vertices.size();
			const size_t k = ( i + 2 ) % vertices.size();
			const Vec3f a( vertices[i].x, vertices[i].y, 0.0f );
			const Vec3f b( vertices[j].x, vertices[j].y, 0.0f );
			const Vec3f c( vertices[k].x, vertices[k].y, 0.0f );
			const Vec3f cross( ( b - a ) % ( c - b ) );
			if( cross.z > 0.0f ) winding++;
			else winding--;
		}

		if( winding < 0 )
		{
			// Vertizes umkehren
			std::vector<Vec2f> newVertices;
			for( size_t i = 0; i < vertices.size(); i++ ) newVertices.push_back( vertices[vertices.size() - 1 - i] );
			vertices = newVertices;
		}
	}
	else
	{
		close = false;
	}

	enlargeBoundingRect( radius );

	// Zylinder erzeugen
	addPolyLine( vertices, close, radius, shift );

	if( dynamic )
	{
		// Masse setzen
		setTotalMass( mass );
	}

	if( !dynamic && !health )
	{
		// Display-Liste erzeugen
		listID = glGenLists( 1 );

		// Display-Liste fllen
		video.setTexture( 0, true );
		glNewList( listID, GL_COMPILE );
		CartoonRenderer& cr = level.getCR();
		cr.setStyle( style );
		cr.setContext( getID() );
		glPushMatrix();
		setupGLMatrix();
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
		if( fill ) cr.fillPolygon( vertices );
		cr.drawPolyLine( vertices, close );
		glPopMatrix();
		glEndList();
		video.setTexture( 0, true );
	}

	oldVisible = isVisible();
	updateCollisionBits();
}

Obj_Wall::~Obj_Wall()
{
	if( listID )
	{
		// Display-Liste freigeben
		glDeleteLists( listID, 1 );
	}
}

void Obj_Wall::update()
{
	GameObject::update();

	if( isVisible() != oldVisible )
	{
		updateCollisionBits();
		oldVisible = isVisible();
	}

	if( health )
	{
		if( getHealth() <= 0 )
		{
			if( fade == 2.0f )
			{
				// Wand zerstrt!
				float area = computePolygonArea( vertices );
				int numEmitters = static_cast<int>( area );
				int numEmittersLeft = numEmitters;
				Vec2f sum( 0.0f, 0.0f );

				while( numEmittersLeft )
				{
					// zufllige Position innerhalb des Bounding-Rects auswhlen
					const Vec2f r( randFloat(), randFloat() );
					const Vec2f pos( getBoundMin() + r * ( getBoundMax() - getBoundMin() ) );

					// Der Punkt muss in der Wand liegen.
					if( !pointInPolygon( pos, vertices ) ) continue;

					SimpleEmitter* p_emitter = level.getPS().createSimpleEmitter();
					p_emitter->setPosition( pos );
					p_emitter->setRate( randFloat( 30.0f, 60.0f ) );
					p_emitter->setStartDelay( randFloat( 0.0f, 0.5f ) );
					p_emitter->setTimeToLive( randFloat( 0.25f, 0.75f ) );
					p_emitter->setVisibilityBehavior( false, 10.0f );
					p_emitter->addParticleStyle( 1 );
					p_emitter->addParticleStyle( 7 );
					p_emitter->setParticleTimeToLive( 0.5f, 0.25f );
					p_emitter->setParticlePositionRadius( 2.0f );
					p_emitter->setParticleSpeed( 4.0f, 2.0f );
					p_emitter->setParticleStartColor( Vec4f( 1.0f, 1.0f, 1.0f, 0.6f ), Vec4f( 0.15f, 0.15f, 0.15f, 0.2f ) );
					p_emitter->setParticleEndColor( Vec4f( 0.0f, 0.0f, 0.0f, 0.0f ), Vec4f( 0.0f, 0.0f, 0.0f, 0.0f ) );
					p_emitter->setParticleStartSize( 1.0f, 0.5f );
					p_emitter->setParticleEndSize( 4.0f, 2.0f );
					p_emitter->setParticleGravity( 0.0f, 0.1f );
					p_emitter->setParticleAnimFrame( 0.0f, 10.0f );
					attachChild( p_emitter, false );

					numEmittersLeft--;
					sum += pos;
				}

				SoundHandle sound = env.engine.getSoundManager().get( "data\\vaporize.ogg" );
				SoundInstance* p_soundInst = sound->createInstance();
				if( p_soundInst )
				{
					attachChild( p_soundInst, false );
					p_soundInst->setPosition( sum / static_cast<float>( numEmitters ) );
					p_soundInst->setVolume( randFloat( 0.9f, 1.0f ) );
					p_soundInst->setPitch( randFloat( 0.75f, 1.25f ) );
					p_soundInst->play();
				}
			}

			fade -= 3.0f * env.dt;
			if( fade <= 0.0f ) kill();
		}
	}
}

void Obj_Wall::render( RenderPass pass )
{
	if( listID )
	{
		if( pass == RP_BACKGROUND_1 )
		{
			glCallList( listID );
			video.setTexture( 0, true );
		}

		return;
	}

	if( ( dynamic && pass == RP_OBJECT_2 ) ||
		( !dynamic && pass == RP_BACKGROUND_1 ) )
	{
		glPushMatrix();
		setupGLMatrix();

		CartoonRenderer& cr = level.getCR();
		Vec4f color( level.getPassColor() );

		if( fade != 1.0f )
		{
			// ausblenden
			color.a *= fade;
		}

		// Stil und Kontext setzen
		cr.setStyle( style );
		cr.setContext( getID() );

		// erst die Fllung
		if( fill )
		{
			glColor4fv( color );
			cr.fillPolygon( vertices );
		}

		if( cracksTexture ) cr.fillPolygon( vertices, true, cracksTexture );

		// jetzt die Umrisse
		cr.drawPolyLine( vertices, close, 1.0f, color );

		glPopMatrix();
	}
}

bool Obj_Wall::reflectsLaser() const
{
	return reflectLaser;
}

void Obj_Wall::updateCollisionBits()
{
	if( dynamic )
	{
		// Dynamische Wnde knnen mit allen Objekten kollidieren.
		setCollisionBits( CB_OBJECT | ( isVisible() ? CB_BLOCK_VIEW : 0 ), blockLaser ? CB_ALL : ~CB_LASER );
	}
	else
	{
		// Wnde knnen nicht mit Wnden kollidieren!
		setCollisionBits( CB_STATIC | ( isVisible() ? CB_BLOCK_VIEW : 0 ), blockLaser ? ~CB_STATIC : ~( CB_STATIC | CB_LASER ) );
	}
}