#include <PCH.h>
#include <Options.h>

Options::Options( Engine& engine,
				  GUI& gui ) : engine( engine ),
							   gui( gui )
{
	// Layouts laden
	gui.loadLayout( "Options.layout" );
	SUBSCRIBE_GUI_EVENT( "Options", CEGUI::Window::EventShown, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window", CEGUI::FrameWindow::EventCloseClicked, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/MasterVolume", CEGUI::Slider::EventValueChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/SFXVolume", CEGUI::Slider::EventValueChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/MusicVolume", CEGUI::Slider::EventValueChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/ShadowQuality", CEGUI::Slider::EventValueChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/ParticleRate", CEGUI::Slider::EventValueChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/MotionBlurring", CEGUI::Checkbox::EventCheckStateChanged, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/OK", CEGUI::PushButton::EventClicked, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/Cancel", CEGUI::PushButton::EventClicked, &Options::handleOptions );
	SUBSCRIBE_GUI_EVENT( "Options/Window/Default", CEGUI::PushButton::EventClicked, &Options::handleOptions );

	// Standardoptionen festlegen
	defaultOptions.set<float>( "MasterVolume", 0.75f );
	defaultOptions.set<float>( "SFXVolume", 1.0f );
	defaultOptions.set<float>( "MusicVolume", 1.0f );
	defaultOptions.set<unsigned int>( "ShadowQuality", 3 );
	defaultOptions.set<float>( "ParticleRate", 1.0f );
	defaultOptions.set<bool>( "MotionBlurring", true );

	options = defaultOptions;

	// versuchen, die Optionen zu laden
	XMLSerializationStream str( "game.cfg" );
	if( !str.getError() )
	{
		// Konfiguration versuchen zu laden
		ParameterMap deserialized;
		if( !Serializer::inst().deserialize( deserialized, str ) )
		{
			// verschmelzen
			options += deserialized;
		}
	}

	// Optionen anwenden
	apply();
}

Options::~Options()
{
}

void Options::show()
{
	// alte Optionen merken
	originalOptions = options;

	// Fenster anzeigen
	gui.getWindow( "Options" )->show();

	updateScreen();
}

void Options::apply()
{
	// Lautstrke setzen
	Audio& audio = engine.getAudio();
	audio.getMasterSoundGroup()->setVolume( options.get<float>( "MasterVolume" ) );
	audio.getSoundGroupByName( "SFX" )->setVolume( options.get<float>( "SFXVolume" ) );
	audio.getSoundGroupByName( "Music" )->setVolume( options.get<float>( "MusicVolume" ) );
}

const ParameterMap& Options::getOptions() const
{
	return options;
}

bool Options::handleOptions( const CEGUI::EventArgs& e )
{
	const CEGUI::WindowEventArgs& we = static_cast<const CEGUI::WindowEventArgs&>( e );
	const std::string windowName( we.window->getName().c_str() );

	if( windowName == "Options" )
	{
		gui.getWindow( "Options/Window" )->activate();
	}
	else if( windowName == "Options/Window" )
	{
		// Schlieen des Fensters bedeutet abbrechen
		cancel();
	}
	else if( windowName == "Options/Window/MasterVolume" ||
			 windowName == "Options/Window/SFXVolume" ||
			 windowName == "Options/Window/MusicVolume" ||
			 windowName == "Options/Window/ShadowQuality" ||
			 windowName == "Options/Window/ParticleRate" ||
			 windowName == "Options/Window/MotionBlurring" )
	{
		updateOptions();
		apply();

		if( windowName == "Options/Window/ShadowQuality" )
		{
			// Qualittsstufe als Text neben den Schieberegler schreiben
			unsigned int quality = static_cast<unsigned int>( static_cast<CEGUI::Slider*>( we.window )->getCurrentValue() );
			std::ostringstream str;
			str << quality;
			gui.getWindow( "Options/Window/ShadowQualityValue" )->setText( str.str() );
		}
		else if( windowName == "Options/Window/ParticleRate" )
		{
			// Partikelrate als Text neben den Schieberegler schreiben
			float particleRate = static_cast<CEGUI::Slider*>( we.window )->getCurrentValue();
			std::ostringstream str;
			str.flags( std::ios::fixed );
			str.width( 3 );
			str.precision( 1 );
			str << particleRate;
			gui.getWindow( "Options/Window/ParticleRateValue" )->setText( str.str() );
		}
	}
	else if( windowName == "Options/Window/OK" )
	{
		// Optionen speichern
		XMLSerializationStream str;
		Serializer::inst().serialize( options, str );
		str.save( "game.cfg" );

		// Fenster verstecken
		gui.getWindow( "Options" )->hide();
	}
	else if( windowName == "Options/Window/Cancel" )
	{
		cancel();
	}
	else if( windowName == "Options/Window/Default" )
	{
		// Standardoptionen wiederherstellen
		options = defaultOptions;
		updateScreen();
		apply();
	}
	else
	{
		return false;
	}

	return true;
}

void Options::updateScreen()
{
	ParameterMap copy( options );

	// Lautstrke
	static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/MasterVolume" ) )->setCurrentValue( copy.get<float>( "MasterVolume" ) );
	static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/SFXVolume" ) )->setCurrentValue( copy.get<float>( "SFXVolume" ) );
	static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/MusicVolume" ) )->setCurrentValue( copy.get<float>( "MusicVolume" ) );

	// Schattenqualitt
	unsigned int shadowQuality = copy.get<unsigned int>( "ShadowQuality" );
	static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/ShadowQuality" ) )->setCurrentValue( static_cast<float>( shadowQuality ) );

	// Partikelrate
	float particleRate = copy.get<float>( "ParticleRate" );
	static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/ParticleRate" ) )->setCurrentValue( particleRate );

	// Motion-Blurring
	static_cast<CEGUI::Checkbox*>( gui.getWindow( "Options/Window/MotionBlurring" ) )->setSelected( copy.get<bool>( "MotionBlurring" ) );
}

void Options::updateOptions()
{
	// Lautstrke
	options.set<float>( "MasterVolume", static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/MasterVolume" ) )->getCurrentValue() );
	options.set<float>( "SFXVolume", static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/SFXVolume" ) )->getCurrentValue() );
	options.set<float>( "MusicVolume", static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/MusicVolume" ) )->getCurrentValue() );

	// Schattenqualitt
	unsigned int shadowQuality = static_cast<unsigned int>( static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/ShadowQuality" ) )->getCurrentValue() );
	options.set<unsigned int>( "ShadowQuality", shadowQuality );

	// Partikelrate
	float particleRate = static_cast<CEGUI::Slider*>( gui.getWindow( "Options/Window/ParticleRate" ) )->getCurrentValue();
	options.set<float>( "ParticleRate", particleRate );

	// Motion-Blurring
	options.set<bool>( "MotionBlurring", static_cast<CEGUI::Checkbox*>( gui.getWindow( "Options/Window/MotionBlurring" ) )->isSelected() );
}

void Options::cancel()
{
	// Fenster verstecken
	gui.getWindow( "Options" )->hide();

	// alte Optionen wiederherstellen
	options = originalOptions;
	apply();
}